@@ -1,3 +1,6 @@
+/* btreei.h */
+/* functions internal to the btree subdirectory */
+
 #ifndef _BTREE_PRIV_H
 #define _BTREE_PRIV_H
 
@@ -10,13 +13,17 @@ void addkey(BTREE, FKEY, RKEY, FKEY);
 BLOCK crtblock(BTREE);
 BLOCK allocblock(void);
 
+/* btree.c */
+RAWRECORD readrec(BTREE btree, BLOCK block, INT i, INT *plen);
+
 /* index.c */
 INDEX crtindex(BTREE);
+void freecache(BTREE);
 INDEX getindex(BTREE, FKEY);
 void initcache(BTREE, INT);
 void putheader(BTREE, BLOCK);
 void putindex(BTREE, INDEX);
-void writeindex(STRING, INDEX);
+void writeindex(BTREE, INDEX);
 
 /* utils.c */
 BOOLEAN newmaster(BTREE, INDEX);
