@@ -288,8 +288,8 @@ do_zformat_date (STRING str, INT dfmt, INT mfmt,
 
 	if (sfmt==12) {
 		/* This is what used to be the shrt flag */
+		zs_free(&zstr);
 		return zshorten_date(str);
-		return zstr;
 	}
 	if (sfmt==14) {
 		zs_sets(zstr, str);
@@ -1317,7 +1317,7 @@ get_date_tok (struct tag_dnum *pdnum)
 			*p = 0;
 		}
 		/* look it up in our big table of GEDCOM keywords */
-		i = valueof_int(keywordtbl, upperascii_s(scratch), 0);
+		i = valueof_int(keywordtbl, upperascii_s(scratch));
 		if (i >= 2001 && i < 2000 + GDV_CALENDARS_IX) {
 			pdnum->val = i - 2000;
 			return CALENDAR_TOK;
@@ -1333,12 +1333,12 @@ get_date_tok (struct tag_dnum *pdnum)
 		} while (sstr[0] && sstr[0]!=')' && !iswhite((uchar)sstr[0]));
 		*p = 0;
 		/* look it up in our big table of GEDCOM keywords */
-		i = valueof_int(keywordtbl, upperascii_s(scratch), 0);
+		i = valueof_int(keywordtbl, upperascii_s(scratch));
 		if (!i) {
 			/* unrecognized word */
 			return CHAR_TOK;
 		}
-		if ((i = valueof_int(keywordtbl, upperascii_s(scratch), 0)) > 0 && i <= 999) {
+		if (i > 0 && i <= 999) {
 			pdnum->val = i % 100;
 			/* TODO: we need to use the fact that calendar is i/100 */
 			/* That is, now we know what calendar this is in */
@@ -1433,7 +1433,7 @@ static void
 init_keywordtbl (void)
 {
 	INT i, j;
-	keywordtbl = create_table(DONTFREE);
+	keywordtbl = create_table_int();
 	/* Load GEDCOM keywords & values into keyword table */
 	for (i=0; i<ARRSIZE(gedkeys); ++i) {
 		j = gedkeys[i].value;
@@ -1873,3 +1873,38 @@ zshorten_date (STRING str)
 	zs_sets(zstr, sht);
 	return zstr;
 }
+/*====================================+
+ * approx_time -- display duration specified in seconds
+ *  eg, approx_time(70000) returns "19h26m"
+ *===================================*/
+ZSTR
+approx_time (INT seconds)
+{
+	INT minutes, hours, days, years;
+	minutes = seconds/60;
+	if (!minutes) {
+		/* TRANSLATORS: seconds time interval */
+		return zs_newf(_("%02ds"), seconds);
+	}
+	seconds = seconds - minutes*60;
+	hours = minutes/60;
+	if (!hours) {
+		/* TRANSLATORS: minutes & seconds time interval */
+		return zs_newf(_("%dm%02ds"), minutes, seconds);
+	}
+	minutes = minutes - hours*60;
+	days = hours/60;
+	if (!days) {
+		/* TRANSLATORS: hours & minutes time interval */
+		return zs_newf(_("%dh%02dm"), hours, minutes);
+	}
+	hours = hours - days*24;
+	years = days/365.2425;
+	if (!years) {
+		/* TRANSLATORS: days & hours time interval */
+		return zs_newf(_("%dd%02dh"), days, hours);
+	}
+	days = days - years*365.2425;
+	/* TRANSLATORS: years & days time interval */
+	return zs_newf( _("%dy%03dd"), years, days);
+}
