@@ -62,8 +62,6 @@
 	all PVALUES for report commands
 */
 
-extern BOOLEAN opt_finnish;		/* Finnish language support */
-
 
 /*********************************************
  * local enums
@@ -76,6 +74,29 @@ extern BOOLEAN opt_finnish;		/* Finnish language support */
 #define ISPRN_FAMSEQ 1
 #define ISPRN_SPOUSESEQ 2
 
+/*********************************************
+ * local types
+ *********************************************/
+
+/*==================================================================
+ * SORTEL -- Data type for indiseq elements; keys are always present
+ *   and belong to the structure; names are always present for
+ *   persons and belong; values do not belong to the structure
+ *================================================================*/
+struct tag_sortel {
+	STRING s_key;	/* person or family key */
+	STRING s_nam;	/* name of person */
+	UNION s_val;	/* any value */
+	STRING s_prn;	/* menu print string */
+	INT s_pri;	/* key as integer (exc valuesort_indiseq puts values here) */
+};
+/* typedef struct tag_sortel *SORTEL; */ /* in indiseq.h */
+#define skey(s) ((s)->s_key)
+#define snam(s) ((s)->s_nam)
+#define sval(s) ((s)->s_val)
+#define sprn(s) ((s)->s_prn)
+#define spri(s) ((s)->s_pri)
+
 /*********************************************
  * local function prototypes
  *********************************************/
@@ -83,7 +104,7 @@ extern BOOLEAN opt_finnish;		/* Finnish language support */
 /* alphabetical */
 static void append_all_tags(INDISEQ, NODE, STRING tagname, BOOLEAN recurse, BOOLEAN nonptrs);
 static void append_indiseq_impl(INDISEQ seq, STRING key, 
-	STRING name, UNION val, BOOLEAN sure, BOOLEAN alloc);
+	CNSTRING name, UNION val, BOOLEAN sure, BOOLEAN alloc);
 static void calc_indiseq_name_el(INDISEQ seq, INT index);
 static INT canonkey_compare(SORTEL el1, SORTEL el2, VPTR param);
 static INT canonkey_order(char c);
@@ -187,7 +208,7 @@ remove_indiseq (INDISEQ seq)
 	for (i = 0; i < n; i++, d++) {
 		stdfree(skey(*d));
 		if (snam(*d)) stdfree(snam(*d));
-		default_delete_value(sval(*d), IValtype(seq));
+		deleteval(seq, sval(*d));
 		if (sprn(*d)) stdfree(sprn(*d));
 		stdfree(*d);
 	}
@@ -311,9 +332,10 @@ append_indiseq_pval (INDISEQ seq,    /* sequence */
                      STRING key,     /* key - not NULL */
                      STRING name,    /* name - may be NULL */
                      VPTR pval,       /* extra val */
-                     BOOLEAN sure,   /* no dupe check? */
-                     BOOLEAN alloc)  /* key alloced? */
+                     BOOLEAN sure)   /* no dupe check? */
+
 {
+	BOOLEAN alloc=FALSE; /* let append_indiseq_impl copy key */
 	UNION u;
 	u.w = pval;
 	check_indiseq_valtype(seq, ISVAL_PTR);
@@ -329,7 +351,7 @@ append_indiseq_pval (INDISEQ seq,    /* sequence */
 void
 append_indiseq_sval (INDISEQ seq,    /* sequence */
                      STRING key,     /* key - not NULL */
-                     STRING name,    /* name - may be NULL */
+                     CNSTRING name,    /* name - may be NULL */
                      STRING sval,       /* extra val */
                      BOOLEAN sure,   /* no dupe check? */
                      BOOLEAN alloc)  /* key alloced? */
@@ -347,7 +369,7 @@ append_indiseq_sval (INDISEQ seq,    /* sequence */
 void
 append_indiseq_null (INDISEQ seq,    /* sequence */
                      STRING key,     /* key - not NULL */
-                     STRING name,    /* name - may be NULL */
+                     CNSTRING name,    /* name - may be NULL */
                      BOOLEAN sure,   /* no dupe check? */
                      BOOLEAN alloc)  /* key alloced? */
 {
@@ -361,13 +383,13 @@ append_indiseq_null (INDISEQ seq,    /* sequence */
  *  all type appends use this implementation
  * INDISEQ seq:      sequence
  * STRING key:       key - not NULL
- * STRING name:      name - may be NULL
+ * CNSTRING name:    name - may be NULL
  * UNION val:        extra val is pointer
  * BOOLEAN sure:     no dupe check?
  * BOOLEAN alloc:    key alloced?
  *================================================*/
 static void
-append_indiseq_impl (INDISEQ seq, STRING key, STRING name, UNION val
+append_indiseq_impl (INDISEQ seq, STRING key, CNSTRING name, UNION val
 	, BOOLEAN sure, BOOLEAN alloc)
 {
 	INT i, m, n;
@@ -559,6 +581,77 @@ element_indiseq_ival (INDISEQ seq, INT index, STRING *pkey, INT *pval
 	*pname = snam(IData(seq)[index]);
 	return TRUE;
 }
+/*================================================
+ * element_key_indiseq -- Return element key from sequence
+ *  or NULL if bad inputs
+ *==============================================*/
+CNSTRING
+element_key_indiseq (INDISEQ seq, INT index)
+{
+	if (!seq) return NULL;
+	if (index<0 || index>=ISize(seq)) return NULL;
+	return skey(IData(seq)[index]);
+}
+/*================================================
+ * element_skey -- Return element key from indiseq element
+ *  Requires valid input
+ *==============================================*/
+CNSTRING
+element_skey (SORTEL el)
+{
+	ASSERT(el);
+	return el->s_key;
+}
+/*================================================
+ * element_name -- Return element name from indiseq element
+ *  Requires valid input
+ *==============================================*/
+CNSTRING
+element_name (SORTEL el)
+{
+	ASSERT(el);
+	return el->s_nam;
+}
+/*================================================
+ * element_sval -- Return string value from indiseq element
+ *  Requires valid input
+ *==============================================*/
+CNSTRING
+element_sval (SORTEL el)
+{
+	ASSERT(el);
+	return el->s_val.w;
+}
+/*================================================
+ * element_ikey -- Return integer key value from indiseq element
+ *  Requires valid input
+ *==============================================*/
+INT
+element_ikey (SORTEL el)
+{
+	ASSERT(el);
+	return el->s_pri;
+}
+/*================================================
+ * element_pval -- Return pointer value from indiseq element
+ *  Requires valid input
+ *==============================================*/
+VPTR
+element_pval (SORTEL el)
+{
+	ASSERT(el);
+	return el->s_val.w;
+}
+/*================================================
+ * set_element_pval -- Change pointer value of element
+ *  Requires valid input
+ *==============================================*/
+void
+set_element_pval (SORTEL el, VPTR ptr)
+{
+	ASSERT(el);
+	el->s_val.w = ptr;
+}
 /*==================================
  * name_compare -- Compare two names
  *================================*/
@@ -1093,30 +1186,28 @@ difference_indiseq (INDISEQ one, INDISEQ two)
 INDISEQ
 parent_indiseq (INDISEQ seq)
 {
-	TABLE tab; /* table of people inserted (values not used) */
-	INDISEQ par;
-	NODE indi, fath, moth;
-	STRING key;
+	TABLE tab=0; /* table of people inserted (values not used) */
+	INDISEQ par=0;
+	NODE indi=0;
+	INT fnum=0, snum=0;
+	STRING key=0;
 	UNION uval;
 	if (!seq) return NULL;
-	tab = create_table(DONTFREE);
+	tab = create_table_vptr();
 	par = create_indiseq_impl(IValtype(seq), IValfnctbl(seq));
 	FORINDISEQ(seq, el, num)
 		indi = key_to_indi(skey(el));
-		fath = indi_to_fath(indi);
-		moth = indi_to_moth(indi);
-		if (fath && !in_table(tab, key = indi_to_key(fath))) {
-			/* indiseq values must be copied with copyval */
-			uval = copyval(seq, sval(el));
-			append_indiseq_impl(par, strsave(key), NULL, uval, TRUE, TRUE);
-			insert_table_int(tab, key, 0);
-		}
-		if (moth && !in_table(tab, key = indi_to_key(moth))) {
-			/* indiseq values must be copied with copyval */
-			uval = copyval(seq, sval(el));
-			append_indiseq_impl(par, strsave(key), NULL, uval, TRUE, TRUE);
-			insert_table_int(tab, key, 0);
-		}
+		FORFAMCS(indi, fam, fath, moth, fnum)
+			FORFAMSPOUSES(fam, spouse, snum)
+				if (spouse && !in_table(tab, key = indi_to_key(spouse))) {
+					/* indiseq values must be copied with copyval */
+					uval = copyval(seq, sval(el));
+					key = strsave(key);
+					append_indiseq_impl(par, key, NULL, uval, TRUE, TRUE);
+					insert_table_ptr(tab, key, 0);
+				}
+			ENDFAMSPOUSES
+		ENDFAMCS
 	ENDINDISEQ
 	destroy_table(tab);
 	return par;
@@ -1129,17 +1220,17 @@ INDISEQ
 child_indiseq (INDISEQ seq)
 {
 	INT num1, num2;
-	TABLE tab; /* table of people already inserted (values not used) */
-	INDISEQ cseq;
-	NODE indi;
-	STRING key;
+	TABLE tab=0; /* table of people already inserted (values not used) */
+	INDISEQ cseq=0;
+	NODE indi=0;
+	STRING key=0;
 	UNION uval;
 	if (!seq) return NULL;
-	tab = create_table(DONTFREE);
+	tab = create_table_vptr();
 	cseq = create_indiseq_impl(IValtype(seq), IValfnctbl(seq));
 	FORINDISEQ(seq, el, num)
 		indi = key_to_indi(skey(el));
-		FORFAMSS(indi, fam, spouse, num1)
+		FORFAMS(indi, fam, num1)
 			FORCHILDRENx(fam, chil, num2)
 				key = indi_to_key(chil);
 				if (!in_table(tab, key)) {
@@ -1147,10 +1238,10 @@ child_indiseq (INDISEQ seq)
 					/* indiseq values must be copied with copyval */
 					uval = copyval(seq, sval(el));
 					append_indiseq_impl(cseq, key, NULL, uval, TRUE, TRUE);
-					insert_table_int(tab, key, 0);
+					insert_table_ptr(tab, key, 0);
 				}
 			ENDCHILDRENx
-		ENDFAMSS
+		ENDFAMS
 	ENDINDISEQ
 	destroy_table(tab);
 	return cseq;
@@ -1167,13 +1258,13 @@ indi_to_children (NODE indi)
 	STRING key;
 	if (!indi) return NULL;
 	seq = create_indiseq_null();
-	FORFAMSS(indi, fam, spouse, num1)
+	FORFAMS(indi, fam, num1)
 		FORCHILDRENx(fam, chil, num2)
 			len++;
 			key = indi_to_key(chil);
 			append_indiseq_null(seq, key, NULL, FALSE, FALSE);
 		ENDCHILDRENx
-	ENDFAMSS
+	ENDFAMS
 	if (len) return seq;
 	remove_indiseq(seq);
 	return NULL;
@@ -1192,24 +1283,16 @@ indi_to_spouses (NODE indi)
 	if (!indi) return NULL;
 	seq = create_indiseq_ival();
 	IPrntype(seq) = ISPRN_SPOUSESEQ;
-	FORFAMSS(indi, fam, spouse, num)
-		FORHUSBS(fam, husb, num1)
-			if(husb != indi) {
+	FORFAMS(indi, fam, num)
+		FORFAMSPOUSES(fam, spouse, num1)
+			if (spouse != indi) {
 				len++;
-				key = indi_to_key(husb);
+				key = indi_to_key(spouse);
 				val = atoi(fam_to_key(fam) + 1);
 				append_indiseq_ival(seq, key, NULL, val, TRUE, FALSE);
 			}
-		ENDHUSBS
-		FORWIFES(fam, wife, num1)
-			if(wife != indi) {
-				len++;
-				key = indi_to_key(wife);
-				val = atoi(fam_to_key(fam) + 1);
-				append_indiseq_ival(seq, key, NULL, val, TRUE, FALSE);
-			}
-		ENDWIFES
-	ENDFAMSS
+		ENDFAMSPOUSES
+	ENDFAMS
 	if (!len) {
 		remove_indiseq(seq);
 		seq=NULL;
@@ -1264,23 +1347,37 @@ indi_to_mothers (NODE indi)
  * indi_to_families -- Create sequence of person's families
  *  values will be spouse keynums
  *  (this is the ISPRN_FAMSEQ style)
+ * indi: [IN]  find families of this person
+ * fams: [IN]  if TRUE, find spousal families, else parental
  *=======================================================*/
 INDISEQ
-indi_to_families (NODE indi,      /* person */
-                  BOOLEAN fams)   /* families as spouse? */
+indi_to_families (NODE indi, BOOLEAN fams)
 {
 	INDISEQ seq;
-	INT num, val;
+	INT num, num2, val;
 	STRING key;
+	INT mykeynum=0;
 	if (!indi) return NULL;
+	mykeynum = atoi(indi_to_key(indi) + 1);
 	seq = create_indiseq_ival();
 	IPrntype(seq) = ISPRN_FAMSEQ;
 	if (fams) {
-		FORFAMSS(indi, fam, spouse, num)
-			key = fam_to_key(fam);
-			val = spouse ? atoi(indi_to_key(spouse) + 1) : 0;
-			append_indiseq_ival(seq, key, NULL, val, TRUE, FALSE);
-		ENDFAMSS
+		FORFAMS(indi, fam, num)
+		{
+			INT spkeynum=0;
+			STRING fkey = strsave(fam_to_key(fam));
+			/* look for a spouse besides indi */
+			FORFAMSPOUSES(fam, spouse, num2)
+			{
+				INT temp = atoi(indi_to_key(spouse) + 1);
+				if (temp && temp != mykeynum)
+					spkeynum = temp;
+			}
+			ENDFAMSPOUSES
+			append_indiseq_ival(seq, fkey, NULL, spkeynum, TRUE, FALSE);
+			strfree(&fkey);
+		}
+		ENDFAMS
 	} else {
 		FORFAMCS(indi, fam, fath, moth, num)
 			key = fam_to_key(fam);
@@ -1324,7 +1421,7 @@ fam_to_fathers (NODE fam)
 	seq = create_indiseq_null();
 	FORHUSBS(fam, husb, num)
 		key = indi_to_key(husb);
-		append_indiseq_null(seq, strsave(key), NULL, TRUE, TRUE);
+		append_indiseq_null(seq, key, NULL, TRUE, FALSE);
 	ENDHUSBS
 	if (num && ISize(seq)) return seq;
 	remove_indiseq(seq);
@@ -1343,7 +1440,7 @@ fam_to_mothers (NODE fam)
 	seq = create_indiseq_null();
 	FORWIFES(fam, wife, num)
 		key = indi_to_key(wife);
-		append_indiseq_null(seq, strsave(key), NULL, TRUE, TRUE);
+		append_indiseq_null(seq, key, NULL, TRUE, FALSE);
 	ENDWIFES
 	if (num && ISize(seq)) return seq;
 	remove_indiseq(seq);
@@ -1351,26 +1448,26 @@ fam_to_mothers (NODE fam)
 }
 /*=========================================================
  * sibling_indiseq -- Create sibling sequence of a sequence
+ * close: if TRUE, do not include duplicates
  *=======================================================*/
 INDISEQ
-sibling_indiseq (INDISEQ seq,
-                 BOOLEAN close)
+sibling_indiseq (INDISEQ seq, BOOLEAN close)
 {
-	INDISEQ fseq, sseq;
-	NODE indi, fam;
-	STRING key, fkey;
-	INT num2;
+	INDISEQ fseq=0, sseq=0;
+	NODE indi=0, fam=0;
+	STRING key=0, fkey=0;
+	INT num2=0;
 	/* table lists people already listed (values unused) */
-	TABLE tab = create_table(DONTFREE);
+	TABLE tab = create_table_vptr();
 	fseq = create_indiseq_null(); /* temporary */
 	sseq = create_indiseq_null();
 	FORINDISEQ(seq, el, num)
 		indi = key_to_indi(skey(el));
 		if ((fam = indi_to_famc(indi))) {
 			fkey = fam_to_key(fam);
-			append_indiseq_null(fseq, strsave(fkey), NULL, FALSE, TRUE);
+			append_indiseq_null(fseq, fkey, NULL, FALSE, FALSE);
 		}
-		if (!close) insert_table_int(tab, skey(el), 0);
+		if (!close) insert_table_ptr(tab, skey(el), 0);
 	ENDINDISEQ
 	FORINDISEQ(fseq, el, num)
 		fam = key_to_fam(skey(el));
@@ -1379,7 +1476,7 @@ sibling_indiseq (INDISEQ seq,
 			if (!in_table(tab, key)) {
 				key = strsave(key);
 				append_indiseq_null(sseq, key, NULL, TRUE, TRUE);
-				insert_table_int(tab, key, 0);
+				insert_table_ptr(tab, key, 0);
 			}
 		ENDCHILDRENx
 	ENDINDISEQ
@@ -1398,14 +1495,15 @@ ancestor_indiseq (INDISEQ seq)
 	/* table lists people already listed (values unused) */
 	TABLE tab;
 	LIST anclist, genlist;
-	INDISEQ anc;
-	NODE indi, fath, moth;
+	INDISEQ anc=0;
+	NODE indi=0;
 	STRING key, pkey;
-	INT gen;
+	INT gen=0;
+	INT fnum=0, snum=0;
 	UNION uval;
 	if (!seq) return NULL;
 		/* table of people already added */
-	tab = create_table(DONTFREE);
+	tab = create_table_vptr();
 		/* paired processing list - see comments in descendant_indiseq code */
 	anclist = create_list();
 	genlist = create_list();
@@ -1418,30 +1516,23 @@ ancestor_indiseq (INDISEQ seq)
 		key = (STRING) dequeue_list(anclist);
 		gen = (INT) dequeue_list(genlist) + 1;
 		indi = key_to_indi(key);
-		fath = indi_to_fath(indi);
-		moth = indi_to_moth(indi);
-		if (fath && !in_table(tab, pkey = indi_to_key(fath))) {
-				/* key copy for seq & list - owned by seq */
-			pkey = strsave(pkey);
-			uval = creategenval(seq, gen);
-			append_indiseq_pval(anc, pkey, NULL, uval.w, TRUE, TRUE);
-			enqueue_list(anclist, (VPTR)pkey);
-			enqueue_list(genlist, (VPTR)gen);
-			insert_table_int(tab, pkey, 0);
-		}
-		if (moth && !in_table(tab, pkey = indi_to_key(moth))) {
-				/* key copy for seq & list - owned by seq */
-			pkey = strsave(pkey);
-			uval = creategenval(seq, gen);
-			append_indiseq_pval(anc, pkey, NULL, uval.w, TRUE, TRUE);
-			enqueue_list(anclist, (VPTR)pkey);
-			enqueue_list(genlist, (VPTR)gen);
-			insert_table_int(tab, pkey, 0);
-		}
+
+		FORFAMCS(indi, fam, fath, moth, fnum)
+			FORFAMSPOUSES(fam, spouse, snum)
+				if (spouse && !in_table(tab, pkey = indi_to_key(spouse))) {
+						/* copy key for list, others make their own copies */
+					uval = creategenval(seq, gen);
+					append_indiseq_pval(anc, pkey, NULL, uval.w, TRUE);
+					enqueue_list(anclist, (VPTR)strsave(pkey));
+					enqueue_list(genlist, (VPTR)gen);
+					insert_table_ptr(tab, pkey, 0);
+				}
+			ENDFAMSPOUSES
+		ENDFAMCS
 	}
 	destroy_table(tab);
-	remove_list(anclist, NULL);
-	remove_list(genlist, NULL);
+	destroy_empty_list(anclist);
+	destroy_empty_list(genlist);
 	return anc;
 }
 /*=============================================================
@@ -1462,10 +1553,10 @@ descendent_indiseq (INDISEQ seq)
 	STRING key, dkey, fkey;
 	UNION uval;
 	if (!seq) return NULL;
-		/* itab = people already added */
-	itab = create_table(DONTFREE);
-		/* ftab = families already added (processed) */
-	ftab = create_table(FREEKEY);
+		/* itab = people already added, value irrelevant */
+	itab = create_table_vptr();
+		/* ftab = families already added (processed), value irrelevant */
+	ftab = create_table_vptr();
 		/*
 		deslist & genlist are paired - 
 		dequeue the person from deslist & the generation
@@ -1488,28 +1579,27 @@ descendent_indiseq (INDISEQ seq)
 		key = (STRING) dequeue_list(deslist);
 		gen = (INT) dequeue_list(genlist) + 1;
 		indi = key_to_indi(key);
-		FORFAMSS(indi, fam, spouse, num1)
+		FORFAMS(indi, fam, num1)
 				/* skip families already processed */
 			if (in_table(ftab, fkey = fam_to_key(fam)))
 				goto a;
-			insert_table_int(ftab, strsave(fkey), 0);
+			insert_table_ptr(ftab, fkey, 0);
 			FORCHILDRENx(fam, child, num2)
 					/* only do people not processed */
 				if (!in_table(itab,
 				    dkey = indi_to_key(child))) {
-						/* key copy for seq & list - owned by seq */
-					dkey = strsave(dkey);
+						/* copy key for list, others make their own copies */
 					uval = creategenval(seq, gen);
 						/* add person to output */
-					append_indiseq_pval(des, strsave(dkey), NULL, uval.w, TRUE, TRUE);
+					append_indiseq_pval(des, strsave(dkey), NULL, uval.w, TRUE);
 						/* also want descendants, so add person to processing list */
-					enqueue_list(deslist, (VPTR)dkey);
+					enqueue_list(deslist, (VPTR)strsave(dkey));
 					enqueue_list(genlist, (VPTR)gen);
-					insert_table_int(itab, dkey, 0);
+					insert_table_ptr(itab, dkey, 0);
 				}
 			ENDCHILDRENx
 		a:;
-		ENDFAMSS
+		ENDFAMS
 	}
 	destroy_table(itab);
 	destroy_table(ftab);
@@ -1527,7 +1617,7 @@ spouse_indiseq (INDISEQ seq)
 	NODE indi;
 	INT num1;
 	if (!seq) return NULL;
-	tab = create_table(DONTFREE);
+	tab = create_table_vptr();
 	sps = create_indiseq_impl(IValtype(seq), IValfnctbl(seq));
 	FORINDISEQ(seq, el, num)
 		indi = key_to_indi(skey(el));
@@ -1538,9 +1628,8 @@ spouse_indiseq (INDISEQ seq)
 					/* key copy for seq & list - owned by seq */
 				spkey = strsave(spkey);
 				u = copyval(seq, sval(el));
-				append_indiseq_impl(sps, spkey, NULL,
-					u, TRUE, TRUE);
-				insert_table_int(tab, spkey, 0);
+				append_indiseq_impl(sps, spkey, NULL, u, TRUE, TRUE);
+				insert_table_ptr(tab, spkey, 0);
 			}
 		ENDSPOUSES
 	ENDINDISEQ
@@ -1560,23 +1649,23 @@ name_to_indiseq (STRING name)
 	list = find_indis_by_name(name);
 	ASSERT(list);
 	if (length_list(list)) {
-		struct tag_list_iter listit;
+		LIST_ITER listit=0;
 		VPTR ptr=0;
 		/* add all entries from list to sequence */
 		seq = create_indiseq_null();
-		begin_list(list, &listit);
-		while (next_list_ptr(&listit, &ptr)) {
+		listit = begin_list(list);
+		while (next_list_ptr(listit, &ptr)) {
 			CNSTRING key = (CNSTRING)ptr;
 			/* skip dupe check, because we do unique_indiseq below */
 			append_indiseq_null(seq, strsave(key), NULL, TRUE, TRUE);
 		}
+		end_list_iter(&listit);
 		/* may have duplicates from multiple soundexes or wildcard searches */
 		unique_indiseq(seq);
 		/* people like to see lists alphabetically by name */
 		namesort_indiseq(seq);
 	}
-	make_list_empty(list);
-	remove_list(list, 0);
+	destroy_list(list);
 	return seq;
 }
 /*===========================================
@@ -1730,6 +1819,7 @@ key_to_indiseq (STRING name, char ctype)
 	rec = id_by_key(name, ctype);
 	if (!rec) return NULL;
 	key = rmvat(nxref(nztop(rec)));
+	release_record(rec); /* done with record */
 	seq = create_indiseq_null();
 	append_indiseq_null(seq, key, NULL, FALSE, FALSE);
 	return seq;
