@@ -36,6 +36,7 @@
 #include "table.h"
 #include "translat.h"
 #include "gedcom.h"
+#include "gedcomi.h"
 #include "version.h"
 #include "lloptions.h"
 #include "codesets.h"
@@ -43,6 +44,9 @@
 #include "zstr.h"
 #include "icvt.h"
 #include "date.h"
+#include "mychar.h"
+#include "charprops.h"
+#include "xlat.h"
 
 /*********************************************
  * global/exported variables
@@ -50,7 +54,6 @@
 
 TABLE tagtable=NULL;		/* table for tag strings */
 TABLE placabbvs=NULL;	/* table for place abbrevs */
-BTREE BTR=NULL;	/* database */
 STRING editstr=NULL; /* edit command to run to edit (has editfile inside of it) */
 STRING editfile=NULL; /* file used for editing, name obtained via mktemp */
 
@@ -60,17 +63,15 @@ STRING editfile=NULL; /* file used for editing, name obtained via mktemp */
 
 extern BOOLEAN writeable;
 extern STRING readpath,readpath_file;
-extern STRING qSdbrecstats;
 extern STRING illegal_char;
-
+extern INT opt_finnish, opt_mychar;
 
 /*********************************************
  * local function prototypes
  *********************************************/
 
-static void add_dbs_to_list(LIST dblist, LIST dbdesclist, STRING dir);
-static STRING getdbdesc(STRING path, STRING userpath);
 static void init_win32_gettext_shim(void);
+static BOOLEAN open_database_impl(LLDATABASE lldb, INT alteration);
 static void update_db_options(void);
 
 /*********************************************
@@ -179,14 +180,14 @@ init_lifelines_global (STRING configfile, STRING * pmsg, void (*notify)(STRING d
 		STRING str = getenv(dirvars[i]);
 		if (!str)
 			str = ".";
-		setoptstr_fallback(dirvars[i], strsave(str));
+		setoptstr_fallback(dirvars[i], str);
 	}
 	/* also check environment variable for editor */
 	{
 		STRING str = getenv("LLEDITOR");
 		if (!str)
 			str = environ_determine_editor(PROGRAM_LIFELINES);
-		setoptstr_fallback("LLEDITOR", strsave(str));
+		setoptstr_fallback("LLEDITOR", str);
 	}
 	/* editor falls back to platform-specific default */
 	e = getoptstr("LLEDITOR", NULL);
@@ -202,6 +203,14 @@ init_lifelines_global (STRING configfile, STRING * pmsg, void (*notify)(STRING d
 	set_usersort(custom_sort);
 	suppress_reload = FALSE;
 	update_useropts(0);
+
+	/* Finnish always uses custom character sets */
+	if (opt_finnish ) {
+		opt_mychar = TRUE;
+		mych_set_table(ISO_Latin1);
+	}
+	
+	
 	return TRUE;
 }
 /*=================================
@@ -289,15 +298,15 @@ BOOLEAN
 init_lifelines_db (void)
 {
 	STRING emsg;
-	TABLE dbopts = create_table(FREEBOTH);
+	TABLE dbopts = create_table_str();
 
-	tagtable = create_table(FREEKEY); /* values are same as keys */
-	placabbvs = create_table(FREEBOTH);
+	tagtable = create_table_str(); /* values are same as keys */
+	placabbvs = create_table_str();
 
 	init_valtab_from_rec("VPLAC", placabbvs, ':', &emsg);
 	init_valtab_from_rec("VUOPT", dbopts, '=', &emsg);
 	set_db_options(dbopts);
-	free_optable(&dbopts);
+	release_table(dbopts);
 	init_caches();
 	init_browse_lists();
 	if (!openxref(readonly))
@@ -330,7 +339,7 @@ get_lifelines_version (INT maxlen)
 void
 close_lifelines (void)
 {
-	close_lldb(); /* make sure database closed */
+	lldb_close(&def_lldb); /* make sure database closed */
 	if (editfile) {
 		unlink(editfile);
 		stdfree(editfile);
@@ -344,34 +353,7 @@ close_lifelines (void)
 	term_date();
 	term_codesets();
 	strfree(&int_codeset);
-}
-/*===================================
- * close_lldb -- Close current database
- *  Safe to call even if not opened
- * fullclose is FALSE when building dblist
- *=================================*/
-void
-close_lldb ()
-{
-	if (tagtable)
-		destroy_table(tagtable);
-	tagtable = 0;
-	/* TODO: reverse the rest of init_lifelines_db -- Perry, 2002.06.05 */
-	if (placabbvs) {
-		destroy_table(placabbvs);
-		placabbvs = NULL;
-	}
-	free_caches();
-	closexref();
-	if (BTR) {
-		closebtree(BTR);
-		BTR=NULL;
-	}
-	if (f_dbnotify)
-		(*f_dbnotify)(readpath, FALSE);
-	transl_free_predefined_xlats(); /* clear any active legacy translation tables */
-	strfree(&readpath_file);
-	strfree(&readpath);
+	xlat_shutdown();
 }
 /*==================================================
  * alterdb -- force open, lock, or unlock a database
@@ -460,37 +442,45 @@ force_open_db_exit:
  * Upon failure, sets bterrno and returns false
  *================================================*/
 static BOOLEAN
-open_database_impl (INT alteration)
+open_database_impl (LLDATABASE lldb, INT alteration)
 {
 	int c;
 	INT writ = !readonly + writeable;
+	BTREE btree = 0;
 
 	/* handle db adjustments (which only affect keyfile) first */
 	if (alteration > 0 && !alterdb(alteration)) return FALSE;
 
 	/* call btree module to do actual open of BTR */
-	if (!(BTR = openbtree(readpath, FALSE, writ, immutable)))
+	if (!(btree = bt_openbtree(readpath, FALSE, writ, immutable)))
 		return FALSE;
+	lldb_set_btree(lldb, btree);
 	/* we have to set the global variable readonly correctly, because
 	it is used widely */
-	readonly = !bwrite(BTR);
-	immutable = bimmut(BTR);
+	readonly = !bwrite(btree);
+	immutable = bimmut(btree);
 	if (readonly && writeable) {
-		int myerr=0;
-		c = bkfile(BTR).k_ostat;
+		c = bkfile(btree).k_ostat;
 		if (c < 0) {
-			myerr = BTERR_WRITER;
+			bterrno = BTERR_WRITER;
 		} else {
 			rdr_count = c-1;
-			myerr = BTERR_READERS;
+			bterrno = BTERR_READERS;
 		}
-		/* close_lldb may have set bterrno itself */
-		close_lldb();
-		bterrno = myerr;
 		return FALSE;
 	}
 	return TRUE;
 }
+/*==================================================
+ * dbnotify_close -- Send notification that default database closed
+ *================================================*/
+void
+dbnotify_close (void)
+{
+
+	if (f_dbnotify)
+		(*f_dbnotify)(readpath, FALSE);
+}
 /*==================================================
  * open_database -- open database
  *  forceopen:    [in] flag to override reader/writer protection
@@ -499,7 +489,8 @@ open_database_impl (INT alteration)
 BOOLEAN
 open_database (INT alteration, STRING dbpath)
 {
-	BOOLEAN rtn;
+	LLDATABASE lldb = lldb_alloc();
+	BOOLEAN rtn = FALSE;
 	char fpath[MAXPATHLEN];
 
 	/* tentatively copy paths into gedlib module versions */
@@ -511,13 +502,14 @@ open_database (INT alteration, STRING dbpath)
 	if (f_dbnotify)
 		(*f_dbnotify)(readpath, TRUE);
 
-	rtn = open_database_impl(alteration);
+	rtn = open_database_impl(lldb, alteration);
 	if (!rtn) {
 		/* open failed so clean up, preserve bterrno */
 		int myerr = bterrno;
-		close_lldb();
+		lldb_close(&lldb);
 		bterrno = myerr;
 	}
+	def_lldb = lldb;
 	return rtn;
 }
 /*==================================================
@@ -527,30 +519,35 @@ open_database (INT alteration, STRING dbpath)
 BOOLEAN
 create_database (STRING dbpath)
 {
+	LLDATABASE lldb = lldb_alloc();
+	BTREE btree = 0;
+
 	/* first test that newdb props are legal */
 	STRING props = getoptstr("NewDbProps", 0);
 	if (props && props[0]) {
-		TABLE dbopts = create_table(FREEBOTH);
+		TABLE dbopts = create_table_str();
 		STRING msg=0;
 		if (!init_valtab_from_string(props, dbopts, '=', &msg)) {
 			bterrno = BTERR_BADPROPS;
 			destroy_table(dbopts);
 			return FALSE;
 		}
-		remove_table(dbopts, FREEBOTH);
+		destroy_table(dbopts);
 	}
 
 	/* tentatively copy paths into gedlib module versions */
 	readpath_file=strsave(lastpathname(dbpath));
 	readpath=strsave(dbpath);
 
-	if (!(BTR = openbtree(dbpath, TRUE, 2, immutable))) {
+	if (!(btree = bt_openbtree(dbpath, TRUE, 2, immutable))) {
 		/* open failed so clean up, preserve bterrno */
 		int myerr = bterrno;
-		close_lldb();
+		lldb_close(&lldb);
 		bterrno = myerr;
 		return FALSE;
 	}
+	def_lldb = lldb;
+	lldb_set_btree(lldb, btree);
 	initxref();
 	if (props)
 		store_record("VUOPT", props, strlen(props));
@@ -659,13 +656,14 @@ update_useropts (VPTR uparm)
 		return;
 	/* deal with db-specific options */
 	/* includes setting int_codeset */
-	if (BTR)
+	if (def_lldb)
 		update_db_options();
 	/* in case user changed any codesets */
 	init_codesets();
 	/* in case user changed locale (need int_codeset already set) */
 	uilocale();
 	/* in case user changed codesets */
+	/* TODO: Isn't this superfluous, as it was called in update_db_options above ? */
 	transl_load_xlats();
 
 	strupdate(&illegal_char, getoptstr("IllegalChar", 0));
@@ -677,7 +675,7 @@ update_useropts (VPTR uparm)
 static void
 update_db_options (void)
 {
-	TABLE opttab = create_table(FREEBOTH);
+	TABLE opttab = create_table_str();
 	STRING str=0;
 	get_db_options(opttab);
 
@@ -694,123 +692,12 @@ update_db_options (void)
 #endif /* ENABLE_NLS */
 		/* need to reload all predefined codeset conversions */
 		transl_load_xlats();
+		if (uu8) {
+			charprops_load_utf8();
+		} else {
+			charprops_load(int_codeset);
+		}
 	}
 
 	destroy_table(opttab);
 }
-/*==================================================
- * get_dblist -- find all dbs on path
- *  path:       [IN]  list of directories to be searched
- *  dblist:     [OUT] list of database paths found
- *  dbdesclist: [OUT] list of descriptions of databases found
- *================================================*/
-INT
-get_dblist (STRING path, LIST * dblist, LIST * dbdesclist)
-{
-	STRING dirs=0;
-	INT ndirs=0;
-	STRING p=0;
-	ASSERT(!(*dblist) && !(*dbdesclist));
-	*dblist = create_list2(LISTDOFREE);
-	*dbdesclist = create_list2(LISTDOFREE);
-	if (!path || !path[0])
-		return 0;
-	dirs = (STRING)stdalloc(strlen(path)+2);
-	/* find directories in dirs & delimit with zeros */
-	ndirs = chop_path(path, dirs);
-	/* now process each directory */
-	for (p=dirs; *p; p+=strlen(p)+1) {
-		add_dbs_to_list(*dblist, *dbdesclist, p);
-	}
-	strfree(&dirs);
-	return llen(*dblist);
-}
-/*==================================================
- * add_dbs_to_list -- Add all dbs in specified dir to list
- *  dblist: [I/O] list of databases found
- *  dir:    [IN]  directory to be searched for more databases
- *================================================*/
-static void
-add_dbs_to_list (LIST dblist, LIST dbdesclist, STRING dir)
-{
-	int n=0;
-	struct dirent **programs=0;
-	char candidate[MAXPATHLEN];
-	char userpath[MAXPATHLEN];
-	STRING dbstr=0;
-	char dirbuf[MAXPATHLEN];
-
-	llstrncpy(dirbuf, dir, sizeof(dirbuf), 0);
-	if (!expand_special_fname_chars(dirbuf, sizeof(dirbuf), uu8))
-		return;
-
-	n = scandir(dirbuf, &programs, 0, 0);
-	if (n < 0) return;
-	while (n--) {
-		concat_path(dirbuf, programs[n]->d_name, uu8, candidate, sizeof(candidate));
-		concat_path(dir, programs[n]->d_name, uu8, userpath, sizeof(userpath));
-		if ((dbstr = getdbdesc(candidate, userpath)) != NULL) {
-			back_list(dblist, strsave(userpath));
-			back_list(dbdesclist, dbstr);
-		}
-		stdfree(programs[n]);
-	}
-	stdfree(programs);
-}
-/*==================================================
- * getdbdesc -- Check a file or directory to see if it
- *  is a lifelines database
- *  returns stdalloc'd string or NULL (if not db)
- *================================================*/
-static STRING
-getdbdesc (STRING path, STRING userpath)
-{
-	BTREE btr;
-	BOOLEAN cflag=FALSE, writ=FALSE, immut=TRUE;
-	char desc[MAXPATHLEN];
-
-	if (f_dbnotify)
-		(*f_dbnotify)(path, TRUE);
-
-	suppress_reload = TRUE; /* do not reload options */
-
-	strcpy(desc, "");
-	btr = openbtree(path, cflag, writ, immut);
-	if (btr) {
-		/* TODO: 'twould be nice to clean up & remove these globals */
-		BTR = btr; /* various code assumes BTR is the btree */
-		readonly = TRUE; /* openxrefs depends on this global */
-		if (init_lifelines_db()) {
-			desc[0]=0;
-			llstrapps(desc, sizeof(desc), uu8, userpath);
-			llstrapps(desc, sizeof(desc), uu8, " <");
-			llstrappf(desc, sizeof(desc), uu8, _(qSdbrecstats)
-				, num_indis(), num_fams(), num_sours(), num_evens(), num_othrs());
-			llstrappf(desc, sizeof(desc), uu8, ">");
-		}
-	}
-	close_lldb();
-	readonly = FALSE;
-	BTR = 0;
-	suppress_reload = FALSE;
-	return desc[0] ? strsave(desc) : 0;
-}
-/*====================================================
- * release_dblist -- free a dblist (caller is done with it)
- *==================================================*/
-void
-release_dblist (LIST dblist)
-{
-	if (dblist) {
-		make_list_empty(dblist);
-		remove_list(dblist, 0);
-	}
-}
-/*====================================================
- * is_db_open -- 
- *==================================================*/
-BOOLEAN
-is_db_open (void)
-{
-	return BTR!=0;
-}
