@@ -22,7 +22,7 @@
    SOFTWARE.
 */
 /*=============================================================
- * keytonod.c -- Convert between keys and node trees
+ * keytonod.c -- Cache for lifelines custom btree database
  * Copyright(c) 1992-94 by T.T. Wetmore IV; all rights reserved
  *   2.3.4 - 24 Jun 93    2.3.5 - 01 Sep 93
  *   3.0.0 - 08 May 94    3.0.2 - 23 Dec 94
@@ -33,6 +33,7 @@
 #include "table.h"
 #include "translat.h"
 #include "gedcom.h"
+#include "gedcomi.h"
 #include "cache.h"
 #include "liflines.h"
 #include "feedback.h"
@@ -45,17 +46,63 @@
 char badkeylist[100] = "";
 int listbadkeys = 0;
 
+
+
+/*===============================
+ * CACHEEL -- Cache element type.
+ *=============================*/
+/* typedef struct tag_cacheel *CACHEEL; */
+struct tag_cacheel {
+	CNSTRING c_magic; /* points to cel_magic */
+	NODE c_node;      /* root node */
+	CACHEEL c_prev;   /* previous el */
+	CACHEEL c_next;   /* next el */
+	STRING c_key;     /* record key */
+	INT c_lock;       /* locked? */
+	RECORD c_record;
+};
+#define cnode(e) ((e)->c_node)
+#define cprev(e) ((e)->c_prev)
+#define cnext(e) ((e)->c_next)
+#define ckey(e)  ((e)->c_key)
+#define cclock(e) ((e)->c_lock)
+#define crecord(e) ((e)->c_record)
+
+/*==============================
+ * CACHE -- Internal cache type.
+ *============================*/
+typedef struct {
+	char c_name[5];
+	TABLE c_data;        /* table of keys */
+	CACHEEL c_firstdir;  /* first direct */
+	CACHEEL c_lastdir;   /* last direct */
+	CACHEEL c_array;     /* big array of cacheels, all alloc'd in a block */
+	CACHEEL c_free;      /* root of free list */
+	INT c_maxdir;        /* max in direct */
+	INT c_sizedir;       /* cur in direct */
+} *CACHE;
+#define cacname(c)     ((c)->c_name)
+#define cacdata(c)     ((c)->c_data)
+#define cacfirstdir(c) ((c)->c_firstdir)
+#define caclastdir(c)  ((c)->c_lastdir)
+#define cacarray(e) ((e)->c_array)
+#define cacfree(e) ((e)->c_free)
+#define cacmaxdir(c)   ((c)->c_maxdir)
+#define cacsizedir(c)  ((c)->c_sizedir)
+
+
 /*********************************************
  * local function prototypes
  *********************************************/
 
-static void add_record_to_direct(CACHE cache, RECORD rec, STRING key);
-static void cache_get_lock_counts(CACHE ca, INT * locks, INT * semilocks);
-static void connect_cel_to_rec(CACHEEL cel, RECORD rec);
-static CACHE create_cache(STRING name, INT dirsize, INT indsize);
+/* static void add_record_to_direct(CACHE cache, RECORD rec, STRING key); */
+static void cache_get_lock_counts(CACHE ca, INT * locks);
+static CACHE create_cache(STRING name, INT dirsize);
 static void delete_cache(CACHE * pcache);
-static void dereference(CACHEEL);
+static void ensure_cel_has_record(CACHEEL cel);
 static ZSTR get_cache_stats(CACHE ca);
+static CACHEEL get_free_cacheel(CACHE cache);
+static void init_cel(CACHEEL cel);
 static CACHEEL key_to_cacheel(CACHE cache, CNSTRING key, STRING tag, INT reportmode);
 static CACHEEL key_to_even_cacheel(CNSTRING key);
 static NODE key_typed_to_node(CACHE cache, CNSTRING key, STRING tag);
@@ -63,23 +110,20 @@ static RECORD key_to_record_impl(CNSTRING key, INT reportmode);
 static RECORD key_typed_to_record(CACHE cache, CNSTRING key, STRING tag);
 static CACHEEL key_to_othr_cacheel(CNSTRING key);
 static CACHEEL key_to_sour_cacheel(CNSTRING key);
-static void prepare_direct_space(CACHE cache);
+static CACHEEL node_to_cache(CACHE, NODE);
+static void put_node_in_cache(CACHE cache, CACHEEL cel, NODE node, STRING key);
+static void remove_cel_from_cache(CACHE cache, CACHEEL cel);
 static NODE qkey_to_node(CACHE cache, CNSTRING key, STRING tag);
 static RECORD qkey_typed_to_record(CACHE cache, CNSTRING key, STRING tag);
-static void record_to_cache(CACHE cache, RECORD rec);
-static void release_all_in_cache(CACHE cache);
+/* static CACHEEL qkey_to_typed_cacheel(STRING key); */
+static void remove_from_cache(CACHE, CNSTRING);
 
 
 INT csz_indi = 200;		/* cache size for indi */
-INT icsz_indi = 3000;		/* indirect cache size for indi */
 INT csz_fam = 200;		/* cache size for fam */
-INT icsz_fam = 2000;		/* indirect cache size for fam */
 INT csz_sour = 200;		/* cache size for sour */
-INT icsz_sour = 2000;		/* indirect cache size for sour */
 INT csz_even = 200;		/* cache size for even */
-INT icsz_even = 2000;		/* indirect cache size for even */
 INT csz_othr = 200;		/* cache size for othr */
-INT icsz_othr = 2000;		/* indirect cache size for othr */
 
 /*********************************************
  * local variables
@@ -87,6 +131,11 @@ INT icsz_othr = 2000;		/* indirect cache size for othr */
 
 static CACHE indicache, famcache, evencache, sourcache, othrcache;
 
+static CNSTRING cel_magic = "CEL_MAGIC"; /* fixed pointer to identify cel */
+
+/* keybuf circular list of last 10 keys we looked up in cache 
+ * kept for printing debug messages in crash log
+ */
 static char keybuf[10][32] = { "", "", "", "", "", "", "", "", "", ""};
 static int keyidx = 0;
 
@@ -250,7 +299,10 @@ keynum_to_record (char ntype, int keynum)
 NODE
 key_to_type (CNSTRING key, INT reportmode)
 {
-	return nztop(key_to_record_impl(key, reportmode));
+	RECORD rec = key_to_record_impl(key, reportmode);
+	NODE node = nztop(rec);
+	release_record(rec); /* avoiding leaking reference */
+	return node;
 }
 /*=====================================
  * qkey_to_type -- Convert key to node
@@ -263,6 +315,7 @@ qkey_to_type (CNSTRING key)
 }
 /*=====================================
  * key_to_record_impl -- Convert key (any type) to RECORD
+ *  returns addref'd record
  *===================================*/
 static RECORD
 key_to_record_impl (CNSTRING key, INT reportmode)
@@ -279,6 +332,7 @@ key_to_record_impl (CNSTRING key, INT reportmode)
 /*=====================================
  * key_to_record -- Convert key (any type) to RECORD
  * ASSERTS if record not found in database
+ *  returns addref'd record
  *===================================*/
 RECORD
 key_to_record (CNSTRING key)
@@ -288,6 +342,7 @@ key_to_record (CNSTRING key)
 /*=====================================
  * qkey_to_record -- Convert key (any type) to RECORD
  * quiet -- that is, returns NULL if record not in database
+ *  returns addref'd record
  *===================================*/
 RECORD
 qkey_to_record (CNSTRING key)
@@ -325,6 +380,7 @@ NODE key_to_othr (CNSTRING key)
  * key_to_???0 -- Convert key to person
  *  (asserts if failure)
  *  5 symmetric versions
+ * returns addref'd record
  *===================================*/
 RECORD key_to_irecord (CNSTRING key)
 {
@@ -376,6 +432,7 @@ NODE qkey_to_othr (CNSTRING key)
  * qkey_to_???0 -- Convert key to node type
  *  report mode (returns NULL if failure)
  *  5 symmetric versions
+ * All return addref'd records
  *======================================*/
 RECORD qkey_to_irecord (CNSTRING key)
 {
@@ -457,11 +514,11 @@ key_to_othr_cacheel (CNSTRING key)
 void
 init_caches (void)
 {
-	indicache = create_cache("INDI", csz_indi, icsz_indi);
-	famcache  = create_cache("FAM", csz_fam, icsz_fam);
-	evencache = create_cache("EVEN", (INT)csz_even, (INT)csz_even);
-	sourcache = create_cache("SOUR", (INT)csz_sour, (INT)icsz_sour);
-	othrcache = create_cache("OTHR", (INT)csz_othr, (INT)icsz_othr);
+	indicache = create_cache("INDI", csz_indi);
+	famcache  = create_cache("FAM", csz_fam);
+	evencache = create_cache("EVEN", csz_even);
+	sourcache = create_cache("SOUR", csz_sour);
+	othrcache = create_cache("OTHR", csz_othr);
 }
 /*======================================
  * free_caches -- Release cache memory
@@ -480,19 +537,36 @@ free_caches (void)
  * create_cache -- Create cache
  *===========================*/
 static CACHE
-create_cache (STRING name, INT dirsize, INT indsize)
+create_cache (STRING name, INT dirsize)
 {
 	CACHE cache;
+	INT i;
 	if (dirsize < 1) dirsize = 1;
-	if (indsize < 1) indsize = 1;
 	cache = (CACHE) stdalloc(sizeof(*cache));
-	llstrncpy(cname(cache), name, sizeof(cname(cache)), uu8);
-	cdata(cache) = create_table(FREEKEY);
-	cfirstdir(cache) = clastdir(cache) = NULL;
-	cfirstind(cache) = clastind(cache) = NULL;
-	csizedir(cache) = csizeind(cache) = 0;
-	cmaxdir(cache) = dirsize;
-	cmaxind(cache) = indsize;
+	memset(cache, 0, sizeof(*cache));
+	llstrncpy(cacname(cache), name, sizeof(cacname(cache)), uu8);
+	/* 
+	It would be nice to set the table hash size larger for large 
+	caches, but right now (2003-10-08), tables do not expose a 
+	method to set their hash size.
+	*/
+	cacdata(cache) = create_table_vptr(); /* pointers to cache elements, owned by cacarray */
+	cacfirstdir(cache) = caclastdir(cache) = NULL;
+	cacsizedir(cache) = 0;
+	cacmaxdir(cache) = dirsize;
+	/* Allocate all the cache elements in a big block */
+	cacarray(cache) = (CACHEEL) stdalloc(cacmaxdir(cache) * sizeof(cacarray(cache)[0]));
+	/* Link all the elements together on the free list */
+	for (i=0; i<cacmaxdir(cache); ++i) {
+		CACHEEL cel = &cacarray(cache)[i];
+		CACHEEL celnext = cacfree(cache);
+		init_cel(cel);
+		if (celnext) {
+			cnext(cel) = celnext;
+			cprev(celnext) = cel;
+		}
+		cacfree(cache) = cel;
+	}
 	return cache;
 }
 /*=============================
@@ -501,13 +575,31 @@ create_cache (STRING name, INT dirsize, INT indsize)
 static void
 delete_cache (CACHE * pcache)
 {
+	INT num=0;
 	CACHE cache = *pcache;
+	CACHEEL frst=0;
 	if (!cache) return;
-	release_all_in_cache(cache);
-	destroy_table(cdata(cache));
+	/* Loop through all cache elements, freeing each */
+	while ((frst = cacfirstdir(cache)) != 0) {
+		remove_cel_from_cache(cache, frst);
+	}
+	/* TODO: Need to delete cache elements on free list */
+	num = get_table_count(cacdata(cache));
+	ASSERT(num == 0);
+	destroy_table(cacdata(cache));
+	stdfree(cacarray(cache));
 	stdfree(cache);
 	*pcache = 0;
 }
+/*=============================
+ * init_cel -- Initialize new cacheel
+ *===========================*/
+static void
+init_cel (CACHEEL cel)
+{
+	memset(cel, 0, sizeof(*cel));
+	cel->c_magic = cel_magic;
+}
 /*=================================================
  * remove_direct -- Unlink CACHEEL from direct list
  *===============================================*/
@@ -519,24 +611,9 @@ remove_direct (CACHE cache, CACHEEL cel)
 	ASSERT(cache && cel);
 	if (prev) cnext(prev) = next;
 	if (next) cprev(next) = prev;
-	if (!prev) cfirstdir(cache) = next;
-	if (!next) clastdir(cache) = prev;
-	csizedir(cache)--;
-}
-/*=====================================================
- * remove_indirect -- Unlink CACHEEL from indirect list
- *===================================================*/
-static void
-remove_indirect (CACHE cache, CACHEEL cel)
-{
-	CACHEEL prev = cprev(cel);
-	CACHEEL next = cnext(cel);
-	ASSERT(cache && cel);
-	if (prev) cnext(prev) = next;
-	if (next) cprev(next) = prev;
-	if (!prev) cfirstind(cache) = next;
-	if (!next) clastind(cache) = prev;
-	csizeind(cache)--;
+	if (!prev) cacfirstdir(cache) = next;
+	if (!next) caclastdir(cache) = prev;
+	cacsizedir(cache)--;
 }
 /*===========================================================
  * first_direct -- Make unlinked CACHEEL first in direct list
@@ -544,127 +621,26 @@ remove_indirect (CACHE cache, CACHEEL cel)
 static void
 first_direct (CACHE cache, CACHEEL cel)
 {
-	CACHEEL frst = cfirstdir(cache);
+	CACHEEL frst = cacfirstdir(cache);
 	ASSERT(cache && cel);
-	csizedir(cache)++;
+	cacsizedir(cache)++;
 	cprev(cel) = NULL;
 	cnext(cel) = frst;
 	if (frst) cprev(frst) = cel;
-	if (!frst) clastdir(cache) = cel;
-	cfirstdir(cache) = cel;
-}
-/*===============================================================
- * first_indirect -- Make unlinked CACHEEL first in indirect list
- *  Does not check for overflow (caller's responsibility)
- *=============================================================*/
-static void
-first_indirect (CACHE cache, CACHEEL cel)
-{
-	CACHEEL frst = cfirstind(cache);
-	ASSERT(cache && cel);
-	csizeind(cache)++;
-	cprev(cel) = NULL;
-	cnext(cel) = frst;
-	if (frst) cprev(frst) = cel;
-	if (!frst) clastind(cache) = cel;
-	cfirstind(cache) = cel;
-}
-/*=======================================================
- * remove_last -- Remove last indirect element from cache
- *=====================================================*/
-static void
-remove_last (CACHE cache)
-{
-	CACHEEL cel=0;
-	STRING key;
-	for (cel = clastind(cache); cel && csemilock(cel); cel = cprev(cel))
-		;
-	if (!cel) {
-		llwprintf("Indirect cache overflow! (cache=%s, size=%d)\n", cname(cache), cmaxind(cache));
-		ASSERT(cel);
-	}
-	ASSERT(!cclock(cel)); /* locked elements should never leave direct */
-	remove_indirect(cache, cel);
-	key = ckey(cel);
-	stdfree(cel);
-	delete_table(cdata(cache), key);
-	stdfree(key);
+	if (!frst) caclastdir(cache) = cel;
+	cacfirstdir(cache) = cel;
 }
 /*============================================================
  * direct_to_first -- Make direct CACHEEL first in direct list
  *==========================================================*/
 static void
-direct_to_first (CACHE cache,
-                 CACHEEL cel)
+direct_to_first (CACHE cache, CACHEEL cel)
 {
 	ASSERT(cache && cel);
-	if (cel == cfirstdir(cache)) return;
+	if (cel == cacfirstdir(cache)) return;
 	remove_direct(cache, cel);
 	first_direct(cache, cel);
 }
-/*==================================================================
- * indirect_to_first -- Make indirect CACHEEL first in direct list
- *================================================================*/
-static void
-indirect_to_first (CACHE cache,
-                   CACHEEL cel)
-{
-	ASSERT(cache && cel);
-	semilock_cache(cel);
-	prepare_direct_space(cache);
-	remove_indirect(cache, cel);
-	dereference(cel);
-	first_direct(cache, cel);
-	unsemilock_cache(cel);
-}
-/*==============================================================
- * direct_to_indirect -- Make last direct CACHEEL first indirect
- *============================================================*/
-static void
-direct_to_indirect (CACHE cache)
-{
-	CACHEEL cel = clastdir(cache);
-	for (cel = clastdir(cache); cel && cclock(cel); cel = cprev(cel))
-		;
-	if (!cel) {
-		llwprintf("Cache overflow! (Cache=%s, size=%d)\n", cname(cache), cmaxdir(cache));
-		ASSERT(cel);
-	}
-	remove_direct(cache, cel);
-	free_rec(crecord(cel)); /* this frees the nodes */
-	connect_cel_to_rec(cel, NULL);
-	first_indirect(cache, cel);
-}
-/*=====================================================
- * connect_cel_to_rec -- Hook record to cache element holder
- * (rec may be NULL)
- *===================================================*/
-static void
-connect_cel_to_rec (CACHEEL cel, RECORD rec)
-{
-	if (cel) {
-		crecord(cel) = rec;
-		cnode(cel) = nztop(rec);
-	}
-	if (rec) {
-		rec->cel = cel;
-	}
-}
-/*=====================================================
- * dereference -- Dereference cel by reading its record
- *===================================================*/
-static void
-dereference (CACHEEL cel)
-{
-	STRING rawrec;
-	INT len;
-	RECORD rec;
-	ASSERT(cel);
-	ASSERT(rawrec = retrieve_raw_record(ckey(cel), &len));
-	ASSERT(rec = string_to_record(rawrec, ckey(cel), len));
-	connect_cel_to_rec(cel, rec);
-	stdfree(rawrec);
-}
 /*========================================================
  * add_to_direct -- Add new CACHEEL to direct part of cache
  * reportmode: if True, then return NULL rather than aborting
@@ -673,24 +649,21 @@ dereference (CACHEEL cel)
  *  cache:      [IN]  which cache to which to add
  *  key:        [IN]  key of record to be added
  *  reportmode: [IN] if non-zero, failures should be silent
+ * Only called if record is not already in cache (caller's responsibility)
  *======================================================*/
 static CACHEEL
 add_to_direct (CACHE cache, CNSTRING key, INT reportmode)
 {
-	STRING rawrec;
-	INT len;
-	CACHEEL cel;
-	RECORD rec;
+	STRING rawrec=0;
+	INT len=0;
+	CACHEEL cel=0;
+	RECORD rec=0;
 	int i, j;
-	STRING keycopy;
 
-#ifdef DEBUG
-	llwprintf("add_to_direct: key == %s\n", key);
-#endif
 	ASSERT(cache && key);
-	prepare_direct_space(cache);
 	rec = NULL;
 	if ((rawrec = retrieve_raw_record(key, &len))) 
+		/* 2003-11-22, we should use string_to_node here */
 		rec = string_to_record(rawrec, key, len);
 	if (!rec)
 	{
@@ -716,16 +689,15 @@ add_to_direct (CACHE cache, CNSTRING key, INT reportmode)
 		/* deliberately fall through to let ASSERT(rec) fail */
 	}
 	ASSERT(rec);
-	ASSERT(csizedir(cache) < cmaxdir(cache));
-	cel = (CACHEEL) stdalloc(sizeof(*cel));
-	keycopy = strsave(key);
-	insert_table_ptr(cdata(cache), keycopy, cel);
-	connect_cel_to_rec(cel, rec);
-	ckey(cel) = keycopy;
-	cclock(cel) = 0;
-	csemilock(cel) = 0;
-	first_direct(cache, cel);
+	/* record was just loaded, nztop should not need to load it */
+	cel = node_to_cache(cache, nztop(rec));
+	ASSERT(!crecord(cel));
+	/* node_to_cache did a first_direct call, so record in cache */
+	record_set_cel(rec, cel);
+	/* our new rec above has one reference, which is held by cel */
+	crecord(cel) = rec;
 	stdfree(rawrec);
+	ASSERT(cel->c_magic == cel_magic);
 	return cel;
 }
 /*======================================================
@@ -740,55 +712,25 @@ key_to_cacheel (CACHE cache, CNSTRING key, STRING tag, INT reportmode)
 	keybuf[keyidx][31] = '\0';
 	keyidx++;
 	if(keyidx >= 10) keyidx = 0;
-	if ((cel = (CACHEEL) valueof_ptr(cdata(cache), key))) {
-		if (cnode(cel))
-			direct_to_first(cache, cel);
-		else
-			indirect_to_first(cache, cel);
+	if ((cel = (CACHEEL) valueof_ptr(cacdata(cache), key))) {
+		ASSERT(cnode(cel));
+		ASSERT(cel->c_magic == cel_magic);
+		direct_to_first(cache, cel);
 		if (tag) {
-#ifdef DEBUG
-			llwprintf("BEFORE ASSERT: tag, ntag(cnode(cel)) = %s, %s\n", tag, ntag(cnode(cel)));
-#endif
 			ASSERT(eqstr(tag, ntag(cnode(cel))));
+			ASSERT(crecord(cel));
+			ASSERT(eqstr(key, nzkey(crecord(cel))));
 		}
 		return cel;
 	}
 	cel = add_to_direct(cache, key, reportmode);
 	if (cel && tag) {
 		ASSERT(eqstr(tag, ntag(cnode(cel))));
+		ASSERT(crecord(cel));
+		ASSERT(eqstr(key, nzkey(crecord(cel))));
 	}
 	return cel;
 }
-/*===============================================================
- * release_all_in_cache -- Release all entries in a cache
- * Created: 2003-02-02 (Perry Rapp)
- *=============================================================*/
-static void
-release_all_in_cache (CACHE cache)
-{
-	while (csizedir(cache)) {
-		CACHEEL cel = cfirstdir(cache);
-		remove_direct(cache, cel);
-		free_rec(crecord(cel)); /* this frees the nodes */
-	}
-	while (csizeind(cache)) {
-		CACHEEL cel = cfirstind(cache);
-		remove_indirect(cache, cel);
-	}
-}
-/*===============================================================
- * prepare_direct_space -- Make space in direct
- *  Moves a direct entry to indirect if necessary
- *=============================================================*/
-static void
-prepare_direct_space (CACHE cache)
-{
-	if (csizedir(cache) >= cmaxdir(cache)) {
-		if (csizeind(cache) >= cmaxind(cache))
-			remove_last(cache);
-		direct_to_indirect(cache);
-	}
-}
 /*===============================================================
  * key_to_node -- Return tree from key; add to cache if not there
  * asserts if failure
@@ -806,6 +748,7 @@ key_typed_to_node (CACHE cache, CNSTRING key, STRING tag)
 /*===============================================================
  * key_typed_to_record -- Return record from key; add to cache if not there
  * asserts if failure
+ * returns addref'd record
  *=============================================================*/
 static RECORD
 key_typed_to_record (CACHE cache, CNSTRING key, STRING tag)
@@ -814,7 +757,39 @@ key_typed_to_record (CACHE cache, CNSTRING key, STRING tag)
 	ASSERT(cache && key);
 	if (!(cel = key_to_cacheel(cache, key, tag, FALSE)))
 		return NULL;
-	return crecord(cel);
+	return get_record_for_cel(cel);
+}
+/*===================================
+ * get_record_for_cel -- get or create new record from cacheel
+ * returns addref'd record
+ *=================================*/
+RECORD
+get_record_for_cel (CACHEEL cel)
+{
+	RECORD rec=0;
+	NODE node=0;
+	STRING key=0;
+
+	ASSERT(cel);
+	if (crecord(cel)) {
+		rec = crecord(cel);
+		addref_record(rec);
+		return rec;
+	}
+	ASSERT(cnode(cel));
+	node = cnode(cel);
+	ASSERT(nxref(node));
+
+	key = ckey(cel);
+
+	rec = make_new_record_with_info(key, cel);
+	record_set_cel(rec, cel);
+	crecord(cel) = rec;
+	key = node_to_key(node);
+
+	set_record_key_info(rec, key);
+	addref_record(rec);
+	return rec;
 }
 /*===============================================================
  * qkey_to_node -- Return tree from key; add to cache if not there
@@ -825,7 +800,8 @@ static NODE
 qkey_to_node (CACHE cache, CNSTRING key, STRING tag)
 {
 	CACHEEL cel;
-	ASSERT(cache && key);
+	ASSERT(cache);
+	if (!key) return NULL;
 	if (!(cel = key_to_cacheel(cache, key, tag, TRUE)))
 		return NULL;
 	return cnode(cel);
@@ -833,39 +809,19 @@ qkey_to_node (CACHE cache, CNSTRING key, STRING tag)
 /*===============================================================
  * qkey_typed_to_record -- Return record from key; add to cache if not there
  * report mode - returns NULL if failure
+ * returns addref'd record
  *=============================================================*/
 static RECORD
 qkey_typed_to_record (CACHE cache, CNSTRING key, STRING tag)
 {
-	CACHEEL cel;
+	CACHEEL cel=0;
+	RECORD rec=0;
+
 	ASSERT(cache && key);
 	if (!(cel = key_to_cacheel(cache, key, tag, TRUE)))
 		return NULL;
-	return crecord(cel);
-}
-/*======================================
- * load_cacheel -- Load CACHEEL into direct cache
- *  if needed & valid
- * (ie, this handles null input, or input already in cache)
- *====================================*/
-void
-load_cacheel (CACHEEL cel)
-{
-	CACHE cache = NULL;
-	if (!cel || cnode(cel)) return;
-	switch (ckey(cel)[0]) {
-	case 'I': cache = indicache; break;
-	case 'F': cache = famcache; break;
-	case 'S': cache = sourcache; break;
-	case 'E': cache = evencache; break;
-	case 'X': cache = othrcache; break;
-	default: {
-		crashlog("Bad cacheel key: <%s>", ckey(cel) ? ckey(cel) : "<0>");
-		ASSERT(0);
-		break;
-		}
-	}
-	indirect_to_first(cache, cel);
+	rec = get_record_for_cel(cel);
+	return rec;
 }
 /*======================================
  * lock_cache -- Lock CACHEEL into direct cache
@@ -877,6 +833,20 @@ lock_cache (CACHEEL cel)
 	cclock(cel)++;
 	ASSERT(cclock(cel)>0);
 }
+/*======================================
+ * lock_record_in_cache -- Load record into cache & lock it
+ *====================================*/
+void
+lock_record_in_cache (RECORD rec)
+{
+	NODE node=0;
+	CACHEEL cel=0;
+	ASSERT(rec);
+	node = nztop(rec); /* force record to be loaded in cache */
+	cel = nzcel(rec);
+	++cclock(cel);
+	ASSERT(cclock(cel) > 0);
+}
 /*==========================================
  * unlock_cache -- Unlock CACHEEL from direct cache
  *========================================*/
@@ -888,33 +858,27 @@ unlock_cache (CACHEEL cel)
 	cclock(cel)--;
 }
 /*======================================
- * semilock_cache -- Lock CACHEEL into cache (indirect is ok)
+ * unlock_record_from_cache -- Remove lock on record
  *====================================*/
 void
-semilock_cache (CACHEEL cel)
+unlock_record_from_cache (RECORD rec)
 {
-	csemilock(cel)++;
-	ASSERT(csemilock(cel)>0);
-}
-/*==========================================
- * unsemilock_cache -- Unlock CACHEEL from cache (indirect is ok)
- *========================================*/
-void
-unsemilock_cache (CACHEEL cel)
-{
-	ASSERT(csemilock(cel)>0);
-	csemilock(cel)--;
+	CACHEEL cel=0;
+	ASSERT(rec);
+	cel = nzcel(rec);
+	ASSERT(cel);
+	ASSERT(cclock(cel) > 0);
+	--cclock(cel);
 }
 /*=========================================
  * cache_get_lock_counts -- Fill in lock counts
  *=======================================*/
 static void
-cache_get_lock_counts (CACHE ca, INT * locks, INT * semilocks)
+cache_get_lock_counts (CACHE ca, INT * locks)
 {
 	CACHEEL cel;
-	for (cel = cfirstdir(ca); cel; cel = cnext(cel)) {
+	for (cel = cacfirstdir(ca); cel; cel = cnext(cel)) {
 		if (cclock(cel) && locks) ++(*locks);
-		if (csemilock(cel) && semilocks) ++(*semilocks);
 	}
 }
 /*=========================================
@@ -925,12 +889,11 @@ static ZSTR
 get_cache_stats (CACHE ca)
 {
 	ZSTR zstr = zs_new();
-	INT lo=0, slo=0;
-	cache_get_lock_counts(ca, &lo, &slo);
+	INT lo=0;
+	cache_get_lock_counts(ca, &lo);
 	zs_appf(zstr
-		, "d:%d/%d (l:%d),i:%d/%d (l:%d)"
-		, csizedir(ca), cmaxdir(ca), lo
-		, csizeind(ca), cmaxind(ca), slo
+		, "d:%d/%d (l:%d)"
+		, cacsizedir(ca), cacmaxdir(ca), lo
 		);
 	return zstr;
 }
@@ -951,21 +914,28 @@ get_cache_stats_fam (void)
 	return get_cache_stats(famcache);
 }
 /*============================================
- * indi_to_cache -- Add person to person cache
+ * ensure_cel_has_record -- Make sure cache element has record
+ *  (node_to_cache, which creates cels, doesn't create records)
  *==========================================*/
-void
-indi_to_cache (RECORD rec)
+static void
+ensure_cel_has_record (CACHEEL cel)
 {
-	record_to_cache(indicache, rec);
+	RECORD rec = get_record_for_cel(cel); /* addref'd */
+	release_record(rec);
 }
 /*============================================
- * indi_to_cache_old -- Add person to person cache
- *  should be obsoleted by indi_to_cache
+ * add_new_indi_to_cache -- Add person to person cache
  *==========================================*/
 void
-indi_to_cache_old (NODE node)
+add_new_indi_to_cache (RECORD rec)
 {
-	node_to_cache(indicache, node);
+	CACHEEL cel=0;
+	NODE root = is_record_temp(rec);
+	cel = node_to_cache(indicache, root);
+	ASSERT(!crecord(cel));
+	record_set_cel(rec, cel);
+	crecord(cel) = rec;
+	addref_record(rec); /* cel holds reference */
 }
 /*===========================================
  * fam_to_cache -- Add family to family cache
@@ -973,7 +943,8 @@ indi_to_cache_old (NODE node)
 void
 fam_to_cache (NODE node)
 {
-	node_to_cache(famcache, node);
+	CACHEEL cel = node_to_cache(famcache, node);
+	ensure_cel_has_record(cel);
 }
 /*==========================================
  * even_to_cache -- Add event to event cache
@@ -981,7 +952,8 @@ fam_to_cache (NODE node)
 void
 even_to_cache (NODE node)
 {
-	node_to_cache(evencache, node);
+	CACHEEL cel = node_to_cache(evencache, node);
+	ensure_cel_has_record(cel);
 }
 /*============================================
  * sour_to_cache -- Add source to source cache
@@ -989,7 +961,8 @@ even_to_cache (NODE node)
 void
 sour_to_cache (NODE node)
 {
-	node_to_cache(sourcache, node);
+	CACHEEL cel = node_to_cache(sourcache, node);
+	ensure_cel_has_record(cel);
 }
 /*===========================================
  * othr_to_cache -- Add other record to cache
@@ -997,61 +970,133 @@ sour_to_cache (NODE node)
 void
 othr_to_cache (NODE node)
 {
-	node_to_cache(othrcache, node);
+	CACHEEL cel = node_to_cache(othrcache, node);
+	ensure_cel_has_record(cel);
 }
 /*========================================
  * node_to_cache -- Add node tree to cache
+ *  This is a high-level entry point, which validates
+ *  and delegates the work
+ *  node tree must be valid, and of the correct type
+ *  (INDI node trees may only be added to INDI cache, etc)
+ *  This puts node into cache (first_direct)
  *======================================*/
-void
-node_to_cache (CACHE cache, NODE node)
+static CACHEEL
+node_to_cache (CACHE cache, NODE top)
 {
-	RECORD rec = create_record(node);
-	record_to_cache(cache, rec);
-}
-/*========================================
- * record_to_cache -- Add record to cache
- *  record expected to be valid
- *  INDI records may only be added to INDI cache, etc
- *======================================*/
-static void
-record_to_cache (CACHE cache, RECORD rec)
-{
-	STRING key;
-	NODE top;
-	ASSERT(cache && rec);
-	top = nztop(rec);
+	STRING key=0;
+	CACHEEL cel=0;
+	ASSERT(cache);
 	ASSERT(top);
-	if (nestr(cname(cache), "OTHR")) {
+	ASSERT(!nparent(top) && !nsibling(top)); /* should be a root */
+	if (nestr(cacname(cache), "OTHR")) {
 		/* only INDI records in INDI cache, etc */
-		if (!eqstr(cname(cache), ntag(top))) {
-			crashlog(_("Bad cache entry <%s> != <%s>"), cname(cache), ntag(top));
+		if (!eqstr(cacname(cache), ntag(top))) {
+			crashlog(_("Bad cache entry <%s> != <%s>"), cacname(cache), ntag(top));
 			ASSERT(0);
 		}
 	}
 	key = node_to_key(top);
 	ASSERT(key);
-	ASSERT(!valueof_ptr(cdata(cache), key));
-	prepare_direct_space(cache);
-	add_record_to_direct(cache, rec, key);
+	/* ASSERT that record is not in cache */
+	/* We're not supposed to be called if record in cache */
+	ASSERT(!valueof_ptr(cacdata(cache), key));
+	cel = get_free_cacheel(cache);
+	put_node_in_cache(cache, cel, top, key);
+	return cel;
 }
 /*=======================================================
- * add_record_to_direct -- Add node to direct part of cache
+ * get_free_cacheel -- Remove and return entry from free list
+ *=====================================================*/
+static CACHEEL
+get_free_cacheel (CACHE cache)
+{
+	CACHEEL cel=0, celnext=0;
+
+	/* If free list is empty, move least recently used entry to free list */
+	if (!cacfree(cache)) {
+		/* find least recently used by unlocked entry */
+		for (cel = caclastdir(cache); cel && cclock(cel); cel = cprev(cel)) {
+		}
+		if (!cel) {
+			llwprintf("Cache overflow! (Cache=%s, size=%d)\n", cacname(cache), cacmaxdir(cache));
+			ASSERT(0);
+		}
+		remove_from_cache(cache, ckey(cel));
+	}
+
+	cel = cacfree(cache);
+	ASSERT(cel);
+
+	/* remove entry from free list */
+	celnext = cnext(cel);
+	cacfree(cache) = celnext;
+	if (celnext)
+		cprev(celnext) = 0;
+	/* reinitialize entry */
+	init_cel(cel);
+
+	return cel;
+}
+/*=======================================================
+ * set_all_nodetree_to_cel -- clear all the cel pointers in a node tree
  *=====================================================*/
 static void
-add_record_to_direct (CACHE cache, RECORD rec, STRING key)
+set_all_nodetree_to_cel (NODE node, CACHEEL cel)
 {
-	CACHEEL cel;
-	STRING keynew;
-	NODE node = nztop(rec);
+	BOOLEAN travdone = FALSE;
+	/* Now set all nodes in tree to point to cache record */
+	while (!travdone) {
+		node->n_cel = cel;
+		/* go to bottom of tree */
+		while (nchild(node)) {
+			node = nchild(node);
+			node->n_cel = cel;
+		}
+		/* find next node in traversal/ascent */
+		while (!nsibling(node)) {
+			if (!nparent(node)) {
+				travdone=TRUE;
+				break;
+			}
+			node = nparent(node);
+		}
+		node = nsibling(node);
+	}
+}
+/*=======================================================
+ * put_node_in_cache -- Low-level work of loading node into cacheel supplied
+ *=====================================================*/
+static void
+put_node_in_cache (CACHE cache, CACHEEL cel, NODE node, STRING key)
+{
+	BOOLEAN travdone = FALSE;
 	ASSERT(cache && node);
-	ASSERT(csizedir(cache) < cmaxdir(cache));
-	cel = (CACHEEL) stdalloc(sizeof(*cel));
-	insert_table_ptr(cdata(cache), keynew=strsave(key), cel);
-	crecord(cel) = rec;
+	ASSERT(cacsizedir(cache) < cacmaxdir(cache));
+	init_cel(cel);
+	insert_table_ptr(cacdata(cache), key, cel);
 	cnode(cel) = node;
-	ckey(cel) = keynew;
+	ckey(cel) = strsave(key);
 	cclock(cel) = FALSE;
 	first_direct(cache, cel);
+	/* Now set all nodes in tree to point to cache record */
+	while (!travdone) {
+		node->n_cel = cel;
+		/* go to bottom of tree */
+		while (nchild(node)) {
+			node = nchild(node);
+			node->n_cel = cel;
+		}
+		/* find next node in traversal/ascent */
+		while (!nsibling(node)) {
+			if (!nparent(node)) {
+				travdone=TRUE;
+				break;
+			}
+			node = nparent(node);
+		}
+		node = nsibling(node);
+	}
 }
 /*==============================================
  * remove_indi_cache -- Remove person from cache
@@ -1070,23 +1115,76 @@ remove_fam_cache (STRING key)
 	remove_from_cache(famcache, key);
 }
 /*=============================================
- * remove_from_cache -- Remove entry from cache
+ * remove_from_cache_by_key -- Remove record from cache
  *===========================================*/
 void
-remove_from_cache (CACHE cache, STRING key)
+remove_from_cache_by_key (CNSTRING key)
 {
-	CACHEEL cel;
+	switch(key[0]) {
+	case 'I': remove_from_cache(indicache, key); break;
+	case 'F': remove_from_cache(famcache, key); break;
+	case 'S': remove_from_cache(sourcache, key); break;
+	case 'E': remove_from_cache(evencache, key); break;
+	case 'X': remove_from_cache(othrcache, key); break;
+	default: ASSERT(0); break;
+	}
+}
+/*=============================================
+ * remove_from_cache -- Move cache entry to free list
+ *===========================================*/
+static void
+remove_from_cache (CACHE cache, CNSTRING key)
+{
+	CACHEEL cel=0;
+
 	if (!key || *key == 0 || !cache)
 		return;
-	if (!(cel = (CACHEEL) valueof_ptr(cdata(cache), key)))
-		return;
-	ASSERT(!cclock(cel) && !csemilock(cel));
-	if (cnode(cel))
-		remove_direct(cache, cel);
-	else
-		remove_indirect(cache, cel);
-	stdfree(cel);
-	delete_table(cdata(cache), key);
+	/* If it has a key, it is in the cache */
+	cel = valueof_ptr(cacdata(cache), key);
+	remove_cel_from_cache(cache, cel);
+}
+/*=============================================
+ * remove_from_cache -- Move cache entry to free list
+ *  Requires non-null input
+ *===========================================*/
+static void
+remove_cel_from_cache (CACHE cache, CACHEEL cel)
+{
+	CACHEEL celnext=0;
+	STRING key = ckey(cel);
+
+	/* caller ensured cache && key are non-null */
+	ASSERT(cel);
+
+	ASSERT(!cclock(cel)); /* not supposed to remove locked elements */
+	ASSERT(cnode(cel));
+	remove_direct(cache, cel);
+
+	/* Clear all node tree info */
+	if (1) {
+		NODE node = cnode(cel);
+		if (node)
+			set_all_nodetree_to_cel(node, 0);
+		cnode(cel) = 0;
+		free_nodes(node);
+	}
+
+	celnext = cacfree(cache);
+	cnext(cel) = celnext;
+	if (celnext)
+		cprev(celnext) = cel;
+	cprev(cel) = 0;
+	ckey(cel) = 0;
+	if (crecord(cel)) {
+		/* cel holds the original reference to the record */
+		RECORD rec = crecord(cel);
+		record_remove_cel(rec, cel);
+		release_record(rec);
+		crecord(cel) = 0;
+	}
+	cacfree(cache) = cel;
+	delete_table_element(cacdata(cache), key);
+	stdfree(key); /* alloc'd when assigned to ckey(cel) */
 }
 /*================================================================
  * value_to_xref -- Converts a string to a record key, if possible
@@ -1110,9 +1208,10 @@ value_to_xref (STRING val)
 CACHEEL
 indi_to_cacheel (RECORD indi)
 {
-	CACHEEL cel;
+	CACHEEL cel=0;
 	if (!indi || !nztop(indi)) return NULL;
-	if (indi->cel) return indi->cel;
+	cel = nzcel(indi);
+	if (cel) return cel;
 	/*
 	This is not efficient, rereading the record
 	But we can't just steal the record given us
@@ -1271,11 +1370,72 @@ CACHEEL qkey_to_othr_cacheel (STRING key)
 	return key_to_cacheel(othrcache, key, NULL, TRUE);
 }
 /*==============================================
- * nztop -- Return first NODE of a RECORD
+ * qkey_to_typed_cacheel -- Lookup/load key
+ *============================================*/
+/* unused
+static CACHEEL
+qkey_to_typed_cacheel (STRING key)
+{
+	char ntype;
+	ASSERT(key && key[0]);
+	ntype = key[0];
+	switch(ntype) {
+	case 'I': return qkey_to_indi_cacheel(key);
+	case 'F': return qkey_to_fam_cacheel(key);
+	case 'S': return qkey_to_even_cacheel(key);
+	case 'E': return qkey_to_sour_cacheel(key);
+	case 'X': return qkey_to_othr_cacheel(key);
+	}
+	ASSERT(0);
+}
+unused */
+/*==============================================
+ * cacheel_to_key -- Return key of record inside of cache element
+ *  handle NULL input
+ *============================================*/
+CNSTRING
+cacheel_to_key (CACHEEL cel)
+{
+	CNSTRING key = cel ? ckey(cel) : 0;
+	return key;
+}
+/*==============================================
+ * cacheel_to_node -- Return root node of record inside of cache element
+ *  loads cache element if needed
  *  handle NULL input
  *============================================*/
 NODE
-nztop (RECORD rec)
+cacheel_to_node (CACHEEL cel)
 {
-	return rec ? rec->top : 0;
+	if (!cel) return NULL;
+	if (!cnode(cel)) {
+		CACHEEL cel2 = key_to_indi_cacheel(ckey(cel));
+		ASSERT(cel2 == cel);
+		ASSERT(cnode(cel));
+	}
+	return cnode(cel);
+}
+/*==============================================
+ * is_cel_loaded -- If cache element is loaded, return root
+ *  handle NULL input
+ *============================================*/
+NODE
+is_cel_loaded (CACHEEL cel)
+{
+	if (!cel) return NULL;
+	return cnode(cel);
+}
+/*==============================================
+ * cel_remove_record -- Our record informing us it is destructing
+ *  Requires non-null inputs
+ *============================================*/
+void
+cel_remove_record (CACHEEL cel, RECORD rec)
+{
+	ASSERT(cel);
+	ASSERT(cel->c_magic == cel_magic);
+	ASSERT(rec);
+	if (crecord(cel) == rec) {
+		crecord(cel) = 0;
+	}
 }
