@@ -102,10 +102,12 @@ ZST qSscnfnm      = N_("Enter pattern to match against full name.");
 ZST qSscnrfn      = N_("Enter pattern to match against refn.");
 ZST qSscantt      = N_("pattern: ");
 
-ZST qSnotone      = N_("Please choose from among these records.");
+ZST qSnotonei     = N_("Please choose from among these people.");
+ZST qSnotonex     = N_("Please choose from among these records.");
 ZST qSiscnew      = N_("Is this the new child? ");
 ZST qSissnew      = N_("Is this the new spouse? ");
-ZST qSifone       = N_("Is this the person?  Select if so.");
+ZST qSifonei      = N_("Is this the person?  Select if so.");
+ZST qSifonex      = N_("Is this the record?  Select if so.");
 ZST qSidcfam      = N_("Select the child the new child precedes or select last.");
 ZST qSidpnxt      = N_("Please identify next person to browse to.");
 ZST qSidnxt       = N_("Please identify record to browse to.");
@@ -184,6 +186,7 @@ ZST qScfrupt      = N_("Do you really want to update this source?");
 ZST qScfeupt      = N_("Do you really want to update this event?");
 ZST qScfxupt      = N_("Do you really want to update this record?");
 ZST qScfpdel      = N_("Are you sure you want to remove the person from the database?");
+ZST qScfodel      = N_("Are you sure you want to remove this record from the database?");
 ZST qScffdel      = N_("Remove this family record?");
 ZST qScffdeld     = N_("(Family %s (%s, %s)");
 ZST qScfpmrg      = N_("Do you really want to merge these two persons?");
@@ -229,6 +232,7 @@ ZST qSnoxmrg      = N_("Two parents of different sexes cannot be merged.");
 ZST qSdhusb       = N_("The families have different fathers/husbands; cannot merge.");
 ZST qSdwife       = N_("The families have different wives/mothers; cannot merge.");
 ZST qSidpdel      = N_("Who do you want to remove from the database?");
+ZST qSidodel      = N_("What record do you want to remove from the database?");
 ZST qSmklast      = N_("Place the child last in the family.");
 ZST qSabverr      = N_("Error in abbreviations file.");
 ZST qSuoperr      = N_("Error in user options file.");
@@ -553,6 +557,7 @@ ZST qSmn_del_chil = N_("c  Child - remove a child from his/her family");
 ZST qSmn_del_spou = N_("s  Spouse - remove a spouse from a family");
 ZST qSmn_del_indi = N_("i  Individual - remove a person completely");
 ZST qSmn_del_fam  = N_("f  Family - remove a family completely");
+ZST qSmn_del_any  = N_("o  Other - remove other record completely");
 
 /* &&end delete menu, begin search menu */
 ZST qSmn_sea_ttl  = N_("How would you like to find a record?");
