@@ -20,16 +20,15 @@
 #include "gedcom.h"
 #include "gedcomi.h"
 #include "feedback.h"
-#include "warehouse.h"
 #include "metadata.h"
 #include "lloptions.h"
 #include "date.h"
+#include "vtable.h"
 
 /*********************************************
  * global/exported variables
  *********************************************/
 
-static BOOLEAN add_metadata = FALSE;
 
 /*********************************************
  * external/imported variables
@@ -56,16 +55,11 @@ enum { NEW_RECORD, EXISTING_LACKING_WH_RECORD };
  * local function prototypes, alphabetical
  *********************************************/
 
-static RECORD alloc_record_from_key(STRING key);
-static void alloc_record_wh(RECORD rec, INT isnew);
 static NODE alloc_node(void);
-static void assign_record(RECORD rec, char ntype, INT keynum);
 static STRING fixup(STRING str);
 static STRING fixtag (STRING tag);
 static RECORD indi_to_prev_sib_impl(NODE indi);
-static void hook_node_to_rec_recurse(RECORD rec, NODE node);
-static void hook_record_to_root_node(RECORD rec, NODE node);
-
+static void node_destructor(VTABLE *obj);
 static INT node_strlen(INT levl, NODE node);
 
 /*********************************************
@@ -86,6 +80,18 @@ NODE parents_nodes(NODE faml);
 
 /* node allocator's free list */
 static NDALLOC first_blck = (NDALLOC) 0;
+static INT live_count = 0;
+
+static struct tag_vtable vtable_for_node = {
+	VTABLE_MAGIC
+	, "node"
+	, &node_destructor
+	, &refcountable_isref
+	, &refcountable_addref
+	, &refcountable_release
+	, 0 /* copy_fnc */
+	, &generic_get_type_name
+};
 
 /*********************************************
  * local function definitions
@@ -103,16 +109,27 @@ fixup (STRING str)
 }
 /*=============================
  * fixtag -- Keep tags in table
+ * returns pointer to table's memory
  *===========================*/
 static STRING
 fixtag (STRING tag)
 {
-	STRING str;
-	if ((str = valueof_str(tagtable, tag))) return str;
-	str = strsave(tag);
-	insert_table_str(tagtable, str, str);
+	STRING str = valueof_str(tagtable, tag);
+	if (!str) {
+		insert_table_str(tagtable, tag, tag);
+		str = valueof_str(tagtable, tag);
+	}
 	return str;
 }
+/*=====================================
+ * change_node_tag -- Give new tag to node
+ *===================================*/
+void
+change_node_tag (NODE node, STRING newtag)
+{
+	/* tag belongs to tagtable, so don't free old one */
+	ntag(node) = fixtag(newtag);
+}
 /*=====================================
  * alloc_node -- Special node allocator
  *===================================*/
@@ -134,6 +151,7 @@ alloc_node (void)
 	}
 	node = (NODE) first_blck;
 	first_blck = first_blck->next;
+	++live_count;
 	return node;
 }
 /*======================================
@@ -151,7 +169,7 @@ free_node (NODE node)
 	*/
 	((NDALLOC) node)->next = first_blck;
 	first_blck = (NDALLOC) node;
-	
+	--live_count;
 }
 /*===========================
  * create_node -- Create NODE
@@ -171,7 +189,7 @@ create_node (STRING xref, STRING tag, STRING val, NODE prnt)
 	nval(node) = fixup(val);
 	nparent(node) = prnt;
 	if (prnt)
-		node->n_rec = prnt->n_rec;
+		node->n_cel = prnt->n_cel;
 	return node;
 }
 /*===========================
@@ -228,152 +246,6 @@ set_temp_node (NODE node, BOOLEAN temp)
 	if (is_temp_node(node) ^ temp)
 		nflag(node) ^= ND_TEMP;
 }
-/*===================================
- * alloc_new_record -- record allocator
- *  perhaps should use special allocator like nodes
- * Created: 2001/01/25, Perry Rapp
- *=================================*/
-RECORD
-alloc_new_record (void)
-{
-	RECORD rec;
-	rec = (RECORD)stdalloc(sizeof(*rec));
-	memset(rec, 0, sizeof(*rec));
-	/* these must be filled in by caller */
-	rec->nkey.key = "";
-	rec->nkey.keynum = 0;
-	rec->nkey.ntype = 0;
-	return rec;
-}
-/*===================================
- * alloc_record_from_key -- allocate record for key
- * Created: 2001/01/25, Perry Rapp
- *=================================*/
-static RECORD
-alloc_record_from_key (STRING key)
-{
-	RECORD rec = alloc_new_record();
-	assign_record(rec, key[0], atoi(key+1));
-	return rec;
-}
-/*===================================
- * assign_record -- put key info into record
- * Created: 2001/02/04, Perry Rapp
- *=================================*/
-static void
-assign_record (RECORD rec, char ntype, INT keynum)
-{
-	char xref[12];
-	char key[9];
-	NODE node;
-	sprintf(key, "%c%d", ntype, keynum);
-	sprintf(xref, "@%s@", key);
-	if ((node = nztop(rec))) {
-		if (nxref(node)) stdfree(nxref(node));
-		nxref(node) = strsave(xref);
-	}
-	rec->nkey.key = strsave(key);
-	rec->nkey.keynum = keynum;
-	rec->nkey.ntype = ntype;
-}
-/*===================================
- * init_new_record -- put key info into record
- *  of brand new record
- * Created: 2001/02/04, Perry Rapp
- *=================================*/
-void
-init_new_record (RECORD rec, char ntype, INT keynum)
-{
-	assign_record(rec, ntype, keynum);
-	alloc_record_wh(rec, NEW_RECORD);
-}
-/*===================================
- * alloc_record_wh -- allocate warehouse for
- *  a record without one (new or existing)
- * Created: 2001/02/04, Perry Rapp
- *=================================*/
-static void
-alloc_record_wh (RECORD rec, INT isnew)
-{
-	LLDATE creation;
-	ASSERT(!rec->mdwh); /* caller must know what it is doing */
-	if (!add_metadata)
-		return;
-	rec->mdwh = (WAREHOUSE)stdalloc(sizeof(*(rec->mdwh)));
-	wh_allocate(rec->mdwh);
-	get_current_lldate(&creation);
-	wh_add_block_var(rec->mdwh, MD_CREATE_DATE, &creation, sizeof(creation));
-	if (isnew == EXISTING_LACKING_WH_RECORD)
-		wh_add_block_int(rec->mdwh, MD_CONVERTED_BOOL, 1);
-}
-/*===================================
- * create_record -- create record to wrap top node
- * Created: 2001/01/29, Perry Rapp
- *=================================*/
-RECORD
-create_record (NODE node)
-{
-	RECORD rec = 0;
-	if (nxref(node))
-		rec = alloc_record_from_key(node_to_key(node));
-	else
-		rec = alloc_new_record();
-	hook_record_to_root_node(rec, node);
-	return rec;
-}
-/*===================================
- * hook_record_to_root_node -- Connect record to node tree
- * Created: 2003-02-04 (Perry Rapp)
- *=================================*/
-static void
-hook_record_to_root_node (RECORD rec, NODE node)
-{
-	rec->top = node;
-	hook_node_to_rec_recurse(rec, node);
-}
-/*===================================
- * hook_node_to_rec_recurse -- Connect node subtree to record
- * Created: 2003-02-04 (Perry Rapp)
- *=================================*/
-static void
-hook_node_to_rec_recurse (RECORD rec, NODE node)
-{
-	while (node) {
-		if (nchild(node))
-			hook_node_to_rec_recurse(rec, nchild(node));
-		node->n_rec = rec;
-		node = nsibling(node);
-	}
-}
-/*===================================
- * init_new_record_and_just_read_node -- 
- *  initializer used by nodeio, after having read a node
- * Created: 2003-02-04 (Perry Rapp)
- *=================================*/
-void
-init_new_record_and_just_read_node (RECORD rec, NODE node, CNSTRING key)
-{
-	hook_record_to_root_node(rec, node);
-	assign_record(rec, key[0], atoi(key+1));
-	if (!rec->mdwh)
-		alloc_record_wh(rec, EXISTING_LACKING_WH_RECORD);
-}
-/*===================================
- * free_rec -- record deallocator
- * Created: 2000/12/30, Perry Rapp
- *=================================*/
-void
-free_rec (RECORD rec)
-{
-	if (rec->top)
-		free_nodes(rec->top);
-	if (rec->mdwh) {
-		stdfree(rec->mdwh);
-	}
-	if (rec->nkey.key[0])
-		stdfree(rec->nkey.key);
-	stdfree(rec);
-}
 /*=====================================
  * free_nodes -- Free all NODEs in tree
  *===================================*/
@@ -543,6 +415,7 @@ fam_to_wife_node (NODE node)
 	if (!node) return NULL;
 	if (!(node = find_tag(nchild(node), "WIFE"))) return NULL;
 	key = rmvat(nval(node));
+	if (!key) return NULL;
 	return qkey_to_indi(key);
 }
 /*========================================
@@ -578,6 +451,32 @@ fam_to_spouse (NODE fam, NODE indi)
 	ENDWIFES
 	return NULL;
 }
+/*===============================================
+ * next_spouse -- Return next spouse of family
+ * NODE *node     [in/out] pass in nchild(fam) to start
+ *                or nsibling(previous node returned from this routine) to continue
+ * RECORD *spouse [out]     next spouse in family
+ * returns 1 for success, -1 if next HUSB/WIFE record is invalid
+ *         0 no more spouses found
+ * returns addref'd record in *spouse
+ *=============================================*/
+int
+next_spouse (NODE *node, RECORD *spouse)
+{
+	CNSTRING key=0;
+	if (!node || !spouse) return 0;
+	while (*node) {
+	    if (eqstr(ntag(*node),"HUSB") || eqstr(ntag(*node),"WIFE")) {
+		key = rmvat(nval(*node));
+		if (!key) return -1;
+		*spouse = qkey_to_irecord(key);
+		if (!*spouse) return -1;
+		return 1;
+	    }
+	    *node = nsibling(*node);
+	}
+	return 0;
+}
 /*==================================================
  * fam_to_first_chil -- Return first child of family
  *================================================*/
@@ -628,6 +527,7 @@ indi_to_moth (NODE node)
 }
 /*==================================================
  * indi_to_prev_sib -- Return previous sib of person
+ *  returns addref'd record
  *================================================*/
 static RECORD
 indi_to_prev_sib_impl (NODE indi)
@@ -661,6 +561,7 @@ indi_to_prev_sib (RECORD irec)
 }
 /*==============================================
  * indi_to_next_sib -- Return next sib of person
+ *  returns addref'd record
  *============================================*/
 static RECORD
 indi_to_next_sib_impl (NODE indi)
@@ -686,7 +587,10 @@ indi_to_next_sib_impl (NODE indi)
 NODE
 indi_to_next_sib_old (NODE indi)
 {
-	return nztop(indi_to_next_sib_impl(indi));
+	RECORD rec = indi_to_next_sib_impl(indi);
+	NODE node = nztop(rec);
+	release_record(rec); /* release ref from indi_to_next_sib_impl */
+	return node;
 }
 RECORD
 indi_to_next_sib (RECORD irec)
@@ -1011,6 +915,14 @@ copy_node (NODE node)
 {
 	return create_node(nxref(node), ntag(node), nval(node), NULL);
 }
+/*========================
+ * copy_node_subtree -- Copy tree
+ *======================*/
+NODE
+copy_node_subtree (NODE node)
+{
+	return copy_nodes(node, TRUE, FALSE);
+}
 /*========================
  * copy_nodes -- Copy tree
  *======================*/
@@ -1226,3 +1138,39 @@ parents_nodes (NODE faml)      /* list of FAMC and/or FAMS nodes */
 	return new;
 }
 #endif /* UNUSED_CODE */
+/*=================================================
+ * node_destructor -- destructor for node
+ *  (destructor entry in vtable)
+ *===============================================*/
+static void
+node_destructor (VTABLE *obj)
+{
+	NODE node = (NODE)obj;
+	ASSERT((*obj) == &vtable_for_node);
+	free_node(node);
+}
+/*=================================================
+ * check_node_leaks -- Called when database closing
+ *  for debugging
+ *===============================================*/
+void
+check_node_leaks (void)
+{
+	if (live_count) {
+		STRING report_leak_path = getoptstr("ReportLeakLog", NULL);
+		FILE * fp=0;
+		if (report_leak_path)
+			fp = fopen(report_leak_path, "at");
+		if (fp) {
+			LLDATE date;
+			get_current_lldate(&date);
+			fprintf(fp, _("Node memory leaks:"));
+			fprintf(fp, " %s", date.datestr);
+			fprintf(fp, "\n  ");
+			fprintf(fp, _pl("%d item leaked", "%d items leaked", live_count), live_count);
+			fprintf(fp, "\n");
+			fclose(fp);
+			fp = 0;
+		}
+	}
+}
