@@ -16,10 +16,10 @@
 #include "gedcom.h"
 #include "gedcomi.h"
 #include "feedback.h"
-#include "warehouse.h"
 #include "metadata.h"
 #include "date.h"
 #include "xlat.h"
+#include "cache.h"
 
 /*********************************************
  * global/exported variables
@@ -38,17 +38,17 @@ extern STRING qSrerbln, qSrernwt, qSrerilv, qSrerwlv, qSunsupunix, qSunsupuniv;
  * local function prototypes, alphabetical
  *********************************************/
 
+/* alphabetical */
 static BOOLEAN buffer_to_line(STRING p, INT *plev, STRING *pxref
 	, STRING *ptag, STRING *pval, STRING *pmsg);
 static NODE do_first_fp_to_node(FILE *fp, BOOLEAN list, XLAT tt
 	, STRING *pmsg, BOOLEAN *peof);
-static RECORD do_first_fp_to_record(FILE *fp, BOOLEAN list, XLAT tt
-	, STRING *pmsg, BOOLEAN *peof);
-static void load_record_wh(RECORD rec, char * whptr, INT whlen);
 static BOOLEAN string_to_line(STRING *ps, INT *plev, STRING *pxref, 
 	STRING *ptag, STRING *pval, STRING *pmsg);
 static STRING swrite_node(INT levl, NODE node, STRING p);
 static STRING swrite_nodes(INT levl, NODE node, STRING p);
+static BOOLEAN write_bom(void);
+static void write_fam_to_file(NODE fam, CNSTRING file, BOOLEAN bom);
 static void write_indi_to_file(NODE indi, CNSTRING file, BOOLEAN bom);
 static void write_node(INT levl, FILE *fp, XLAT ttm,
 	NODE node, BOOLEAN indent);
@@ -208,16 +208,22 @@ gettag:
  * ttm:  [IN]  character translation table
  * pmsg: [OUT] possible error message
  * pemp: [OUT] set true if file is empty
+ * returns addref'd record
  *===============================================*/
 RECORD
 file_to_record (STRING fname, XLAT ttm, STRING *pmsg, BOOLEAN *pemp)
 {
 	NODE node = file_to_node(fname, ttm, pmsg, pemp);
-	RECORD rec = 0;
-	if (node) {
-		rec = create_record(node);
+	if (!node)
+		return 0;
+	if (nxref(node)) {
+		CACHEEL cel = node_to_cacheel_old(node);
+		RECORD rec = get_record_for_cel(cel);
+		return rec;
+	} else {
+		RECORD rec = create_record_for_unkeyed_node(node);
+		return rec;
 	}
-	return rec;
 }
 /*=================================================
  * file_to_node -- Convert GEDCOM file to NODE tree
@@ -308,36 +314,6 @@ do_first_fp_to_node (FILE *fp, BOOLEAN list, XLAT ttm,
 	lahead = TRUE;
 	return next_fp_to_node(fp, list, ttm, pmsg, peof);
 }
-/*================================================================
- * do_first_fp_to_record -- Convert first GEDCOM record in file to tree
- *
- *  fp:    [IN]  file that holds GEDCOM record/s
- *  list:  [IN]  can be list at level 0?
- *  tt:    [IN]  character translation table
- *  *pmsg: [OUT] possible error message
- *  *peof: [OUT] set true if file is at end of file
- * Called after unicode header processed
- *==============================================================*/
-static RECORD
-do_first_fp_to_record (FILE *fp, BOOLEAN list, XLAT ttm,
-	STRING *pmsg,  BOOLEAN *peof)
-{
-	INT rc;
-	ateof = FALSE;
-	flineno = 0;
-	*pmsg = NULL;
-	*peof = FALSE;
-	rc = file_to_line(fp, ttm, &lev, &xref, &tag, &val, pmsg);
-	if (rc == DONE) {
-		*peof = ateof = TRUE;
-		*pmsg = _(qSfileof);
-		return NULL;
-	} else if (rc == ERROR)
-		return NULL;
-	lev0 = lev;
-	lahead = TRUE;
-	return next_fp_to_record(fp, list, ttm, pmsg, peof);
-}
 /*==============================================================
  * next_fp_to_record -- Convert next GEDCOM record in file to tree
  *
@@ -346,13 +322,14 @@ do_first_fp_to_record (FILE *fp, BOOLEAN list, XLAT ttm,
  *  ttm:  [IN]  character translation table
  *  pmsg: [OUT] possible error message
  *  peof: [OUT] set true if file is at end of file
+ * returns addref'd record
  *============================================================*/
 RECORD
 next_fp_to_record (FILE *fp, BOOLEAN list, XLAT ttm,
 	STRING *pmsg, BOOLEAN *peof)
 {
 	NODE node = next_fp_to_node(fp, list, ttm, pmsg, peof);
-	return create_record(node);
+	return create_record_for_keyed_node(node, 0);
 }
 /*==============================================================
  * next_fp_to_node -- Convert next GEDCOM record in file to tree
@@ -370,7 +347,7 @@ next_fp_to_node (FILE *fp, BOOLEAN list, XLAT ttm,
 {
 	INT curlev, bcode, rc;
 	NODE root, node, curnode;
-	char scratch[100];
+	static char scratch[100]; /*we return errors with this, keep it around*/
 	*pmsg = NULL;
 	*peof = FALSE;
 	if (ateof) {
@@ -447,14 +424,9 @@ next_fp_to_node (FILE *fp, BOOLEAN list, XLAT ttm,
 /*============================================
  * string_to_record -- Read record from data block
  *  (modifies string -- inserts 0 between lines)
- *  This is the layout for metadata nodes:
- *   Q___      (four bytes, but only first char used)
- *   0016      (offset to 0 INDI... line)
- *   whhdr     (warehouse header)
- *   whdata    (warehouse data)
- *   0 INDI    (traditional node data)
  *  This is the layout for traditional nodes:
  *   0 INDI    (or 0 FAM or 0 SOUR etc)
+ * returns addref'd record
  *==========================================*/
 RECORD
 string_to_record (STRING str, CNSTRING key, INT len)
@@ -463,29 +435,11 @@ string_to_record (STRING str, CNSTRING key, INT len)
 	NODE node = 0;
 	len=len; /* unused */
 
-	/* create it now, & release it at bottom if we fail */
-	rec = alloc_new_record();
-	/* we must fill in the top & mdwh fields */
+	/* we must fill in the top field */
 
-	if (*str == '0') /* traditional node, no metadata */
+	if (*str == '0') { /* traditional node, no metadata */
+		/* actually no metadata was ever used in any version */
 		node = string_to_node(str);
-	else if (*str == 'Q') {
-		/*
-		first four bytes just used for the Q flag
-		second four bytes contain node_offset from str
-		NB: UNTESTED because it isn't yet being written
-		Perry, 2001/01/15
-		*/
-#ifdef CODE_TO_BE_DELETED
-		INT * ptr = (INT *)str;
-		INT node_offset = ptr[1]; /* in characters */
-		char * whptr = (char *)&ptr[2];
-		const char * node_ptr = str + node_offset;
-		INT whlen = node_offset - 8;
-		ASSERT(0); /* not yet being written */
-		load_record_wh(rec, whptr, whlen);
-		node = string_to_node(node_ptr);
-#endif
 	} else {
 		if (!strcmp(str, "DELE\n")) {
 			/* should have been filtered out in getrecord */
@@ -495,10 +449,7 @@ string_to_record (STRING str, CNSTRING key, INT len)
 		}
 	}
 	if (node) {
-		init_new_record_and_just_read_node(rec, node, key);
-	} else { /* node==0, we failed, clean up */
-		free_rec(rec);
-		rec = 0;
+		rec = create_record_for_keyed_node(node, key);
 	}
 	return rec;
 }
@@ -687,18 +638,6 @@ node_to_string (NODE node)      /* root */
 	(void) swrite_nodes(0, node, str);
 	return str;
 }
-/*===================================
- * load_record_wh -- load existing warehouse
- *  for a record
- *  warehouse is opaque lump of metadata
- * Created: 2001/02/04, Perry Rapp
- *=================================*/
-static void
-load_record_wh (RECORD rec, char * whptr, INT whlen)
-{
-	rec->mdwh = (WAREHOUSE)stdalloc(sizeof(*(rec->mdwh)));
-	wh_assign_from_blob(rec->mdwh, whptr, whlen);
-}
 /*=====================================
  * write_indi_to_file - write node tree into GEDCOM
  * (no user interaction)
@@ -726,6 +665,19 @@ write_indi_to_file (NODE indi, CNSTRING file, BOOLEAN bom)
 	fclose(fp);
 	join_indi(indi, name, refn, sex, body, famc, fams);
 }
+/*=====================================
+ * write_bom - whether to write Unicode BOM
+ * (byte order marker)
+ *===================================*/
+static BOOLEAN
+write_bom (void)
+{
+#ifdef WIN32
+	return TRUE;
+#else
+	return FALSE;
+#endif
+}
 /*=====================================
  * write_indi_to_file_for_edit - write node tree into GEDCOM
  *  file to be edited
@@ -734,24 +686,35 @@ write_indi_to_file (NODE indi, CNSTRING file, BOOLEAN bom)
 void
 write_indi_to_file_for_edit (NODE indi, CNSTRING file, RFMT rfmt)
 {
-	BOOLEAN bom = FALSE;
-#ifdef WIN32
-	bom = TRUE;
-#endif
+	BOOLEAN bom = write_bom();
 	annotate_with_supplemental(indi, rfmt);
-	write_indi_to_file(indi, editfile, bom);
+	write_indi_to_file(indi, file, bom);
 	resolve_refn_links(indi);
 }
+/*=====================================
+ * write_indi_to_file_for_edit - write node tree into GEDCOM
+ *  file to be edited
+ * (no user interaction)
+ *===================================*/
+void
+write_fam_to_file_for_edit (NODE fam, CNSTRING file, RFMT rfmt)
+{
+	BOOLEAN bom = write_bom();
+	annotate_with_supplemental(fam, rfmt);
+	write_fam_to_file(fam, file, bom);
+	resolve_refn_links(fam);
+}
 /*=====================================
  * write_fam_to_file -- write node tree into GEDCOM
  * (no user interaction)
  *===================================*/
-void
-write_fam_to_file (NODE fam, CNSTRING file)
+static void
+write_fam_to_file (NODE fam, CNSTRING file, BOOLEAN bom)
 {
 	FILE *fp;
 	XLAT ttmo = transl_get_predefined_xlat(MINED);
 	NODE refn, husb, wife, chil, body;
+	bom = bom; /* unused */
 
 	ASSERT(fp = fopen(file, LLWRITETEXT));
 	split_fam(fam, &refn, &husb, &wife, &chil, &body);
