@@ -36,6 +36,7 @@
 #include "gedcom.h"
 #include "lloptions.h"
 #include "zstr.h"
+#include "cache.h"
 
 /*********************************************
  * external/imported variables
@@ -51,7 +52,7 @@ static void annotate_node(NODE node, BOOLEAN expand_refns, BOOLEAN annotate_poin
 static BOOLEAN is_annotated_xref(CNSTRING val, INT * len);
 static STRING symbolic_link(CNSTRING);
 static void parserefnrec(RKEY rkey, CNSTRING p);
-static RKEY refn2rkey(STRING);
+static RKEY refn2rkey(CNSTRING refn);
 static BOOLEAN resolve_node(NODE node, BOOLEAN annotate_pointers);
 
 /*********************************************
@@ -150,13 +151,13 @@ parserefnrec (RKEY rkey, CNSTRING p)
  * getrefnrec -- Read refn record and store in globals
  *==================================================*/
 BOOLEAN
-getrefnrec (STRING refn)
+getrefnrec (CNSTRING refn)
 {
 	STRING p;
 /* Convert refn to key and read refn record */
 	RRkey = refn2rkey(refn);
 	if (RRrec) stdfree(RRrec);
-	p = RRrec = getrecord(BTR, &RRkey, &RRsize);
+	p = RRrec = bt_getrecord(BTR, &RRkey, &RRsize);
 	if (!RRrec) {
 		RRcount = 0;
 		if (RRmax == 0) {
@@ -174,7 +175,7 @@ getrefnrec (STRING refn)
  * refn2rkey - Convert refn to refn record key
  *==========================================*/
 static RKEY
-refn2rkey (STRING refn)
+refn2rkey (CNSTRING refn)
 {
 	RKEY rkey;
 	rkey.r_rkey[0] = rkey.r_rkey[1] = ' ';
@@ -213,10 +214,11 @@ refn_hi (void)
 }
 /*=========================================
  * add_refn -- Add new entry to refn record
+ *  refn: [IN]  record's user refn key
+ *  key:  [IN]  record's GEDCOM key
  *=======================================*/
 BOOLEAN
-add_refn (STRING refn,  /* record's user refn key */
-          STRING key)   /* record's GEDCOM key */
+add_refn (CNSTRING refn, CNSTRING key)
 {
 	STRING rec, p;
 	INT i, len, off;
@@ -255,7 +257,7 @@ add_refn (STRING refn,  /* record's user refn key */
 		p += strlen(RRrefns[i]) + 1;
 		len += strlen(RRrefns[i]) + 1;
 	}
-	addrecord(BTR, RRkey, rec, len);
+	bt_addrecord(BTR, RRkey, rec, len);
 	stdfree(rec);
 	return TRUE;
 }
@@ -263,8 +265,8 @@ add_refn (STRING refn,  /* record's user refn key */
  * remove_refn -- Remove entry from refn record
  *===========================================*/
 BOOLEAN
-remove_refn (STRING refn,       /* record's refn */
-             STRING key)        /* record's GEDCOM key */
+remove_refn (CNSTRING refn,       /* record's refn */
+             CNSTRING key)        /* record's GEDCOM key */
 {
 	STRING rec, p;
 	INT i, len, off;
@@ -308,7 +310,7 @@ remove_refn (STRING refn,       /* record's refn */
 		p += strlen(RRrefns[i]) + 1;
 		len += strlen(RRrefns[i]) + 1;
 	}
-	addrecord(BTR, RRkey, rec, len);
+	bt_addrecord(BTR, RRkey, rec, len);
 	stdfree(rec);
 	return TRUE;
 }
@@ -371,17 +373,24 @@ INT
 resolve_refn_links (NODE node)
 {
 	INT unresolved = 0;
-	struct tag_node_iter nodeit;
-	NODE child=0;
 	BOOLEAN annotate_pointers = (getoptint("AnnotatePointers", 0) > 0);
+	NODE child=0;
+	CACHEEL cel = node ? node->n_cel: 0;
+	struct tag_node_iter nodeit;
 
 	if (!node) return 0;
 
+	if (cel) lock_cache(cel); /* ensure node doesn't fall out of cache */
+
+	/* resolve all descendant nodes */
 	begin_node_it(node, &nodeit);
 	while ((child = next_node_it_ptr(&nodeit)) != NULL) {
 		if (!resolve_node(child, annotate_pointers))
 			++unresolved;
 	}
+
+	if (cel) unlock_cache(cel);
+
 	return unresolved;
 }
 /*=======================================================
@@ -411,7 +420,7 @@ resolve_node (NODE node, BOOLEAN annotate_pointers)
 		INT i=0,len=0;
 		if (is_annotated_xref(nval(node), &len)) {
 			char newval[20];
-			ASSERT(len < sizeof(newval));
+			ASSERT(len < (INT)sizeof(newval));
 			for (i=0; i<len; ++i) {
 				newval[i] = nval(node)[i];
 			}
@@ -461,11 +470,18 @@ annotate_with_supplemental (NODE node, RFMT rfmt)
 	BOOLEAN expand_refns = (getoptint("ExpandRefnsDuringEdit", 0) > 0);
 	BOOLEAN annotate_pointers = (getoptint("AnnotatePointers", 0) > 0);
 	NODE child=0;
+	CACHEEL cel = node->n_cel;
 	struct tag_node_iter nodeit;
+
+	if (cel) lock_cache(cel); /* ensure node doesn't fall out of cache */
+
+	/* annotate all descendant nodes */
 	begin_node_it(node, &nodeit);
 	while ((child = next_node_it_ptr(&nodeit)) != NULL) {
 		annotate_node(child, expand_refns, annotate_pointers, rfmt);
 	}
+
+	if (cel) unlock_cache(cel);
 }
 /*=======================================================
  * annotate_node -- Alter a node by
@@ -568,6 +584,9 @@ record_letter (CNSTRING tag)
 }
 /*=========================================
  * key_possible_to_record -- Returns record with key
+ *  str:  string that may be a key
+ *  let:  if string starts with a letter, it must be this (eg, 'I' for indi)
+ * This returns NULL upon failure
  *=======================================*/
 RECORD key_possible_to_record (STRING str, /* string that may be a key */
                     INT let)    /* if string starts with letter it
@@ -587,16 +606,32 @@ RECORD key_possible_to_record (STRING str, /* string that may be a key */
 	kbuf[i] = 0;
 	if (!isrecord(BTR, str2rkey(kbuf))) return NULL;
 	switch (let) {
-	case 'I': return key_to_irecord(kbuf);
-	case 'F': return key_to_frecord(kbuf);
-	case 'S': return key_to_srecord(kbuf);
-	case 'E': return key_to_erecord(kbuf);
-	case 'X': return key_to_orecord(kbuf);
+	case 'I': return qkey_to_irecord(kbuf);
+	case 'F': return qkey_to_frecord(kbuf);
+	case 'S': return qkey_to_srecord(kbuf);
+	case 'E': return qkey_to_erecord(kbuf);
+	case 'X': return qkey_to_orecord(kbuf);
 	default:  FATAL();
 	}
 	FATAL();
 	return NULL;
 }
+/*================================================
+ * is_record_missing_data_entry - Check if record is 
+ * orphaned in index (is in index, but lacks data)
+ *==============================================*/
+BOOLEAN
+is_record_missing_data_entry (CNSTRING key)
+{
+	RECORD rec=0;
+	if (!isrecord(BTR, str2rkey(key)))
+		return FALSE;
+	if ((rec = qkey_to_record(key))) {
+		release_record(rec);
+		return FALSE;
+	}
+	return TRUE;
+}
 /*================================================
  * refn_to_record - Get record from user reference
  *  ukey: [IN]  refn key found
@@ -641,7 +676,7 @@ typedef struct
 } TRAV_REFN_PARAM;
 /* see above */
 static BOOLEAN
-traverse_refn_callback (RKEY rkey, CNSTRING data, INT len, void *param)
+traverse_refn_callback (RKEY rkey, STRING data, INT len, void *param)
 {
 	TRAV_REFN_PARAM *tparam = (TRAV_REFN_PARAM *)param;
 	INT i;
