@@ -19,6 +19,10 @@
 #include "gedcomi.h"
 #include "xlat.h" 
 #include "arch.h"
+#include "lloptions.h"
+#include "log.h"
+#include "vtable.h"
+
 
 /*********************************************
  * local types
@@ -35,11 +39,15 @@ struct tag_xlat {
 	INT uparam; /* opaque number used by client */
 };
 /* dynamically loadable translation table, entry in dyntt list */
-typedef struct tag_dyntt {
+struct tag_dyntt {
+	struct tag_vtable *vtable; /* generic object */
+	INT refcnt; /* ref-countable object */
+	STRING name;
 	STRING path;
 	TRANTABLE tt; /* when loaded */
 	BOOLEAN loadfailure;
-} *DYNTT;
+};
+typedef struct tag_dyntt *DYNTT;
 
 /* step of a translation, either iconv_src or trantble is NULL */
 typedef struct xlat_step_s {
@@ -60,7 +68,9 @@ static XLSTEP create_iconv_step(CNSTRING src, CNSTRING dest);
 static XLSTEP create_dyntt_step(DYNTT dyntt);
 static XLAT create_null_xlat(BOOLEAN adhoc);
 static XLAT create_xlat(CNSTRING src, CNSTRING dest, BOOLEAN adhoc);
-static DYNTT create_dyntt(TRANTABLE tt, STRING path);
+static DYNTT create_dyntt(TRANTABLE tt, CNSTRING name, CNSTRING path);
+static void destroy_dyntt(DYNTT dyntt);
+static void dyntt_destructor(VTABLE *obj);
 static void free_dyntts(void);
 static void free_xlat(XLAT xlat);
 static DYNTT get_conversion_dyntt(CNSTRING src, CNSTRING dest);
@@ -68,7 +78,6 @@ static DYNTT get_subcoding_dyntt(CNSTRING codeset, CNSTRING subcoding);
 static void load_dyntt_if_needed(DYNTT dyntt);
 static void load_dynttlist_from_dir(STRING dir);
 static int select_tts(const struct dirent *entry);
-static void zero_dyntt(DYNTT dyntt);
 
 /*********************************************
  * local variables
@@ -78,6 +87,17 @@ static LIST f_xlats=0; /* cache of conversions */
 static TABLE f_dyntts=0; /* cache of dynamic translation tables */
 static char f_ttext[] = ".tt";
 
+static struct tag_vtable vtable_for_dyntt = {
+	VTABLE_MAGIC
+	, "dyntt"
+	, &dyntt_destructor
+	, &refcountable_isref
+	, &refcountable_addref
+	, &refcountable_release
+	, 0 /* copy_fnc */
+	, &generic_get_type_name
+};
+
 /*********************************************
  * local & exported function definitions
  * body of module
@@ -132,8 +152,10 @@ free_xlat (XLAT xlat)
 		strfree(&xstep->iconv_dest);
 		xstep->dyntt = 0; /* f_dyntts owns dyntt memory */
 	ENDLIST
-	make_list_empty(xlat->steps);
-	remove_list(xlat->steps, 0);
+	destroy_list(xlat->steps);
+	strfree(&xlat->src);
+	strfree(&xlat->dest);
+	stdfree(xlat);
 }
 /*==========================================================
  * create_iconv_step -- Create an iconv step of a translation chain
@@ -167,11 +189,14 @@ create_dyntt_step (DYNTT dyntt)
  * Created: 2002/12/10 (Perry Rapp)
  *========================================================*/
 static DYNTT
-create_dyntt (TRANTABLE tt, STRING path)
+create_dyntt (TRANTABLE tt, CNSTRING name, CNSTRING path)
 {
 	DYNTT dyntt = (DYNTT)stdalloc(sizeof(*dyntt));
 	memset(dyntt, 0, sizeof(*dyntt));
+	dyntt->vtable = &vtable_for_dyntt;
+	dyntt->refcnt = 1;
 	dyntt->tt = tt;
+	dyntt->name = strsave(name);
 	dyntt->path = strsave(path);
 	return dyntt;
 }
@@ -240,10 +265,24 @@ xl_get_xlat (CNSTRING src, CNSTRING dest, BOOLEAN adhoc)
 		XLSTEP xstep = create_iconv_step(zs_str(zsrc), zs_str(zdest));
 		enqueue_list(xlat->steps, xstep);
 	} else {
-		DYNTT dyntt = get_conversion_dyntt(zs_str(zsrc), zs_str(zdest));
-		if (dyntt)
+		STRING src = zs_str(zsrc), dest = zs_str(zdest);
+		BOOLEAN foundit = FALSE;
+		DYNTT dyntt = get_conversion_dyntt(src, dest);
+		if (dyntt) {
 			add_dyntt_step(xlat, dyntt);
-		else
+			foundit = TRUE;
+		}
+		else if (!eqstr(src, "UTF-8") && !eqstr(dest, "UTF-8")) {
+			/* try going through UTF-8 as intermediate step */
+			DYNTT dyntt1 = get_conversion_dyntt(src, "UTF-8");
+			DYNTT dyntt2 = get_conversion_dyntt("UTF-8", dest);
+			if (dyntt1 && dyntt2) {
+				add_dyntt_step(xlat, dyntt1);
+				add_dyntt_step(xlat, dyntt2);
+				foundit = TRUE;
+			}
+		}
+		if (!foundit)
 			xlat->valid = FALSE; /* missing main conversion */
 	}
 
@@ -298,7 +337,14 @@ get_conversion_dyntt (CNSTRING src, CNSTRING dest)
 	ZSTR zttname = zs_news(src);
 	zs_appc(zttname, '_');
 	zs_apps(zttname, dest);
-	dyntt = (DYNTT)valueof_ptr(f_dyntts, zs_str(zttname));
+	dyntt = (DYNTT)valueof_obj(f_dyntts, zs_str(zttname));
+	if (getoptint("TTPATH.debug", 0)) {
+		log_outf("ttpath.dbg",
+			_("ttpath trying to convert from <%s> to <%s>: %s"),
+			src, dest,
+			dyntt ? _("succeeded") : _("failed")
+			);
+	}
 	zs_free(&zttname);
 	return dyntt;
 }
@@ -313,7 +359,7 @@ get_subcoding_dyntt (CNSTRING codeset, CNSTRING subcoding)
 	ZSTR zttname = zs_news(codeset);
 	zs_apps(zttname, "__");
 	zs_apps(zttname, subcoding);
-	dyntt = (DYNTT)valueof_ptr(f_dyntts, zs_str(zttname));
+	dyntt = (DYNTT)valueof_obj(f_dyntts, zs_str(zttname));
 	zs_free(&zttname);
 	return dyntt;
 }
@@ -327,7 +373,7 @@ load_dyntt_if_needed (DYNTT dyntt)
 	ZSTR zerr=zs_new();
 	if (dyntt->tt || dyntt->loadfailure)
 		return;
-	if (!init_map_from_file(dyntt->path, dyntt->path, &dyntt->tt, zerr)) {
+	if (!init_map_from_file(dyntt->path, dyntt->name, &dyntt->tt, zerr)) {
 		dyntt->loadfailure = TRUE;
 	}
 	zs_free(&zerr);
@@ -358,13 +404,13 @@ xl_do_xlat (XLAT xlat, ZSTR zstr)
 		} else if (xstep->dyntt) {
 			/* a custom translation table step */
 			if (xstep->dyntt->tt)
-				custom_translate(zstr, xstep->dyntt->tt);
+				custom_translatez(zstr, xstep->dyntt->tt);
 		}
 	ENDLIST
 	return cvtd;
 }
 /*==========================================================
- * xl_load_all_tts -- Load internal list of available translation
+ * xl_load_all_dyntts -- Load internal list of available translation
  *  tables (based on *.tt files in TTPATH)
  * Created: 2002/11/27 (Perry Rapp)
  *========================================================*/
@@ -373,9 +419,15 @@ xl_load_all_dyntts (CNSTRING ttpath)
 {
 	STRING dirs,p;
 	free_dyntts();
+	if (getoptint("TTPATH.debug", 0)) {
+		if (!ttpath ||  !ttpath[0])
+			log_outf("ttpath.dbg", _("No TTPATH config variable"));
+		else
+			log_outf("ttpath.dbg", "ttpath: %s", ttpath);
+	}
 	if (!ttpath ||  !ttpath[0])
 		return;
-	f_dyntts = create_table(FREEBOTH);
+	f_dyntts = create_table_obj();
 	dirs = (STRING)stdalloc(strlen(ttpath)+2);
 	/* find directories in dirs & delimit with zeros */
 	chop_path(ttpath, dirs);
@@ -397,6 +449,9 @@ load_dynttlist_from_dir (STRING dir)
 	struct dirent **programs;
 	INT n = scandir(dir, &programs, select_tts, alphasort);
 	INT i;
+	if (getoptint("TTPATH.debug", 0)) {
+		log_outf("ttpath.dbg", _("ttpath checking dir <%s>"), dir);
+	}
 	for (i=0; i<n; ++i) {
 		CNSTRING ttfile = programs[i]->d_name;
 		/* filename without extension */
@@ -408,13 +463,17 @@ load_dynttlist_from_dir (STRING dir)
 			UTF-8_ISO-8859-1 (type 1; code conversion)
 			UTF-8__HTML (type 2; subcoding)
 		*/
+		if (getoptint("TTPATH.debug", 0)) {
+			log_outf("ttpath.dbg", _("ttpath file <%s> typed as %d"), ttfile, ntype);
+		}
 		if (ntype==1 || ntype==2) {
-			if (!valueof_ptr(f_dyntts, zs_str(zfile))) {
+			if (!valueof_obj(f_dyntts, zs_str(zfile))) {
 				TRANTABLE tt=0; /* will be loaded when needed */
 				STRING path = concat_path_alloc(dir, ttfile);
-				DYNTT dyntt = create_dyntt(tt, path);
+				DYNTT dyntt = create_dyntt(tt, ttfile, path);
 				strfree(&path);
-				insert_table_ptr(f_dyntts, strsave(zs_str(zfile)), dyntt);
+				insert_table_obj(f_dyntts, zs_str(zfile), dyntt);
+				--dyntt->refcnt; /* leave table as sole owner */
 			}
 		}
 		zs_free(&zfile);
@@ -489,8 +548,8 @@ select_tts (const struct dirent *entry)
 void
 xl_free_adhoc_xlats (void)
 {
-    XLAT xlattemp;
-	LIST newlist;
+    XLAT xlattemp=0;
+	LIST newlist=0;
 	if (!f_xlats)
 		return;
 	/* we don't have a way to delete items from a list,
@@ -504,8 +563,7 @@ xl_free_adhoc_xlats (void)
 			back_list(newlist, xlattemp);
 		}
 	ENDLIST
-	make_list_empty(f_xlats);
-	remove_list(f_xlats, 0);
+	destroy_list(f_xlats);
 	f_xlats = newlist;
 }
 /*==========================================================
@@ -515,15 +573,14 @@ xl_free_adhoc_xlats (void)
 void
 xl_free_xlats (void)
 {
-	XLAT xlattemp;
+	XLAT xlattemp=0;
 	if (!f_xlats)
 		return;
 	FORLIST(f_xlats, el)
 		xlattemp = (XLAT)el;
 		free_xlat(xlattemp);
 	ENDLIST
-	make_list_empty(f_xlats);
-	remove_list(f_xlats, 0);
+	destroy_list(f_xlats);
 	f_xlats = 0;
 }
 /*==========================================================
@@ -534,32 +591,31 @@ static void
 free_dyntts (void)
 {
 	if (f_dyntts) {
-		struct tag_table_iter tabit;
-		xl_free_xlats(); /* xlats point into f_dyntts */
-		if (begin_table(f_dyntts, &tabit)) {
-			VPTR ptr;
-			STRING key;
-			while (next_table_ptr(&tabit, &key, &ptr)) {
-				DYNTT dyntt = (DYNTT)ptr;
-				zero_dyntt(dyntt);
-			}
-		}
 		destroy_table(f_dyntts);
 		f_dyntts = 0;
 	}
 }
 /*==========================================================
- * zero_dyntt -- Empty (free) contents of a dyntt
- *  NB: This does not free the dyntt memory itself
- *  b/c that is handled in the f_dyntts cache table
- * Created: 2002/12/13 (Perry Rapp)
+ * xlat_shutdown -- Clear and free allocated objects
  *========================================================*/
-static void
-zero_dyntt (DYNTT dyntt)
+void
+xlat_shutdown (void)
 {
+	free_dyntts();
+}
+/*=================================================
+ * destroy_table -- destroy all element & memory for table
+ *===============================================*/
+static void
+destroy_dyntt (DYNTT dyntt)
+{
+	if (!dyntt) return;
+	ASSERT(dyntt->vtable == &vtable_for_dyntt);
+	strfree(&dyntt->name);
 	strfree(&dyntt->path);
 	remove_trantable(dyntt->tt);
 	dyntt->tt = 0;
+	stdfree(dyntt);
 }
 /*==========================================================
  * xl_parse_codeset -- Parse out subcode suffixes of a codeset
@@ -685,4 +741,14 @@ xl_get_dest_codeset (XLAT xlat)
 {
 	return xlat->dest;
 }
-
+/*=================================================
+ * dyntt_destructor -- destructor for dyntt
+ *  (destructor entry in vtable)
+ *===============================================*/
+static void
+dyntt_destructor (VTABLE *obj)
+{
+	DYNTT tab = (DYNTT)obj;
+	ASSERT(tab->vtable == &vtable_for_dyntt);
+	destroy_dyntt(tab);
+}
