@@ -23,6 +23,9 @@
 */
 /*=============================================================
  * xreffile.c -- Handle the xref file
+ * This module manages the lists of free record numbers
+ * For example, if I1, I2, I4, and I6 are in use
+ *  then irecs contains I3 and I5 (free record numbers)
  * Copyright(c) 1991-94 by T.T. Wetmore IV; all rights reserved
  * pre-SourceForge version information:
  *   2.3.4 - 24 Jun 93    2.3.5 - 02 Sep 93
@@ -32,13 +35,13 @@
 #include "sys_inc.h"
 #include "llstdlib.h"
 #include "btree.h"
+#include "gedcom.h"
+#include "gedcomi.h"
 #include "table.h"
 #include "translat.h"
-#include "gedcom.h"
 
 extern BTREE BTR;
 
-
 /*===================================================================
  * First five words in xrefs file are number of INDI, FAM, EVEN, SOUR
  *   and other keys in file; remaining words are keys, in respective
@@ -48,7 +51,13 @@ extern BTREE BTR;
  * nixrefs==2 means there is one deleted INDI key (ixrefs[1])
  *=================================================================*/
 
-/*==================================== 
+/*********************************************
+ * local types
+ *********************************************/
+
+typedef enum { DUPSOK, NODUPS } DUPS;
+
+ /*==================================== 
  * deleteset -- set of deleted records
  *  NB: storage order is IFESX
  *  whereas canonical order is IFSEX
@@ -63,14 +72,31 @@ struct deleteset_s
 typedef struct deleteset_s *DELETESET;
 
 
-static BOOLEAN readxrefs(void);
-static void readrecs(DELETESET set);
+/*********************************************
+ * local function prototypes
+ *********************************************/
+
+/* alphabetical */
+static BOOLEAN addxref_impl(CNSTRING key, DUPS dups);
+static INT find_slot(INT keynum, DELETESET set);
+static void freexref(DELETESET set);
+static DELETESET get_deleteset_from_type(char ctype);
 static STRING getxref(DELETESET set);
-static void addxref(INT key, DELETESET set);
 static void growxrefs(DELETESET set);
-static INT num_set(DELETESET set);
 static STRING newxref(STRING xrefp, BOOLEAN flag, DELETESET set);
+static INT num_set(DELETESET set);
+static void readrecs(DELETESET set);
+static BOOLEAN readxrefs(void);
 static INT xref_last(DELETESET set);
+static BOOLEAN addixref_impl(INT key, DUPS dups);
+static BOOLEAN addfxref_impl(INT key, DUPS dups);
+static BOOLEAN addsxref_impl(INT key, DUPS dups);
+static BOOLEAN addexref_impl(INT key, DUPS dups);
+static BOOLEAN addxxref_impl(INT key, DUPS dups);
+
+/*********************************************
+ * local variables
+ *********************************************/
 
 /* INDI, FAM, EVEN, SOUR, other sets */
 static struct deleteset_s irecs, frecs, srecs, erecs, xrecs;
@@ -80,6 +106,11 @@ static BOOLEAN xrefReadonly = FALSE;
 
 static INT maxkeynum=-1; /* cache value of largest key extant (-1 means not sure) */
 
+/*********************************************
+ * local & exported function definitions
+ * body of module
+ *********************************************/
+
 /*==================================== 
  * initdset -- Initialize a delete set
  *==================================*/
@@ -150,6 +181,11 @@ closexref (void)
 	if (xreffp) {
 		fclose(xreffp); xreffp = 0;
 	}
+	freexref(&irecs);
+	freexref(&frecs);
+	freexref(&srecs);
+	freexref(&erecs);
+	freexref(&xrecs);
 }
 /*=========================================
  * getxrefnum -- Return new keynum for type
@@ -222,8 +258,8 @@ sortxref (DELETESET set)
 				set->recs[j] = set->recs[i];
 				set->recs[i] = temp;
 			}
-			if (i==1 && !ct) return; /* already sorted */
 		}
+		if (i==1 && !ct) return; /* already sorted */
 	}
 }
 /*======================================
@@ -258,8 +294,8 @@ readxrefs (void)
 	ASSERT(xrecs.n > 0);
 	if (irecs.n > irecs.max) growxrefs(&irecs);
 	if (frecs.n > frecs.max) growxrefs(&frecs);
-	if (erecs.n > erecs.max) growxrefs(&erecs);
 	if (srecs.n > srecs.max) growxrefs(&srecs);
+	if (erecs.n > erecs.max) growxrefs(&erecs);
 	if (xrecs.n > xrecs.max) growxrefs(&xrecs);
 	readrecs(&irecs);
 	readrecs(&frecs);
@@ -301,50 +337,110 @@ writexrefs (void)
 	return TRUE;
 }
 /*=====================================
- * addxref -- Add deleted key to xrefs.
+ * find_slot -- Find slot at which to add key
+ *===================================*/
+static INT
+find_slot (INT keynum, DELETESET set)
+{
+	INT lo=1;
+	INT hi=(set->n)-1;
+	/* binary search to find where to insert key */
+	while (lo<=hi) {
+		INT md = (lo + hi)/2;
+		if (keynum > (set->recs)[md])
+			hi=--md;
+		else if (keynum < (set->recs)[md])
+			lo=++md;
+		else {
+			return md;
+		}
+	}
+	return lo;
+}
+/*=====================================
+ * add_xref_to_set_impl -- Add deleted key to xrefs.
  *  generic for all types
  *===================================*/
-static void
-addxref (INT key, DELETESET set)
+static BOOLEAN
+add_xref_to_set_impl (INT keynum, DELETESET set, DUPS dups)
 {
-	INT lo,hi,md, i;
-	if (key <= 0 || !xreffp || (set->n) < 1) FATAL();
+	INT lo, i;
+	if (keynum <= 0 || !xreffp || (set->n) < 1) FATAL();
 	if (set->n >= set->max)
 		growxrefs(set);
 	ASSERT(set->n < set->max);
-	lo=1;
-	hi=(set->n)-1;
-	/* binary search to find where to insert key */
-	while (lo<=hi) {
-		md = (lo + hi)/2;
-		if (key>(set->recs)[md])
-			hi=--md;
-		else if (key<(set->recs)[md])
-			lo=++md;
-		else {
-			char msg[64];
-			sprintf(msg, "Tried to add already-deleted record (%d) to xref (%c)!"
-				, key, set->ctype);
-			FATAL2(msg); /* deleting a deleted record! */
-		}
+
+	lo = find_slot(keynum, set);
+	if ((set->recs)[lo] == keynum) {
+		/* key is already free */
+		char msg[96];
+		if (dups==DUPSOK) 
+			return FALSE;
+		sprintf(msg, "Tried to add already-deleted record (%d) to xref (%c)!"
+			, keynum, set->ctype);
+		FATAL2(msg); /* deleting a deleted record! */
 	}
 	/* key replaces xrefs[lo] - push lo+ up */
 	for (i=set->n-1; i>=lo; --i)
 		(set->recs)[i+1] = (set->recs)[i];
-	(set->recs)[lo] = key;
+	(set->recs)[lo] = keynum;
 	(set->n)++;
 	ASSERT(writexrefs());
 	maxkeynum=-1;
+	return TRUE;
 }
 /*===================================================
- * add?xref -- Wrappers for each type to addxref (qv)
+ * add?xref_impl -- Wrappers for each type to add_xref_to_set (qv)
  *  5 symmetric versions
  *=================================================*/
-void addixref (INT key) { addxref(key, &irecs); }
-void addfxref (INT key) { addxref(key, &frecs); }
-void addsxref (INT key) { addxref(key, &srecs); }
-void addexref (INT key) { addxref(key, &erecs); }
-void addxxref (INT key) { addxref(key, &xrecs); }
+static BOOLEAN addixref_impl (INT key, DUPS dups) { return add_xref_to_set_impl(key, &irecs, dups); }
+static BOOLEAN addfxref_impl (INT key, DUPS dups) { return add_xref_to_set_impl(key, &frecs, dups); }
+static BOOLEAN addsxref_impl (INT key, DUPS dups) { return add_xref_to_set_impl(key, &srecs, dups); }
+static BOOLEAN addexref_impl (INT key, DUPS dups) { return add_xref_to_set_impl(key, &erecs, dups); }
+static BOOLEAN addxxref_impl (INT key, DUPS dups) { return add_xref_to_set_impl(key, &xrecs, dups); }
+/*===================================================
+ * add?xref -- Wrappers for each type to add_xref_to_set (qv)
+ *  5 symmetric versions
+ *=================================================*/
+void addixref (INT key) { addixref_impl(key, NODUPS); }
+void addfxref (INT key) { addfxref_impl(key, NODUPS); }
+void addsxref (INT key) { addsxref_impl(key, NODUPS); }
+void addexref (INT key) { addexref_impl(key, NODUPS); }
+void addxxref (INT key) { addxxref_impl(key, NODUPS); }
+/*===================================================
+ * addxref_impl -- Mark key free (accepts string key, any type)
+ *  key:    [IN]  key to delete (add to free set)
+ *  silent: [IN]  if FALSE, ASSERT if record is already free
+ *=================================================*/
+static BOOLEAN
+addxref_impl (CNSTRING key, DUPS dups)
+{
+	INT keyint = atoi(key + 1);
+	switch(key[0]) {
+	case 'I': return addixref_impl(keyint, dups);
+	case 'F': return addfxref_impl(keyint, dups);
+	case 'S': return addsxref_impl(keyint, dups);
+	case 'E': return addexref_impl(keyint, dups);
+	case 'X': return addxxref_impl(keyint, dups);
+	default: ASSERT(0); return FALSE;
+	}
+}
+/*===================================================
+ * addxref -- Mark key free (accepts string key, any type)
+ * ASSERT if key is already free
+ *=================================================*/
+void addxref (CNSTRING key)
+{
+	addxref_impl(key, NODUPS);
+}
+/*===================================================
+ * addxref_if_missing -- Mark key free (accepts string key, any type)
+ * Does nothing if key already free
+ *=================================================*/
+BOOLEAN addxref_if_missing (CNSTRING key)
+{
+	return addxref_impl(key, DUPSOK);
+}
 /*==========================================
  * growxrefs -- Grow memory for xrefs array.
  *  generic for all types
@@ -353,10 +449,10 @@ static void
 growxrefs (DELETESET set)
 {
 	INT i, m = set->max, *newp;
-	if (set->n < 500)
-		set->max = set->n + 10;
-	else
-		set->max = set->n + 100;
+	if (set->max == 0)
+		set->max = 64;
+	while (set->max <= set->n)
+		set->max = set->max << 1;
 	newp = (INT *) stdalloc((set->max)*sizeof(INT));
 	if (m) {
 		for (i = 0; i < set->n; i++)
@@ -365,12 +461,75 @@ growxrefs (DELETESET set)
 	}
 	set->recs = newp;
 }
+/*==========================================
+ * get_deleteset_from_type -- Return deleteset
+ *  of type specified
+ *========================================*/
+static DELETESET
+get_deleteset_from_type (char ctype)
+{
+	switch(ctype) {
+	case 'I': return &irecs;
+	case 'F': return &frecs;
+	case 'S': return &srecs;
+	case 'E': return &erecs;
+	case 'X': return &xrecs;
+	}
+	ASSERT(0); return 0;
+}
+/*==========================================
+ * delete_xref_if_present -- If record is listed
+ *  as free, remove it from the free list
+ *========================================*/
+BOOLEAN
+delete_xref_if_present (CNSTRING key)
+{
+	DELETESET set=0;
+	INT keynum=0;
+	INT lo=0;
+	INT i=0;
+
+	ASSERT(key && key[0] && key[1]);
+	set = get_deleteset_from_type(key[0]);
+	keynum = atoi(key + 1);
+	ASSERT(keynum>0);
+	lo = find_slot(keynum, set);
+	if ((set->recs)[lo] != keynum)
+		return FALSE;
+	/* removing xrefs[lo] -- move lo+ down */
+	for (i=lo; i+1<set->n-1; ++i)
+		(set->recs)[i] = (set->recs)[i+1];
+	--(set->n);
+	ASSERT(writexrefs());
+	maxkeynum=-1;
+	return TRUE;
+
+}
+/*==========================================
+ * freexref -- Free memory & clear xrefs array
+ *  Called when database is closed
+ *========================================*/
+static void
+freexref (DELETESET set)
+{
+	ASSERT(set);
+	if (set->recs) {
+		stdfree(set->recs);
+		set->recs = 0;
+		set->max = 0;
+		set->n = 0;
+	} else {
+		ASSERT(set->max == 0);
+		ASSERT(set->n == 0);
+	}
+}
 /*==========================================================
  * num_????s -- Return number of type of things in database.
  *  5 symmetric versions
  *========================================================*/
 static INT num_set (DELETESET set)
 {
+	ASSERT(set);
 	return set->recs[0] - set->n;
 }
 INT num_indis (void) { return num_set(&irecs); }
@@ -417,7 +576,7 @@ xref_max_any (void)
  * flag = use the current key
  *  returns static buffer
  *==============================================*/
-STRING
+static STRING
 newxref (STRING xrefp, BOOLEAN flag, DELETESET set)
 {
 	INT keynum;
@@ -438,8 +597,8 @@ newxref (STRING xrefp, BOOLEAN flag, DELETESET set)
 	return(getxref(set));
 }
 /*================================================
- * newixref -- Return original or next ixref value
- * xrefp = key of the individual
+ * new?xref -- Return original or next ?xref value
+ * xrefp = key of the record
  * flag = use the current key
  *==============================================*/
 STRING
@@ -447,41 +606,21 @@ newixref (STRING xrefp, BOOLEAN flag)
 {
 	return newxref(xrefp, flag, &irecs);
 }
-/*================================================
- * newfxref -- Return original or next fxref value
- * xrefp = key of the individual
- * flag = use the current key
- *==============================================*/
 STRING
 newfxref (STRING xrefp, BOOLEAN flag)
 {
 	return newxref(xrefp, flag, &frecs);
 }
-/*================================================
- * newsxref -- Return original or next sxref value
- * xrefp = key of the individual
- * flag = use the current key
- *==============================================*/
 STRING
 newsxref (STRING xrefp, BOOLEAN flag)
 {
 	return newxref(xrefp, flag, &srecs);
 }
-/*================================================
- * newexref -- Return original or next exref value
- * xrefp = key of the individual
- * flag = use the current key
- *==============================================*/
 STRING
 newexref (STRING xrefp, BOOLEAN flag)
 {
 	return newxref(xrefp, flag, &erecs);
 }
-/*================================================
- * newxxref -- Return original or next xxref value
- * xrefp = key of the individual
- * flag = use the current key
- *==============================================*/
 STRING
 newxxref (STRING xrefp, BOOLEAN flag)
 {
@@ -610,3 +749,44 @@ INT xref_lastf (void) { return xref_last(&frecs); }
 INT xref_lasts (void) { return xref_last(&srecs); }
 INT xref_laste (void) { return xref_last(&erecs); }
 INT xref_lastx (void) { return xref_last(&xrecs); }
+/*=======================================
+ * xrefs_get_counts_from_unopened_db --
+ *  read record counts out of file on disk
+ * returns FALSE if specified path is not the root of a traditional lifelines database
+ *=====================================*/
+BOOLEAN
+xrefs_get_counts_from_unopened_db (CNSTRING path, INT *nindis, INT *nfams
+	, INT *nsours, INT *nevens, INT *nothrs)
+{
+	char scratch[100];
+	STRING fmode;
+	FILE * fp = 0;
+	INT i;
+	INT ndels[5], nmax[5];
+
+	ASSERT(!xreffp);
+	sprintf(scratch, "%s/xrefs", path);
+	fmode = LLREADBINARY;
+	if (!(fp = fopen(scratch, fmode))) {
+		return FALSE;
+	}
+	for (i=0; i<5; ++i) {
+		ASSERT(fread(&ndels[i], sizeof(INT), 1, fp) == 1);
+	}
+	for (i=0; i<5; ++i) {
+		INT j;
+		for (j=0; j<ndels[i]; ++j) {
+			INT k;
+			ASSERT(fread(&k, sizeof(INT), 1, fp) == 1);
+			if (!j)
+				nmax[i] = k;
+		}
+	}
+	*nindis = nmax[0] - ndels[0];
+	*nfams = nmax[1] - ndels[1];
+	*nevens = nmax[2] - ndels[2];
+	*nsours = nmax[3] - ndels[3];
+	*nothrs = nmax[4] - ndels[4];
+	fclose(fp);
+	return TRUE;
+}
