@@ -44,7 +44,11 @@
 #define NAMESEP '/'
 #define MAXGEDNAMELEN 512
 /* keys run up to 9,999,999, eg I9999999 */
+/* These are GEDCOM identifier numbers, eg, for @I510@,
+ the key is 510 */
+/* btree databases use this as part of the database key */
 #define MAXKEYWIDTH 8
+#define MAXKEYNUMBER 9999999
 
 #define OKAY  1
 #define ERROR 0
@@ -66,9 +70,12 @@
  its NODE tree). (E.g., its parent might be a NODE representing
  "1 BIRT".)
 */
+typedef struct tag_cacheel *CACHEEL;
 typedef struct tag_node *NODE;
-typedef struct tag_record *RECORD;
 struct tag_node {
+	/* a NODE is an OBJECT */
+	struct tag_vtable * vtable; /* generic object table (see vtable.h) */
+	int    n_refcnt;    /* refcount for temp nodes */
 	STRING n_xref;      /* cross ref */
 	STRING n_tag;       /* tag */
 	STRING n_val;       /* value */
@@ -76,8 +83,7 @@ struct tag_node {
 	NODE   n_child;     /* first child */
 	NODE   n_sibling;   /* sibling */
 	INT    n_flag;      /* eg, ND_TEMP */
-	int    n_refcnt;    /* refcount for temp nodes */
-	RECORD n_rec;       /* pointer to record, if in a record's node tree */
+	CACHEEL n_cel;      /* pointer to cacheel, if node is inside cache */
 };
 #define nxref(n)    ((n)->n_xref)
 #define ntag(n)     ((n)->n_tag)
@@ -89,7 +95,7 @@ struct tag_node {
 #define nrefcnt(n)  ((n)->n_refcnt)
 enum { ND_TEMP=1 };
 
-struct tag_nkey { char ntype; INT keynum; STRING key; };
+struct tag_nkey { char ntype; INT keynum; char key[MAXKEYWIDTH+1]; };
 typedef struct tag_nkey NKEY;
 
 /*=====================================
@@ -104,16 +110,19 @@ typedef struct tag_nkey NKEY;
  LifeLines is very RECORD-oriented.
 */
 
-struct tag_record { /* RECORD */
-	NODE top;
-	NKEY nkey;
-	WAREHOUSE mdwh; /* metadata */
-	struct tag_cacheel * cel; /* cache wrapper */
-};
-NODE nztop(RECORD); /* function so it can handle NULL input */
-#define nzkey(n)    ((n)->nkey.key)
-#define nzkeynum(n) ((n)->nkey.keynum)
-#define nztype(n)   ((n)->nkey.ntype)
+typedef struct tag_record *RECORD;
+
+NODE nztop(RECORD rec); /* handles NULL, also reloads from cache */
+CNSTRING nzkey(RECORD rec);
+INT nzkeynum(RECORD rec);
+char nztype(RECORD rec);
+CACHEEL nzcel(RECORD rec);
+
+/*=====================================
+ * LLDATABASE types -- LifeLines database
+ *===================================*/
+
+typedef struct tag_lldatabase *LLDATABASE;
 
 /*
  reformating functions - format date or place for display
@@ -178,8 +187,10 @@ extern BOOLEAN readonly;
 extern BOOLEAN immutable;
 extern STRING editstr;
 extern STRING editfile;
+/* tabtable & placabbvs should be moved into LLDATABASE */
 extern TABLE tagtable;		/* table for GEDCOM tags */
 extern TABLE placabbvs;		/* table for place abbrvs */
+extern LLDATABASE def_lldb;        /* default database */
 
 
 /*=====================
@@ -193,13 +204,16 @@ typedef enum { SURFIRST, REGORDER } SURORDER;
  *===================*/
 
 STRING addat(STRING);
-void addixref(INT);
-void addexref(INT);
-void addfxref(INT);
-void addsxref(INT);
-void addxxref(INT);
-BOOLEAN add_refn(STRING, STRING);
+void addixref(INT key);
+void addexref(INT key);
+void addfxref(INT key);
+void addref_record(RECORD rec);
+void addsxref(INT key);
+void addxref(CNSTRING key);
+void addxxref(INT key);
+BOOLEAN add_refn(CNSTRING refn, CNSTRING key);
 BOOLEAN are_locales_supported(void);
+void change_node_tag(NODE node, STRING newtag);
 RECORD choose_child(RECORD irec, RECORD frec, STRING msg0, STRING msgn, ASK1Q ask1);
 void choose_and_remove_family(void);
 RECORD choose_father(RECORD irec, RECORD frec, STRING msg0, STRING msgn, ASK1Q ask1);
@@ -212,15 +226,13 @@ RECORD choose_spouse(RECORD irec, STRING msg0, STRING msgn);
 void classify_nodes(NODE*, NODE*, NODE*);
 void closexref(void);
 void close_lifelines(void);
-void close_lldb(void);
 NODE convert_first_fp_to_node(FILE*, BOOLEAN, XLAT, STRING*, BOOLEAN*);
 NODE copy_node(NODE);
-NODE copy_nodes(NODE, BOOLEAN, BOOLEAN);
+NODE copy_node_subtree(NODE node);
+NODE copy_nodes(NODE node, BOOLEAN kids, BOOLEAN sibs);
 BOOLEAN create_database(STRING dbused);
-RECORD create_record(NODE node);
 NODE create_node(STRING, STRING, STRING, NODE);
 NODE create_temp_node(STRING, STRING, STRING, NODE);
-void del_in_dbase(STRING key);
 void delete_metarec(STRING key);
 BOOLEAN edit_mapping(INT);
 BOOLEAN edit_valtab_from_db(STRING, TABLE*, INT sep, STRING, STRING (*validator)(TABLE tab));
@@ -241,12 +253,12 @@ NODE fam_to_first_chil(NODE);
 NODE fam_to_last_chil(NODE);
 STRING fam_to_list_string(NODE fam, INT len, STRING delim);
 NODE fam_to_spouse(NODE, NODE);
+int next_spouse(NODE *node, RECORD *spouse);
 RECORD file_to_record(STRING fname, XLAT ttm, STRING *pmsg, BOOLEAN *pemp);
 NODE file_to_node(STRING, XLAT, STRING*, BOOLEAN*);
 INT file_to_line(FILE*, XLAT, INT*, STRING*, STRING*, STRING*, STRING*);
 NODE find_node(NODE, STRING, STRING, NODE*);
 NODE find_tag(NODE, CNSTRING);
-void free_rec(RECORD);
 void free_node(NODE);
 void free_nodes(NODE);
 void free_temp_node_tree(NODE);
@@ -255,7 +267,6 @@ ZSTR get_cache_stats_fam(void);
 ZSTR get_cache_stats_indi(void);
 STRING get_current_locale_collate(void);
 STRING get_current_locale_msgs(void);
-INT get_dblist(STRING path, LIST * dblist, LIST * dbdesclist);
 INT get_decimal(STRING);
 INT get_hexidecimal(STRING);
 STRING get_lifelines_version(INT maxlen);
@@ -268,17 +279,14 @@ STRING getfxref(void);
 INT getixrefnum(void);
 STRING getsxref(void);
 STRING getxxref(void);
-BOOLEAN getrefnrec(STRING);
+BOOLEAN getrefnrec(CNSTRING refn);
 void growexrefs(void);
 void growfxrefs(void);
 void growixrefs(void);
 void growsxrefs(void);
 void growxxrefs(void);
 INT hexvalue(INT);
-BOOLEAN in_string(INT, STRING);
 void index_by_refn(NODE, STRING);
-void indi_to_cache(RECORD rec);
-void indi_to_cache_old(NODE);
 void indi_to_dbase(NODE);
 STRING indi_to_event(NODE, STRING tag, STRING head, INT len, RFMT);
 NODE indi_to_famc(NODE);
@@ -299,14 +307,14 @@ void init_disp_reformat(void);
 BOOLEAN init_lifelines_db(void);
 BOOLEAN init_lifelines_global(STRING configfile, STRING * pmsg, void (*notify)(STRING db, BOOLEAN opening));
 CNSTRING init_get_config_file(void);
-void init_new_record(RECORD rec, char ntype, INT keynum);
+void init_new_record(RECORD rec, CNSTRING key);
 BOOLEAN init_valtab_from_file(STRING, TABLE, XLAT, INT sep, STRING*);
 BOOLEAN init_valtab_from_rec(CNSTRING, TABLE, INT sep, STRING*);
 BOOLEAN init_valtab_from_string(CNSTRING, TABLE, INT sep, STRING*);
-BOOLEAN is_db_open(void);
 BOOLEAN is_codeset_utf8(CNSTRING codeset);
 BOOLEAN is_iconv_supported(void);
 BOOLEAN is_nls_supported(void);
+BOOLEAN is_record_missing_data_entry(CNSTRING key);
 BOOLEAN is_temp_node(NODE);
 BOOLEAN iso_list(NODE, NODE);
 BOOLEAN iso_nodes(NODE, NODE, BOOLEAN, BOOLEAN);
@@ -369,6 +377,10 @@ BOOLEAN node_to_nkey(NODE node, NKEY * nkey);
 NODE node_to_node(NODE, INT*);
 RECORD node_to_record(NODE node);
 STRING node_to_tag(NODE node, STRING tag, INT len);
+void normalize_fam(NODE fam);
+void normalize_irec(RECORD irec);
+void normalize_indi(NODE indi);
+void normalize_rec(RECORD rec);
 INT num_evens(void);
 INT num_fam_xrefs(NODE fam);
 INT num_fams(void);
@@ -381,7 +393,6 @@ BOOLEAN openxref(BOOLEAN readonly);
 STRING other_to_list_string(NODE node, INT len, STRING delim);
 void othr_to_cache(NODE);
 void othr_to_dbase(NODE);
-BOOLEAN place_to_list(STRING, LIST, INT*);
 BOOLEAN pointer_value(STRING);
 NODE qkey_to_even(CNSTRING key);
 RECORD qkey_to_erecord(CNSTRING key);
@@ -403,14 +414,17 @@ NODE record_to_first_event(RECORD record, CNSTRING tag);
 NODE refn_to_record(STRING, INT);
 void register_uicodeset_callback(CALLBACK_FNC fncptr, VPTR uparm);
 void register_uilang_callback(CALLBACK_FNC fncptr, VPTR uparm);
-void release_dblist(LIST dblist);
+void release_record(RECORD rec);
 BOOLEAN remove_child(NODE indi, NODE fam);
 BOOLEAN remove_empty_fam(NODE);
-void remove_indi(NODE);
-void remove_indi_cache(STRING);
-void remove_fam_cache(STRING);
+BOOLEAN remove_any_record(RECORD record);
+void remove_indi_by_root(NODE);
+void remove_indi_cache(STRING key);
+void remove_fam_cache(STRING key);
+void remove_from_cache_by_key(CNSTRING key);
+BOOLEAN remove_fam_record(RECORD frec);
 void remove_from_browse_lists(STRING);
-BOOLEAN remove_refn(STRING, STRING);
+BOOLEAN remove_refn(CNSTRING refn, CNSTRING key);
 void rename_from_browse_lists(STRING);
 BOOLEAN remove_spouse(NODE indi, NODE fam);
 void replace_fam(NODE fam1, NODE fam2);
@@ -438,12 +452,9 @@ void split_fam(NODE, NODE*, NODE*, NODE*, NODE*, NODE*);
 void split_indi_old(NODE, NODE*, NODE*, NODE*, NODE*, NODE*, NODE*);
 void split_othr(NODE node, NODE *prefn, NODE *prest);
 BOOLEAN store_file_to_db(STRING key, STRING file);
-BOOLEAN store_record(STRING key, STRING rec, INT len);
-BOOLEAN store_text_file_to_db(STRING key, CNSTRING file, TRANSLFNC);
+BOOLEAN store_record(CNSTRING key, STRING rec, INT len);
 RECORD string_to_record(STRING str, CNSTRING key, INT len);
 void termlocale(void);
-void traverse_db_key_recs(BOOLEAN(*func)(CNSTRING key, RECORD, void *param), void *param);
-void traverse_db_rec_keys(CNSTRING lo, CNSTRING hi, BOOLEAN(*func)(CNSTRING key, STRING data, INT, void *param), void * param);
 BOOLEAN traverse_nodes(NODE node, BOOLEAN (*func)(NODE, VPTR), VPTR param);
 void traverse_refns(BOOLEAN(*func)(CNSTRING key, CNSTRING refn, BOOLEAN newset, void *param), void *param);
 INT tree_strlen(INT, NODE);
@@ -463,10 +474,91 @@ BOOLEAN valid_to_list(STRING, LIST, INT*, STRING);
 BOOLEAN valid_even_tree(NODE, STRING*, NODE);
 BOOLEAN valid_othr_tree(NODE, STRING*, NODE);
 INT val_to_sex(NODE);
-BOOLEAN value_to_list(STRING, LIST, INT*, STRING);
 STRING value_to_xref(STRING);
 BOOLEAN writexrefs(void);
 void write_node_to_editfile(NODE); /* used by Ethel */
+
+/* dblist.c */
+INT get_dblist(STRING path, LIST * dblist, LIST * dbdesclist);
+void release_dblist(LIST dblist);
+
+/* interface.c */
+void del_in_dbase(CNSTRING key);
+void delete_record_missing_data_entry(CNSTRING key);
+BOOLEAN mark_deleted_record_as_deleted(CNSTRING key);
+BOOLEAN mark_live_record_as_live(CNSTRING key);
+BOOLEAN store_text_file_to_db(STRING key, CNSTRING file, TRANSLFNC);
+void traverse_db_rec_keys(CNSTRING lo, CNSTRING hi, BOOLEAN(*func)(CNSTRING key, STRING data, INT, void *param), void * param);
+void traverse_db_key_recs(BOOLEAN(*func)(CNSTRING key, RECORD, void *param), void *param);
+
+/* keytonod.c */
+void add_new_indi_to_cache(RECORD rec);
+RECORD get_record_for_cel(CACHEEL cel);
+
+/* gstrings.c */
+STRING generic_to_list_string(NODE node, STRING key, INT len, STRING delim, RFMT rfmt, BOOLEAN appkey);
+STRING *get_child_strings(NODE, RFMT, INT*, STRING**);
+STRING indi_to_list_string(NODE indi, NODE fam, INT len, RFMT rfmt, BOOLEAN appkey);
+
+/* lldatabase.c */
+void lldb_adderror(LLDATABASE lldb, int errnum, CNSTRING errstr);
+LLDATABASE lldb_alloc(void);
+void lldb_close(LLDATABASE *plldb);
+void lldb_set_btree(LLDATABASE lldb, void * btree);
+
+/* misc.c */
+INT xrefval(char ntype, STRING str);
+
+/* names.c */
+void add_name(CNSTRING name, CNSTRING key);
+LIST find_indis_by_name(CNSTRING name);
+CNSTRING getasurname(CNSTRING);
+CNSTRING getsxsurname(CNSTRING);
+CNSTRING givens(CNSTRING);
+STRING manip_name(STRING name, SURCAPTYPE captype, SURORDER surorder, INT len);
+LIST name_to_list(CNSTRING name, INT *plen, INT *psind);
+STRING name_string(STRING);
+int namecmp(STRING, STRING);
+void remove_name(STRING name, CNSTRING key);
+void traverse_names(BOOLEAN(*func)(CNSTRING key, CNSTRING name, BOOLEAN newset, void *param), void *param);
+STRING trim_name(STRING, INT);
+
+/* node.c */
+/* node tree iterator */
+struct tag_node_iter {
+	NODE start;
+	NODE next;
+};
+typedef struct tag_node_iter * NODE_ITER;
+void begin_node_it(NODE node, NODE_ITER nodeit);
+INT fam_to_husb(RECORD frec, RECORD * prec);
+NODE fam_to_husb_node(NODE);
+INT fam_to_wife(RECORD frec, RECORD * prec);
+NODE fam_to_wife_node(NODE);
+NODE next_node_it_ptr(NODE_ITER nodeit);
+
+/* nodeio.c */
+STRING node_to_string(NODE);
+NODE string_to_node(STRING);
+void write_indi_to_file_for_edit(NODE indi, CNSTRING file, RFMT rfmt);
+void write_fam_to_file_for_edit(NODE fam, CNSTRING file, RFMT rfmt);
+void write_nodes(INT, FILE*, XLAT, NODE, BOOLEAN, BOOLEAN, BOOLEAN);
+
+/* place.c */
+LIST place_to_list(STRING, INT*);
+LIST value_to_list (STRING str, INT *plen, STRING dlm);
+
+/* refns.c */
+void annotate_with_supplemental(NODE node, RFMT rfmt);
+
+/* soundex.c */
+CNSTRING trad_soundex(CNSTRING);
+INT soundex_count(void);
+CNSTRING soundex_get(INT i, CNSTRING name);
+
+/* xreffile.c */
+BOOLEAN addxref_if_missing (CNSTRING key);
+BOOLEAN delete_xref_if_present(CNSTRING key);
 INT xref_firste(void);
 INT xref_firstf(void);
 INT xref_firsti(void);
@@ -495,57 +587,6 @@ INT xref_prevf(INT);
 INT xref_previ(INT);
 INT xref_prevs(INT);
 INT xref_prevx(INT);
-INT xrefval(char ntype, STRING str);
-
-/* gstrings.c */
-STRING generic_to_list_string(NODE node, STRING key, INT len, STRING delim, RFMT rfmt, BOOLEAN appkey);
-STRING *get_child_strings(NODE, RFMT, INT*, STRING**);
-STRING indi_to_list_string(NODE indi, NODE fam, INT len, RFMT rfmt, BOOLEAN appkey);
-
-/* names.c */
-void add_name(STRING, STRING);
-void free_name_list(LIST list);
-void free_string_list(LIST list);
-LIST find_indis_by_name(CNSTRING name);
-CNSTRING getasurname(CNSTRING);
-CNSTRING getsxsurname(CNSTRING);
-CNSTRING givens(CNSTRING);
-STRING manip_name(STRING name, SURCAPTYPE captype, SURORDER surorder, INT len);
-BOOLEAN name_to_list(CNSTRING, LIST, INT*, INT*);
-STRING name_string(STRING);
-int namecmp(STRING, STRING);
-void remove_name(STRING, STRING);
-void traverse_names(BOOLEAN(*func)(STRING key, CNSTRING name, BOOLEAN newset, void *param), void *param);
-STRING trim_name(STRING, INT);
-
-/* node.c */
-/* node tree iterator */
-struct tag_node_iter {
-	NODE start;
-	NODE next;
-};
-typedef struct tag_node_iter * NODE_ITER;
-void begin_node_it(NODE node, NODE_ITER nodeit);
-INT fam_to_husb(RECORD frec, RECORD * prec);
-NODE fam_to_husb_node(NODE);
-INT fam_to_wife(RECORD frec, RECORD * prec);
-NODE fam_to_wife_node(NODE);
-NODE next_node_it_ptr(NODE_ITER nodeit);
-
-/* nodeio.c */
-STRING node_to_string(NODE);
-NODE string_to_node(STRING);
-void write_indi_to_file_for_edit(NODE indi, CNSTRING file, RFMT rfmt);
-void write_fam_to_file(NODE fam, CNSTRING file);
-void write_nodes(INT, FILE*, XLAT, NODE, BOOLEAN, BOOLEAN, BOOLEAN);
-
-/* refns.c */
-void annotate_with_supplemental(NODE node, RFMT rfmt);
-
-/* soundex.c */
-CNSTRING trad_soundex(CNSTRING);
-INT soundex_count(void);
-CNSTRING soundex_get(INT i, CNSTRING name);
 
 /*******************
  various Macros
@@ -597,123 +638,197 @@ CNSTRING soundex_get(INT i, CNSTRING name);
 #define num_children(fam)  (length_nodes(CHIL(fam)))
 
 /*
-TODO: Change all FORCHILDRENx to FORCHILDREN loops
-This means changing use of NODE to RECORD
-Perry, 2002.06.24
+  Possibly all of these should lock their main argument in the cache
+  Instead of having some of the client code do it, and some not
+  Now that NODE has a pointer to its parent RECORD,
+  and RECORD has a pointer to the cacheel, it is possible 
 */
+/*
+ FORCHILDRENx takes a node as its first arg
+ FORCHILDREN takes a record as its first arg
+*/
+
 #define FORCHILDRENx(fam,child,num) \
 	{\
 	NODE __node = find_tag(nchild(fam), "CHIL");\
+	RECORD irec=0;\
 	NODE child=0;\
 	STRING __key=0;\
 	num = 0;\
 	while (__node) {\
+		if (!eqstr(ntag(__node), "CHIL")) break;\
 		__key = rmvat(nval(__node));\
-		if (!__key || !(child = key_to_indi(__key))) {\
-			++num;\
-			__node = nsibling(__node);\
+		__node = nsibling(__node);\
+		++num;\
+		if (!__key || !(irec=qkey_to_irecord(__key)) || !(child=nztop(irec))) {\
 			continue;\
 		}\
-		++num;\
 		{
 
 #define ENDCHILDRENx \
 		}\
-		__node = nsibling(__node);\
-		if (__node && nestr(ntag(__node), "CHIL")) __node = NULL;\
+		release_record(irec);\
 	}}
 
 #define FORCHILDREN(fam,child,num) \
 	{\
 	NODE __node = find_tag(nchild(fam), "CHIL");\
-	RECORD child=0;\
+	RECORD child=0, irec=0;\
 	STRING __key=0;\
 	num = 0;\
 	while (__node) {\
+		if (!eqstr(ntag(__node), "CHIL")) break;\
 		__key = rmvat(nval(__node));\
+		__node = nsibling(__node);\
+		++num;\
 		if (!__key || !(child = key_to_irecord(__key))) {\
-			++num;\
-			__node = nsibling(__node);\
 			continue;\
 		}\
-		++num;\
+		irec=child;\
 		{
 
 #define ENDCHILDREN \
 		}\
-		__node = nsibling(__node);\
-		if (__node && nestr(ntag(__node), "CHIL")) __node = NULL;\
+		release_record(irec);\
 	}}
 
+/* FORSPOUSES iterate over all FAMS nodes & all spouses of indi
+ * if there are multiple spouses, user will see same fam with multiple spouses
+ * if there are no spouses for a particular family the family is not returned.
+ */
 #define FORSPOUSES(indi,spouse,fam,num) \
 	{\
-	NODE __node = FAMS(indi);\
-	INT __sex = SEX(indi);\
-	NODE spouse=0,fam=0;\
+	NODE __node = find_tag(nchild(indi),"FAMS");\
+	NODE __node1=0, fam=0;\
 	STRING __key=0;\
 	num = 0;\
 	while (__node) {\
-		__key = rmvat(nval(__node));\
-		if (!__key || !(fam = qkey_to_fam(__key))) {\
-			++num;\
-			__node = nsibling(__node);\
+		if (!eqstr(ntag(__node), "FAMS")) break;\
+	    __key = rmvat(nval(__node));\
+	    __node = nsibling(__node);\
+	    if (!__key || !(fam = qkey_to_fam(__key))) {\
 			continue;\
 		}\
-		if (__sex == SEX_MALE)\
-			spouse = fam_to_wife_node(fam);\
-		else if (__sex == SEX_FEMALE)\
-			spouse = fam_to_husb_node(fam);\
-		else    spouse = fam_to_spouse(fam, indi);\
-		if (spouse != NULL) {\
+		__node1 = nchild(fam);\
+		/* Now loop through fam node tree looking for spouses */\
+		while (__node1) {\
+			NODE spouse=0;\
+			INT __hits=0;\
+			if (eqstr(ntag(__node1), "HUSB")||eqstr(ntag(__node1), "WIFE")) ++__hits;\
+			else if (__hits)\
+				/* Its not HUSB or WIFE, and we've seen a HUSB or WIFE before */ \
+				/* So we must be out of the HUSB & WIFE section of the node tree */ \
+				break;\
+			__key = rmvat(nval(__node1));\
+			__node1 = nsibling(__node1);\
+			if (!__hits || !__key || !(spouse = qkey_to_indi(__key))||spouse==indi){\
+				/* If it was a valid HUSB or WIFE tag, count it */\
+				if (__hits && spouse!=indi) ++num;\
+				continue;\
+			}\
+			/* It is a valid HUSB or WIFE, not self, has key, and in database */\
 			++num;\
 			{
 
 #define ENDSPOUSES \
-		}}\
-		__node = nsibling(__node);\
-		if (__node && nestr(ntag(__node), "FAMS")) __node = NULL;\
+			}\
+	    }\
 	}}
 
-#define FORFAMSS(indi,fam,spouse,num) \
+/* FORFAMS iterate over all FAMS nodes of indi
+ *    this is an optimization of FORFAMSS for cases where spouse is not used
+ *    or computed in other ways.
+ */
+#define FORFAMS(indi,fam,num) \
 	{\
-	NODE __node = FAMS(indi);\
-	INT __sex = SEX(indi);\
-	NODE fam=0, spouse=0;\
+	RECORD frec=0; \
+	NODE __node = find_tag(nchild(indi),"FAMS");\
+	NODE fam=0;\
 	STRING __key=0;\
 	num = 0;\
 	while (__node) {\
+		if (!eqstr(ntag(__node), "FAMS")) break;\
 		__key = rmvat(nval(__node));\
-		if (!__key || !(fam = qkey_to_fam(__key))) {\
-			++num;\
-			__node = nsibling(__node);\
+		__node = nsibling(__node);\
+		++num;\
+	    if (!__key || !(frec=qkey_to_frecord(__key)) || !(fam=nztop(frec))) {\
 			continue;\
 		}\
-		if (__sex == SEX_MALE)\
-			spouse = fam_to_wife_node(fam);\
-		else if (__sex == SEX_FEMALE)\
-			spouse = fam_to_husb_node(fam);\
-		else \
-			spouse = fam_to_spouse(fam, indi);\
-		++num;\
 		{
 
-#define ENDFAMSS \
+#define ENDFAMS \
 		}\
-		__node = nsibling(__node);\
-		if (__node && nestr(ntag(__node), "FAMS")) __node = NULL;\
+		release_record(frec); \
 	}}
 
+/* FORFAMSS iterate over all FAMS nodes & all spouses of indi
+ * if there are multiple spouses, user will see same fam with multiple spouses
+ * if there are no spouses for a particular family NULL is returned for spouse
+ */
+#define FORFAMSS(indi,fam,spouse,num) \
+	{\
+	INT first_sp=0; /* have reported spouse in current family? */\
+	RECORD frec=0; \
+	NODE __node = find_tag(nchild(indi),"FAMS");\
+	NODE __node1=0, fam=0;\
+	STRING __key=0;\
+	num = 0;\
+	while (__node) {\
+		if (!eqstr(ntag(__node), "FAMS")) break;\
+	    __key = rmvat(nval(__node));\
+	    __node = nsibling(__node);\
+	    if (__key && (frec=qkey_to_frecord(__key)) && (fam=nztop(frec))) {\
+			first_sp=0; /* not yet reported this family */\
+			__node1 = nchild(fam);\
+			while (__node1 || !first_sp) {\
+				NODE spouse=0;\
+				if (__node1) { \
+				    if (eqstr("HUSB", ntag(__node1)) || eqstr("WIFE", ntag(__node1))) { \
+					__key = rmvat(nval(__node1));\
+					__node1 = nsibling(__node1);\
+					if (!__key || !(spouse = qkey_to_indi(__key))||spouse==indi) {\
+						    /* invalid spouse, or self */\
+						    continue;\
+					}\
+				    } else {\
+					    /* Not a spouse (not HUSB or WIFE) */\
+					    if (first_sp)\
+						    break; /* finished HUSB/WIFE section & reported already */\
+						__node1 = nsibling(__node1);\
+					    if (__node1) continue;\
+						/* fall through here for end of family not yet reported */\
+					}\
+				/* } else { \
+				 * !__node1 && !first_sp fall thru to report * fam*/ \
+				}\
+				++num; \
+				++first_sp; /* reporting this family */\
+				{
+
+#define ENDFAMSS \
+				}\
+			}\
+			release_record(frec); \
+		}\
+	}}
+
+/* FORFAMCS iterate over all parent families of indi
+ * Up to one father and mother are given for each
+ * (This ignores non-traditional parents)
+ */
 #define FORFAMCS(indi,fam,fath,moth,num) \
 	{\
-	NODE __node = FAMC(indi);\
+	RECORD frec=0; \
+	NODE __node = find_tag(nchild(indi),"FAMC");\
 	NODE fam, fath, moth;\
 	STRING __key=0;\
 	num = 0;\
 	while (__node) {\
+		if (!eqstr(ntag(__node), "FAMC")) break;\
 		__key = rmvat(nval(__node));\
-		if (!__key || !(fam = qkey_to_fam(__key))) {\
-			 ++num;\
-			 __node = nsibling(__node);\
+		 __node = nsibling(__node);\
+		 ++num;\
+	    if (!__key || !(frec=qkey_to_frecord(__key)) || !(fam=nztop(frec))) {\
 			 continue;\
 		}\
 		fath = fam_to_husb_node(fam);\
@@ -723,10 +838,12 @@ Perry, 2002.06.24
 
 #define ENDFAMCS \
 		}\
-		__node = nsibling(__node);\
-		if (__node && nestr(ntag(__node), "FAMC")) __node = NULL;\
+		release_record(frec); \
 	}}
 
+/* FORHUSBS iterate over all husbands in one family
+ * (This handles more than one husband in a family)
+ */
 #define FORHUSBS(fam,husb,num) \
 	{\
 	NODE __node = find_tag(nchild(fam), "HUSB");\
@@ -749,6 +866,9 @@ Perry, 2002.06.24
 		if (__node && nestr(ntag(__node), "HUSB")) __node = NULL;\
 	}}
 
+/* FORWIFES iterate over all wives in one family
+ * (This handles more than one wife in a family)
+ */
 #define FORWIFES(fam,wife,num) \
 	{\
 	NODE __node = find_tag(nchild(fam), "WIFE");\
@@ -757,9 +877,10 @@ Perry, 2002.06.24
 	num = 0;\
 	while (__node) {\
 		__key = rmvat(nval(__node));\
-		if (!__key || !(wife = key_to_indi(__key))) {\
+		if (!__key || !(wife = qkey_to_indi(__key))) {\
 			++num;\
 			__node = nsibling(__node);\
+			if (__node && nestr(ntag(__node), "WIFE")) __node = NULL;\
 			continue;\
 		}\
 		ASSERT(wife);\
@@ -772,6 +893,34 @@ Perry, 2002.06.24
 		if (__node && nestr(ntag(__node), "WIFE")) __node = NULL;\
 	}}
 
+/* FORFAMSPOUSES iterate over all spouses in one family
+ * (All husbands and wives)
+ */
+#define FORFAMSPOUSES(fam,spouse,num) \
+	{\
+	NODE __node = nchild(fam);\
+	NODE spouse=0;\
+	STRING __key=0;\
+	num = 0;\
+	while (__node) {\
+		if (!eqstr(ntag(__node), "HUSB") && !eqstr(ntag(__node), "WIFE")) {\
+			__node = nsibling(__node);\
+			continue;\
+		}\
+		__key = rmvat(nval(__node));\
+		if (!__key || !(spouse = qkey_to_indi(__key))) {\
+			++num;\
+			__node = nsibling(__node);\
+			continue;\
+		}\
+		++num;\
+		{
+
+#define ENDFAMSPOUSES \
+		}\
+		__node = nsibling(__node);\
+	}}
+
 #define FORTAGVALUES(root,tag,node,value)\
 	{\
 	NODE node, __node = nchild(root);\
