@@ -44,6 +44,7 @@
 #include "liflines.h"
 #include "codesets.h"
 #include "zstr.h"
+#include "vtable.h"
 
 /*********************************************
  * global/exported variables
@@ -68,6 +69,8 @@ STRING nonnodx     = 0;
 STRING nonvar1     = 0;
 STRING nonvarx     = 0;
 STRING nontabx     = 0;
+STRING nonset1     = 0;
+STRING nonsetx     = 0;
 STRING nonboox     = 0;
 STRING nonlst1     = 0;
 STRING nonlstx     = 0;
@@ -104,6 +107,7 @@ static void clear_pnode(PNODE node);
 static PNODE create_pnode(PACTX pactx, INT type);
 static void delete_pnode(PNODE node);
 static void free_pnode_memory(PNODE node);
+static void rptinfo_destructor(VTABLE *obj);
 static void set_parents(PNODE body, PNODE node);
 static void verify_builtins(void);
 
@@ -117,6 +121,17 @@ static STRING interp_locale = 0;
 static INT live_pnodes = 0;
 static TABLE f_rptinfos=0;
 
+static struct tag_vtable vtable_for_rptinfo = {
+	VTABLE_MAGIC
+	, "rptinfo"
+	, &rptinfo_destructor
+	, &refcountable_isref
+	, &refcountable_addref
+	, &refcountable_release
+	, 0 /* copy_fnc */
+	, &generic_get_type_name
+};
+
 /*********************************************
  * local & exported function definitions
  * body of module
@@ -183,7 +198,6 @@ void
 free_all_pnodes (void)
 {
 	PN_BLOCK block;
-	/* live_count is how many were leaked */
 	while ((block = block_list)) {
 		PN_BLOCK next = block->next;
 		free_list = 0;
@@ -192,7 +206,6 @@ free_all_pnodes (void)
 			for (i=0; i<(INT)BLOCK_NODES; i++) {
 				PNODE node1=&block->nodes[i];
 				if (itype(node1) != IFREED) {
-					/* leaked */
 					delete_pnode(node1);
 				}
 			}
@@ -227,11 +240,77 @@ create_pnode (PACTX pactx, INT type)
 static void
 clear_pnode (PNODE node)
 {
-	node=node; /* unused */
-	/*
-	just points to string inside record, which is
-	in the cache, so nothing to free here.
-	*/
+	if (node->i_flags & PN_IVALUEX_PVALUE) {
+		PVALUE val = ivaluex(node);
+		if (val) {
+			delete_pvalue(val);
+			ivaluex(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_INAME_HSTR) {
+		STRING str = iname(node);
+		if (str) {
+			stdfree(str);
+			iname(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_ICHILD_HPTR) {
+		STRING str = ichild(node);
+		if (str) {
+			stdfree(str);
+			ichild(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_INUM_HPTR) {
+		STRING str = inum(node);
+		if (str) {
+			stdfree(str);
+			inum(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_ISPOUSE_HPTR) {
+		STRING str = ispouse(node);
+		if (str) {
+			stdfree(str);
+			ispouse(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_IFAMILY_HPTR) {
+		STRING str = ifamily(node);
+		if (str) {
+			stdfree(str);
+			ifamily(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_IELEMENT_HPTR) {
+		STRING str = ielement(node);
+		if (str) {
+			stdfree(str);
+			ielement(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_IPARENT_HPTR) {
+		STRING str = iiparent(node);
+		if (str) {
+			stdfree(str);
+			iiparent(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_IVALVAR_HPTR) {
+		STRING str = ivalvar(node);
+		if (str) {
+			stdfree(str);
+			ivalvar(node) = 0;
+		}
+	}
+	if (node->i_flags & PN_IIDENT_HSTR) {
+		STRING str = iident(node);
+		if (str) {
+			stdfree(str);
+			iident(node) = 0;
+		}
+	}
+
 }
 /*==================================
  * delete_pnode -- Create PNODE node
@@ -254,6 +333,7 @@ string_node (PACTX pactx, STRING str)
 	PNODE node = create_pnode(pactx, ISCONS);
 	ASSERT(str); /* we're not converting NULL to "" because nobody passes us NULL */
 	ivaluex(node) = create_pvalue_from_string(str);
+	node->i_flags = PN_IVALUEX_PVALUE;
 	return node;
 }
 /*========================================
@@ -272,6 +352,7 @@ children_node (PACTX pactx, PNODE fexpr, STRING cvar, STRING nvar, PNODE body)
 	ichild(node) = (VPTR) cvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_ICHILD_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -293,6 +374,7 @@ spouses_node (PACTX pactx, PNODE pexpr, STRING svar, STRING fvar, STRING nvar, P
 	ifamily(node) = (VPTR) fvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_ISPOUSE_HPTR + PN_IFAMILY_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -314,6 +396,7 @@ families_node (PACTX pactx, PNODE pexpr, STRING fvar, STRING svar, STRING nvar,
 	ispouse(node) = (VPTR) svar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IFAMILY_HPTR + PN_ISPOUSE_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -335,6 +418,7 @@ fathers_node (PACTX pactx, PNODE pexpr, STRING pvar, STRING fvar, STRING nvar, P
 	ifamily(node) = (VPTR) fvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IPARENT_HPTR + PN_IFAMILY_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -356,6 +440,7 @@ mothers_node (PACTX pactx, PNODE pexpr, STRING pvar, STRING fvar, STRING nvar, P
 	ifamily(node) = (VPTR) fvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IPARENT_HPTR + PN_IFAMILY_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -375,6 +460,7 @@ parents_node (PACTX pactx, PNODE pexpr, STRING fvar, STRING nvar, PNODE body)
 	ifamily(node) = (VPTR) fvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IFAMILY_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -396,6 +482,7 @@ forindiset_node (PACTX pactx, PNODE iexpr, STRING ivar, STRING vvar, STRING nvar
 	ivalvar(node) = (VPTR) vvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_IVALVAR_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -415,6 +502,7 @@ forlist_node (PACTX pactx, PNODE iexpr, STRING evar, STRING nvar, PNODE body)
 	ielement(node) = (VPTR) evar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -432,6 +520,7 @@ forindi_node (PACTX pactx, STRING ivar, STRING nvar, PNODE body)
 	ielement(node) = (VPTR) ivar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -449,6 +538,7 @@ forsour_node (PACTX pactx, STRING svar, STRING nvar, PNODE body)
 	ielement(node) = (VPTR) svar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -466,6 +556,7 @@ foreven_node (PACTX pactx, STRING evar, STRING nvar, PNODE body)
 	ielement(node) = (VPTR) evar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -483,6 +574,7 @@ forothr_node (PACTX pactx, STRING ovar, STRING nvar, PNODE body)
 	ielement(node) = (VPTR) ovar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -500,6 +592,7 @@ forfam_node (PACTX pactx, STRING fvar, STRING nvar, PNODE body)
 	ielement(node) = (VPTR) fvar;
 	inum(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR + PN_INUM_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -517,6 +610,7 @@ fornotes_node (PACTX pactx, PNODE nexpr, STRING vvar, PNODE body)
 	iloopexp(node) = (VPTR) nexpr;
 	ielement(node) = (VPTR) vvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -534,6 +628,7 @@ fornodes_node (PACTX pactx, PNODE nexpr, STRING nvar, PNODE body)
 	iloopexp(node) = (VPTR) nexpr;
 	ielement(node) = (VPTR) nvar;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -553,6 +648,7 @@ traverse_node (PACTX pactx, PNODE nexpr, STRING snode, STRING levv, PNODE body)
 	ielement(node) = (VPTR) snode;
 	ilev(node) = (VPTR) levv;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_IELEMENT_HPTR;
 	set_parents(body, node);
 	return node;
 }
@@ -564,6 +660,7 @@ iden_node (PACTX pactx, STRING iden)
 {
 	PNODE node = create_pnode(pactx, IIDENT);
 	iident(node) = (VPTR) iden;
+	node->i_flags = PN_IIDENT_HSTR;
 	return node;
 }
 /*==================================
@@ -574,6 +671,7 @@ icons_node (PACTX pactx, INT ival)
 {
 	PNODE node = create_pnode(pactx, IICONS);
 	ivaluex(node) = create_pvalue_from_int(ival);
+	node->i_flags = PN_IVALUEX_PVALUE;
 	return node;
 }
 /*===================================
@@ -584,6 +682,7 @@ fcons_node (PACTX pactx, FLOAT fval)
 {
 	PNODE node = create_pnode(pactx, IFCONS);
 	ivaluex(node) = create_pvalue_from_float(fval);
+	node->i_flags = PN_IVALUEX_PVALUE;
 	return node;
 }
 /*===================================
@@ -600,6 +699,7 @@ proc_node (PACTX pactx, CNSTRING name, PNODE parms, PNODE body)
 	iname(node) = (VPTR) name;
 	iargs(node) = (VPTR) parms;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_INAME_HSTR;
 	set_parents(body, node);
 	return node;
 }
@@ -617,6 +717,7 @@ fdef_node (PACTX pactx, CNSTRING name, PNODE parms, PNODE body)
 	iname(node) = (VPTR) name;
 	iargs(node) = (VPTR) parms;
 	ibody(node) = (VPTR) body;
+	node->i_flags = PN_INAME_HSTR;
 	set_parents(body, node);
 	return node;
 }
@@ -625,6 +726,7 @@ fdef_node (PACTX pactx, CNSTRING name, PNODE parms, PNODE body)
  *  pactx: [I/O] pointer to parseinfo structure (parse globals)
  *  name:  [IN]  function name
  *  elist: [IN]  param(s)
+ * consumes name heap pointer
  *=====================================================*/
 PNODE
 func_node (PACTX pactx, STRING name, PNODE elist)
@@ -641,6 +743,7 @@ func_node (PACTX pactx, STRING name, PNODE elist)
 		node = create_pnode(pactx, IFCALL);
 		iname(node) = (VPTR) name;
 		iargs(node) = (VPTR) elist;
+		node->i_flags = PN_INAME_HSTR;
 		ifunc(node) = func;
 		return node;
 	} else if (count) {
@@ -680,6 +783,7 @@ func_node (PACTX pactx, STRING name, PNODE elist)
 		iname(node) = (VPTR) name;
 		iargs(node) = (VPTR) elist;
 		ifunc(node) = (VPTR) builtins[md].ft_eval;
+		node->i_flags = PN_INAME_HSTR;
 		return node;
 		
 	}
@@ -730,6 +834,8 @@ clear_error_strings (void)
 	strfree(&nonlst1);
 	strfree(&nonlstx);
 	strfree(&nontabx);
+	strfree(&nonset1);
+	strfree(&nonsetx);
 	strfree(&badargs);
 	strfree(&badarg1);
 	strfree(&badargx);
@@ -792,6 +898,8 @@ interp_load_lang (void)
 	nonlst1     = strsave(_("%s: the arg must be a list."));
 	nonlstx     = strsave(_("%s: the arg #%s must be a list."));
 	nontabx     = strsave(_("%s: the arg #%s must be a table."));
+	nonset1     = strsave(_("%s: the arg must be a set."));
+	nonsetx     = strsave(_("%s: the arg #%s must be a set."));
 	badargs     = strsave(_("%s: Bad arguments"));
 	badarg1     = strsave(_("%s: the arg had a major error."));
 	badargx     = strsave(_("%s: the arg #%s had a major error."));
@@ -866,6 +974,7 @@ call_node (PACTX pactx, STRING name, PNODE args)
 	PNODE node = create_pnode(pactx, IPCALL);
 	iname(node) = (VPTR) name;
 	iargs(node) = (VPTR) args;
+	node->i_flags = PN_INAME_HSTR;
 	return node;
 }
 /*================================
@@ -1056,25 +1165,38 @@ debug_show_one_pnode (PNODE node)     /* node to print */
 		break;
 	}
 }
+/*==========================================================
+ * create_rptinfo -- Create new empty report info object
+ * returns addref'd rptinfo
+ *========================================================*/
+static RPTINFO
+create_rptinfo (void)
+{
+	RPTINFO rptinfo = (RPTINFO)stdalloc(sizeof(*rptinfo));
+	memset(rptinfo, 0, sizeof(*rptinfo));
+	rptinfo->vtable = &vtable_for_rptinfo;
+	rptinfo->refcnt = 1;
+	return rptinfo;
+}
 /*==========================================================
  * get_rptinfo -- Fetch info about report file
- * Created: 2002/11/30 (Perry Rapp)
+ *  create if not yet known
  *========================================================*/
 RPTINFO
 get_rptinfo (CNSTRING fullpath)
 {
 	RPTINFO rptinfo;
 	if (!f_rptinfos)
-		f_rptinfos = create_table_old();
-	rptinfo = (RPTINFO)valueof_ptr(f_rptinfos, fullpath);
+		f_rptinfos = create_table_obj();
+	rptinfo = (RPTINFO)valueof_obj(f_rptinfos, fullpath);
 	if (!rptinfo) {
 		STRING filename=0;
 		ZSTR zstr=0;
 
-		rptinfo = (RPTINFO)stdalloc(sizeof(*rptinfo));
+		rptinfo = create_rptinfo();
 		rptinfo->fullpath = strsave(fullpath);
-		rptinfo->functab = create_table_old();
-		rptinfo->proctab = create_table_old();
+		rptinfo->functab = create_table_vptr(); /* PNODES owned elsewhere */
+		rptinfo->proctab = create_table_vptr(); /* PNODES owned elsewhere */
 		rptinfo->codeset = strsave(report_codeset_in);
 
 		/* calculate localpath & localepath for report gettext */
@@ -1086,30 +1208,39 @@ get_rptinfo (CNSTRING fullpath)
 		strfree(&filename);
 		rptinfo->textdomain = zs_news("llreports"); /* for now, fixed textdomain */
 		
-		insert_table_ptr(f_rptinfos, strsave(fullpath), rptinfo);
+		insert_table_obj(f_rptinfos, fullpath, rptinfo);
+		--rptinfo->refcnt; /* release our reference on rptinfo */
+		ASSERT(rptinfo->refcnt>0);
 	}
 	return rptinfo;
 }
 /*==========================================================
- * clear_rptinfos -- Fetch info about report file
- * Created: 2002/11/30 (Perry Rapp)
+ * clear_rptinfos -- Delete all allocated rptinfos
  *========================================================*/
 void
 clear_rptinfos (void)
 {
 	if (f_rptinfos) {
-		struct tag_table_iter tabit;
-		STRING key=0;
-		VPTR ptr=0;
-		begin_table(f_rptinfos, &tabit);
-		while (next_table_ptr(&tabit, &key, &ptr)) {
-			RPTINFO rptinfo = (RPTINFO)ptr;
-			remove_table(rptinfo->proctab, FREEKEY); /* values are PNODES */
-			remove_table(rptinfo->functab, FREEKEY); /* values are PNODES */
-			strfree(&rptinfo->fullpath);
-			strfree(&rptinfo->codeset);
-		}
-		remove_table(f_rptinfos, FREEBOTH);
+		destroy_table(f_rptinfos);
 		f_rptinfos = 0;
 	}
 }
+/*=================================================
+ * rptinfo_destructor -- destructor for rptinfo
+ *  (destructor entry in vtable)
+ *===============================================*/
+static void
+rptinfo_destructor (VTABLE *obj)
+{
+	RPTINFO rptinfo = (RPTINFO)obj;
+	ASSERT(rptinfo->vtable == &vtable_for_rptinfo);
+
+	destroy_table(rptinfo->proctab); /* values are vptr PNODES */
+	destroy_table(rptinfo->functab); /* values are vptr PNODES */
+	strfree(&rptinfo->fullpath);
+	strfree(&rptinfo->codeset);
+	zs_free(&rptinfo->localpath);
+	zs_free(&rptinfo->localepath);
+	zs_free(&rptinfo->textdomain);
+	stdfree(rptinfo);
+}
