@@ -45,6 +45,7 @@
 #include "parse.h"
 #include "zstr.h"
 #include "icvt.h"
+#include "date.h"
 
 #ifndef INCLUDED_STDARG_H
 #include <stdarg.h>
@@ -62,14 +63,14 @@
  that parseinfo currently has
 */
 TABLE gproctab=0, gfunctab=0;
-SYMTAB globtab; /* assume all zero is null SYMTAB */
-PATHINFO cur_pathinfo = 0; /* program currently being parsed or run */
-FILE *Poutfp = NULL;          /* file to write program output to */
-STRING Poutstr = NULL;	      /* string to write program output to */
+SYMTAB globtab = NULL;
+PATHINFO cur_pathinfo = 0;	/* program currently being parsed or run */
+FILE *Poutfp = NULL;		/* file to write program output to */
+STRING Poutstr = NULL;		/* string to write program output to */
 INT Perrors = 0;
-LIST Plist = NULL;     /* list of program files still to read */
-PNODE Pnode = NULL;           /* node being interpreted */
-BOOLEAN explicitvars = FALSE; /* all vars must be declared */
+LIST Plist = NULL;		/* list of program files still to read */
+PNODE Pnode = NULL;		/* node being interpreted */
+BOOLEAN explicitvars = FALSE;	/* all vars must be declared */
 BOOLEAN rpt_cancelled = FALSE;
 
 /*********************************************
@@ -90,13 +91,13 @@ extern STRING qSunsupuniv;
  * local function prototypes
  *********************************************/
 
-static ZSTR approx_time(INT seconds);
 static STRING check_rpt_requires(PACTX pactx, STRING fname);
 static void disp_symtab(STRING title, SYMTAB stab);
 static BOOLEAN disp_symtab_cb(STRING key, PVALUE val, VPTR param);
 static void enqueue_parse_error(const char * fmt, ...);
-static BOOLEAN find_program(STRING fname, STRING localdir, STRING *pfull);
+static BOOLEAN find_program(STRING fname, STRING localdir, STRING *pfull,BOOLEAN include);
 static void init_pactx(PACTX pactx);
+static BOOLEAN interpret_prog(PNODE begin, SYMTAB stab);
 static void parse_file(PACTX pactx, STRING fname, STRING fullpath);
 static void print_report_duration(INT duration, INT uiduration);
 static void progmessage(MSG_LEVEL level, STRING);
@@ -128,6 +129,7 @@ initinterp (void)
 	initrassa();
 	Perrors = 0;
 	rpt_cancelled = FALSE;
+	explicitvars = FALSE;
 	/* clear previous information */
 	vprog_prevfile[0] = 0;
 	vprog_prevline=-1;
@@ -221,8 +223,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 	, STRING ofile, BOOLEAN picklist)
 {
 	LIST plist=0, donelist=0;
-	SYMTAB stab = null_symtab();
-	PVALUE dummy;
+	SYMTAB stab = NULL;
 	INT i;
 	INT nfiles = length_list(lifiles);
 	PNODE first, parm;
@@ -244,7 +245,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 		for (i = 1; i < nfiles+1; i++) {
 			STRING fullpath = 0;
 			STRING progfile = get_list_element(lifiles, i, NULL);
-			if (find_program(progfile, 0, &fullpath)) {
+			if (find_program(progfile, 0, &fullpath,FALSE)) {
 				PATHINFO pathinfo = new_pathinfo(progfile, fullpath);
 				strfree(&fullpath);
 				enqueue_list(plist, pathinfo);
@@ -266,7 +267,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 			}
 			strfree(&fname);
 			strfree(&fullpath);
-			goto interp_program_exit;
+			goto interp_program_notfound;
 		}
 		pathinfo = new_pathinfo(fname, fullpath);
 		strfree(&fname);
@@ -281,16 +282,16 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 	/* Parse each file in the list -- don't reparse any file */
 	/* (paths are resolved before files are enqueued, & stored in pathinfo) */
 
-	gproctab = create_table_old();
-	create_symtab(&globtab);
-	gfunctab = create_table_old();
+	gproctab = create_table_obj();
+	globtab = create_symtab();
+	gfunctab = create_table_obj();
 	initinterp();
 
 	while (!is_empty_list(plist)) {
 		cur_pathinfo = (PATHINFO) dequeue_list(plist);
 		if (!in_table(pactx->filetab, cur_pathinfo->fullpath)) {
 			STRING str;
-			insert_table_ptr(pactx->filetab, strsave(cur_pathinfo->fullpath), 0);
+			insert_table_obj(pactx->filetab, cur_pathinfo->fullpath, 0);
 			Plist = plist;
 			parse_file(pactx, cur_pathinfo->fname, cur_pathinfo->fullpath);
 			if ((str = check_rpt_requires(pactx, cur_pathinfo->fullpath)) != 0) {
@@ -304,7 +305,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 		}
 		cur_pathinfo = 0;
 	}
-	remove_list(plist, NULL);
+	destroy_list(plist);
 	plist=NULL;
 
 
@@ -315,8 +316,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 			prog_error(NULL, str);
 			++Perrors;
 		ENDLIST
-		make_list_empty(outstanding_parse_errors);
-		remove_list(outstanding_parse_errors, 0);
+		destroy_list(outstanding_parse_errors);
 		outstanding_parse_errors=0;
 	}
 
@@ -334,10 +334,10 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 
    /* Open output file if name is provided */
 
-	if (ofile && !(Poutfp = fopen(ofile, LLWRITETEXT))) {
-		msg_error(_("Error: output file <%s> could not be created.\n")
-			, ofile);
-		goto interp_program_exit;
+	if (ofile) {
+		if (!start_output_file(ofile)) {
+			goto interp_program_exit;
+		}
 	}
 	if (Poutfp) setbuf(Poutfp, NULL);
 
@@ -349,7 +349,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 			proc, num_params(parm), nargs);
 		goto interp_program_exit;
 	}
-	create_symtab(&stab);
+	stab = create_symtab();
 	for (i = 0; i < nargs; i++) {
 		insert_symtab(stab, iident(parm), args[0]);
 		parm = inext(parm);
@@ -361,18 +361,7 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 	progrunning = TRUE;
 	progerror = 0;
 	progmessage(MSG_STATUS, _("is running..."));
-	switch (interpret((PNODE) ibody(first), stab, &dummy)) {
-	case INTOKAY:
-	case INTRETURN:
-		progmessage(MSG_INFO, _("was run successfully.\n"));
-		break;
-	default:
-		if (rpt_cancelled)
-			progmessage(MSG_STATUS, _("was cancelled.\n"));
-		else
-			progmessage(MSG_STATUS, _("was not run because of errors.\n"));
-		break;
-	}
+	ranit = interpret_prog((PNODE) ibody(first), stab);
 
    /* Clean up and return */
 
@@ -383,9 +372,15 @@ interp_program_list (STRING proc, INT nargs, VPTR *args, LIST lifiles
 
 interp_program_exit:
 
-	
 	remove_tables(pactx);
-	remove_symtab(&stab);
+	if (stab) {
+		remove_symtab(stab);
+		stab = NULL;
+	}
+
+interp_program_notfound:
+
+	symbol_tables_end();
 	pvalues_end();
 	wipe_pactx(pactx);
 
@@ -400,9 +395,35 @@ interp_program_exit:
 		delete_pathinfo(&pathinfo);
 	}
 	strfree(&rootfilepath);
-	remove_list(donelist, NULL);
+	destroy_list(donelist);
+	destroy_list(plist);
 	return ranit;
 }
+/*===============================================
+ * interpret_prog -- execute a report program
+ *=============================================*/
+static BOOLEAN
+interpret_prog (PNODE begin, SYMTAB stab)
+{
+	PVALUE dummy=0;
+	INT rtn = interpret(begin, stab, &dummy);
+
+	delete_pvalue(dummy);
+	dummy=0;
+
+	switch(rtn) {
+	case INTOKAY:
+	case INTRETURN:
+		progmessage(MSG_INFO, _("was run successfully.\n"));
+		return TRUE;
+	default:
+		if (rpt_cancelled) {
+			progmessage(MSG_STATUS, _("was cancelled.\n"));
+		} else
+			progmessage(MSG_STATUS, _("was not run because of errors.\n"));
+		return FALSE;
+	}
+}
 /*===============================================
  * init_pactx -- initialize global parsing context
  *=============================================*/
@@ -410,7 +431,7 @@ static void
 init_pactx (PACTX pactx)
 {
 	memset(pactx, 0, sizeof(*pactx));
-	pactx->filetab = create_table_old();
+	pactx->filetab = create_table_obj(); /* table of tables */
 }
 /*===============================================
  * wipe_pactx -- destroy global parsing context
@@ -418,23 +439,10 @@ init_pactx (PACTX pactx)
 static void
 wipe_pactx (PACTX pactx)
 {
-	/* 2003-03-24, Perry: */
-	/* leaking the file property tables optionally hanging off file entries */
-	remove_table(pactx->filetab, FREEBOTH);
+	destroy_table(pactx->filetab);
 	pactx->filetab=NULL;
 	memset(pactx, 0, sizeof(*pactx));
 }
-/*===========================================+
- * freelistentry -- Clear list in entry in table
- * This is used as a callback with traverse_table
- * Created: 2002/11/30 (Perry Rapp)
- *==========================================*/
-static void
-freelistentry (ENTRY ent)
-{
-	LIST list = (LIST)ent->uval.w;
-	make_list_empty(list);
-}
 /*===========================================+
  * remove_tables -- Remove interpreter's tables
  *==========================================*/
@@ -442,12 +450,11 @@ static void
 remove_tables (PACTX pactx)
 {
 	pactx=pactx; /* unused */
-	traverse_table(gproctab, &freelistentry);
-	remove_table(gproctab, FREEBOTH); /* values are empty lists */
+	destroy_table(gproctab);
 	gproctab=NULL;
-	remove_symtab(&globtab);
-	traverse_table(gfunctab, &freelistentry);
-	remove_table(gfunctab, FREEBOTH); /* values are PNODES */
+	remove_symtab(globtab);
+	globtab = NULL;
+	destroy_table(gfunctab);
 	gfunctab=NULL;
 }
 /*======================================+
@@ -457,7 +464,7 @@ remove_tables (PACTX pactx)
  * Returns TRUE if found
  *=====================================*/
 static BOOLEAN
-find_program (STRING fname, STRING localdir, STRING *pfull)
+find_program (STRING fname, STRING localdir, STRING *pfull,BOOLEAN include)
 {
 	STRING programsdir = getoptstr("LLPROGRAMS", ".");
 	FILE * fp = 0;
@@ -470,6 +477,14 @@ find_program (STRING fname, STRING localdir, STRING *pfull)
 		zs_apps(zstr, LLSTRPATHSEPARATOR);
 	}
 	zs_apps(zstr, programsdir);
+	if (include) {
+	    fp = fopenpath(fname, LLREADTEXT, zs_str(zstr), ".li", uu8, pfull);
+	    if (fp) {
+		    fclose(fp);
+		    rtn = TRUE;
+		    goto end_find_program;
+	    }
+	}
 	fp = fopenpath(fname, LLREADTEXT, zs_str(zstr), ".ll", uu8, pfull);
 	if (fp) {
 		fclose(fp);
@@ -520,35 +535,6 @@ parse_file (PACTX pactx, STRING fname, STRING fullpath)
 	pactx->lineno = 0;
 	pactx->charpos = 0;
 }
-/*====================================+
- * approx_time -- make time in seconds more legible
- *===================================*/
-static ZSTR
-approx_time (INT seconds)
-{
-	ZSTR zti=zs_new();
-	INT minutes, hours, days, years;
-	minutes = seconds/60;
-	seconds = seconds - minutes*60;
-	hours = minutes/60;
-	minutes = minutes - hours*60;
-	days = hours/60;
-	hours = hours - days*24;
-	years = days/365.2425;
-	days = days - years*365.2425;
-	if (years) {
-		zs_setf(zti, _("%dy%03dd"), years, days);
-	} else if (days) {
-		zs_setf(zti, ("%dd%03dh"), days, hours);
-	} else if (hours) {
-		zs_setf(zti, ("%dh%03dm"), hours, minutes);
-	} else if (minutes) {
-		zs_setf(zti, _("%dm%03ds"), minutes, seconds);
-	} else {
-		zs_setf(zti, _("%02ds"), seconds);
-	}
-	return zti;
-}
 /*====================================+
  * report_duration -- print report duration
  *===================================*/
@@ -952,8 +938,8 @@ aloop:	;
 	ENDCHILDRENx
 	irc = INTOKAY;
 aleave:
-	delete_symtab(stab, ichild(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ichild(node));
+	delete_symtab_element(stab, inum(node));
 	unlock_cache(fcel);
 	return irc;
 }
@@ -1007,9 +993,9 @@ bloop:	;
 	ENDSPOUSES
 	irc = INTOKAY;
 bleave:
-	delete_symtab(stab, ispouse(node));
-	delete_symtab(stab, ifamily(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ispouse(node));
+	delete_symtab_element(stab, ifamily(node));
+	delete_symtab_element(stab, inum(node));
 	unlock_cache(icel);
 	return irc;
 }
@@ -1066,9 +1052,9 @@ cloop:	;
 	ENDFAMSS
 	irc = INTOKAY;
 cleave:
-	delete_symtab(stab, ifamily(node));
-	delete_symtab(stab, ispouse(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ifamily(node));
+	delete_symtab_element(stab, ispouse(node));
+	delete_symtab_element(stab, inum(node));
 	unlock_cache(icel);
 	return irc;
 }
@@ -1106,7 +1092,7 @@ interp_fathers (PNODE node, SYMTAB stab, PVALUE *pval)
 		fval = create_pvalue_from_fam(fam);
 		fcel = pvalue_to_cel(fval);
 		insert_symtab(stab, ifamily(node), fval);
-		insert_symtab(stab, iiparent(node), create_pvalue_from_cel(scel));
+		insert_symtab(stab, iiparent(node), create_pvalue_from_cel(PINDI, scel));
 		insert_symtab(stab, inum(node), create_pvalue_from_int(ncount++));
 		lock_cache(fcel);
 		lock_cache(scel);
@@ -1125,9 +1111,9 @@ dloop:	;
 	ENDFAMCS
 	irc = INTOKAY;
 dleave:
-	delete_symtab(stab, ifamily(node));
-	delete_symtab(stab, iiparent(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ifamily(node));
+	delete_symtab_element(stab, iiparent(node));
+	delete_symtab_element(stab, inum(node));
 	unlock_cache(icel);
 	return irc;
 }
@@ -1165,7 +1151,7 @@ interp_mothers (PNODE node, SYMTAB stab, PVALUE *pval)
 		fval = create_pvalue_from_fam(fam);
 		fcel = pvalue_to_cel(fval);
 		insert_symtab(stab, ifamily(node), fval);
-		insert_symtab(stab, iiparent(node), create_pvalue_from_cel(scel));
+		insert_symtab(stab, iiparent(node), create_pvalue_from_cel(PINDI, scel));
 		insert_symtab(stab, inum(node), create_pvalue_from_int(ncount++));
 		lock_cache(fcel);
 		lock_cache(scel);
@@ -1183,9 +1169,9 @@ eloop:	;
 	ENDFAMCS
 	irc = INTOKAY;
 eleave:
-	delete_symtab(stab, ifamily(node));
-	delete_symtab(stab, iiparent(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ifamily(node));
+	delete_symtab_element(stab, iiparent(node));
+	delete_symtab_element(stab, inum(node));
 	unlock_cache(icel);
 	return irc;
 }
@@ -1233,8 +1219,8 @@ floop:	;
 	ENDFAMCS
 	irc = INTOKAY;
 fleave:
-	delete_symtab(stab, ifamily(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ifamily(node));
+	delete_symtab_element(stab, inum(node));
 	unlock_cache(icel);
 	return INTOKAY;
 }
@@ -1269,7 +1255,7 @@ gloop:      ;
 	ENDTAGVALUES
 	irc = INTOKAY;
 gleave:
-	delete_symtab(stab, ielement(node));
+	delete_symtab_element(stab, ielement(node));
 	return irc;
 }
 /*==========================================+
@@ -1308,7 +1294,7 @@ hloop: ;
 	}
 	irc = INTOKAY;
 hleave:
-	delete_symtab(stab, ielement(node));
+	delete_symtab_element(stab, ielement(node));
 	return irc;
 }
 /*========================================+
@@ -1347,6 +1333,10 @@ interp_forindi (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 		ival = create_pvalue_from_indi_keynum(count);
 		icel = pvalue_to_cel(ival);
+		if (!icel) { /* apparently missing record */
+			delete_pvalue(ival);
+			continue;
+		}
 		icount++;
 		lock_cache(icel); /* keep current indi in cache during loop body */
 		/* set loop variables */
@@ -1364,8 +1354,8 @@ interp_forindi (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 	}
 ileave:
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, inum(node));
 	return irc;
 }
 /*========================================+
@@ -1392,6 +1382,10 @@ interp_forsour (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 		sval = create_pvalue_from_sour_keynum(count);
 		scel = pvalue_to_cel(sval);
+		if (!scel) { /* apparently missing record */
+			delete_pvalue(sval);
+			continue;
+		}
 		scount++;
 		lock_cache(scel); /* keep current source in cache during loop body */
 		/* set loop variables */
@@ -1410,8 +1404,8 @@ interp_forsour (PNODE node, SYMTAB stab, PVALUE *pval)
 	}
 sourleave:
 	/* remove loop variables from symbol table */
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, inum(node));
 	return irc;
 }
 /*========================================+
@@ -1435,6 +1429,10 @@ interp_foreven (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 		eval = create_pvalue_from_even_keynum(count);
 		ecel = pvalue_to_cel(eval);
+		if (!ecel) { /* apparently missing record */
+			delete_pvalue(eval);
+			continue;
+		}
 		ecount++;
 		lock_cache(ecel); /* keep current event in cache during loop body */
 		/* set loop variables */
@@ -1453,8 +1451,8 @@ interp_foreven (PNODE node, SYMTAB stab, PVALUE *pval)
 	}
 evenleave:
 	/* remove loop variables from symbol table */
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, inum(node));
 	return irc;
 }
 /*========================================+
@@ -1478,6 +1476,10 @@ interp_forothr (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 		xval = create_pvalue_from_othr_keynum(count);
 		xcel = pvalue_to_cel(xval);
+		if (!xcel) { /* apparently missing record */
+			delete_pvalue(xval);
+			continue;
+		}
 		xcount++;
 		lock_cache(xcel); /* keep current source in cache during loop body */
 		/* set loop variables */
@@ -1495,8 +1497,8 @@ interp_forothr (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 	}
 othrleave:
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, inum(node));
 	return irc;
 }
 /*======================================+
@@ -1520,6 +1522,10 @@ interp_forfam (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 		fval = create_pvalue_from_fam_keynum(count);
 		fcel = pvalue_to_cel(fval);
+		if (!fcel) { /* apparently missing record */
+			delete_pvalue(fval);
+			continue;
+		}
 		fcount++;
 		lock_cache(fcel);
 		insert_symtab(stab, ielement(node), fval);
@@ -1535,8 +1541,8 @@ interp_forfam (PNODE node, SYMTAB stab, PVALUE *pval)
 		}
 	}
 mleave:
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, inum(node));
 	return irc;
 }
 /*============================================+
@@ -1549,7 +1555,7 @@ interp_indisetloop (PNODE node, SYMTAB stab, PVALUE *pval)
 {
 	BOOLEAN eflg = FALSE;
 	INTERPTYPE irc;
-	PVALUE indival, loopval;
+	PVALUE indival=0, loopval=0;
 	INDISEQ seq = NULL;
 	PVALUE val = evaluate(iloopexp(node), stab, &eflg);
 	if (eflg || !val || ptype(val) != PSET) {
@@ -1566,10 +1572,10 @@ interp_indisetloop (PNODE node, SYMTAB stab, PVALUE *pval)
 	insert_symtab(stab, inum(node), create_pvalue_from_int(0));
 	FORINDISEQ(seq, el, ncount)
 		/* put current indi in symbol table */
-		indival = create_pvalue_from_indi_key(skey(el));
+		indival = create_pvalue_from_indi_key(element_skey(el));
 		insert_symtab(stab, ielement(node), indival);
 		/* put current indi's value in symbol table */
-		loopval = sval(el).w;
+		loopval = element_pval(el);
 		if (loopval)
 			loopval = copy_pvalue(loopval);
 		else
@@ -1589,9 +1595,9 @@ hloop:	;
 	irc = INTOKAY;
 hleave:
 	delete_pvalue(val); /* delete temp evaluated val - may destruct seq */
-	delete_symtab(stab, ielement(node)); /* remove indi */
-	delete_symtab(stab, ivalvar(node)); /* remove indi's value */
-	delete_symtab(stab, inum(node)); /* remove counter */
+	delete_symtab_element(stab, ielement(node)); /* remove indi */
+	delete_symtab_element(stab, ivalvar(node)); /* remove indi's value */
+	delete_symtab_element(stab, inum(node)); /* remove counter */
 	return irc;
 }
 /*=====================================+
@@ -1638,8 +1644,8 @@ iloop:	;
 ileave:
 	delete_pvalue(val); /* delete temp evaluated val - may destruct list */
 	/* remove element & counter from symbol table */
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, inum(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, inum(node));
 	return irc;
 }
 /*===================================+
@@ -1689,7 +1695,7 @@ get_proc_node (CNSTRING procname, TABLE loctab, TABLE gtab, INT * count)
 	PNODE proc = valueof_ptr(loctab, procname);
 	if (proc) return proc;
 	/* now look for global proc, and must be unique */
-	list = valueof_ptr(gtab, procname);
+	list = valueof_obj(gtab, procname);
 	if (!list) {
 		*count = 0;
 		return NULL;
@@ -1708,7 +1714,7 @@ get_proc_node (CNSTRING procname, TABLE loctab, TABLE gtab, INT * count)
 INTERPTYPE
 interp_call (PNODE node, SYMTAB stab, PVALUE *pval)
 {
-	SYMTAB newstab=null_symtab();
+	SYMTAB newstab = NULL;
 	INTERPTYPE irc=INTERROR;
 	PNODE arg=NULL, parm=NULL, proc;
 	STRING procname = iname(node);
@@ -1723,7 +1729,7 @@ interp_call (PNODE node, SYMTAB stab, PVALUE *pval)
 		irc = INTERROR;
 		goto call_leave;
 	}
-	create_symtab(&newstab);
+	newstab = create_symtab();
 	arg = (PNODE) iargs(node);
 	parm = (PNODE) iargs(proc);
 	while (arg && parm) {
@@ -1757,7 +1763,10 @@ interp_call (PNODE node, SYMTAB stab, PVALUE *pval)
 	}
 
 call_leave:
-	remove_symtab(&newstab);
+	if (newstab) {
+		remove_symtab(newstab);
+		newstab = NULL;
+	}
 	return irc;
 }
 /*==============================================+
@@ -1813,8 +1822,8 @@ interp_traverse (PNODE node, SYMTAB stab, PVALUE *pval)
 	}
 	irc = INTOKAY;
 traverse_leave:
-	delete_symtab(stab, ielement(node));
-	delete_symtab(stab, ilev(node));
+	delete_symtab_element(stab, ielement(node));
+	delete_symtab_element(stab, ilev(node));
 	delete_pvalue(val);
 	val=NULL;
 	return irc;
@@ -1869,10 +1878,10 @@ prog_var_error (PNODE node, SYMTAB stab, PNODE arg, PVALUE val, STRING fmt, ...)
 		ZSTR zstr=zs_new();
 		INT n=0,i=0;
 		/* report debugger: Option to display list of local symbols */
-		n = (stab.tab ? get_table_count(stab.tab) : 0);
+		n = (stab->tab ? get_table_count(stab->tab) : 0);
 		zs_setf(zstr, _("Display locals (%d)"), n);
 		choices[i++] = strsave(zs_str(zstr));
-		n = (globtab.tab ? get_table_count(globtab.tab) : 0);
+		n = (globtab->tab ? get_table_count(globtab->tab) : 0);
 		zs_setf(zstr, _("Display globals (%d)"), n);
 		choices[i++] = strsave(zs_str(zstr));
 		choices[i++] = strsave(_("Pop one level"));
@@ -1901,8 +1910,8 @@ dbgloop:
 static void
 disp_symtab (STRING title, SYMTAB stab)
 {
-	struct tag_symtab_iter symtabits;
-	INT n = (stab.tab ? get_table_count(stab.tab) : 0);
+	SYMTAB_ITER symtabit=0;
+	INT n = (stab->tab ? get_table_count(stab->tab) : 0);
 	struct dbgsymtab_s sdata;
 	INT bytes = n * sizeof(STRING);
 	if (!n) return;
@@ -1911,12 +1920,14 @@ disp_symtab (STRING title, SYMTAB stab)
 	sdata.locals = (STRING *)stdalloc(bytes);
 	memset(sdata.locals, 0, bytes);
 	/* Now traverse & print the actual entries via disp_symtab_cb() */
-	if (begin_symtab(stab, &symtabits)) {
+	symtabit = begin_symtab_iter(stab);
+	if (symtabit) {
 		STRING key=0;
 		PVALUE pval=0;
-		while (next_symtab_entry(&symtabits, &key, &pval)) {
+		while (next_symtab_entry(symtabit, (CNSTRING *)&key, &pval)) {
 			disp_symtab_cb(key, pval, &sdata.locals);
 		}
+		end_symtab_iter(&symtabit);
 	}
 	/* Title of report debugger's list of local symbols */
 	/* TODO: 2003-01-19, we could allow drilldown on lists, tables & sets here */
@@ -2109,7 +2120,7 @@ pa_handle_include (PACTX pactx, PNODE node)
 		localpath = zs_str(irptinfo(node)->localpath);
 	}
 
-	if (find_program(newfname, localpath, &fullpath)) {
+	if (find_program(newfname, localpath, &fullpath,TRUE)) {
 		pathinfo = new_pathinfo(newfname, fullpath);
 		strfree(&fullpath);
 		enqueue_list(Plist, pathinfo);
@@ -2134,14 +2145,15 @@ pa_handle_require (PACTX pactx, PNODE node)
 	ASSERT(ptype(pval)==PSTRING);
 	pactx=pactx; /* unused */
 
-	tab = (TABLE)valueof_ptr(pactx->filetab, pactx->fullpath);
+	tab = (TABLE)valueof_obj(pactx->filetab, pactx->fullpath);
 	if (!tab) {
-		tab = create_table(FREEBOTH);
-		insert_table_ptr(pactx->filetab, strsave(cur_pathinfo->fullpath), tab);
+		tab = create_table_str();
+		insert_table_obj(pactx->filetab, cur_pathinfo->fullpath, tab);
+		release_table(tab); /* release our reference, pactx->filetab owns now */
 	}
 
 	str = pvalue_to_string(pval);
-	insert_table_str(tab, propstr, strsave(str));
+	insert_table_str(tab, propstr, str);
 }
 /*=============================================+
  * pa_handle_proc -- proc declaration (parse time)
@@ -2160,14 +2172,16 @@ pa_handle_proc (PACTX pactx, CNSTRING procname, PNODE nd_args, PNODE nd_body)
 		enqueue_parse_error(_("Duplicate proc %s (lines %d and %d) in report: %s")
 			, procname, iline(procnode)+1, iline(nd_body)+1, pactx->fullpath);
 	}
+	/* consumes procname */
 	procnode = proc_node(pactx, procname, nd_args, nd_body);
-	insert_table_ptr(rptinfo->proctab, strsave(procname), (VPTR)procnode);
+	insert_table_ptr(rptinfo->proctab, procname, procnode);
 
 	/* add to global proc table */
-	list = (LIST)valueof_ptr(gproctab, procname);
+	list = (LIST)valueof_obj(gproctab, procname);
 	if (!list) {
 		list = create_list2(LISTNOFREE);
-		insert_table_ptr(gproctab, procname, list);
+		insert_table_obj(gproctab, procname, list);
+		release_list(list); /* now table owns list */
 	}
 	enqueue_list(list, procnode);
 }
@@ -2179,8 +2193,8 @@ void
 pa_handle_func (PACTX pactx, CNSTRING procname, PNODE nd_args, PNODE nd_body)
 {
 	RPTINFO rptinfo = get_rptinfo(pactx->fullpath);
-	PNODE procnode;
-	LIST list;
+	PNODE procnode=0;
+	LIST list=0;
 
 	/* check for local duplicates, else add to local proc table */
 	procnode = (PNODE)valueof_ptr(rptinfo->functab, procname);
@@ -2188,14 +2202,16 @@ pa_handle_func (PACTX pactx, CNSTRING procname, PNODE nd_args, PNODE nd_body)
 		enqueue_parse_error(_("Duplicate func %s (lines %d and %d) in report: %s")
 			, procname, iline(procnode)+1, iline(nd_body)+1, pactx->fullpath);
 	}
+	/* consumes procname */
 	procnode = fdef_node(pactx, procname, nd_args, nd_body);
-	insert_table_ptr(rptinfo->functab, strsave(procname), (VPTR)procnode);
+	insert_table_ptr(rptinfo->functab, procname, procnode);
 
 	/* add to global proc table */
-	list = (LIST)valueof_ptr(gfunctab, procname);
+	list = (LIST)valueof_obj(gfunctab, procname);
 	if (!list) {
 		list = create_list2(LISTNOFREE);
-		insert_table_ptr(gfunctab, procname, list);
+		insert_table_obj(gfunctab, procname, list);
+		release_list(list); /* now table owns list */
 	}
 	enqueue_list(list, procnode);
 }
@@ -2220,7 +2236,7 @@ parse_error (PACTX pactx, STRING str)
 static STRING
 check_rpt_requires (PACTX pactx, STRING fullpath)
 {
-	TABLE tab = (TABLE)valueof_ptr(pactx->filetab, fullpath);
+	TABLE tab = (TABLE)valueof_obj(pactx->filetab, fullpath);
 	STRING str;
 	STRING propstr = "requires_lifelines-reports.version:";
 	INT ours=0, desired=0;
