@@ -57,6 +57,8 @@ typedef struct tag_pathinfo {
 } *PATHINFO;
 
 struct tag_rptinfo {
+	struct tag_vtable *vtable; /* generic object */
+	INT refcnt; /* ref-countable object */
 	STRING fullpath; /* fully qualified path to file */
 	ZSTR localpath;  /* directory file is in */
 	ZSTR localepath; /* directory for message catalogs */
@@ -71,26 +73,27 @@ typedef struct tag_rptinfo *RPTINFO;
 /* Symbol Table Structures and Prototypes                               */
 /************************************************************************/
 
+
 /* symbol table data is just table data */
 typedef struct tag_symtab {
 	TABLE tab;
-} SYMTAB;
+} *SYMTAB;
 
-/* symbol table iterator data is just table iterator data */
-typedef struct tag_symtab_iter {
-	struct tag_table_iter tabiters;
-} *SYMTAB_ITER;
-
-BOOLEAN begin_symtab(SYMTAB stab, SYMTAB_ITER stabit);
-void create_symtab(SYMTAB * stab);
-void delete_symtab(SYMTAB stab, STRING iden);
+SYMTAB create_symtab(void);
+void delete_symtab_element(SYMTAB stab, STRING iden);
 BOOLEAN in_symtab(SYMTAB stab, STRING key);
 void insert_symtab(SYMTAB stab, STRING iden, PVALUE val);
-BOOLEAN next_symtab_entry(SYMTAB_ITER tabit, STRING *pkey, PVALUE *ppval);
-SYMTAB null_symtab(void);
-void remove_symtab(SYMTAB *);
+void remove_symtab(SYMTAB stab);
+void symbol_tables_end(void);
 PVALUE symtab_valueofbool(SYMTAB, STRING, BOOLEAN*);
 
+
+/* symbol table iteration */
+typedef struct tag_symtab_iter *SYMTAB_ITER;
+SYMTAB_ITER begin_symtab_iter(SYMTAB stab);
+BOOLEAN next_symtab_entry(SYMTAB_ITER tabit, CNSTRING *pkey, PVALUE *ppval);
+void end_symtab_iter(SYMTAB_ITER * psymtabit);
+
 /************************************************************************/
 /* Interpreter Structures and Functions                                 */
 /************************************************************************/
@@ -101,6 +104,7 @@ struct tag_pnode {
 	PNODE    i_prnt;       /* parent of this node */
 	INT      i_line;       /* line where text of this node begins */
 	RPTINFO  i_rptinfo;    /* information about this report file */
+	INT      i_flags;
 	PNODE    i_next;       /* next node */
 	VPTR     i_word1;      /* variable data associated with node type */
 	VPTR     i_word2;      /* ... */
@@ -109,6 +113,7 @@ struct tag_pnode {
 	VPTR     i_word5;
 };
 
+/* pnode types */
 #define IICONS       1   /* integer constant */
 #define IFCONS       2   /* floating constant */
 #define ILCONS       3   /* long integer constant */
@@ -142,6 +147,20 @@ struct tag_pnode {
 #define INOTES      31   /* notes loop */
 #define IFREED      99   /* returned to free list */
 
+/* pnode flags */
+enum {
+	PN_IVALUEX_PVALUE = 0x1 /* ivaluex is a pvalue */
+	, PN_INAME_HSTR = 0x2 /* iname is a heap string */
+	, PN_ICHILD_HPTR = 0x4 /* ichild is a heap string */
+	, PN_INUM_HPTR = 0x8 /* inum is a heap string */
+	, PN_ISPOUSE_HPTR = 0x10 /* ispouse is a heap string */
+	, PN_IFAMILY_HPTR = 0x20 /* ifamily is a heap string */
+	, PN_IELEMENT_HPTR = 0x40 /* ielement is a heap string */
+	, PN_IPARENT_HPTR = 0x80 /* iiparent is a heap string */
+	, PN_IVALVAR_HPTR = 0x100 /* ivalvar is a heap string */
+	, PN_IIDENT_HSTR = 0x200 /* iident is a heap string */
+};
+
 #define itype(i)     ((i)->i_type)  /* node type - all nodes */
 #define iprnt(i)     ((i)->i_prnt)  /* parent node - all nodes */
 #define inext(i)     ((i)->i_next)  /* next node - all nodes */
@@ -160,7 +179,7 @@ struct tag_pnode {
 #define ichild(i)    ((i)->i_word2)     /* var in children loop */
 #define ispouse(i)   ((i)->i_word2)     /* var in families and spouses loop */
 #define ifamily(i)   ((i)->i_word3)     /* var in all families type loops */
-#define iiparent(i)  ((i)->i_word2)    /* var in some families type loops */
+#define iiparent(i)  ((i)->i_word2)     /* var in some families type loops */
 #define ivalvar(i)   ((i)->i_word3)     /* var in indiset loop */
 #define iname(i)     ((i)->i_word1)     /* proc, func and builtin names */
 #define ilev(i)      ((i)->i_word3)     /* var traverse loop */
@@ -238,6 +257,7 @@ PVALUE __choosefam(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __chooseindi(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __choosespouse(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __choosesubset(PNODE, SYMTAB, BOOLEAN *);
+PVALUE __clear(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __col(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __complexdate(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __complexformat(PNODE, SYMTAB, BOOLEAN *);
@@ -296,7 +316,7 @@ PVALUE __getindi(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __getindiset(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __getint(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __getproperty(PNODE, SYMTAB, BOOLEAN *);
-PVALUE __getrecord(PNODE, SYMTAB, BOOLEAN *);
+PVALUE __dereference(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __getstr(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __gettext(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __gettoday(PNODE, SYMTAB, BOOLEAN *);
@@ -404,6 +424,7 @@ PVALUE __surname(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __runsystem(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __table(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __tag(PNODE, SYMTAB, BOOLEAN *);
+PVALUE __test(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __titl(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __titlcase(PNODE, SYMTAB, BOOLEAN *);
 PVALUE __trim(PNODE, SYMTAB, BOOLEAN *);
@@ -453,7 +474,6 @@ extern INT nobuiltins;
 	}
 
 void dolock_node_in_cache(NODE, BOOLEAN lock);
-void dosemilock_record_in_cache(RECORD rec, BOOLEAN lock);
 
 /* Prototypes */
 void assign_iden(SYMTAB, STRING, PVALUE);
@@ -463,6 +483,7 @@ PNODE children_node(PACTX pactx, PNODE, STRING, STRING, PNODE);
 void clear_rptinfos(void);
 PNODE continue_node(PACTX pactx);
 void debug_show_one_pnode(PNODE);
+CNSTRING get_pvalue_type_name(INT ptype);
 PVALUE evaluate(PNODE, SYMTAB, BOOLEAN*);
 BOOLEAN evaluate_cond(PNODE, SYMTAB, BOOLEAN*);
 PVALUE evaluate_func(PNODE, SYMTAB, BOOLEAN*);
@@ -485,6 +506,7 @@ PNODE fornodes_node(PACTX pactx, PNODE, STRING, PNODE);
 PNODE fornotes_node(PACTX pactx, PNODE, STRING, PNODE);
 PNODE forothr_node(PACTX pactx, STRING, STRING, PNODE);
 PNODE forsour_node(PACTX pactx, STRING, STRING, PNODE);
+void free_iden(void *iden);
 void free_all_pnodes(void);
 void free_pnode_tree(PNODE);
 PNODE func_node(PACTX pactx, STRING, PNODE);
@@ -516,6 +538,7 @@ void set_rptfile_prop(PACTX pactx, STRING fname, STRING key, STRING value);
 void show_pnode(PNODE);
 void show_pnodes(PNODE);
 PNODE spouses_node(PACTX pactx, PNODE, STRING, STRING, STRING, PNODE);
+BOOLEAN start_output_file (STRING outfname);
 PNODE string_node(PACTX pactx, STRING);
 void trace_endl(void);
 void trace_out(STRING fmt, ...);
