@@ -37,6 +37,12 @@
 #include "indiseq.h"
 #include "gengedc.h"
 
+/*********************************************
+ * external/imported variables
+ *********************************************/
+
+extern STRING nonindx,nonvar1,nonset1,nonsetx,nonboox;
+
 /*********************************************
  * local function prototypes
  *********************************************/
@@ -72,18 +78,18 @@ __indiset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	INDISEQ newseq=0;
 	PVALUE newval=0;
-	PNODE var = (PNODE) iargs(node);
-	if (!iistype(var, IIDENT)) {
+	PNODE arg1 = (PNODE) iargs(node);
+	if (!iistype(arg1, IIDENT)) {
 		*eflg = TRUE;
-		/* TODO: change to prog_var_error */
-		prog_error(node, "the arg to indiset is not a variable.");
+		prog_var_error(node, stab, arg1, NULL, nonvar1, "indiset");
 		return NULL;
 	}
 	*eflg = FALSE;
 	newseq = create_indiseq_pval();
 	set_indiseq_value_funcs(newseq, &pvseq_fnctbl);
-	newval = create_pvalue(PSET, newseq);
-	assign_iden(stab, iident(var), newval);
+	newval = create_pvalue_from_seq(newseq);
+	assign_iden(stab, iident(arg1), newval);
+	/* gave val1 to stab, so don't clear it */
 	return NULL;
 }
 /*==================================+
@@ -93,59 +99,56 @@ __indiset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __addtoset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	NODE indi;
-	STRING key;
-	INDISEQ seq;
+	NODE indi=0;
+	STRING key=0;
+	INDISEQ seq=0;
 	PNODE arg1 = (PNODE) iargs(node), arg2 = inext(arg1),
 	    arg3 = inext(arg2);
 	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	PVALUE val2=0;
 	if (*eflg) {
-		prog_error(node, "1st arg to addtoset is not a set.");
+		prog_var_error(node, stab, arg1, val1, nonsetx, "addtoset", "1");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
 	indi = eval_indi(arg2, stab, eflg, NULL);
 	if (*eflg) {
-		prog_error(node, "2nd arg to addtoset must be a person.");
-		return NULL;
+		prog_var_error(node, stab, arg2, NULL, nonindx, "addtoset","2");
+		goto ats_exit;
 	}
-	if (!indi) return NULL;
+	if (!indi) goto ats_exit;
 	*eflg = TRUE;
 	if (!(key = strsave(rmvat(nxref(indi))))) {
 		prog_error(node, "major error in addtoset.");
-		return NULL;
+		goto ats_exit;
 	}
 	*eflg = FALSE;
 	val2 = evaluate(arg3, stab, eflg);
 	if (*eflg) {
 		prog_error(node, "3rd arg to addtoset is in error.");
-		return NULL;
+		goto ats_exit;
 	}
-	append_indiseq_pval(seq, key, NULL, val2, FALSE, TRUE);
-	/* delay to last minute lest it is a temp owning seq,
-	eg, addtoset(ancestorset(i),j) */
-	delete_pvalue(val1);
+	append_indiseq_pval(seq, key, NULL, val2, FALSE);
+ats_exit:
+	if (key) strfree(&key); /* append made its own copy */
+	/* delay to last minute val1 cleanup lest it is a temp owning seq,
+	    eg, addtoset(ancestorset(i),j) */
+	if (val1) delete_pvalue(val1);
 	return NULL;
 }
 /*======================================+
  * lengthset -- Find length of an INDISEQ
  *   lengthset(SET) -> INT
+ * Implementation Detail:
+ * - implemented using __length(), which
+ *   was changed to accept LISTs, SETs
+ *   and TABLEs
+ * - this function is DEPRECATED
  *=====================================*/
 PVALUE
 __lengthset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	INT len;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
-	if (*eflg) {
-		prog_error(node, "the arg to lengthset must be a set.");
-		return NULL;
-	}
-	seq = pvalue_to_seq(val);
-	len = seq ? length_indiseq(seq) : 0;
-	delete_pvalue(val);
-	return create_pvalue_from_int(len);
+	return __length(node, stab, eflg);
 }
 /*====================================+
  * inset -- See if person is in INDISEQ
@@ -157,29 +160,36 @@ __inset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	NODE indi;
 	STRING key;
 	INDISEQ seq;
+	BOOLEAN rel;
 	PNODE arg1 = (PNODE) iargs(node), arg2 = inext(arg1);
 	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	PVALUE valr=0;
 	if (*eflg ||!val1 || !(seq = pvalue_to_seq(val1))) {
 		*eflg = TRUE;
-		prog_error(node, "1st arg to inset must be a set.");
-		return NULL;
+		prog_var_error(node, stab, arg1, val1, nonsetx, "inset", "1");
+		goto inset_exit;
 	}
 	indi = eval_indi(arg2, stab, eflg, NULL);
 	if (*eflg) {
-		prog_error(node, "2nd arg to inset must be a person.");
-		return NULL;
+		prog_var_error(node, stab, arg2, NULL, nonindx, "inset", "2");
+		goto inset_exit;
 	}
-	if (!indi) return create_pvalue(PBOOL, (VPTR) FALSE);
-	if (!(key = strsave(rmvat(nxref(indi))))) {
-		*eflg = TRUE;
-		prog_error(node, "major error in inset.");
-		return NULL;
+	if (!indi) {
+		rel = FALSE;
+        } else { 
+		if (!(key = strsave(rmvat(nxref(indi))))) {
+			*eflg = TRUE;
+			prog_error(node, "major error in inset.");
+			goto inset_exit;
+		}
+		rel = in_indiseq(seq, key);
 	}
-	valr = create_pvalue(PBOOL, (VPTR) in_indiseq(seq, key));
-	/* delay to last minute lest it is a temp owning seq,
-	eg, inset(ancestorset(i),j) */
-	delete_pvalue(val1);
+	valr = create_pvalue_from_bool(rel);
+inset_exit:
+	/* delay delete of val1 to last minute lest it is a temp owning seq,
+	    eg, inset(ancestorset(i),j) */
+	if (val1) delete_pvalue(val1);
+	if (key) strfree(&key);
 	return valr;
 }
 /*===========================================+
@@ -198,35 +208,37 @@ __deletefromset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	PVALUE val3=0;
 	if (*eflg) {
-		prog_error(node, "1st arg to deletefromset must be a set.");
-		return NULL;
+		prog_var_error(node, stab, arg1, val1, nonsetx, "deletefromset", "1");
+		goto dfs_exit;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
 	indi = eval_indi(arg2, stab, eflg, NULL);
 	if (*eflg) {
-		prog_error(node, "2nd arg to deletefromset must be a person.");
-		return NULL;
+		prog_var_error(node, stab, arg2, NULL, nonindx, "deletefromset", "2");
+		goto dfs_exit;
 	}
-	if (!indi) return NULL;
+	if (!indi) goto dfs_exit;
 	*eflg = TRUE;
 	if (!(key = strsave(rmvat(nxref(indi))))) {
 		prog_error(node, "major error in deletefromset.");
-		return NULL;
+		goto dfs_exit;
 	}
 	*eflg = FALSE;
 	val3 = eval_and_coerce(PBOOL, arg3, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "3rd arg to deletefromset must be boolean.");
-		return NULL;
+		prog_var_error(node, stab, arg2, NULL, nonboox, "deletefromset", "3");
+		goto dfs_exit;
 	}
 	all = pvalue_to_bool(val3);
 	delete_pvalue(val3);
 	do {
 		rc = delete_indiseq(seq, key, NULL, 0);
 	} while (rc && all);
-	/* delay to last minute lest it is a temp owning seq,
-	eg, deletefromset(ancestorset(i),j) */
-	delete_pvalue(val1);
+dfs_exit:
+	/* delay delete of val1 to last minute lest it is a temp owning seq,
+	    eg, deletefromset(ancestorset(i),j) */
+	if (val1) delete_pvalue(val1);
+	if (key) strfree(&key);
 	return NULL;
 }
 /*================================+
@@ -237,9 +249,10 @@ PVALUE
 __namesort (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to namesort must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "namesort");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
@@ -257,9 +270,10 @@ PVALUE
 __keysort (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to namesort must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "namesort");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
@@ -277,15 +291,16 @@ PVALUE
 __valuesort (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to valuesort must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "valuesort");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
 	valuesort_indiseq(seq,eflg);
 	if (*eflg) {
-		prog_error(node, "missing or incorrect value for sort");
+		prog_error(node, _("missing or incorrect value for sort"));
 		return NULL;
 	}
 	/* delay to last minute lest it is a temp owning seq,
@@ -300,10 +315,11 @@ __valuesort (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __uniqueset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to uniqueset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "uniqueset");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
@@ -321,23 +337,23 @@ PVALUE
 __union (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	PNODE arg1 = (PNODE) iargs(node), arg2 = inext(arg1);
-	INDISEQ op2, op1;
+	INDISEQ op2=0, op1=0;
 	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	PVALUE val2=0;
 	if (*eflg) {
-		prog_error(node, "1st arg to union must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonsetx, "union", "1");
 		return NULL;
 	}
 	/* NULL indiseqs are possible, because of getindiset */
 	op1 = pvalue_to_seq(val1);
 	val2 = eval_and_coerce(PSET, arg2, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "2nd arg to union must be a set.");
+		prog_var_error(node, stab, arg2, val2, nonsetx, "union", "2");
 		return NULL;
 	}
 	op2 = pvalue_to_seq(val2);
 	op2 = union_indiseq(op1, op2);
-	set_pvalue(val1, PSET, op2);
+	set_pvalue_seq(val1, op2);
 	/* delay to last minute lest it is a temp owning seq,
 	eg, intersect(ancestorset(i),ancestorset(j)) */
 	delete_pvalue(val2);
@@ -351,22 +367,24 @@ PVALUE
 __intersect (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	PNODE arg1 = (PNODE) iargs(node), arg2 = inext(arg1);
-	INDISEQ op2, op1;
+	INDISEQ op2=0, op1=0;
 	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	PVALUE val2=0;
 	if (*eflg) {
-		prog_error(node, "1st arg to intersect must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonsetx, "intersect", "1");
 		return NULL;
 	}
 	/* NULL indiseqs are possible, because of getindiset */
 	op1 = pvalue_to_seq(val1);
 	val2 = eval_and_coerce(PSET, arg2, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "2nd arg to intersect must be a set.");
+		prog_var_error(node, stab, arg2, val2, nonsetx, "intersect", "2");
 		return NULL;
 	}
 	op2 = pvalue_to_seq(val2);
-	set_pvalue(val1, PSET, op2 = intersect_indiseq(op1, op2));
+	/* do actual interset */
+	op2 = intersect_indiseq(op1, op2);
+	set_pvalue_seq(val1, op2);
 	/* delay to last minute lest it is a temp owning seq,
 	eg, intersect(ancestorset(i),ancestorset(j)) */
 	delete_pvalue(val2);
@@ -380,26 +398,28 @@ PVALUE
 __difference (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	PNODE arg1 = (PNODE) iargs(node), arg2 = inext(arg1);
-	INDISEQ op2, op1;
-	PVALUE val = eval_and_coerce(PSET, arg1, stab, eflg);
-	PVALUE val2;
+	INDISEQ op2=0, op1=0;
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
+	PVALUE val2=0;
 	if (*eflg) {
-		prog_error(node, "1st arg to difference must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonsetx, "difference", "1");
 		return NULL;
 	}
 	/* NULL indiseqs are possible, because of getindiset */
-	op1 = pvalue_to_seq(val);
+	op1 = pvalue_to_seq(val1);
 	val2 = eval_and_coerce(PSET, arg2, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "2nd arg to difference must be a set.");
+		prog_var_error(node, stab, arg2, val2, nonsetx, "difference", "2");
 		return NULL;
 	}
 	op2 = pvalue_to_seq(val2);
-	set_pvalue(val, PSET, op2 = difference_indiseq(op1, op2));
+	/* do actual difference */
+	op2 = difference_indiseq(op1, op2);
+	set_pvalue_seq(val1, op2);
 	/* delay to last minute lest it is a temp owning seq,
 	eg, difference(ancestorset(i),ancestorset(j)) */
 	delete_pvalue(val2);
-	return val;
+	return val1;
 }
 /*=========================================+
  * parentset -- Create parent set of INDISEQ
@@ -408,17 +428,19 @@ __difference (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __parentset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to parentset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "parentset");
 		return NULL;
 	}
 	/* NULL indiseqs are possible, because of getindiset */
-	seq = pvalue_to_seq(val);
+	seq = pvalue_to_seq(val1);
+	/* do actual construction of parent set */
 	seq = parent_indiseq(seq);
-	set_pvalue(val, PSET, seq);
-	return val;
+	set_pvalue_seq(val1, seq);
+	return val1;
 }
 /*==========================================+
  * childset -- Create child set of an INDISEQ
@@ -427,15 +449,18 @@ __parentset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __childset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to childset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "childset");
 		return NULL;
 	}
-	ASSERT(seq = pvalue_to_seq(val));
-	set_pvalue(val, PSET, seq = child_indiseq(seq));
-	return val;
+	ASSERT(seq = pvalue_to_seq(val1));
+	/* do actual construction of child set */
+	seq = child_indiseq(seq);
+	set_pvalue_seq(val1, seq);
+	return val1;
 }
 /*==============================================+
  * siblingset -- Create sibling set of an INDISEQ
@@ -444,15 +469,17 @@ __childset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __siblingset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to siblingset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "siblingset");
 		return NULL;
 	}
-	ASSERT(seq = pvalue_to_seq(val));
-	set_pvalue(val, PSET, seq = sibling_indiseq(seq, FALSE));
-	return val;
+	ASSERT(seq = pvalue_to_seq(val1));
+	seq = sibling_indiseq(seq, FALSE);
+	set_pvalue_seq(val1, seq);
+	return val1;
 }
 /*============================================+
  * spouseset -- Create spouse set of an INDISEQ
@@ -461,15 +488,17 @@ __siblingset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __spouseset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to spouseset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "spouseset");
 		return NULL;
 	}
-	ASSERT(seq = pvalue_to_seq(val));
-	set_pvalue(val, PSET, seq = spouse_indiseq(seq));
-	return val;
+	ASSERT(seq = pvalue_to_seq(val1));
+	seq = spouse_indiseq(seq);
+	set_pvalue_seq(val1, seq);
+	return val1;
 }
 /*================================================+
  * ancestorset -- Create ancestor set of an INDISEQ
@@ -478,16 +507,17 @@ __spouseset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __ancestorset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to ancestorset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "ancestorset");
 		return NULL;
 	}
-	ASSERT(seq = pvalue_to_seq(val));
+	ASSERT(seq = pvalue_to_seq(val1));
 	seq = ancestor_indiseq(seq);
-	set_pvalue(val, PSET, seq);
-	return val;
+	set_pvalue_seq(val1, seq);
+	return val1;
 }
 /*====================================================+
  * descendentset -- Create descendent set of an INDISEQ
@@ -496,16 +526,17 @@ __ancestorset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __descendentset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to descendentset must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "descendentset");
 		return NULL;
 	}
-	ASSERT(seq = pvalue_to_seq(val));
+	ASSERT(seq = pvalue_to_seq(val1));
 	seq = descendent_indiseq(seq);
-	set_pvalue(val, PSET, seq);
-	return val;
+	set_pvalue_seq(val1, seq);
+	return val1;
 }
 /*===================================================+
  * gengedcom -- Generate GEDCOM output from an INDISEQ
@@ -514,10 +545,11 @@ __descendentset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __gengedcom (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to gengedcom must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "gengedcom");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
@@ -535,10 +567,11 @@ __gengedcom (PNODE node, SYMTAB stab, BOOLEAN *eflg)
  *==================================================*/
 PVALUE __gengedcomweak (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to gengedcomweak must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "gengedcomweak");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
@@ -556,10 +589,11 @@ PVALUE __gengedcomweak (PNODE node, SYMTAB stab, BOOLEAN *eflg)
  *==================================================*/
 PVALUE __gengedcomstrong (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	INDISEQ seq;
-	PVALUE val1 = eval_and_coerce(PSET, iargs(node), stab, eflg);
+	INDISEQ seq=0;
+	PNODE arg1 = (PNODE) iargs(node);
+	PVALUE val1 = eval_and_coerce(PSET, arg1, stab, eflg);
 	if (*eflg) {
-		prog_error(node, "the arg to gengedcomstrong must be a set.");
+		prog_var_error(node, stab, arg1, val1, nonset1, "gengedcomstrong");
 		return NULL;
 	}
 	ASSERT(seq = pvalue_to_seq(val1));
