@@ -50,7 +50,7 @@
  * external/imported variables
  *********************************************/
 
-extern STRING qSnotone, qSifone;
+extern STRING qSnotonei, qSifonei;
 extern STRING nonint1,nonintx,nonstr1,nonstrx,nonlstx,nonvarx,nonnodx;
 extern STRING nonind1,nonindx,nonfam1,nonrecx,nonnod1,nonnodx;
 
@@ -86,7 +86,7 @@ BOOLEAN prog_trace = FALSE;
 PVALUE
 __extractnames (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	LIST list, temp;
+	LIST list=0;
 	STRING str, str2;
 	INT len, sind;
 	PNODE nexp = (PNODE) iargs(node);
@@ -128,14 +128,14 @@ __extractnames (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	/* now create all the values, whether or not we found a NAME line */
 	*eflg = FALSE;
 	str = (line ? nval(line) : 0);
-	temp = create_list();
 	if (str && str[0]) {
-		name_to_list(str, temp, &len, &sind);
+		LIST temp = name_to_list(str, &len, &sind);
 		/* list has string elements */
 		FORLIST(temp, el)
 			str2 = (STRING)el;
 			push_list(list, create_pvalue_from_string(str2));
 		ENDLIST
+		destroy_list(temp);
 	} else {
 		/* no NAME line or empty NAME line */
 		len = 0;
@@ -152,7 +152,7 @@ __extractnames (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __extractplaces (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	LIST list, temp;
+	LIST list=0, temp=0;
 	STRING str, str2;
 	INT len;
 	PNODE nexp = (PNODE) iargs(node);
@@ -174,7 +174,7 @@ __extractplaces (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 		return NULL;
 	}
 	list = pvalue_to_list(val);
-	delete_pvalue(val);
+	delete_pvalue(val); /* Could this inadvertently delete the list? */
 	if (list)
 		make_list_empty(list);
 	else
@@ -190,8 +190,7 @@ __extractplaces (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	if (strcmp("PLAC", ntag(line)) && !(line = PLAC(line))) return NULL;
 	str = nval(line);
 	if (!str || *str == 0) return NULL;
-	temp = create_list();
-	place_to_list(str, temp, &len);
+	temp = place_to_list(str, &len);
 	FORLIST(temp, el)
 		str2 = (STRING)el; /* place_to_list made list of strings */
 		push_list(list, create_pvalue_from_string(str2));
@@ -206,7 +205,7 @@ __extractplaces (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 PVALUE
 __extracttokens (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
-	LIST list, temp;
+	LIST list=0, temp=0;
 	INT len;
 	STRING str, dlm;
 	PNODE sexp = (PNODE) iargs(node);
@@ -244,8 +243,7 @@ __extracttokens (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	}
 	*eflg = FALSE;
 	insert_symtab(stab, iident(lvar), create_pvalue_from_int(0));
-	temp = create_list();
-	value_to_list(str, temp, &len, dlm);
+	temp = value_to_list(str, &len, dlm);
 	FORLIST(temp, el)
 		push_list(list, create_pvalue_from_string((STRING)el));
 	ENDLIST
@@ -461,8 +459,8 @@ allocsubstring (STRING s, INT i, INT j)
 PVALUE
 __chooseindi (PNODE node, SYMTAB stab, BOOLEAN * eflg)
 {
-	NODE indi;
-	INDISEQ seq;
+	NODE indi=0;
+	INDISEQ seq=0;
 	PVALUE val = eval_and_coerce(PSET, iargs(node), stab, eflg);
 	if (*eflg) {
 		prog_error(node, "the arg to chooseindi is not a set of persons");
@@ -471,7 +469,7 @@ __chooseindi (PNODE node, SYMTAB stab, BOOLEAN * eflg)
 	seq = pvalue_to_seq(val);
 	delete_pvalue(val);
 	if (!seq || length_indiseq(seq) < 1) return NULL;
-	indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifone), _(qSnotone)));
+	indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifonei), _(qSnotonei)));
 	if (!indi) return NULL;
 	return create_pvalue_from_indi(indi);
 }
@@ -493,12 +491,12 @@ __choosesubset (PNODE node, SYMTAB stab, BOOLEAN * eflg)
 	delete_pvalue(val);
 	if (!seq || length_indiseq(seq) < 1) return NULL;
 	newseq = copy_indiseq(seq);
-	msg = (length_indiseq(newseq) > 1) ? _(qSnotone ): _(qSifone);
+	msg = (length_indiseq(newseq) > 1) ? _(qSnotonei): _(qSifonei);
 	if (-1 == choose_list_from_indiseq(msg, newseq)) {
 		remove_indiseq(newseq);
 		newseq = NULL;
 	}
-	return create_pvalue(PSET, (VPTR)newseq);
+	return create_pvalue_from_seq(newseq);
 }
 /*=========================================================+
  * choosechild -- Have user choose child of person or family
@@ -508,9 +506,7 @@ PVALUE
 __choosechild (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	INT type;
-	STRING key;
-	NODE indi, fam;
-	INDISEQ seq;
+	CNSTRING key=0;
 	CACHEEL cel;
 	PVALUE val = evaluate(iargs(node), stab, eflg);
 	if (*eflg || !val || ((type = which_pvalue_type(val)) != PINDI && type != PFAM)) {
@@ -521,21 +517,22 @@ __choosechild (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	cel = pvalue_to_cel(val);
 	delete_pvalue(val);
 	if (!cel) return create_pvalue_from_indi(NULL);
-	key = ckey(cel);
+	key = cacheel_to_key(cel);
 	if (*key == 'I') {
-		indi = cnode(cel);
-		seq = indi_to_children(indi);
+		NODE indi = cacheel_to_node(cel);
+		INDISEQ seq = indi_to_children(indi);
 		if (!seq || length_indiseq(seq) < 1)
 			return create_pvalue_from_indi(NULL);
-		indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifone), _(qSnotone)));
+		indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifonei), _(qSnotonei)));
 		remove_indiseq(seq);
 		return create_pvalue_from_indi(indi); /* indi may be NULL */
 	} else if (*key == 'F') {
-		fam = key_to_fam(key);
-		seq = fam_to_children(fam);
+		NODE fam = key_to_fam(key);
+		NODE indi=0;
+		INDISEQ seq = fam_to_children(fam);
 		if (!seq || length_indiseq(seq) < 1)
 			return create_pvalue_from_indi(NULL);
-		indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifone), _(qSnotone)));
+		indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifonei), _(qSnotonei)));
 		remove_indiseq(seq);
 		return create_pvalue_from_indi(indi); /* indi may be NULL */
 	}
@@ -559,7 +556,7 @@ __choosespouse (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	seq = indi_to_spouses(indi);
 	if (!seq || length_indiseq(seq) < 1)
 		return create_pvalue_from_indi(NULL);
-	indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifone), _(qSnotone)));
+	indi = nztop(choose_from_indiseq(seq, DOASK1, _(qSifonei), _(qSnotonei)));
 	remove_indiseq(seq);
 	return create_pvalue_from_indi(indi); /* indi may be NULL */
 }
@@ -579,7 +576,7 @@ __choosefam (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	seq = indi_to_families(indi, TRUE);
 	if (!seq || length_indiseq(seq) < 1)
 		return create_pvalue_from_fam(NULL);
-	fam = nztop(choose_from_indiseq(seq, DOASK1, _(qSifone), _(qSnotone)));
+	fam = nztop(choose_from_indiseq(seq, DOASK1, _(qSifonei), _(qSnotonei)));
 	remove_indiseq(seq);
 	return create_pvalue_from_fam(fam); /* fam may be NULL */
 }
@@ -594,10 +591,7 @@ makestring (PVALUE val, STRING str, INT len, BOOLEAN *eflg)
 	str[0]=0;
 
 	switch(which_pvalue_type(val)) {
-		case PNONE: 
-			llstrapps(str, len, uu8, "<NONE>");
-			break;
-		case PANY:
+		case PNULL:
 			llstrapps(str, len, uu8, "<NULL>");
 			break;
 		case PINT:
@@ -628,7 +622,7 @@ makestring (PVALUE val, STRING str, INT len, BOOLEAN *eflg)
 		case PEVEN:
 		case POTHR:
 			{
-				RECORD rec = pvalue_to_rec(val);
+				RECORD rec = pvalue_to_record(val);
 				NODE node = nztop(rec);
 				STRING txt = generic_to_list_string(node, NULL, len, " ", NULL, TRUE);
 				llstrapps(str, len, uu8, txt);
@@ -869,13 +863,13 @@ __lastfam (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	return create_pvalue_from_fam_keynum(xref_lastf());
 }
 /*=============================================+
- * getrecord -- Read GEDCOM record from database
- *  usage: getrecord(STRING) -> NODE
+ * __dereference -- Read top node of GEDCOM record from database
+ *  usage: dereference(STRING) -> NODE
  *  usage: dereference(STRING) -> NODE
  *  NOTE: persons and families NOT cached!
  *============================================*/
 PVALUE
-__getrecord (PNODE node, SYMTAB stab, BOOLEAN *eflg)
+__dereference (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	STRING key;
 	PNODE arg = iargs(node);
@@ -884,11 +878,12 @@ __getrecord (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	STRING rawrec = NULL;
 	NODE node2 = NULL;
 	if (*eflg) {
-		prog_var_error(node, stab, arg, val, nonstr1, "getrecord");
+		prog_var_error(node, stab, arg, val, nonstr1, "dereference");
 		return NULL;
 	}
 	key = pvalue_to_string(val);
 	if (*key == '@') key = rmvat(key);
+	if (!key) key=""; /* rmvat can return null */
 	if (*key == 'I' || *key == 'F' || *key == 'S' ||
 	    *key == 'E' || *key == 'X') {
 		rawrec = retrieve_raw_record(key, &len);
@@ -945,11 +940,14 @@ __rjustify (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	}
 	len = pvalue_to_int(val2);
 	delete_pvalue(val2);
-	set_pvalue_string(val1, rightjustify(str, len));
+	str = rightjustify(str, len); /* newly alloc'd */
+	set_pvalue_string(val1, str);
+	strfree(&str);
 	return val1;
 }
 /*===========================================
  * rightjustify -- Right justify string value
+ *  returns heap-allocated string
  *=========================================*/
 static STRING
 rightjustify (STRING str, INT len)
@@ -1030,7 +1028,8 @@ __savenode (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 	}
 	line = pvalue_to_node(val);
 	if (!line) return val;
-	set_pvalue(val, PGNODE, (VPTR) copy_nodes(line, TRUE, TRUE));
+	line = copy_nodes(line, TRUE, TRUE);
+	set_pvalue_node(val, line);
 	return val;
 }
 /*===================================================+
@@ -1042,6 +1041,7 @@ __genindiset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 {
 	PNODE arg = (PNODE) iargs(node);
 	STRING name;
+	PVALUE seqval=0;
 	PVALUE val1 = eval_and_coerce(PSTRING, arg, stab, eflg);
 	if (*eflg) {
 		prog_var_error(node, stab, arg, val1, nonstrx, "genindiset" , "1");
@@ -1057,10 +1057,11 @@ __genindiset (PNODE node, SYMTAB stab, BOOLEAN *eflg)
 		prog_error(node, "2nd arg to genindiset must be a variable");
 		return NULL;
 	}
-	assign_iden(stab, iident(arg), create_pvalue(PSET, (VPTR)NULL));
+	seqval = create_pvalue_from_seq(NULL);
+	assign_iden(stab, iident(arg), seqval);
 	if (!name || *name == 0) return NULL;
-	assign_iden(stab, iident(arg), create_pvalue(PSET,
-	    (VPTR)str_to_indiseq(name, 'I')));
+	seqval = create_pvalue_from_seq(str_to_indiseq(name, 'I'));
+	assign_iden(stab, iident(arg), seqval);
 	return NULL;
 }
 /*POINT*/
@@ -1285,17 +1286,18 @@ sortimpl (PNODE node, SYMTAB stab, BOOLEAN *eflg, BOOLEAN fwd)
 
 	/* reorder the values collection (1st argument) */
 	if (list_vals) {
-		struct tag_list_iter listit;
+		LIST_ITER listit=0;
 		VPTR ptr=0;
 		i=0;
 		if (fwd)
-			begin_list_rev(list_vals, &listit);
+			listit = begin_list_rev(list_vals);
 		else
-			begin_list(list_vals, &listit);
-		while (next_list_ptr(&listit, &ptr)) {
-			change_list_ptr(&listit, index[i]->value);
+			listit = begin_list(list_vals);
+		while (next_list_ptr(listit, &ptr)) {
+			change_list_ptr(listit, index[i]->value);
 			++i;
 		}
+		end_list_iter(&listit);
 	} else {
 		INT j;
 		ASSERT(arr_vals);
@@ -1306,17 +1308,18 @@ sortimpl (PNODE node, SYMTAB stab, BOOLEAN *eflg, BOOLEAN fwd)
 		}
 	}
 	if (list_keys) {
-		struct tag_list_iter listit;
+		LIST_ITER listit=0;
 		VPTR ptr=0;
 		i=0;
 		if (fwd)
-			begin_list_rev(list_keys, &listit);
+			listit = begin_list_rev(list_keys);
 		else
-			begin_list(list_keys, &listit);
-		while (next_list_ptr(&listit, &ptr)) {
-			change_list_ptr(&listit, index[i]->key);
+			listit = begin_list(list_keys);
+		while (next_list_ptr(listit, &ptr)) {
+			change_list_ptr(listit, index[i]->key);
 			++i;
 		}
+		end_list_iter(&listit);
 	} else if (arr_keys) {
 		INT j;
 		for (i=0; i<nsort; ++i) {
