@@ -40,6 +40,7 @@
 #include "zstr.h"
 #include "vtable.h"
 #include "array.h"
+#include "object.h"
 
 /*********************************************
  * local function prototypes
@@ -50,24 +51,28 @@ static FLOAT bool_to_float(BOOLEAN);
 static INT bool_to_int(BOOLEAN);
 static void clear_pv_indiseq(INDISEQ seq);
 static void clear_pvalue(PVALUE val);
+
 static PVALUE create_pvalue_from_keynum_impl(INT i, INT ptype);
-static PVALUE create_pvalue_from_key_impl(STRING key, INT ptype);
+static PVALUE create_pvalue_from_key_impl(CNSTRING key, INT ptype);
 static PVALUE create_pvalue_from_record(RECORD rec, INT ptype);
-static BOOLEAN eq_pstrings(PVALUE val1, PVALUE val2);
+/* static BOOLEAN eq_pstrings(PVALUE val1, PVALUE val2); */
 static int float_to_int(float f);
 static void free_float_pvalue(PVALUE val);
+static BOOLEAN is_record_pvaltype(INT valtype);
 static OBJECT pvalue_copy(OBJECT obj, int deep);
 static void pvalue_destructor(VTABLE *obj);
-static void table_pvcleaner(ENTRY ent);
+static void table_pvcleaner(CNSTRING key, UNION uval);
 
 /*********************************************
  * local variables
  *********************************************/
 
+/* These names are offset by the number of their type */
+/* PFLOAT == 4, so "PFLOAT" must be at array offset 4 */
 static char *ptypes[] = {
-	"PNONE", "PANY", "PINT", "PLONG", "PFLOAT", "PBOOL", "PSTRING",
+	"?", "PNULL", "PINT", "PLONG", "PFLOAT", "PBOOL", "PSTRING",
 	"PGNODE", "PINDI", "PFAM", "PSOUR", "PEVEN", "POTHR", "PLIST",
-	"PTABLE", "PSET"
+	"PTABLE", "PSET", "PARRAY"
 };
 static struct tag_vtable vtable_for_pvalue = {
 	VTABLE_MAGIC
@@ -105,40 +110,59 @@ create_pvalue (INT type, VPTR value)
 void
 set_pvalue (PVALUE val, INT type, VPTR value)
 {
-	if (type == ptype(val) && value == pvalue(val)) {
+	if (type == ptype(val) && value == pvalvv(val)) {
 		/* self-assignment */
+		/* already have the value, or hold the pointer & reference */
 		return;
 	}
+
+	/* record pvalues each have their own RECORD on the heap */
+
 	clear_pvalue(val);
 
 	/* sanity check */
 	switch(type) {
-	case PNONE: 
-	case PANY:
+	case PNULL:
 		ASSERT(!value);
 		break;
 	}
 	/* types that don't simply assign pointer */
+	/* old refers to value passed in */
+	/* new refers to newly allocated copy to set */
 	switch(type) {
 	case PSTRING:
 		{
-		/* always copies string so caller doesn't have to */
-		if (value)
-			value = (VPTR) strsave((STRING) value);
+			/* always copies string so caller doesn't have to */
+			if (value) {
+				STRING strold = (STRING)value;
+				STRING strnew = strsave(strold);
+				value = strnew;
+			}
 		break;
 		}
 	case PFLOAT:
 		{
+			float valold, *ptrnew;
+			ASSERT(value);
 			/* floats don't fit into VPTR, so we're using heap copies */
-			float *ptr = (float *)stdalloc(sizeof(*ptr));
-			*ptr = *(float *)value;
-			value = ptr;
+			valold = *(float *)value;
+			/* allocate new pointer & copy float into place */
+			ptrnew = (float *)stdalloc(sizeof(*ptrnew));
+			*ptrnew = valold;
+			value = ptrnew;
 		}
 		break;
 	}
 
 	ptype(val) = type;
-	pvalue(val) = value;
+	pvalvv(val) = value;
+
+	if (is_record_pvaltype(type)) {
+		RECORD rec = pvalue_to_record(val);
+		if (rec) {
+			addref_record(rec);
+		}
+	}
 
 	/* reference counted types and so forth */
 	switch(type) {
@@ -154,13 +178,13 @@ set_pvalue (PVALUE val, INT type, VPTR value)
 	case PLIST:
 		{
 			LIST list = pvalue_to_list(val);
-			++lrefcnt(list);
+			addref_list(list);
 		}
 		break;
 	case PTABLE:
 		{
 			TABLE table = pvalue_to_table(val);
-			++table->refcnt;
+			addref_table(table);
 		}
 		break;
 	case PSET:
@@ -173,28 +197,6 @@ set_pvalue (PVALUE val, INT type, VPTR value)
 		}
 		break;
 	}
-	if (is_record_pvalue(val)) {
-		/* instead of reference counting records, we lock them */
-		dosemilock_record_in_cache(pvalue_to_rec(val), TRUE);
-	}
-}
-/*========================================
- * dosemilock_record_in_cache -- semi-Lock/unlock record
- *  (if possible)
- * Created: 2003-02-06 (Perry Rapp)
- *======================================*/
-void
-dosemilock_record_in_cache (RECORD rec, BOOLEAN lock)
-{
-	if (rec) {
-		CACHEEL cel = rec->cel;
-		if (cel) {
-				if (lock)
-					semilock_cache(cel);
-				else
-					unsemilock_cache(cel);
-		}
-	}
 }
 /*========================================
  * dolock_node_in_cache -- Lock/unlock node in cache
@@ -204,6 +206,9 @@ dosemilock_record_in_cache (RECORD rec, BOOLEAN lock)
 void
 dolock_node_in_cache (NODE node, BOOLEAN lock)
 {
+	node = node;	/* NOTUSED */
+	lock = lock;	/* NOTUSED */
+
 #if NOT_WORKING_ON_LARGE_DATA_SETS
 /* This leads to cache overflow, so there is something
 wrong here - Perry, 2003-03-07 */
@@ -234,10 +239,10 @@ clear_pvalue (PVALUE val)
 	switch (ptype(val)) {
 	/*
 	embedded values have no referenced memory to clear
-	PINT, PBOOLEAN  (PLONG is unused)
+	PINT, PBOOLEAN 
 	*/
 	/*
-	PANY is a null value
+	PNULL is a null value
 	*/
 	case PGNODE:
 		{
@@ -251,7 +256,6 @@ clear_pvalue (PVALUE val)
 			}
 		}
 		return;
-	/* nodes from cache handled below switch - PINDI, PFAM, PSOUR, PEVEN, POTHR */
 	case PFLOAT:
 		free_float_pvalue(val);
 		return;
@@ -266,21 +270,13 @@ clear_pvalue (PVALUE val)
 	case PLIST:
 		{
 			LIST list = pvalue_to_list(val);
-			--lrefcnt(list);
-			if (!lrefcnt(list)) {
-				remove_list(list, delete_vptr_pvalue);
-			}
+			release_list(list);
 		}
 		return;
 	case PTABLE:
 		{
 			TABLE table = pvalue_to_table(val);
-			--table->refcnt;
-			if (!table->refcnt) {
-				/* TODO: this will go away when pvalues go away */
-				traverse_table(table, table_pvcleaner);
-				destroy_table(table);
-			}
+			release_table(table);
 		}
 		return;
 	case PSET:
@@ -296,13 +292,11 @@ clear_pvalue (PVALUE val)
 			}
 		}
 		return;
+	/* record nodes handled below (PINDI, PFAM, PSOUR, PEVEN, POTHR) */
 	}
 	if (is_record_pvalue(val)) {
-		/*
-		unlock any cache elements
-		don't worry about memory - it is owned by cache
-		*/
-		dosemilock_record_in_cache(pvalue_to_rec(val), FALSE);
+		RECORD rec = pvalue_to_record(val);
+		release_record(rec);
 	}
 }
 /*========================================
@@ -316,10 +310,10 @@ clear_pv_indiseq (INDISEQ seq)
 	/* NUL value indiseqs can get into reports via getindiset */
 	ASSERT(IValtype(seq) == ISVAL_PTR || IValtype(seq) == ISVAL_NUL);
 	FORINDISEQ(seq, el, ncount)
-		val = (PVALUE) sval(el).w;
+		val = (PVALUE)element_pval(el);
 		if (val) {
 			delete_pvalue(val);
-			sval(el).w = NULL;
+			set_element_pval(el, NULL);
 		}
 	ENDINDISEQ
 }
@@ -329,11 +323,12 @@ clear_pv_indiseq (INDISEQ seq)
  * Created: 2001/03/24, Perry Rapp
  *======================================*/
 static void
-table_pvcleaner (ENTRY ent)
+table_pvcleaner (CNSTRING key, UNION uval)
 {
-	PVALUE val = ent->uval.w;
+	PVALUE val = uval.w;
+	key=key; /* unused */
 	delete_pvalue(val);
-	ent->uval.w = NULL;
+	uval.w = NULL;
 }
 /*========================================
  * delete_vptr_pvalue -- Delete a program value
@@ -347,16 +342,21 @@ delete_vptr_pvalue (VPTR ptr)
 	delete_pvalue(val);
 }
 /*========================================
- * delete_pvalue_wrapper -- Delete the pvalue
- * shell, but not the value inside
- * Created: 2003-02-02 (Perry Rapp)
+ * remove_node_and_delete_pvalue -- Remove
+ *  node inside pvalue, and delete pvalue
  *======================================*/
-void
-delete_pvalue_wrapper (PVALUE val)
+NODE
+remove_node_and_delete_pvalue (PVALUE * pval)
 {
-	if (!val) return;
-	pvalue(val) = 0; /* remove pointer to payload */
-	delete_pvalue(val);
+	NODE node=0;
+	if (*pval) {
+		PVALUE vl= *pval;
+		node = pvalue_to_node(vl);
+		pvalvv(vl) = 0; /* remove pointer to payload */
+		delete_pvalue(vl);
+	}
+	*pval = 0;
+	return node;
 }
 /*========================================
  * delete_pvalue -- Delete a program value
@@ -384,13 +384,14 @@ delete_pvalue_ptr (PVALUE * valp)
 /*====================================
  * copy_pvalue -- Create a new pvalue & copy into it
  *  handles NULL
+ * delegates all the real work to create_pvalue
  *==================================*/
 PVALUE
 copy_pvalue (PVALUE val)
 {
 	if (!val)
 		return NULL;
-	return create_pvalue(ptype(val), pvalue(val));
+	return create_pvalue(ptype(val), pvalvv(val));
 }
 /*=====================================================
  * create_pvalue_from_indi -- Return indi as pvalue
@@ -400,8 +401,10 @@ copy_pvalue (PVALUE val)
 PVALUE
 create_pvalue_from_indi (NODE indi)
 {
-	CACHEEL cel = indi ? indi_to_cacheel_old(indi) : NULL;
-	return create_pvalue(PINDI, cel);
+	if (indi)
+		return create_pvalue_from_indi_key(indi_to_key(indi));
+	else
+		return create_pvalue(PINDI, 0);
 }
 /*=====================================================
  * create_pvalue_from_indi_key
@@ -409,18 +412,31 @@ create_pvalue_from_indi (NODE indi)
  * Created: 2000/12/30, Perry Rapp
  *===================================================*/
 PVALUE
-create_pvalue_from_indi_key (STRING key)
+create_pvalue_from_indi_key (CNSTRING key)
 {
 	return create_pvalue_from_key_impl(key, PINDI);
 }
+/*=====================================================
+ * create_pvalue_from_fam_key
+ *  handles NULL
+ *===================================================*/
+PVALUE
+create_pvalue_from_fam_key (STRING key)
+{
+	return create_pvalue_from_key_impl(key, PFAM);
+}
 /*=====================================================
  * create_pvalue_from_cel
  * Created: 2002/02/17, Perry Rapp
  *===================================================*/
 PVALUE
-create_pvalue_from_cel (CACHEEL cel)
+create_pvalue_from_cel (int type, CACHEEL cel)
 {
-	return create_pvalue(PINDI, cel);
+	PVALUE val=0;
+	RECORD rec = cel ? get_record_for_cel(cel) : 0;
+	val = create_pvalue(type, rec);
+	release_record(rec); /* ownership transferred to pvalue */
+	return val;
 }
 /*=====================================================
  * create_pvalue_from_indi_keynum -- Return indi as pvalue
@@ -442,7 +458,7 @@ PVALUE
 create_pvalue_from_fam (NODE fam)
 {
 	CACHEEL cel = fam ? fam_to_cacheel_old(fam) : NULL;
-	return create_pvalue(PFAM, cel);
+	return create_pvalue_from_cel(PFAM, cel);
 }
 /*====================================================
  * create_pvalue_from_fam_keynum -- Return indi as pvalue
@@ -485,13 +501,13 @@ create_pvalue_from_othr_keynum (INT i)
 /*=====================================================
  * create_pvalue_from_record -- Create pvalue from any node
  *  handles NULL
- * Created: 2001/03/20, Perry Rapp
+ * If rec is not null, it is given to new pvalue to own
  *===================================================*/
 static PVALUE
 create_pvalue_from_record (RECORD rec, INT ptype)
 {
-	CACHEEL cel = rec ? record_to_cacheel(rec) : NULL;
-	return create_pvalue(ptype, cel);
+	/* record pvalues simply point to their heap-alloc'd record */
+	return create_pvalue(ptype, rec);
 }
 /*====================================================
  * create_pvalue_from_keynum_impl -- Create pvalue for any type
@@ -523,7 +539,7 @@ create_pvalue_from_keynum_impl (INT i, INT ptype)
 static void
 free_float_pvalue (PVALUE val)
 {
-	float *ptr = (float *)pvalue(val);
+	float *ptr = (float *)pvalvv(val);
 	stdfree(ptr);
 }
 /*==================================
@@ -531,11 +547,12 @@ free_float_pvalue (PVALUE val)
  * Created: 2001/03/20, Perry Rapp
  *================================*/
 static PVALUE
-create_pvalue_from_key_impl (STRING key, INT ptype)
+create_pvalue_from_key_impl (CNSTRING key, INT ptype)
 {
 	/* report mode, so may return NULL */
-	RECORD rec = qkey_to_record(key);
+	RECORD rec = qkey_to_record(key); /* addref'd record */
 	PVALUE val = create_pvalue_from_record(rec, ptype);
+	release_record(rec); /* release our reference, now only pvalue holds */
 	return val;
 }
 /*==================================================
@@ -545,7 +562,7 @@ BOOLEAN
 is_numeric_pvalue (PVALUE val)
 {
 	INT type = ptype(val);
-	return type == PINT || type == PFLOAT || type == PANY;
+	return type == PINT || type == PFLOAT || type == PNULL;
 }
 /*===========================================================
  * eq_conform_pvalues -- Make the types of two values conform
@@ -557,14 +574,14 @@ eq_conform_pvalues (PVALUE val1, PVALUE val2, BOOLEAN *eflg)
 
 	ASSERT(val1 && val2);
 	if (ptype(val1) == ptype(val2)) return;
-	if (ptype(val1) == PANY && pvalue(val1) == NULL)
+	if (ptype(val1) == PNULL)
 		ptype(val1) = ptype(val2);
-	if (ptype(val2) == PANY && pvalue(val2) == NULL)
+	if (ptype(val2) == PNULL)
 		ptype(val2) = ptype(val1);
 	if (ptype(val1) == ptype(val2)) return;
-	if (ptype(val1) == PINT && pvalue(val1) == 0 && !is_numeric_pvalue(val2))
+	if (ptype(val1) == PINT && pvalue_to_int(val1) == 0 && !is_numeric_pvalue(val2))
 		ptype(val1) = ptype(val2);
-	if (ptype(val2) == PINT && pvalue(val2) == 0 && !is_numeric_pvalue(val1))
+	if (ptype(val2) == PINT && pvalue_to_int(val2) == 0 && !is_numeric_pvalue(val1))
 		ptype(val2) = ptype(val1);
 	if (ptype(val1) == ptype(val2)) return;
 	if (is_numeric_pvalue(val1) && is_numeric_pvalue(val2)) {
@@ -591,21 +608,21 @@ coerce_pvalue (INT type, PVALUE val, BOOLEAN *eflg)
 
 	if (type == PBOOL) {
 		/* Anything is convertible to PBOOL */
-		BOOLEAN boo = (pvalue(val) != NULL);
+		BOOLEAN boo = (pvalvv(val) != NULL);
 		set_pvalue_bool(val, boo);
 		return;
 	}
-	/* Anything is convertible to PANY */
+	/* Anything is convertible to PNULL */
 	/* Perry, 2002.02.16: This looks suspicious to me, but I 
 	don't know how it is used -- it might be used in some
 	eq_conform_pvalues call(s) ? */
-	if (type == PANY) {
-		ptype(val) = PANY;
+	if (type == PNULL) {
+		ptype(val) = PNULL;
 		return;
 	}
 
-	/* PANY or PINT with NULL (0) value is convertible to any scalar (1995.07.31) */
-	if ((ptype(val) == PANY || ptype(val) == PINT) && pvalue(val) == NULL) {
+	/* PNULL or PINT with NULL (0) value is convertible to any scalar (1995.07.31) */
+	if (ptype(val) == PNULL || (ptype(val) == PINT && pvalue_to_int(val) == 0)) {
 		if (type == PSET || type == PTABLE || type == PLIST) goto bad;
 		/*
 		  INTs convert to FLOATs numerically further down, no special 
@@ -669,7 +686,7 @@ coerce_pvalue (INT type, PVALUE val, BOOLEAN *eflg)
 		}
 		break;
 	/* Nothing else is convertible to anything else */
-	/* record types (PINDI...), PANY, PGNODE */
+	/* record types (PINDI...), PNULL, PGNODE */
 	}
 
 	/* fall through to failure */
@@ -693,7 +710,15 @@ which_pvalue_type (PVALUE val)
 BOOLEAN
 is_record_pvalue (PVALUE value)
 {
-	switch (ptype(value)) {
+	return is_record_pvaltype(ptype(value));
+}
+/*========================================
+ * is_record_pvaltype -- Does pvalue contain record ?
+ *======================================*/
+static BOOLEAN
+is_record_pvaltype (INT valtype)
+{
+	switch (valtype) {
 	case PINDI: case PFAM: case PSOUR: case PEVEN: case POTHR:
 		return TRUE;
 	}
@@ -728,18 +753,34 @@ eqv_pvalues (VPTR ptr1, VPTR ptr2)
 	BOOLEAN rel = FALSE;
 	if(val1 && val2 && (ptype(val1) == ptype(val2))) {
 		switch (ptype(val1)) {
+		/* types with value semantics do value comparison */
 		case PSTRING:
-			v1 = pvalue(val1);
-			v2 = pvalue(val2);
+			v1 = pvalue_to_string(val1);
+			v2 = pvalue_to_string(val2);
 			if(v1 && v2) rel = eqstr(v1, v2);
 			else rel = (v1 == v2);
 			break;
 		case PFLOAT:
 			rel = (pvalue_to_float(val1) == pvalue_to_float(val2));
 			break;
+		case PINT:
+			rel = (pvalue_to_int(val1) == pvalue_to_int(val2));
+			break;
+		case PBOOL:
+			rel = (pvalue_to_bool(val1) == pvalue_to_bool(val2));
+			break;
+		case PINDI: case PFAM: case PSOUR: case PEVEN: case POTHR:
+		{
+		    RECORD rec1,rec2;
+		    rec1 = pvalue_to_record(val1);
+		    rec2 = pvalue_to_record(val2);
+		    if (rec1 && rec2) rel = eqstrn(nzkey(rec1),nzkey(rec2),MAXKEYWIDTH+1);
+		    else rel = (rec1  == rec2);
+		    break;
+		 }
 		/* for everything else, just compare value pointer */
 		default:
-			rel = (pvalue(val1) == pvalue(val2));
+			rel = (pvalvv(val1) == pvalvv(val2));
 			break;
 		}
 	}
@@ -766,14 +807,17 @@ bad_type_error (CNSTRING op, ZSTR *zerr, PVALUE val1, PVALUE val2)
  * eq_pstrings -- Compare two PSTRINGS
  *  Caller is responsible for ensuring these are PSTRINGS
  *=============================================*/
+/* unused
 static BOOLEAN
 eq_pstrings (PVALUE val1, PVALUE val2)
 {
-	STRING str1 = pvalue(val1), str2 = pvalue(val2);
+	STRING str1 = pvalue_to_string(val1);
+	STRING str2 = pvalue_to_string(val2);
 	if (!str1) str1 = "";
 	if (!str2) str2 = "";
 	return eqstr(str1, str2);
 }
+unused */
 /*===========================================
  * eq_pvalues -- See if two PVALUEs are equal
  * Result into val1, deletes val2
@@ -789,18 +833,9 @@ eq_pvalues (PVALUE val1, PVALUE val2, BOOLEAN *eflg, ZSTR * zerr)
 		bad_type_error("eq", zerr, val1, val2);
 		return;
 	}
-	switch (ptype(val1)) {
-	case PSTRING:
-		rel = eq_pstrings(val1, val2);
-		break;
-	case PFLOAT:
-		rel = (pvalue_to_float(val1) == pvalue_to_float(val2));
-		break;
-		/* for everything else, just compare value pointer */
-	default:
-		rel = (pvalue(val1) == pvalue(val2));
-		break;
-	}
+	rel = eqv_pvalues(val1, val2);
+
+	/* Now store answer into val1, and delete val2 */
 	set_pvalue_bool(val1, rel);
 	delete_pvalue(val2);
 }
@@ -819,17 +854,9 @@ ne_pvalues (PVALUE val1, PVALUE val2, BOOLEAN *eflg, ZSTR * zerr)
 		bad_type_error("ne", zerr, val1, val2);
 		return;
 	}
-	switch (ptype(val1)) {
-	case PSTRING:
-		rel = !eq_pstrings(val1, val2);
-		break;
-	case PFLOAT:
-		rel = (pvalue_to_float(val1) != pvalue_to_float(val2));
-		break;
-	default:
-		rel = (pvalue(val1) != pvalue(val2));
-		break;
-	}
+	rel = !eqv_pvalues(val1, val2);
+
+	/* Now store answer into val1, and delete val2 */
 	set_pvalue_bool(val1, rel);
 	delete_pvalue(val2);
 }
@@ -843,6 +870,19 @@ show_pvalue (PVALUE val)
 	llwprintf(zs_str(zstr));
 	zs_free(&zstr);
 }
+/*=================================================
+ * get_pvalue_type_name -- Return static string name of pvalue type
+ *  eg, get_pvalue_type_name(PTABLE) => "PTABLE"
+ *===============================================*/
+CNSTRING
+get_pvalue_type_name (INT ptype)
+{
+	if (ptype >= 0 && ptype <= ARRSIZE(ptypes)) {
+		return ptypes[ptype];
+	} else {
+		return "INVALID ptype";
+	}
+}
 /*======================================================
  * debug_pvalue_as_string -- DEBUG routine that shows a PVALUE
  *  returns static buffer
@@ -863,9 +903,9 @@ describe_pvalue (PVALUE val)
 	}
 	type = ptype(val);
 	zs_appc(zstr, '<');
-	zs_apps(zstr, ptypes[type]);
+	zs_apps(zstr, get_pvalue_type_name(type));
 	zs_appc(zstr, ',');
-	if (pvalue(val) == NULL) {
+	if (pvalvv(val) == NULL) {
 		zs_apps(zstr, "NULL>");
 		return zstr;
 	}
@@ -880,16 +920,16 @@ describe_pvalue (PVALUE val)
 		zs_appf(zstr, "\"%s\"", pvalue_to_string(val));
 		break;
 	case PINDI:
+	case PFAM:
+	case PSOUR:
+	case PEVEN:
+	case POTHR:
 		{
-			NODE node;
-			CACHEEL cel = (CACHEEL) pvalue(val);
-			STRING nam;
-			if (!cnode(cel))
-				cel = key_to_indi_cacheel(ckey(cel));
-       		node = cnode(cel);
-			node = NAME(node);
-			nam = node ? nval(node) : _("{NoName}");
-			zs_appf(zstr, nam);
+			RECORD rec = pvalue_to_record(val);
+			if (rec)
+				zs_appf(zstr, nzkey(rec));
+			else
+				zs_appf(zstr, "NULL");
 		}
 		break;
 	case PLIST:
@@ -913,38 +953,45 @@ describe_pvalue (PVALUE val)
 			zs_appf(zstr, _pl("%d record", "%d records", n), n);
 		}
 		break;
+	case PARRAY:
+		{
+			ARRAY arr = pvalue_to_array(val);
+			INT n = get_array_size(arr);
+			zs_appf(zstr, _pl("%d element", "%d elements", n), n);
+		}
+		break;
 	default:
-		zs_appf(zstr, "%p", pvalue(val));
+		zs_appf(zstr, "%p", pvalvv(val));
 		break;
 	}
 	zs_appc(zstr, '>');
 	return zstr;
 }
 /*==================================
- * PANY: pvalue with no content
+ * PNULL: pvalue with no content
  *================================*/
 PVALUE
 create_pvalue_any (void)
 {
-	return create_pvalue(PANY, NULL);
+	return create_pvalue(PNULL, NULL);
 }
 /*==================================
- * PBOOL: pvalue containing a boolean
+ * PINT: pvalue containing an int
  *================================*/
 PVALUE
-create_pvalue_from_bool (BOOLEAN bval)
+create_pvalue_from_int (INT ival)
 {
-	return create_pvalue_from_int(bval);
+	return create_pvalue(PINT, (VPTR) ival);
 }
 void
-set_pvalue_bool (PVALUE val, BOOLEAN bnum)
+set_pvalue_int (PVALUE val, INT inum)
 {
-	set_pvalue(val, PBOOL, (VPTR)bnum);
+	set_pvalue(val, PINT, (VPTR)inum);
 }
-BOOLEAN
-pvalue_to_bool (PVALUE val)
+INT
+pvalue_to_int (PVALUE val)
 {
-	return (BOOLEAN)pvalue(val);
+	return (INT)pvalvv(val);
 }
 /*==================================
  * PFLOAT: pvalue containing a float
@@ -965,104 +1012,25 @@ float
 pvalue_to_float (PVALUE val)
 {
 	/* TODO: change when ptag goes to UNION */
-	return *(float*)pvalue(val);
-}
-float*
-pvalue_to_pfloat (PVALUE val)
-{
-	/* convenience for math */
-	/* TODO: change when ptag goes to UNION */
-	return (float*)pvalue(val);
+	return *(float*)pvalvv(val);
 }
 /*==================================
- * PGNODE: pvalue containing a GEDCOM node
+ * PBOOL: pvalue containing a boolean
  *================================*/
 PVALUE
-create_pvalue_from_node (NODE node)
+create_pvalue_from_bool (BOOLEAN bval)
 {
-	return create_pvalue(PGNODE, node);
+	return create_pvalue_from_int(bval);
 }
 void
-set_pvalue_node (PVALUE val, NODE node)
+set_pvalue_bool (PVALUE val, BOOLEAN bnum)
 {
-	set_pvalue(val, PGNODE, (VPTR)node);
+	set_pvalue(val, PBOOL, (VPTR)bnum);
 }
-NODE
-pvalue_to_node (PVALUE val)
+BOOLEAN
+pvalue_to_bool (PVALUE val)
 {
-	return (NODE)pvalue(val);
-}
-/*==================================
- * PINT: pvalue containing an int
- *================================*/
-PVALUE
-create_pvalue_from_int (INT ival)
-{
-	return create_pvalue(PINT, (VPTR) ival);
-}
-void
-set_pvalue_int (PVALUE val, INT inum)
-{
-	set_pvalue(val, PINT, (VPTR)inum);
-}
-INT
-pvalue_to_int (PVALUE val)
-{
-	return (INT)pvalue(val);
-}
-INT*
-pvalue_to_pint (PVALUE val)
-{
-	/* convenience for math */
-	return (INT *)&pvalue(val);
-}
-/*==================================
- * ARRAY: pvalue containing an array
- *================================*/
-ARRAY
-pvalue_to_array (PVALUE val)
-{
-	return (ARRAY)pvalue(val);
-}
-/*==================================
- * LIST: pvalue containing a list
- *================================*/
-LIST
-pvalue_to_list (PVALUE val)
-{
-	return (LIST)pvalue(val);
-}
-/*==================================
- * record pvalues (PINDI, PFAM, ...)
- *================================*/
-CACHEEL
-pvalue_to_cel (PVALUE val)
-{
-	/* also load record into direct cache */
-	CACHEEL cel = pvalue(val); /* may be NULL */
-	ASSERT(is_record_pvalue(val));
-	load_cacheel(cel); /* handles null cel ok */
-	return cel;
-}
-RECORD
-pvalue_to_rec (PVALUE val)
-{
-	CACHEEL cel = pvalue_to_cel(val);
-	if (!cel) return NULL;
-	return crecord(cel);
-}
-/*==================================
- * PSET: pvalue containing a set (INDISEQ)
- *================================*/
-PVALUE
-create_pvalue_from_set (INDISEQ seq)
-{
-	return create_pvalue(PSET, seq);
-}
-INDISEQ
-pvalue_to_seq (PVALUE val)
-{
-	return (INDISEQ)pvalue(val);
+	return (BOOLEAN)pvalvv(val);
 }
 /*==================================
  * PSTRING: pvalue containing a string
@@ -1080,12 +1048,96 @@ set_pvalue_string (PVALUE val, CNSTRING str)
 STRING
 pvalue_to_string (PVALUE val)
 {
-	return (STRING)pvalue(val);
+	return (STRING)pvalvv(val);
+}
+/*==================================
+ * PGNODE: pvalue containing a GEDCOM node
+ *================================*/
+PVALUE
+create_pvalue_from_node (NODE node)
+{
+	return create_pvalue(PGNODE, node);
+}
+void
+set_pvalue_node (PVALUE val, NODE node)
+{
+	set_pvalue(val, PGNODE, (VPTR)node);
+}
+NODE
+pvalue_to_node (PVALUE val)
+{
+	return (NODE)pvalvv(val);
+}
+/*==================================
+ * record pvalues (PINDI, PFAM, ...)
+ *================================*/
+RECORD
+pvalue_to_record (PVALUE val)
+{
+	RECORD rec = pvalvv(val); /* may be NULL */
+	ASSERT(is_record_pvalue(val));
+	return rec;
+}
+CACHEEL
+pvalue_to_cel (PVALUE val)
+{
+	RECORD rec = pvalue_to_record(val);
+	NODE root = nztop(rec); /* force record into cache */
+	CACHEEL cel = nzcel(rec);
+	root = root;	/* NOTUSED */
+	return cel;
+}
+/*==================================
+ * LIST: pvalue containing a list
+ *================================*/
+PVALUE
+create_pvalue_from_list (LIST list)
+{
+	return create_pvalue(PLIST, list);
+}
+LIST
+pvalue_to_list (PVALUE val)
+{
+	return (LIST)pvalvv(val);
+}
+/*==================================
+ * TABLE: pvalue containing a table
+ *================================*/
+PVALUE
+create_pvalue_from_table (TABLE tab)
+{
+	return create_pvalue(PTABLE, tab);
 }
 TABLE
 pvalue_to_table (PVALUE val)
 {
-	return (TABLE)pvalue(val);
+	return (TABLE)pvalvv(val);
+}
+/*==================================
+ * PSET: pvalue containing a set (INDISEQ)
+ *================================*/
+PVALUE
+create_pvalue_from_seq (INDISEQ seq)
+{
+	return create_pvalue(PSET, seq);
+}
+void
+set_pvalue_seq (PVALUE val, INDISEQ seq)
+{
+	set_pvalue(val, PSET, (VPTR)seq);
+}
+INDISEQ
+pvalue_to_seq (PVALUE val)
+{
+	return (INDISEQ)pvalvv(val);
+}
+/*==================================
+ * ARRAY: pvalue containing an array
+ *================================*/
+ARRAY
+pvalue_to_array (PVALUE val)
+{
+	return (ARRAY)pvalvv(val);
 }
 /*========================================
  * init_pvalue_vtable -- set vtable (for allocator in pvalalloc.c)
@@ -1143,4 +1195,33 @@ pvalues_collate (PVALUE val1, PVALUE val2)
 	}
 	return 0; /* TODO: what about other types ? */
 }
-
+/*=============================================
+ * create_new_pvalue_table -- Create new table inside new pvalue
+ *============================================*/
+PVALUE
+create_new_pvalue_table (void)
+{
+	TABLE tab = create_table_custom_vptr(delete_vptr_pvalue);
+	PVALUE val = create_pvalue_from_table(tab);
+	release_table(tab);
+	return val;
+}
+/*=============================================
+ * create_new_pvalue_list -- Create new list inside new pvalue
+ *============================================*/
+PVALUE
+create_new_pvalue_list (void)
+{
+	LIST list = create_list3(delete_vptr_pvalue);
+	PVALUE val = create_pvalue_from_list(list);
+	release_list(list); /* release our ref to list */
+	return val;
+}
+/*=============================================
+ * set_pvalue_to_pvalue -- Set val to be same value as src
+ *============================================*/
+void
+set_pvalue_to_pvalue (PVALUE val, const PVALUE src)
+{
+	set_pvalue(val, ptype(src), pvalvv(src));
+}
