@@ -377,12 +377,12 @@ static const yysigned_char yyrhs[] =
 /* YYRLINE[YYN] -- source line where rule number YYN was defined.  */
 static const unsigned short yyrline[] =
 {
-       0,    67,    67,    68,    71,    72,    73,    77,    90,    95,
-      99,   102,   106,   109,   114,   117,   122,   127,   132,   137,
-     142,   147,   152,   157,   162,   167,   171,   175,   180,   185,
-     190,   194,   198,   214,   219,   223,   227,   231,   235,   239,
-     242,   246,   249,   254,   259,   262,   266,   270,   274,   277,
-     280,   284,   287,   291,   294,   299,   302,   306
+       0,    67,    67,    68,    71,    72,    73,    79,    92,    98,
+     103,   106,   110,   114,   120,   123,   128,   134,   140,   146,
+     152,   158,   164,   170,   176,   182,   188,   194,   200,   206,
+     212,   217,   222,   238,   243,   248,   252,   256,   260,   264,
+     267,   271,   274,   279,   284,   287,   291,   296,   301,   304,
+     307,   311,   314,   318,   321,   326,   329,   333
 };
 #endif
 
@@ -1373,11 +1373,13 @@ yyreduce:
     {
 			if (eqstr("global", (STRING) yyvsp[-3]))
 				pa_handle_global((STRING) yyvsp[-1]);
+				free_iden(yyvsp[-3]);
+				free_iden(yyvsp[-1]);
 		}
     break;
 
   case 7:
-#line 77 "../../../src/interp/yacc.y"
+#line 79 "../../../src/interp/yacc.y"
     {
 			if (eqstr("include", (STRING) yyvsp[-3]))
 				pa_handle_include(pactx, (PNODE) yyvsp[-1]);
@@ -1387,62 +1389,66 @@ yyreduce:
 				pa_handle_char_encoding(pactx, (PNODE) yyvsp[-1]);
 			if (eqstr("require", (STRING) yyvsp[-3]))
 				pa_handle_require(pactx, (PNODE) yyvsp[-1]);
-
+			free_iden(yyvsp[-3]);
 		}
     break;
 
   case 8:
-#line 90 "../../../src/interp/yacc.y"
+#line 92 "../../../src/interp/yacc.y"
     {
+			/* consumes $2 */
 			pa_handle_proc(pactx, (STRING) yyvsp[-6], (PNODE) yyvsp[-4], (PNODE) yyvsp[-1]);
 		}
     break;
 
   case 9:
-#line 95 "../../../src/interp/yacc.y"
+#line 98 "../../../src/interp/yacc.y"
     {
+			/* consumes $2 */
 			pa_handle_func(pactx, (STRING) yyvsp[-6], (PNODE) yyvsp[-4], (PNODE) yyvsp[-1]);
 		}
     break;
 
   case 10:
-#line 99 "../../../src/interp/yacc.y"
+#line 103 "../../../src/interp/yacc.y"
     {
 			yyval = 0;
 		}
     break;
 
   case 11:
-#line 102 "../../../src/interp/yacc.y"
+#line 106 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 12:
-#line 106 "../../../src/interp/yacc.y"
+#line 110 "../../../src/interp/yacc.y"
     {
+			/* consumes $1 */
 			yyval = iden_node(pactx, (STRING)yyvsp[0]);
 		}
     break;
 
   case 13:
-#line 109 "../../../src/interp/yacc.y"
+#line 114 "../../../src/interp/yacc.y"
     {
+			/* consumes $1 */
 			yyval = iden_node(pactx, (STRING)yyvsp[-2]);
 			inext(((PNODE)yyval)) = (PNODE) yyvsp[0];
 		}
     break;
 
   case 14:
-#line 114 "../../../src/interp/yacc.y"
+#line 120 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 15:
-#line 117 "../../../src/interp/yacc.y"
+#line 123 "../../../src/interp/yacc.y"
     {
 			join((PNODE) yyvsp[-1], (PNODE) yyvsp[0]);
 			yyval = yyvsp[-1];
@@ -1450,135 +1456,151 @@ yyreduce:
     break;
 
   case 16:
-#line 123 "../../../src/interp/yacc.y"
+#line 129 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 */
 			yyval = children_node(pactx, (PNODE)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT)yyvsp[-10];
 		}
     break;
 
   case 17:
-#line 128 "../../../src/interp/yacc.y"
+#line 135 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 and $10 */
 			yyval = spouses_node(pactx, (PNODE)yyvsp[-10], (STRING)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT)yyvsp[-12];
 		}
     break;
 
   case 18:
-#line 133 "../../../src/interp/yacc.y"
+#line 141 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 and $10 */
 			yyval = families_node(pactx, (PNODE)yyvsp[-10], (STRING)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT)yyvsp[-12];
 		}
     break;
 
   case 19:
-#line 138 "../../../src/interp/yacc.y"
+#line 147 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 and $10 */
 			yyval = fathers_node(pactx, (PNODE)yyvsp[-10], (STRING)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT)yyvsp[-12];
 		}
     break;
 
   case 20:
-#line 143 "../../../src/interp/yacc.y"
+#line 153 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 and $10 */
 			yyval = mothers_node(pactx, (PNODE)yyvsp[-10], (STRING)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT)yyvsp[-12];
 		}
     break;
 
   case 21:
-#line 148 "../../../src/interp/yacc.y"
+#line 159 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 */
 			yyval = parents_node(pactx, (PNODE)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT)yyvsp[-10];
 		}
     break;
 
   case 22:
-#line 153 "../../../src/interp/yacc.y"
+#line 165 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 and $10 */
 			yyval = forindiset_node(pactx, (PNODE)yyvsp[-10], (STRING)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-12];
 		}
     break;
 
   case 23:
-#line 158 "../../../src/interp/yacc.y"
+#line 171 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 and $8 */
 			yyval = forlist_node(pactx, (PNODE)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-10];
 		}
     break;
 
   case 24:
-#line 163 "../../../src/interp/yacc.y"
+#line 177 "../../../src/interp/yacc.y"
     {
+			/* consumes $4 and $6 */
 			yyval = forindi_node(pactx, (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
 		}
     break;
 
   case 25:
-#line 167 "../../../src/interp/yacc.y"
+#line 183 "../../../src/interp/yacc.y"
     {
-                        yyval = fornotes_node(pactx, (PNODE)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
-                        ((PNODE)yyval)->i_line = (INT) yyvsp[-8];
-                }
+			/* consumes $6 */
+			yyval = fornotes_node(pactx, (PNODE)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
+			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
+		}
     break;
 
   case 26:
-#line 171 "../../../src/interp/yacc.y"
+#line 189 "../../../src/interp/yacc.y"
     {
+			/* consumes $4 and $6 */
 			yyval = forfam_node(pactx, (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
 		}
     break;
 
   case 27:
-#line 176 "../../../src/interp/yacc.y"
+#line 195 "../../../src/interp/yacc.y"
     {
+			/* consumes $4 and $6 */
 			yyval = forsour_node(pactx, (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
 		}
     break;
 
   case 28:
-#line 181 "../../../src/interp/yacc.y"
+#line 201 "../../../src/interp/yacc.y"
     {
+			/* consumes $4 and $6 */
 			yyval = foreven_node(pactx, (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
 		}
     break;
 
   case 29:
-#line 186 "../../../src/interp/yacc.y"
+#line 207 "../../../src/interp/yacc.y"
     {
+			/* consumes $4 and $6 */
 			yyval = forothr_node(pactx, (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
 		}
     break;
 
   case 30:
-#line 190 "../../../src/interp/yacc.y"
+#line 212 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 */
 			yyval = traverse_node(pactx, (PNODE)yyvsp[-8], (STRING)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-10];
 		}
     break;
 
   case 31:
-#line 194 "../../../src/interp/yacc.y"
+#line 217 "../../../src/interp/yacc.y"
     {
+			/* consumes $6 */
 			yyval = fornodes_node(pactx, (PNODE)yyvsp[-6], (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-8];
 		}
     break;
 
   case 32:
-#line 198 "../../../src/interp/yacc.y"
+#line 222 "../../../src/interp/yacc.y"
     {
 			inext(((PNODE)yyvsp[-7])) = (PNODE)yyvsp[-6];
 			prev = NULL;  this = (PNODE)yyvsp[-1];
@@ -1598,7 +1620,7 @@ yyreduce:
     break;
 
   case 33:
-#line 214 "../../../src/interp/yacc.y"
+#line 238 "../../../src/interp/yacc.y"
     {
 			inext(((PNODE)yyvsp[-5])) = (PNODE)yyvsp[-4];
 			yyval = while_node(pactx, (PNODE)yyvsp[-5], (PNODE)yyvsp[-1]);
@@ -1607,15 +1629,16 @@ yyreduce:
     break;
 
   case 34:
-#line 219 "../../../src/interp/yacc.y"
+#line 243 "../../../src/interp/yacc.y"
     {
+			/* consumes $2 */
 			yyval = call_node(pactx, (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-3];
 		}
     break;
 
   case 35:
-#line 223 "../../../src/interp/yacc.y"
+#line 248 "../../../src/interp/yacc.y"
     {
 			yyval = break_node(pactx);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-2];
@@ -1623,7 +1646,7 @@ yyreduce:
     break;
 
   case 36:
-#line 227 "../../../src/interp/yacc.y"
+#line 252 "../../../src/interp/yacc.y"
     {
 			yyval = continue_node(pactx);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-2];
@@ -1631,7 +1654,7 @@ yyreduce:
     break;
 
   case 37:
-#line 231 "../../../src/interp/yacc.y"
+#line 256 "../../../src/interp/yacc.y"
     {
 			yyval = return_node(pactx, (PNODE)yyvsp[-1]);
 			((PNODE)yyval)->i_line = (INT) yyvsp[-3];
@@ -1639,35 +1662,35 @@ yyreduce:
     break;
 
   case 38:
-#line 235 "../../../src/interp/yacc.y"
+#line 260 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 39:
-#line 239 "../../../src/interp/yacc.y"
+#line 264 "../../../src/interp/yacc.y"
     {
 			yyval = 0;
 		}
     break;
 
   case 40:
-#line 242 "../../../src/interp/yacc.y"
+#line 267 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 41:
-#line 246 "../../../src/interp/yacc.y"
+#line 271 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 42:
-#line 249 "../../../src/interp/yacc.y"
+#line 274 "../../../src/interp/yacc.y"
     {
 			ielse((PNODE)yyvsp[-1]) = (VPTR)yyvsp[0];
 			yyval = yyvsp[-1];
@@ -1675,7 +1698,7 @@ yyreduce:
     break;
 
   case 43:
-#line 254 "../../../src/interp/yacc.y"
+#line 279 "../../../src/interp/yacc.y"
     {
 			inext(((PNODE)yyvsp[-5])) = (PNODE)yyvsp[-4];
 			yyval = if_node(pactx, (PNODE)yyvsp[-5], (PNODE)yyvsp[-1], (PNODE)NULL);
@@ -1683,79 +1706,81 @@ yyreduce:
     break;
 
   case 44:
-#line 259 "../../../src/interp/yacc.y"
-    {
-			yyval = 0;
-		}
-    break;
-
-  case 45:
-#line 262 "../../../src/interp/yacc.y"
-    {
-			yyval = yyvsp[-1];
-		}
-    break;
-
-  case 46:
-#line 266 "../../../src/interp/yacc.y"
-    {
-			yyval = iden_node(pactx, (STRING)yyvsp[0]);
-			iargs(((PNODE)yyval)) = NULL;
-		}
-    break;
-
-  case 47:
-#line 270 "../../../src/interp/yacc.y"
-    {
-			yyval = func_node(pactx, (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
-			((PNODE)yyval)->i_line = (INT)yyvsp[-3];
-		}
-    break;
-
-  case 48:
-#line 274 "../../../src/interp/yacc.y"
-    {
-			yyval = yyvsp[0];
-		}
-    break;
-
-  case 49:
-#line 277 "../../../src/interp/yacc.y"
-    {
-			yyval = icons_node(pactx, Yival);
-		}
-    break;
-
-  case 50:
-#line 280 "../../../src/interp/yacc.y"
-    {
-			yyval = fcons_node(pactx, Yfval);
-		}
-    break;
-
-  case 51:
 #line 284 "../../../src/interp/yacc.y"
     {
 			yyval = 0;
 		}
     break;
 
-  case 52:
+  case 45:
 #line 287 "../../../src/interp/yacc.y"
+    {
+			yyval = yyvsp[-1];
+		}
+    break;
+
+  case 46:
+#line 291 "../../../src/interp/yacc.y"
+    {
+			/* consumes $1 */
+			yyval = iden_node(pactx, (STRING)yyvsp[0]);
+			iargs(((PNODE)yyval)) = NULL;
+		}
+    break;
+
+  case 47:
+#line 296 "../../../src/interp/yacc.y"
+    {
+			/* consumes $1 */
+			yyval = func_node(pactx, (STRING)yyvsp[-4], (PNODE)yyvsp[-1]);
+			((PNODE)yyval)->i_line = (INT)yyvsp[-3];
+		}
+    break;
+
+  case 48:
+#line 301 "../../../src/interp/yacc.y"
+    {
+			yyval = yyvsp[0];
+		}
+    break;
+
+  case 49:
+#line 304 "../../../src/interp/yacc.y"
+    {
+			yyval = icons_node(pactx, Yival);
+		}
+    break;
+
+  case 50:
+#line 307 "../../../src/interp/yacc.y"
+    {
+			yyval = fcons_node(pactx, Yfval);
+		}
+    break;
+
+  case 51:
+#line 311 "../../../src/interp/yacc.y"
+    {
+			yyval = 0;
+		}
+    break;
+
+  case 52:
+#line 314 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 53:
-#line 291 "../../../src/interp/yacc.y"
+#line 318 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 54:
-#line 294 "../../../src/interp/yacc.y"
+#line 321 "../../../src/interp/yacc.y"
     {
 			inext(((PNODE)yyvsp[-2])) = (PNODE) yyvsp[0];
 			yyval = yyvsp[-2];
@@ -1763,21 +1788,21 @@ yyreduce:
     break;
 
   case 55:
-#line 299 "../../../src/interp/yacc.y"
+#line 326 "../../../src/interp/yacc.y"
     {
 			yyval = 0;
 		}
     break;
 
   case 56:
-#line 302 "../../../src/interp/yacc.y"
+#line 329 "../../../src/interp/yacc.y"
     {
 			yyval = yyvsp[0];
 		}
     break;
 
   case 57:
-#line 306 "../../../src/interp/yacc.y"
+#line 333 "../../../src/interp/yacc.y"
     {
 			yyval = (YYSTYPE)((PACTX)pactx)->lineno;
 		}
@@ -1787,7 +1812,7 @@ yyreduce:
     }
 
 /* Line 999 of yacc.c.  */
-#line 1791 "yacc.c"
+#line 1816 "yacc.c"
 
   yyvsp -= yylen;
   yyssp -= yylen;
@@ -1993,7 +2018,7 @@ yyreturn:
 }
 
 
-#line 310 "../../../src/interp/yacc.y"
+#line 337 "../../../src/interp/yacc.y"
 
 
 void
