@@ -50,6 +50,12 @@ extern STRING qSidsadd, qSidsinf, qSkchild, qSiscinf, qSnotopp, qSidsps1, qSidsp
 extern STRING qSnosex,  qShashsb, qShaswif, qSidchld, qSgdfadd, qScfcadd, qSiredit;
 extern STRING qScfpadd, qScfsadd, qSgdpadd, qSgdcadd, qSgdsadd, qSronlya, qSronlye;
 
+/*********************************************
+ * local function prototypes
+ *********************************************/
+
+static void add_new_indi_to_db(RECORD indi0);
+static void add_new_fam_to_db(NODE fam2, NODE spouse1, NODE spouse2, NODE child);
 
 /*==========================================================
  * get_unresolved_ref_error -- get string for unresolved reference(s)
@@ -65,6 +71,7 @@ get_unresolved_ref_error_string (INT count)
 /*==========================================================
  * add_indi_by_edit -- Add new person to database by editing
  * (with user interaction)
+ * returns addref'd record
  *========================================================*/
 RECORD
 add_indi_by_edit (RFMT rfmt)
@@ -99,6 +106,10 @@ add_indi_by_edit (RFMT rfmt)
 	do_edit();
 	while (TRUE) {
 		INT cnt;
+		if (indi0) {
+			release_record(indi0);
+			indi0=0;
+		}
 		indi0 = file_to_record(editfile, ttmi, &msg, &emp);
 		if (!indi0) {
 			if (ask_yes_or_no_msg(msg, _(qSiredit))) {
@@ -116,7 +127,7 @@ add_indi_by_edit (RFMT rfmt)
 				do_edit();
 				continue;
 			}
-			free_rec(indi0);
+			release_record(indi0);
 			indi0 = NULL;
 			break;
 		}
@@ -135,38 +146,44 @@ add_indi_by_edit (RFMT rfmt)
 		break;
 	}
 	if (!indi0 || !ask_yes_or_no(_(qScfpadd))) {
-		if (indi0) free_rec(indi0);
+		if (indi0) release_record(indi0);
 		return NULL;
 	}
-	add_new_indi(indi0);
+	
+	/* add the new record to the database */
+	add_new_indi_to_db(indi0);
+
 	msg_status(_(qSgdpadd), indi_to_name(nztop(indi0), 35));
 	return indi0;
 }
 /*==========================================================
- * add_new_indi -- Add newly created person to database
+ * add_new_indi_to_db -- Add newly created person to database
  * (no user interaction)
  * creates record & adds to cache
  *========================================================*/
-void
-add_new_indi (RECORD indi0)
+static void
+add_new_indi_to_db (RECORD indi0)
 {
 	NODE name, refn, sex, body, dumb, node;
-	STRING key;
-	INT keynum;
+	char key[MAXKEYWIDTH]="";
+	INT keynum=0;
 	NODE indi = nztop(indi0);
 
 	split_indi_old(indi, &name, &refn, &sex, &body, &dumb, &dumb);
 	keynum = getixrefnum();
-	init_new_record(indi0, 'I', keynum);
-	key = nzkey(indi0);
-	for (node = name; node; node = nsibling(node))
+	sprintf(key, "I%d", keynum);
+	init_new_record(indi0, key);
+	for (node = name; node; node = nsibling(node)) {
 		add_name(nval(node), key);
-	for (node = refn; node; node = nsibling(node))
-		if (nval(node)) add_refn(nval(node), key);
+	}
+	for (node = refn; node; node = nsibling(node)) {
+		if (nval(node))
+			add_refn(nval(node), key);
+	}
 	join_indi(indi, name, refn, sex, body, NULL, NULL);
 	resolve_refn_links(indi);
 	indi_to_dbase(indi);
-	indi_to_cache(indi0);
+	add_new_indi_to_cache(indi0);
 }
 /*================================================================
  * add_indi_no_cache -- Add new person to database
@@ -483,20 +500,19 @@ add_members_to_family (STRING xref, NODE spouse1, NODE spouse2, NODE child)
 	}
 }
 /*=========================================
- * add_family -- Add new family to database
+ * add_family_by_edit -- Add new family to database
  * (with user interaction)
  *=======================================*/
 RECORD
-add_family (RECORD sprec1, RECORD sprec2, RECORD chrec)
+add_family_by_edit (RECORD sprec1, RECORD sprec2, RECORD chrec, RFMT rfmt)
 {
 	INT sex1 = 0;
 	INT sex2 = 0;
 	NODE spouse1, spouse2, child;
-	NODE fam1, fam2=0, refn, husb, wife, chil, body;
-	NODE node;
+	NODE fam1, fam2=0, husb, wife, chil;
 	XLAT ttmi = transl_get_predefined_xlat(MEDIN);
 	XLAT ttmo = transl_get_predefined_xlat(MINED);
-	STRING xref, msg, key, str;
+	STRING msg, key, str;
 	BOOLEAN emp;
 	FILE *fp;
 
@@ -528,7 +544,8 @@ add_family (RECORD sprec1, RECORD sprec2, RECORD chrec)
 	if (!sprec2)
 		sprec2 = ask_for_indi(_(qSidsps2), NOCONFIRM, DOASK1);
 	if (sprec2) {
-		if ((sex2 = SEX(nztop(sprec2))) == SEX_UNKNOWN || sex1 == sex2) {
+		if ((sex2 = SEX(nztop(sprec2))) == SEX_UNKNOWN || 
+			(traditional && sex1 == sex2)) {
 			message(_(qSnotopp));
 			return NULL;
 		}
@@ -542,17 +559,23 @@ editfam:
 
 	fam1 = create_node(NULL, "FAM", NULL, NULL);
 	husb = wife = chil = NULL;
-	if (spouse1) {
-		if (sex1 == SEX_MALE)
-			husb = create_node(NULL, "HUSB", nxref(spouse1), fam1);
-		else
-			wife = create_node(NULL, "WIFE", nxref(spouse1), fam1);
+	/* support non-traditional families if traditional is false
+	 * to do this we make slightly fib about the use of the
+	 * terms husb and wife in setting up spouse nodes
+	 */
+	if (sex1 == sex2 && sex2 == SEX_FEMALE) {
+	    husb = create_node(NULL, "WIFE", nxref(spouse1), fam1);
+	} else if (sex1 == SEX_MALE ) {
+	    husb = create_node(NULL, "HUSB", nxref(spouse1), fam1);
+	} else if (sex2 == SEX_MALE && sex1 != SEX_MALE) {
+	    husb = create_node(NULL, "HUSB", nxref(spouse2), fam1);
 	}
-	if (spouse2) {
-		if (sex2 == SEX_MALE)
-			husb = create_node(NULL, "HUSB", nxref(spouse2), fam1);
-		else
-			wife = create_node(NULL, "WIFE", nxref(spouse2), fam1);
+	if (sex1 == SEX_FEMALE && sex2 != SEX_FEMALE) {
+	    wife = create_node(NULL, "WIFE", nxref(spouse1), fam1);
+	} else if (sex2 == SEX_FEMALE) {
+	    wife = create_node(NULL, "WIFE", nxref(spouse2), fam1);
+	} else if (sex1 == sex2 && sex2 == SEX_MALE) {
+	    wife = create_node(NULL, "HUSB", nxref(spouse2), fam1);
 	}
 	if (child)
 		chil = create_node(NULL, "CHIL", nxref(child), fam1);
@@ -604,7 +627,7 @@ editfam:
 			llstrncpyf(msgb, sizeof(msgb), uu8
 				, get_unresolved_ref_error_string(cnt), cnt);
 			if (ask_yes_or_no_msg(msgb, _(qSfreditopt))) {
-				write_fam_to_file(fam2, editfile);
+				write_fam_to_file_for_edit(fam2, editfile, rfmt);
 				do_edit();
 				continue;
 			}
@@ -619,7 +642,29 @@ editfam:
 		free_nodes(fam2);
 		return NULL;
 	}
-	nxref(fam2) = strsave(xref = getfxref());
+
+	/* Add the new record to the database */
+	add_new_fam_to_db(fam2, spouse1, spouse2, child);
+
+	message(_(qSgdfadd));
+
+	key = rmvat(nxref(fam2));
+	return key_to_record(key);
+}
+/*==========================================================
+ * add_new_fam_to_db -- Add newly created family to database
+ * (no user interaction)
+ * creates record & adds to cache
+ *========================================================*/
+static void
+add_new_fam_to_db (NODE fam2, NODE spouse1, NODE spouse2, NODE child)
+{
+	NODE refn, husb, wife, chil, body;
+	NODE node;
+	STRING key=0;
+	STRING xref = getfxref();
+
+	nxref(fam2) = strsave(xref);
 
 /* Modify spouse/s and/or child */
 
@@ -632,6 +677,7 @@ editfam:
 	for (node = refn; node; node = nsibling(node))
 		if (nval(node)) add_refn(nval(node), key);
 	join_fam(fam2, refn, husb, wife, chil, body);
+	resolve_refn_links(fam2);
 	resolve_refn_links(spouse1);
 	resolve_refn_links(spouse2);
 	resolve_refn_links(child);
@@ -640,8 +686,6 @@ editfam:
 	if (spouse1) indi_to_dbase(spouse1);
 	if (spouse2) indi_to_dbase(spouse2);
 	if (child) indi_to_dbase(child);
-	message(_(qSgdfadd));
-	return key_to_record(key);
 }
 #ifdef ETHEL
 /*=========================================
@@ -695,26 +739,7 @@ add_family_to_db (NODE spouse1, NODE spouse2, NODE child)
 
 	free_nodes(fam1);
 
-	nxref(fam2) = strsave(xref = getfxref());
-
-/* Modify spouse/s and/or child */
-
-	add_members_to_family(xref, spouse1, spouse2, child);
-
-	split_fam(fam2, &refn, &husb, &wife, &chil, &body);
-	key = rmvat(nxref(fam2));
-	for (node = refn; node; node = nsibling(node))
-		if (nval(node)) add_refn(nval(node), key);
-	join_fam(fam2, refn, husb, wife, chil, body);
-	resolve_refn_links(fam2);
-	resolve_refn_links(spouse1);
-	resolve_refn_links(spouse2);
-	resolve_refn_links(child);
-	fam_to_dbase(fam2);
-	fam_to_cache(fam2);
-	if (spouse1) indi_to_dbase(spouse1);
-	if (spouse2) indi_to_dbase(spouse2);
-	if (child) indi_to_dbase(child);
+	add_new_fam_to_db(fam2, spouse1, spouse2, child);
 
 	return fam2;
 }
