@@ -111,6 +111,7 @@ static RECORD pick_create_new_family(RECORD current, RECORD save, STRING * addst
 static void pick_remove_spouse_from_family(RECORD frec);
 static void save_hist_lists(void);
 static void save_nkey_list(STRING key, struct hist * histp);
+static void setrecord(RECORD * dest, RECORD * src);
 
 /*********************************************
  * local variables
@@ -149,7 +150,7 @@ static struct hist chist; /* records changed */
  *  code:  [I/O]  current browse type
  *  pseq:  [OUT]  current sequence
  * Sets either *prec or *pseq & sets *code to appropriate browse type
- * Created: 2001/02/25, Perry Rapp
+ *  returns addref'd record
  *=======================================*/
 static void
 prompt_for_browse (RECORD * prec, INT * code, INDISEQ * pseq)
@@ -195,7 +196,7 @@ main_browse (RECORD rec1, INT code)
 {
 	RECORD rec2=0;
 	INDISEQ seq = NULL;
-	STRING key;
+	STRING key=0;
 
 	if (!rec1)
 		prompt_for_browse(&rec1, &code, &seq);
@@ -231,44 +232,60 @@ main_browse (RECORD rec1, INT code)
 			}
 		}
 	}
+	setrecord(&rec1, NULL);
+	setrecord(&rec2, NULL);
 }
 /*================================================
  * goto_indi_child - jump to child by number
+ * returns addref'd record
  *==============================================*/
 static RECORD
 goto_indi_child (RECORD irec, int childno)
 {
 	INT num1, num2, i = 0;
 	RECORD answer = 0;
+	INT akeynum=0; /* answer key */
 	NODE indi = nztop(irec);
 	if (!irec) return NULL;
-	FORFAMSS(indi, fam, spouse, num1)
+	FORFAMS(indi, fam, num1)
 		FORCHILDREN(fam, chil, num2)
 			i++;
-			if (i == childno) answer = chil;
+			if (i == childno) 
+				akeynum = nzkeynum(chil);
 		ENDCHILDREN
-	ENDFAMSS
+	ENDFAMS
+	if (akeynum) {
+		answer = keynum_to_irecord(akeynum);
+		addref_record(answer);
+	}
 	return answer;
 }
 /*================================================
  * goto_fam_child - jump to child by number
+ * returns addref'd record
  *==============================================*/
 static RECORD
-goto_fam_child(RECORD frec, int childno)
+goto_fam_child (RECORD frec, int childno)
 {
 	INT num, i = 0;
 	RECORD answer = 0;
+	INT akeynum=0;
 	NODE fam = nztop(frec);
 	if (!frec) return NULL;
 	FORCHILDREN(fam, chil, num)
 		i++;
-		if (i == childno) answer = chil;
+		if (i == childno) 
+			akeynum = nzkeynum(chil);
 	ENDCHILDREN
+	if (akeynum) {
+		answer = keynum_to_irecord(akeynum);
+		addref_record(answer);
+	}
 	return answer;
 }
 /*===============================================
  * pick_create_new_family -- 
- *  pulled out of browse_indi, 2001/02/04, Perry Rapp
+ * returns addref'd record
  *=============================================*/
 static RECORD
 pick_create_new_family (RECORD current, RECORD save, STRING * addstrings)
@@ -282,8 +299,9 @@ pick_create_new_family (RECORD current, RECORD save, STRING * addstrings)
 	}
 	i = choose_from_array(_(qSidfcop), 2, addstrings);
 	if (i == -1) return NULL;
-	if (i == 0) rec = add_family(NULL, NULL, current);
-	else if (save) {
+	if (i == 0) {
+		rec = add_family_by_edit(NULL, NULL, current, &disp_long_rfmt);
+	} else if (save) {
 		char scratch[100];
 		STRING name = indi_to_name(nztop(save), 55);
 		llstrncpyf(scratch, sizeof(scratch), uu8, "%s%s", _(qSissnew), name);
@@ -292,13 +310,31 @@ pick_create_new_family (RECORD current, RECORD save, STRING * addstrings)
 			llstrappf(scratch, sizeof(scratch), uu8, " (%s)", key);
 		}
 		if (ask_yes_or_no(scratch))
-			rec = add_family(current, save, NULL);
+			rec = add_family_by_edit(current, save, NULL, &disp_long_rfmt);
 		else
-			rec = add_family(current, NULL, NULL);
+			rec = add_family_by_edit(current, NULL, NULL, &disp_long_rfmt);
 	} else
-		rec = add_family(current, NULL, NULL);
+		rec = add_family_by_edit(current, NULL, NULL, &disp_long_rfmt);
 	return rec;
 }
+/*====================================================
+ * setrecord -- Move record in src to dest
+ *  Handles releasing old reference
+ *==================================================*/
+static void
+setrecord (RECORD * dest, RECORD * src)
+{
+	ASSERT(dest);
+	if (*dest) {
+		release_record(*dest);
+	}
+	if (src) {
+		*dest = *src;
+		*src = 0;
+	} else {
+		*dest = 0;
+	}
+}
 /*====================================================
  * browse_indi_modes -- Handle person/pedigree browse.
  *  prec1 [I/O]  current record (or upper in tandem screens)
@@ -316,19 +352,22 @@ browse_indi_modes (RECORD *prec1, RECORD *prec2, INDISEQ *pseq, INT indimode)
 	RECORD save=0, tmp=0, tmp2=0;
 	INDISEQ seq = NULL;
 	char c2;
+	INT rtn=0; /* return code */
 
 	ASSERT(prec1 && *prec1 && nztype(*prec1)=='I');
 	ASSERT(!*prec2 && !*pseq);
-	current = *prec1;
 
-	*prec1 = 0;
-	*prec2 = 0;
+	/* move working record into current */
+	setrecord(&current, prec1);
+	setrecord(prec1, 0);
+	setrecord(prec2, 0);
 
 	show_reset_scroll();
 	nkeyp = 0;
 	indimodep = indimode;
 
 	while (TRUE) {
+		setrecord(&tmp, NULL);
 		if (nzkeynum(current) != nkeyp 
 			|| indimode != indimodep) {
 			show_reset_scroll();
@@ -352,33 +391,41 @@ reprocess_indi_cmd: /* so one command can forward to another */
 		i = handle_history_cmds(c, prec1);
 		if (i == 1)
 			continue; /* history cmd handled, stay here */
-		if (i == -1)
-			return BROWSE_UNK; /* history cmd handled, leave page */
+		if (i == -1) {
+			/* history cmd handled, leave page */
+			rtn = BROWSE_UNK;
+			goto exitbrowse;
+		}
 		switch (c)
 		{
 		case CMD_EDIT:	/* Edit this person */
 			edit_indi(current, &disp_long_rfmt);
 			break;
 		case CMD_FAMILY: 	/* Browse to person's family */
-			if ((*prec1 = choose_family(current, _(qSntprnt)
-				,  _(qSidfbrs), TRUE)))
-				return BROWSE_FAM;
+			if ((tmp = choose_family(current, _(qSntprnt)
+				,  _(qSidfbrs), TRUE))) {
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_FAM;
+				goto exitbrowse;
+			}
 			break;
 		case CMD_TANDEM_FAMILIES:
 			if ((tmp = choose_family(current, _(qSntprnt),
 				_(qSid1fbr), TRUE)) != 0) {
 				if ((tmp2 = choose_family(current, _(qSntprnt),
 					_(qSid2fbr), TRUE)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_2FAM;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_2FAM;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_FATHER: 	/* Browse to person's father */
 			if ((tmp = choose_father(current, NULL, _(qSnofath),
 				_(qSidhbrs), NOASK1)) != 0) {
-				current = tmp;
+				setrecord(&current, &tmp);
 			}
 			break;
 		case CMD_TANDEM_FATHERS:	/* Tandem Browse to person's fathers */
@@ -386,16 +433,18 @@ reprocess_indi_cmd: /* so one command can forward to another */
 				_(qSid1hbr), NOASK1)) != 0) {
 				if ((tmp2 = choose_father(current, NULL, _(qSnofath),
 					_(qSid2hbr), NOASK1)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_MOTHER:	/* Browse to person's mother */
 			if ((tmp = choose_mother(current, NULL, _(qSnomoth),
 				_(qSidwbrs), NOASK1)) != 0) {
-				current = tmp;
+				setrecord(&current, &tmp);
 			}
 			break;
 		case CMD_TANDEM_MOTHERS:	/* Tandem Browse to person's mothers */
@@ -403,43 +452,49 @@ reprocess_indi_cmd: /* so one command can forward to another */
 				_(qSid1wbr), NOASK1)) != 0) {
 				if ((tmp2 = choose_mother(current, NULL, _(qSnomoth),
 					_(qSid2wbr), NOASK1)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_BROWSE_ZIP_INDI:	/* Zip browse another person */
 			if ((tmp = ask_for_indi(_(qSidpnxt), NOCONFIRM, NOASK1)) != 0)
-				current = tmp;
+				setrecord(&current, &tmp);
 			break;
 		case CMD_BROWSE_ZIP_ANY:	/* Zip browse any record */
 			if ((tmp = ask_for_any(_(qSidnxt), NOCONFIRM, NOASK1)) != 0) {
 				if (nztype(tmp) != 'I') {
-					*prec1 = tmp;
-					return BROWSE_UNK;
+					setrecord(prec1, &tmp);
+					rtn = BROWSE_UNK;
+					goto exitbrowse;
 				} else {
-					current = tmp;
+					setrecord(&current, &tmp);
 				}
 			}
 			break;
 		case CMD_SPOUSE:	/* Browse to person's spouse */
 			if ((tmp = choose_spouse(current, _(qSnospse), _(qSidsbrs))) != 0)
-				current = tmp;
+				setrecord(&current, &tmp);
 			break;
 		case CMD_TANDEM_SPOUSES:	/* browse to tandem spouses */
 			if ((tmp = choose_spouse(current, _(qSnospse), _(qSid1sbr))) != 0) {
 				if ((tmp2 = choose_spouse(current, _(qSnospse), _(qSid2sbr))) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_CHILDREN:	/* Browse to person's child */
 			if ((tmp = choose_child(current, NULL, _(qSnocofp),
-				_(qSidcbrs), NOASK1)) != 0)
-				current = tmp;
+				_(qSidcbrs), NOASK1)) != 0) {
+				setrecord(&current, &tmp);
+			}
 			break;
 		case CMD_TOGGLE_PEDTYPE:       /* toggle pedigree mode (ancestors/descendants) */
 			pedigree_toggle_mode();
@@ -463,7 +518,7 @@ reprocess_indi_cmd: /* so one command can forward to another */
 		case CMD_CHILD_DIRECT0+8:
 		case CMD_CHILD_DIRECT0+9:
 			if ((tmp = goto_indi_child(current, c-CMD_CHILD_DIRECT0)) != 0)
-				current = tmp;
+				setrecord(&current, &tmp);
 			else
 				message(_(qSnochil));
 			break;
@@ -472,10 +527,12 @@ reprocess_indi_cmd: /* so one command can forward to another */
 				_(qSid1cbr), NOASK1)) != 0) {
 				if ((tmp2 = choose_child(current, NULL, _(qSnocofp),
 					_(qSid2cbr), NOASK1)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_PEDIGREE:	/* Switch to pedigree mode */
@@ -486,21 +543,22 @@ reprocess_indi_cmd: /* so one command can forward to another */
 			break;
 		case CMD_UPSIB:	/* Browse to older sib */
 			if ((tmp = indi_to_prev_sib(current)) != 0)
-				current = tmp;
+				setrecord(&current, &tmp);
 			else
 				message(_(qSnoosib));
 			break;
 		case CMD_DOWNSIB:	/* Browse to younger sib */
 			if ((tmp = indi_to_next_sib(current)) != 0)
-				current = tmp;
+				setrecord(&current, &tmp);
 			else
 				message(_(qSnoysib));
 			break;
 		case CMD_PARENTS:	/* Browse to parents' family */
 			if ((tmp = choose_family(current, _(qSnoprnt),
 				_(qSidfbrs), FALSE)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_FAM;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_FAM;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_TANDEM_PARENTS:	/* tandem browse to two parents families*/
@@ -508,10 +566,12 @@ reprocess_indi_cmd: /* so one command can forward to another */
 				_(qSid1fbr), FALSE)) != 0) {
 				if ((tmp2 = choose_family(current, _(qSnoprnt),
 					_(qSid2fbr), FALSE)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_2FAM;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_2FAM;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_BROWSE: 	/* Browse new list of persons */
@@ -519,22 +579,26 @@ reprocess_indi_cmd: /* so one command can forward to another */
 			if (!seq) break;
 			if (length_indiseq(seq) == 1) {
 				element_indiseq(seq, 0, &key, &name);
-				current = key_to_record(key);
+				tmp = key_to_record(key);
+				setrecord(&current, &tmp);
 				remove_indiseq(seq);
 				seq=NULL;
 				if (nztype(current) != 'I') {
-					*prec1 = current;
-					return BROWSE_UNK;
+					setrecord(prec1, &current);
+					rtn = BROWSE_UNK;
+					goto exitbrowse;
 				}
 			} else {
 				*pseq = seq;
-				return BROWSE_LIST;
+				rtn = BROWSE_LIST;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_NEWPERSON:	/* Add new person */
-			if (!(tmp = add_indi_by_edit(&disp_long_rfmt))) break;
-			save = current;
-			current = tmp;
+			if (!(tmp = add_indi_by_edit(&disp_long_rfmt))) 
+				break;
+			setrecord(&save, &current);
+			setrecord(&current, &tmp);
 			break;
 		case CMD_NEWFAMILY:	/* Add family for current person */
 			{
@@ -542,9 +606,10 @@ reprocess_indi_cmd: /* so one command can forward to another */
 				addstrings[0] = _(qScrtcfm);
 				addstrings[1] = _(qScrtsfm);
 				if ((tmp = pick_create_new_family(current, save, addstrings)) != 0) {
-					save = NULL;
-					*prec1 = tmp;
-					return BROWSE_FAM;
+					setrecord(&save, NULL);
+					setrecord(prec1, &tmp);
+					rtn = BROWSE_FAM;
+					goto exitbrowse;
 				}
 			}
 			break;
@@ -557,15 +622,17 @@ reprocess_indi_cmd: /* so one command can forward to another */
 					c = CMD_EDIT;
 					goto reprocess_indi_cmd; /* forward to edit */
 				}
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_UNK;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_TANDEM:	/* Switch to tandem browsing */
 			if ((tmp = ask_for_indi(_(qSidp2br), NOCONFIRM, NOASK1)) != 0) {
-				*prec1 = current;
-				*prec2 = tmp;
-				return BROWSE_TAND;
+				setrecord(prec1, &current);
+				setrecord(prec2, &tmp);
+				rtn = BROWSE_TAND;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_SWAPFAMILIES: 	/* Swap families of current person */
@@ -595,41 +662,55 @@ reprocess_indi_cmd: /* so one command can forward to another */
 		case CMD_NEXT:	/* Go to next indi in db */
 			{
 				i = xref_nexti(nkeyp);
-				if (i)
-					current = keynum_to_irecord(i);
-				else message(_(qSnopers));
+				if (i) {
+					tmp = keynum_to_irecord(i);
+					setrecord(&current, &tmp);
+				} else {
+					message(_(qSnopers));
+				}
 			}
 			break;
 		case CMD_PREV:	/* Go to prev indi in db */
 			{
 				i = xref_previ(nkeyp);
-				if (i)
-					current = keynum_to_irecord(i);
-				else message(_(qSnopers));
+				if (i) {
+					tmp = keynum_to_irecord(i);
+					setrecord(&current, &tmp);
+				} else {
+					message(_(qSnopers));
+				}
 			}
 			break;
 		case CMD_SOURCES:	/* Browse to sources */
 			if ((tmp = choose_source(current, _(qSnosour), _(qSidsour))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_AUX;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_AUX;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_NOTES:	/* Browse to notes */
 			if ((tmp = choose_note(current, _(qSnonote), _(qSidnote))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_AUX;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_AUX;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_POINTERS:	/* Browse to references */
 			if ((tmp = choose_pointer(current, _(qSnoptr), _(qSidptr))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_AUX;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_QUIT:
-			return BROWSE_QUIT;
+			rtn = BROWSE_QUIT;
+			goto exitbrowse;
 		}
 	}
+exitbrowse:
+	setrecord(&tmp, NULL);
+	setrecord(&current, NULL);
+	return rtn;
 }
 /*==========================================
  * display_aux -- Show aux node in current mode
@@ -653,17 +734,22 @@ display_aux (RECORD rec, INT mode, BOOLEAN reuse)
 static INT
 browse_aux (RECORD *prec1, RECORD *prec2, INDISEQ *pseq)
 {
-	RECORD current;
+	RECORD current=0;
 	INT i, c;
 	BOOLEAN reuse=FALSE; /* flag to reuse same display strings */
 	INT nkeyp=0, auxmode=0, auxmodep=0;
 	char ntype=0, ntypep=0;
 	RECORD tmp=0;
 	char c2;
+	INT rtn=0; /* return code */
 
 	ASSERT(prec1 && *prec1);
 	ASSERT(!*prec2 && !*pseq);
-	current = *prec1;
+
+	/* move working record into current */
+	setrecord(&current, prec1);
+	setrecord(prec1, 0);
+	setrecord(prec2, 0);
 
 
 	auxmode = 'x';
@@ -717,54 +803,68 @@ reprocess_aux_cmd:
 					c = CMD_EDIT;
 					goto reprocess_aux_cmd; /* forward to edit */
 				} else {
-					*prec1 = tmp;
-					return BROWSE_UNK;
+					setrecord(prec1, &tmp);
+					rtn = BROWSE_UNK;
+					goto exitbrowse;
 				}
 			}
 			break;
 		case CMD_BROWSE_ZIP_INDI:	/* Zip browse to new person */
 			if ((tmp = ask_for_indi(_(qSidpnxt), NOCONFIRM, NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_UNK;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_BROWSE_ZIP_ANY:	/* Zip browse any record */
 			if ((tmp = ask_for_any(_(qSidnxt), NOCONFIRM, NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_UNK;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_NOTES:	/* Browse to notes */
 			if ((tmp = choose_note(current, _(qSnonote), _(qSidnote))) != 0) {
-				current = tmp;
+				setrecord(&current, &tmp);
 			}
 			break;
 		case CMD_POINTERS:	/* Browse to references */
 			if ((tmp = choose_pointer(current, _(qSnoptr), _(qSidptr))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_UNK;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_NEXT:	/* Go to next in db */
 			{
 				i = xref_next(ntype, nkeyp);
-				if (i)
-					current = keynum_to_record(ntype, i);
-				else message(_(qSnorec));
+				if (i) {
+					tmp = keynum_to_record(ntype, i);
+					setrecord(&current, &tmp);
+				} else {
+					message(_(qSnorec));
+				}
 				break;
 			}
 		case CMD_PREV:	/* Go to prev in db */
 			{
 				i = xref_prev(ntype, nkeyp);
-				if (i)
-					current = keynum_to_record(ntype, i);
-				else message(_(qSnorec));
+				if (i) {
+					tmp = keynum_to_record(ntype, i);
+					setrecord(&current, &tmp);
+				} else {
+					message(_(qSnorec));
+				}
 				break;
 			}
 		case CMD_QUIT:
-			return BROWSE_QUIT;
+			rtn = BROWSE_QUIT;
+			goto exitbrowse;
 		}
 	}
+exitbrowse:
+	setrecord(&current, NULL);
+	return rtn;
 }
 /*================================================
  * browse_indi -- Handle person browse operations.
@@ -877,9 +977,15 @@ prompt_add_child_check_save (NODE fam, NODE save)
 	}
 	if (save) {
 		if (keyflag)
-			sprintf(scratch, "%s%s (%s)", _(qSiscnew),
-				 indi_to_name(save, 56),
-				 rmvat(nxref(save))+1);
+			if(getoptint("DisplayKeyTags", 0) > 0) {
+				sprintf(scratch, "%s%s (i%s)", _(qSiscnew),
+				 	indi_to_name(save, 56),
+				 	rmvat(nxref(save))+1);
+			} else {
+				sprintf(scratch, "%s%s (%s)", _(qSiscnew),
+				 	indi_to_name(save, 56),
+				 	rmvat(nxref(save))+1);
+			}
 		else
 			sprintf(scratch, "%s%s", _(qSiscnew),
 				 indi_to_name(save, 56));
@@ -911,19 +1017,22 @@ browse_fam (RECORD *prec1, RECORD *prec2, INDISEQ *pseq)
 	INDISEQ seq;
 	STRING key, name;
 	char c2;
+	INT rtn=0; /* return code */
 
 	ASSERT(prec1 && *prec1 && nztype(*prec1)=='F');
 	ASSERT(!*prec2 && !*pseq);
-	current = *prec1;
 
-	*prec1 = 0;
-	*prec2 = 0;
+	/* move working record into current */
+	setrecord(&current, prec1);
+	setrecord(prec1, 0);
+	setrecord(prec2, 0);
 
 	show_reset_scroll();
 	nkeyp = 0;
 	fammodep = fammode;
 
 	while (TRUE) {
+		setrecord(&tmp, NULL);
 		if (nzkeynum(current) != nkeyp
 			|| fammode != fammodep) {
 			show_reset_scroll();
@@ -947,8 +1056,11 @@ reprocess_fam_cmd: /* so one command can forward to another */
 		i = handle_history_cmds(c, prec1);
 		if (i == 1)
 			continue; /* history cmd handled, stay here */
-		if (i == -1)
-			return BROWSE_UNK; /* history cmd handled, leave page */
+		if (i == -1) {
+			/* history cmd handled, leave page */
+			rtn = BROWSE_UNK;
+			goto exitbrowse;
+		}
 		switch (c) 
 		{
 		case CMD_ADVANCED:	/* Advanced family edit */
@@ -957,8 +1069,9 @@ reprocess_fam_cmd: /* so one command can forward to another */
 		case CMD_BROWSE_FAM:
 			if (ask_for_int(_(qSidfamk), &i) && (i>0)) {
 				if ((tmp = qkeynum_to_frecord(i)))
-					current = tmp;
-				else message(_(qSnofam));
+					setrecord(&current, &tmp);
+				else
+					message(_(qSnofam));
 			}
 			break;
 		case CMD_EDIT:	/* Edit family's record */
@@ -967,8 +1080,9 @@ reprocess_fam_cmd: /* so one command can forward to another */
 		case CMD_FATHER:	/* Browse to family's father */
 			if ((tmp = choose_father(NULL, current, _(qSnohusb),
 				_(qSidhbrs), NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_INDI;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_INDI;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_TANDEM_FATHERS:	/* Tandem Browse to family's fathers */
@@ -976,17 +1090,20 @@ reprocess_fam_cmd: /* so one command can forward to another */
 				_(qSid1hbr), NOASK1)) != 0) {
 				if ((tmp2 = choose_father(NULL, current, _(qSnohusb),
 					_(qSid2hbr), NOASK1)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_MOTHER:	/* Browse to family's mother */
 			if ((tmp = choose_mother(NULL, current, _(qSnowife),
 				_(qSidwbrs), NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_INDI;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_INDI;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_TANDEM_MOTHERS:	/* Tandem Browse to family's mother */
@@ -994,17 +1111,20 @@ reprocess_fam_cmd: /* so one command can forward to another */
 				_(qSid1wbr), NOASK1)) != 0) {
 				if ((tmp2 = choose_mother(NULL, current, _(qSnowife), 
 					_(qSid2wbr), NOASK1)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_CHILDREN:	/* Browse to a child */
 			if ((tmp = choose_child(NULL, current, _(qSnocinf),
 				_(qSidcbrs), NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_INDI;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_INDI;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_TANDEM_CHILDREN:	/* browse to tandem children */
@@ -1012,10 +1132,12 @@ reprocess_fam_cmd: /* so one command can forward to another */
 				_(qSid1cbr), NOASK1)) != 0) {
 				if ((tmp2 = choose_child(NULL, current, _(qSnocinf),
 					_(qSid2cbr), NOASK1)) != 0) {
-					*prec1 = tmp;
-					*prec2 = tmp2;
-					return BROWSE_TAND;
+					setrecord(prec1, &tmp);
+					setrecord(prec2, &tmp2);
+					rtn = BROWSE_TAND;
+					goto exitbrowse;
 				}
+				setrecord(&tmp, 0);
 			}
 			break;
 		case CMD_REMOVECHILD:	/* Remove a child */
@@ -1024,18 +1146,21 @@ reprocess_fam_cmd: /* so one command can forward to another */
 				break;
 			}
 			if ((tmp = choose_child(NULL, current, _(qSnocinf),
-			    _(qSidcrmv), DOASK1)) != 0)
+				_(qSidcrmv), DOASK1)) != 0) {
 				choose_and_remove_child(tmp, current, TRUE);
+				setrecord(&tmp, 0);
+			}
 			break;
 		case CMD_ADDSPOUSE:	/* Add spouse to family */
 			prompt_add_spouse_with_candidate(current, save);
-			save = NULL;
+			setrecord(&save, 0);
 			break;
 		case CMD_REMOVESPOUSE:	/* Remove spouse from family */
 			pick_remove_spouse_from_family(current);
 			break;
 		case CMD_NEWPERSON:	/* Add person to database */
-			save = add_indi_by_edit(&disp_long_rfmt);
+			tmp = add_indi_by_edit(&disp_long_rfmt);
+			setrecord(&save, &tmp);
 			break;
 		case CMD_ADD_SOUR: /* add source */
 		case CMD_ADD_EVEN: /* add event */
@@ -1046,45 +1171,52 @@ reprocess_fam_cmd: /* so one command can forward to another */
 					c = CMD_EDIT;
 					goto reprocess_fam_cmd; /* forward to edit */
 				} else {
-					*prec1 = tmp;
-					return BROWSE_UNK;
+					setrecord(prec1, &tmp);
+					rtn = BROWSE_UNK;
+					goto exitbrowse;
 				}
 			}
 			break;
 		case CMD_ADDCHILD:	/* Add child to family */
 			prompt_add_child_check_save(nztop(current), nztop(save));
-			save = NULL;
+			setrecord(&save, 0);
 			break;
 		case CMD_BROWSE: 	/* Browse to new list of persons */
 			seq = ask_for_indiseq(_(qSidplst), 'I', &rc);
 			if (!seq) break;
 			if (length_indiseq(seq) == 1) {
 				element_indiseq(seq, 0, &key, &name);
-				*prec1 = key_to_irecord(key);
+				tmp = key_to_irecord(key);
+				setrecord(prec1, &tmp);
 				remove_indiseq(seq);
 				seq=NULL;
-				return BROWSE_INDI;
+				rtn = BROWSE_INDI;
+				goto exitbrowse;
 			}
 			*pseq = seq;
-			return BROWSE_LIST;
+			rtn = BROWSE_LIST;
+			goto exitbrowse;
 			break;
 		case CMD_BROWSE_ZIP_INDI:	/* Zip browse to new person */
 			if ((tmp = ask_for_indi(_(qSidpnxt), NOCONFIRM, NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_INDI;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_INDI;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_BROWSE_ZIP_ANY:	/* Zip browse any record */
 			if ((tmp = ask_for_any(_(qSidnxt), NOCONFIRM, NOASK1)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_UNK;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_TANDEM:	/* Enter family tandem mode */
 			if ((tmp = ask_for_fam(_(qSids2fm), _(qSidc2fm))) != 0) {
-				*prec1 = current;
-				*prec2 = tmp;
-				return BROWSE_2FAM;
+				setrecord(prec1, &current);
+				setrecord(prec2, &tmp);
+				rtn = BROWSE_2FAM;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_SWAPCHILDREN:	/* Swap two children */
@@ -1106,49 +1238,62 @@ reprocess_fam_cmd: /* so one command can forward to another */
 		case CMD_CHILD_DIRECT0+8:
 		case CMD_CHILD_DIRECT0+9:
 			if ((tmp = goto_fam_child(current, c-CMD_CHILD_DIRECT0)) != 0) {
-				*prec1 = tmp;
-				return BROWSE_INDI;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_INDI;
+				goto exitbrowse;
 			}
 			message(_(qSnochil));
 			break;
 		case CMD_NEXT:	/* Go to next fam in db */
 			{
 				i = xref_nextf(nkeyp);
-				if (i)
-					current = keynum_to_frecord(i);
-				else message(_(qSnofam));
+				if (i && (tmp = qkeynum_to_frecord(i))) {
+					setrecord(&current, &tmp);
+				} else {
+					message(_(qSnofam));
+				}
 				break;
 			}
 		case CMD_PREV:	/* Go to prev fam in db */
 			{
 				i = xref_prevf(nkeyp);
-				if (i)
-					current = keynum_to_frecord(i);
-				else message(_(qSnofam));
+				if (i) {
+					tmp = keynum_to_frecord(i);
+					setrecord(&current, &tmp);
+				} else {
+					message(_(qSnofam));
+				}
 				break;
 			}
 		case CMD_SOURCES:	/* Browse to sources */
 			if ((tmp = choose_source(current, _(qSnosour), _(qSidsour))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_AUX;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_AUX;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_NOTES:	/* Browse to notes */
 			if ((tmp = choose_note(current, _(qSnonote), _(qSidnote))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_AUX;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_AUX;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_POINTERS:	/* Browse to references */
 			if ((tmp = choose_pointer(current, _(qSnoptr), _(qSidptr))) != 0) {
-				*prec1 = tmp;
-				return BROWSE_UNK;
+				setrecord(prec1, &tmp);
+				rtn = BROWSE_UNK;
+				goto exitbrowse;
 			}
 			break;
 		case CMD_QUIT:
-			return BROWSE_QUIT;
+			rtn = BROWSE_QUIT;
+			goto exitbrowse;
 		}
 	}
+exitbrowse:
+	setrecord(&current, NULL);
+	return rtn;
 }
 /*======================================================
  * handle_menu_cmds -- Handle menuing commands
@@ -1410,16 +1555,17 @@ load_nkey_list (STRING key, struct hist * histp)
 	count = temp;
 	if (count > histp->size) count = histp->size;
 	for (i=0,temp=0; i<count; ++i) {
-		char key[12];
+		char key[MAXKEYWIDTH+1];
 		char ntype = *ptr++;
 		INT keynum = *ptr++;
 		if (!ntype || !keynum)
 			continue;
+		if (keynum<1 || keynum>MAXKEYNUMBER)
+			continue;
+		snprintf(key, sizeof(key), "%c%d", ntype, keynum);
+		strcpy(histp->list[temp].key, key);
 		histp->list[temp].ntype = ntype;
 		histp->list[temp].keynum = keynum;
-		/* We could sanity check these */
-		snprintf(key, sizeof(key), "%c%d", ntype, keynum);
-		histp->list[temp].key = strsave(key);
 		++temp;
 	}
 	count = temp;
@@ -1812,6 +1958,9 @@ autoadd_xref (RECORD rec, NODE newnode)
 		}
 		nsibling(prev) = xref;
 	}
+
+	normalize_rec(rec);
+	
 	unknown_node_to_dbase(node);
 }
 /*==================================================
