@@ -21,7 +21,6 @@
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
 */
-/* modified 05 Jan 2000 by Paul B. McBride (pmcbride@tiac.net) */
 /*=============================================================
  * main.c -- Main program of LifeLines
  * Copyright(c) 1992-95 by T.T. Wetmore IV; all rights reserved
@@ -65,11 +64,11 @@ extern STRING qSmtitle,qSnorwandro,qSnofandl,qSbdlkar;
 extern STRING qSusgFinnOpt,qSusgFinnAlw,qSusgNorm;
 extern STRING qSbaddb;
 
-extern INT csz_indi, icsz_indi;
-extern INT csz_fam, icsz_fam;
-extern INT csz_sour, icsz_sour;
-extern INT csz_even, icsz_even;
-extern INT csz_othr, icsz_othr;
+extern INT csz_indi;
+extern INT csz_fam;
+extern INT csz_sour;
+extern INT csz_even;
+extern INT csz_othr;
 extern INT winx, winy;
 
 extern int opterr;
@@ -81,6 +80,7 @@ extern int opterr;
 
 static STRING usage = "";      /* usage string */
 int opt_finnish  = FALSE;      /* Finnish Language sorting order if TRUE */
+int opt_mychar = FALSE;        /* Custom character set handling (bypass libc) */
 BOOLEAN debugmode = FALSE;     /* no signal handling, so we can get coredump */
 BOOLEAN opt_nocb  = FALSE;     /* no cb. data is displayed if TRUE */
 BOOLEAN keyflag   = TRUE;      /* show key values */
@@ -137,6 +137,10 @@ main (INT argc, char **argv)
 	STRING progout=NULL;
 	BOOLEAN graphical=TRUE;
 	STRING configfile=0;
+	STRING crashlog=NULL;
+
+	/* initialize all the low-level library code */
+	init_stdlib();
 
 #if HAVE_SETLOCALE
 	/* initialize locales */
@@ -162,18 +166,23 @@ main (INT argc, char **argv)
 				if(isasciiletter((uchar)*optarg) && isupper((uchar)*optarg))
 					*optarg = tolower((uchar)*optarg);
 				if(*optarg == 'i') {
+					INT icsz_indi=0;
 					sscanf(optarg+1, "%d,%d", &csz_indi, &icsz_indi);
 				}
 				else if(*optarg == 'f') {
+					INT icsz_fam=0;
 					sscanf(optarg+1, "%d,%d", &csz_fam, &icsz_fam);
 				}
 				else if(*optarg == 's') {
+					INT icsz_sour=0;
 					sscanf(optarg+1, "%d,%d", &csz_sour, &icsz_sour);
 				}
 				else if(*optarg == 'e') {
+					INT icsz_even=0;
 					sscanf(optarg+1, "%d,%d", &csz_even, &icsz_even);
 				}
 				else if((*optarg == 'o') || (*optarg == 'x')) {
+					INT icsz_othr=0;
 					sscanf(optarg+1, "%d,%d", &csz_othr, &icsz_othr);
 				}
 				optarg++;
@@ -246,9 +255,9 @@ main (INT argc, char **argv)
 				parse_arg(optarg, &optname, &optval);
 				if (optname && optval) {
 					if (!exargs) {
-						exargs = create_table(FREEBOTH); /* TODO: destroy this */
+						exargs = create_table_str();
 					}
-					insert_table_str(exargs, strdup(optname), strdup(optval));
+					insert_table_str(exargs, optname, optval);
 				}
 				strfree(&optname);
 				strfree(&optval);
@@ -269,6 +278,7 @@ main (INT argc, char **argv)
 			break;
 		}
 	}
+
 prompt_for_db:
 
 	/* catch any fault, so we can close database */
@@ -288,7 +298,9 @@ prompt_for_db:
 		goto finish;
 	}
 	/* setup crashlog in case init_screen fails (eg, bad menu shortcuts) */
-	crash_setcrashlog(getoptstr("CrashLog_llines", NULL));
+        crashlog = getoptstr("CrashLog_llexec", NULL);
+        if (!crashlog) { crashlog = "CrashLog_llines.log"; }
+        crash_setcrashlog(crashlog);
 	/* initialize curses interface */
 	if (!init_screen(graphical))
 		goto finish;
@@ -385,13 +397,14 @@ prompt_for_db:
 	init_browse_module();
 	if (exargs) {
 		set_cmd_options(exargs);
+		release_table(exargs);
+		exargs = 0;
 	}
 	if (exprogs) {
 		BOOLEAN picklist = FALSE;
 		BOOLEAN timing = FALSE;
 		interp_main(exprogs, progout, picklist, timing);
-		make_list_empty(exprogs);
-		remove_list(exprogs, 0);
+		destroy_list(exprogs);
 	} else {
 		alldone = 0;
 		while (!alldone)
@@ -600,6 +613,7 @@ load_usage (void)
 static void
 main_db_notify (STRING db, BOOLEAN opening)
 {
+	/* store name away for reporting in case of crash later */
 	if (opening)
 		crash_setdb(db);
 	else
@@ -612,10 +626,12 @@ main_db_notify (STRING db, BOOLEAN opening)
 static BOOLEAN
 init_curses_ui (void)
 {
-        WINDOW *win = initscr();
-        if (!win) return FALSE;
-        noecho();
-        keypad(win, 1);
-        return TRUE;
+	WINDOW *win;
+
+	win = initscr();
+	if (!win) return FALSE;
+	noecho();
+	keypad(win, 1);
+	return TRUE;
 }
 
