@@ -294,6 +294,7 @@ static void
 init_display_indi (RECORD irec, INT width)
 {
 	NODE pers=nztop(irec);
+	NODE this_fam = 0;
 	INT nsp, nch, num, nm;
 	STRING s;
 	NODE fth;
@@ -304,8 +305,6 @@ init_display_indi (RECORD irec, INT width)
 
 
 	ASSERT(pers);
-	fth = indi_to_fath(pers);
-	mth = indi_to_moth(pers);
 
 	disp_person_name(Spers, _(qSdspl_indi), irec, width);
 
@@ -313,10 +312,12 @@ init_display_indi (RECORD irec, INT width)
 
 	disp_person_birthdeath(Sdeat, irec, f_death_tags, &disp_long_rfmt);
 
+	fth = indi_to_fath(pers);
 	s = person_display(fth, NULL, width-13);
 	if (s) llstrncpyf(Sfath, liwidth, uu8, "  %s: %s", _(qSdspl_fath), s);
 	else llstrncpyf(Sfath, liwidth, uu8, "  %s:", _(qSdspl_fath));
 
+	mth = indi_to_moth(pers);
 	s = person_display(mth, NULL, width-13);
 	if (s) llstrncpyf(Smoth, liwidth, uu8, "  %s: %s", _(qSdspl_moth), s);
 	else llstrncpyf(Smoth, liwidth, uu8, "  %s:", _(qSdspl_moth));
@@ -327,9 +328,12 @@ init_display_indi (RECORD irec, INT width)
 	lock_cache(icel);
 	FORFAMSS(pers, fam, sp, num)
 		if (sp) add_spouse_line(++nsp, sp, fam, width);
-		FORCHILDREN(fam, chld, nm)
-			if(chld) add_child_line(++nch, chld, width);
-		ENDCHILDREN
+	        if (this_fam != fam) {
+		        this_fam = fam; /* only do each family once */
+			FORCHILDREN(fam, chld, nm)
+				if(chld) add_child_line(++nch, chld, width);
+			ENDCHILDREN
+		}
 	ENDFAMSS
 	unlock_cache(icel);
 }
@@ -449,15 +453,42 @@ init_display_fam (RECORD frec, INT width)
 	STRING father = _(qSdspl_fath);
 	STRING mother = _(qSdspl_moth);
 	RECORD ihusb=0, iwife=0;
-	INT husbstatus = fam_to_husb(frec, &ihusb);
-	INT wifestatus = fam_to_wife(frec, &iwife);
+	INT husbstatus = 0;
+	INT wifestatus = 0;
+	NODE fnode;
 
-	husb = nztop(ihusb);
-	wife = nztop(iwife);
+	/* Get the first two spouses in the family and use them rather than
+	 * displaying first husband and first mother
+	 * This causes a more reasonable presentation of non-traditional
+	 * familes.  Also it will display first hustband and first wife
+	 * for traditional families (as there's only one) and the db routines
+	 * insert HUSB records before WIFE records.
+	 */
+	if (fam) {
+		fnode = nchild(fam);
+		husbstatus = next_spouse(&fnode,&ihusb);
+		husb = nztop(ihusb);
+		if (fnode) {
+			fnode = nsibling(fnode);
+			wifestatus = next_spouse(&fnode,&iwife);
+			wife = nztop(iwife);
+		}
+	}
+	/* if the only spouse is female, list in second slot
+	 * hiding the non-traditional behavior
+	 */
+	if (!wife && husb && SEX(husb) == SEX_FEMALE) {
+		wife = husb;
+		husb = 0;
+		iwife = ihusb;
+		ihusb = 0;
+		wifestatus = husbstatus;
+		husbstatus  = 0;
+	}
 
 	if (husbstatus == 1) {
 		INT avail = width - zs_len(famkey) - 3;
-		disp_person_name(Shusb, father, ihusb, avail);
+		disp_person_name(Shusb, SEX(husb)==SEX_MALE?father:mother, ihusb, avail);
 	} else {
 		zs_setf(Shusb, "%s:", father);
 		if (husbstatus == -1)
@@ -475,7 +506,7 @@ init_display_fam (RECORD frec, INT width)
 
 	if (wifestatus == 1) {
 		INT avail = width;
-		disp_person_name(Swife, mother, iwife, avail);
+		disp_person_name(Swife, SEX(wife)==SEX_MALE?father:mother, iwife, avail);
 	} else {
 		zs_setf(Swife, "%s:", mother);
 		if (wifestatus == -1)
@@ -502,6 +533,8 @@ init_display_fam (RECORD frec, INT width)
 	FORCHILDREN(fam, chld, nm)
 		add_child_line(++nch, chld, width);
 	ENDCHILDREN
+	release_record(ihusb);
+	release_record(iwife);
 }
 /*===================================
  * show_fam_vitals -- Display family
