@@ -45,6 +45,7 @@
 #include "lloptions.h"
 #include "zstr.h"
 #include "codesets.h"
+#include "btree.h"
 
 #ifndef INCLUDED_STDARG_H
 #include <stdarg.h>
@@ -90,6 +91,7 @@ static char f_logpath[MAXPATHLEN] = "import.log";
 
 static STRING qSmisval      = N_("Line %d: This %s line is missing a value field.");
 static STRING qSundrec      = N_("Record %s is referred to but not defined.");
+static STRING qSlinlev1     = N_("Line %d: Tag %s found in unexpected record: %s.");
 
 ELMNT *index_data = NULL;
 
@@ -110,6 +112,7 @@ static int check_akey (int firstchar, STRING keyp, INT *maxp);
 static void check_even_links(IMPORT_FEEDBACK ifeed, ELMNT);
 static void check_fam_links(IMPORT_FEEDBACK ifeed, ELMNT);
 static void check_indi_links(IMPORT_FEEDBACK ifeed, ELMNT per);
+static void check_level1_tag(IMPORT_FEEDBACK ifeed, CNSTRING tag0, CNSTRING tag, CNSTRING val, INT line);
 static void check_othr_links(IMPORT_FEEDBACK ifeed, ELMNT);
 static void check_references(IMPORT_FEEDBACK ifeed);
 static void check_sour_links(IMPORT_FEEDBACK ifeed, ELMNT src);
@@ -135,17 +138,17 @@ validate_gedcom (IMPORT_FEEDBACK ifeed, FILE *fp)
 	ELMNT el;
 	XLAT xlat = transl_get_predefined_xlat(MGDIN);
 	STRING xref, tag, val, msg;
+	STRING tag0=0;
 
 	nhead = ntrlr = nindi = nfam = nsour = neven = nothr = 0;
 	num_errors = num_warns = 0;
 	f_logopen = FALSE;
 	f_flog = 0;
 	set_import_log(getoptstr("ImportLog", "errs.log"));
-	flineno = 0;
 	defline = 0;
 	curlev = 0;
 	clear_structures();
-	convtab = create_table(DONTFREE);
+	convtab = create_table_int();
 
 
 	rc = file_to_line(fp, xlat, &lev, &xref, &tag, &val, &msg);
@@ -184,6 +187,7 @@ validate_gedcom (IMPORT_FEEDBACK ifeed, FILE *fp)
 				}
 			}
 			defline = flineno;
+			strupdate(&tag0, tag); /* store current level 0 tag */
 			if (eqstr("HEAD", tag))  {
 				rec_type = (nhead==0 ? HEAD_REC : IGNR_REC);
 			} else if (eqstr("TRLR", tag)) {
@@ -219,6 +223,7 @@ validate_gedcom (IMPORT_FEEDBACK ifeed, FILE *fp)
 					ifeed->validated_rec_fnc(tag[0], tag, count);
 			}
 		} else {
+			/* specific handling for specific record types */
 			if (rec_type == HEAD_REC)
 				handle_head_lev1(ifeed, tag, val, flineno);
 			else if (rec_type == TRLR_REC)
@@ -229,6 +234,8 @@ validate_gedcom (IMPORT_FEEDBACK ifeed, FILE *fp)
 				handle_fam_lev1(ifeed, tag, val, flineno);
 			else
 				handle_value(val, flineno);
+			/* specific handling for specific tag types */
+			check_level1_tag(ifeed, tag0, tag, val, flineno);
 		}
 		curlev = lev;
 		rc = file_to_line(fp, xlat, &lev, &xref, &tag, &val, &msg);
@@ -588,8 +595,42 @@ handle_fam_lev1 (IMPORT_FEEDBACK ifeed, STRING tag, STRING val, INT line)
 	} else
 		handle_value(val, line);
 }
+/*==========================================================
+ * check_level1_tag -- Warnings for specific tags at level 1
+ *========================================================*/
+static void
+check_level1_tag (IMPORT_FEEDBACK ifeed, CNSTRING tag0, CNSTRING tag, CNSTRING val, INT line)
+{
+	/*
+	lifelines expects lineage-linking records (FAMS, FAMC, HUSB, & WIFE)
+	to be correct, so warn if any of them occur in unusual locations
+	*/
+	if (eqstr(tag, "FAMS")) {
+		if (!eqstr(tag0, "INDI"))
+			handle_warn(ifeed, qSlinlev1, line, tag, tag0);
+		return;
+	}
+	if (eqstr(tag, "FAMC")) {
+		if (!eqstr(tag0, "INDI"))
+			handle_warn(ifeed, qSlinlev1, line, tag, tag0);
+		return;
+	}
+	if (eqstr(tag, "HUSB")) {
+		if (!eqstr(tag0, "FAM"))
+			handle_warn(ifeed, qSlinlev1, line, tag, tag0);
+		return;
+	}
+	if (eqstr(tag, "WIFE")) {
+		if (!eqstr(tag0, "FAM"))
+			handle_warn(ifeed, qSlinlev1, line, tag, tag0);
+		return;
+	}
+}
 /*=================================================
  * check_akey -- Check for a standard format key
+ *               lifelines assumes key's are of form 
+ *               a letter followed by a number with no leading zeros 
+ *               e.g. I248
  *===============================================*/
 static int
 check_akey (int firstchar,
@@ -599,7 +640,7 @@ check_akey (int firstchar,
     INT val;
     if(keyp && (*keyp == firstchar)) {
 	keyp++;
-	if(*keyp && isdigit((uchar)*keyp)) {
+	if(*keyp && (*keyp != '0') && isdigit(*keyp)) {
 	    val = atoi(keyp);
 	    if(val > *maxp) *maxp = val;
 	    while(*keyp && isdigit((uchar)*keyp)) keyp++;
@@ -623,26 +664,32 @@ check_stdkeys (void)
 	gd_xmax = 0; gd_xtot = 0;
 	for (i = 0; retval && (i < struct_len); i++) {
 		ELMNT el = index_data[i];
+		STRING skey = Key(el);
+		/* check that key is not too long */
+		if (strlen(skey)>RKEYLEN) {
+			retval = FALSE;
+			break;
+		}
 		switch (Type(el)) {
 		case INDI_REC:
 		    gd_itot++;
-		    retval = check_akey('I', Key(el), &gd_imax);
+		    retval = check_akey('I', skey, &gd_imax);
 		    break;
 		case FAM_REC:
 		    gd_ftot++;
-		    retval = check_akey('F', Key(el), &gd_fmax);
+		    retval = check_akey('F', skey, &gd_fmax);
 		    break;
 		case EVEN_REC:
 		    gd_etot++;
-		    retval = check_akey('E', Key(el), &gd_emax);
+		    retval = check_akey('E', skey, &gd_emax);
 		    break;
 		case SOUR_REC:
 		    gd_stot++;
-		    retval = check_akey('S', Key(el), &gd_smax);
+		    retval = check_akey('S', skey, &gd_smax);
 		    break;
 		case OTHR_REC:
 		    gd_xtot++;
-		    retval = check_akey('X', Key(el), &gd_xmax);
+		    retval = check_akey('X', skey, &gd_xmax);
 		    break;
 		default: retval = FALSE; break;
 		}
@@ -965,13 +1012,13 @@ BOOLEAN
 scan_header (FILE * fp, TABLE metadatatab, ZSTR * zerr)
 {
 	STRING parents[2] = { 0, 0 };
-	INT linno, head=0, lev=-1, curlev;
+	INT linno, head=0, lev=-1, curlev,i;
 	INT lastoff=0;
 	ZSTR zpath = zs_new();
 	*zerr = 0;
 	for (linno=1; 1; ++linno) {
 		XLAT xlat=0; /* no codeset translation yet */
-		INT rc, i;
+		INT rc;
 		STRING xref, tag, val, msg;
 		lastoff = ftell(fp);
 		curlev = lev;
@@ -1005,6 +1052,7 @@ scan_header (FILE * fp, TABLE metadatatab, ZSTR * zerr)
 				break;
 			} else {
 				fseek(fp, lastoff, SEEK_SET);
+				flineno--;
 				/* finished head */
 				break;
 			}
@@ -1014,7 +1062,7 @@ scan_header (FILE * fp, TABLE metadatatab, ZSTR * zerr)
 			for (i=1; i<lev && i<ARRSIZE(parents); ++i)
 				append_path(zpath, '.', parents[i-1]);
 			append_path(zpath, '.', tag);
-			insert_table_str(metadatatab, strsave(zs_str(zpath)), strsave(val));
+			insert_table_str(metadatatab, zs_str(zpath), val);
 		}
 		if (lev>0 && lev-1<ARRSIZE(parents))
 			strupdate(&parents[lev-1], tag);
@@ -1024,6 +1072,9 @@ scan_header (FILE * fp, TABLE metadatatab, ZSTR * zerr)
 		}
 	}
 
+	for (i=0; i<ARRSIZE(parents); ++i) {
+		strupdate(&parents[i], 0);
+	}
 	zs_free(&zpath);
 	return !(*zerr);
 }
