@@ -32,6 +32,9 @@
 
 #include "llstdlib.h"
 /* llstdlib.h pulls in standard.h, config.h, sys_inc.h */
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif
 #include "gedcom.h"
 #include "btree.h"
 #include "indiseq.h"
@@ -63,6 +66,7 @@ BOOLEAN readonly = FALSE;	/* normally defined in liflines/main.c */
 BOOLEAN writeable = FALSE;	/* normally defined in liflines/main.c */
 BOOLEAN immutable = FALSE;  /* normally defined in liflines/main.c */
 int opt_finnish = 0;
+int opt_mychar = 0;
 
 /*********************************************
  * local types
@@ -85,15 +89,15 @@ struct work {
 	INT find_ghosts;
 	INT fix_ghosts;
 	INT check_indis;
-	INT fix_indis;
 	INT check_fams;
-	INT fix_fams;
 	INT check_sours;
-	INT fix_sours;
 	INT check_evens;
-	INT fix_evens;
 	INT check_othes;
-	INT fix_othes;
+	INT fix_deletes;
+	INT fix_alter_pointers;
+	INT check_missing_data_records; /* record in index, but no data */
+	INT fix_missing_data_records;
+	INT pass; /* =1 is checking, =2 is fixing */
 };
 /*=======================================
  * NAMEREFN_REC -- holds one name or refn
@@ -114,28 +118,38 @@ typedef struct
 static NAMEREFN_REC * alloc_namerefn(CNSTRING namerefn, CNSTRING key, INT err);
 static BOOLEAN cgn_callback(CNSTRING key, CNSTRING name, BOOLEAN newset, void *param);
 static BOOLEAN cgr_callback(CNSTRING key, CNSTRING refn, BOOLEAN newset, void *param);
-static BOOLEAN check_block(BTREE btr, BLOCK block, RKEY * lo, RKEY * hi);
+static void check_and_fix_records(void);
+static BOOLEAN check_block(BLOCK block, RKEY * lo, RKEY * hi);
 static BOOLEAN check_btree(BTREE btr);
 static BOOLEAN check_even(CNSTRING key, RECORD rec);
 static BOOLEAN check_fam(CNSTRING key, RECORD rec);
 static void check_ghosts(void);
-static BOOLEAN check_keys(BTREE btr, BLOCK block, RKEY * lo, RKEY * hi);
+static BOOLEAN check_keys(BLOCK block, RKEY * lo, RKEY * hi);
 static BOOLEAN check_index(BTREE btr, INDEX index, TABLE fkeytab, RKEY * lo, RKEY * hi);
 static BOOLEAN check_indi(CNSTRING key, RECORD rec);
 static void check_node(CNSTRING key, NODE node, INT level);
 static void check_nodes(void);
+static void check_missing_data_records(void);
 static void check_pointers(CNSTRING key, RECORD rec);
 static void check_set(INDISEQ seq, char ctype);
 static BOOLEAN check_sour(CNSTRING key, RECORD rec);
+static void check_typed_missing_data_records(char ntype);
 static BOOLEAN check_othe(CNSTRING key, RECORD rec);
 static BOOLEAN find_xref(CNSTRING key, NODE node, CNSTRING tag1, CNSTRING tag2);
 static void finish_and_delete_nameset(void);
 static void finish_and_delete_refnset(void);
+static void fix_nodes(void);
 static void free_namerefn(NAMEREFN_REC * rec);
+static BOOLEAN fix_bad_pointer(CNSTRING key, RECORD rec, NODE node);
 static BOOLEAN nodes_callback(CNSTRING key, RECORD rec, void *param);
-static void printblock(BTREE btr, BLOCK block);
+static void printblock(BLOCK block);
+static CNSTRING printkey(CNSTRING key);
 static void print_usage(void);
+static void process_fam(RECORD rec);
+static void process_indi(RECORD rec);
+static void process_record(RECORD rec);
 static void report_error(INT err, STRING fmt, ...);
+static void report_fix(INT err, STRING fmt, ...);
 static void report_progress(STRING fmt, ...);
 static void report_results(void);
 static void validate_errs(void);
@@ -148,11 +162,12 @@ enum {
 	ERR_ORPHANNAME,  ERR_GHOSTNAME, ERR_DUPNAME, ERR_DUPREFN
 	, ERR_NONINDINAME, ERR_DUPINDI, ERR_DUPFAM
 	, ERR_DUPSOUR, ERR_DUPEVEN, ERR_DUPOTHE
-	, ERR_MISSING, ERR_DELETED, ERR_BADNAME
+	, ERR_UNDELETED, ERR_DELETED, ERR_BADNAME
 	, ERR_BADFAMREF, ERR_MISSINGCHILD, ERR_MISSINGSPOUSE
 	, ERR_BADHUSBREF, ERR_BADWIFEREF, ERR_BADCHILDREF
 	, ERR_EXTRAHUSB, ERR_EXTRAWIFE, ERR_EXTRACHILD
 	, ERR_EMPTYFAM, ERR_SOLOFAM, ERR_BADPOINTER
+	, ERR_MISSINGREC
 };
 
 static struct errinfo errs[] = {
@@ -166,7 +181,7 @@ static struct errinfo errs[] = {
 	, { ERR_DUPSOUR, 0, 0, N_("Duplicate sources") }
 	, { ERR_DUPEVEN, 0, 0, N_("Duplicate events") }
 	, { ERR_DUPOTHE, 0, 0, N_("Duplicate others") }
-	, { ERR_MISSING, 0, 0, N_("Missing records") }
+	, { ERR_UNDELETED, 0, 0, N_("Missing records (from deleteset)") }
 	, { ERR_DELETED, 0, 0, N_("Deleted records") }
 	, { ERR_BADNAME, 0, 0, N_("Bad name") }
 	, { ERR_BADFAMREF, 0, 0, N_("Bad family reference") }
@@ -181,9 +196,10 @@ static struct errinfo errs[] = {
 	, { ERR_EMPTYFAM, 0, 0, N_("Empty family") }
 	, { ERR_SOLOFAM, 0, 0, N_("Single person family") }
 	, { ERR_BADPOINTER, 0, 0, N_("Bad pointer") }
+	, { ERR_MISSINGREC, 0, 0, N_("Missing data records") }
 };
 static struct work todo;
-static LIST tofix;
+static LIST tofix=0;
 /* sequence of NAMEs or REFNs in the same block */
 static INDISEQ soundexseq=0;
 static BOOLEAN noisy=FALSE;
@@ -193,6 +209,7 @@ static STRING lineage_tags[] = {
 	, "FAMS"
 	, "WIFE"
 	, "HUSB"
+	, "CHIL"
 };
 /*********************************************
  * local & exported function definitions
@@ -215,31 +232,50 @@ print_usage (void)
 #endif
 	llstrncpyf(verstr, sizeof(verstr), uu8, title
 		, get_lifelines_version(sizeof(verstr)-1-strlen(title)));
-	printf(
-		_("usage: dbverify -(flags) <btree>\n"
-		"flags:\n"
-		"\t-a = Perform all checks (does not include fixes)\n"
-		"\t-g = Check for ghosts (names/refns)\n"
-		"\t-G = Check for & fix ghosts (names/refns)\n"
-		"\t-i = Check individuals\n"
-		"\t-f = Check families\n"
-		"\t-s = Check sours\n"
-		"\t-e = Check events\n"
-		"\t-x = Check others\n"
-		"\t-n = Noisy (echo every record processed)\n")
-		);
+	printf(_("usage: dbverify -(flags) <btree>\n"));
+	printf(_("flags:\n"));
+	printf(_("\t-a = Perform all checks (does not include fixes)\n"));
+	printf(_("\t-g = Check for ghosts (names/refns)\n"));
+	printf(_("\t-G = Check for & fix ghosts (names/refns)\n"));
+	printf(_("\t-i = Check individuals\n"));
+	printf(_("\t-f = Check families\n"));
+	printf(_("\t-F = Alter any bad family lineage pointers (to _badptr)\n"));
+	printf(_("\t-s = Check sours\n"));
+	printf(_("\t-e = Check events\n"));
+	printf(_("\t-x = Check others\n"));
+	printf(_("\t-l = Check database structure\n"));
+	printf(_("\t-m = Check for records missing data entries\n"));
+	printf(_("\t-M = Fix records missing data entries\n"));
+	printf(_("\t-D = Fix bad delete entries\n"));
+	printf(_("\t-n = Noisy (echo every record processed)\n"));
 	printf(_("example: dbverify -ifsex \"%s\"\n"), fname);
 	printf("%s\n", verstr);
 }
 /*========================================
  * report_error -- report some error found
- * Created: 2001/01/01, Perry Rapp
+ *  increment error count
  *======================================*/
 static void
 report_error (INT err, STRING fmt, ...)
 {
 	va_list args;
-	errs[err].err_count++;
+	if (err>=0)
+		++errs[err].err_count;
+	va_start(args, fmt);
+	printf("! ");
+	vprintf(fmt, args);
+	printf("\n");
+	va_end(args);
+}
+/*========================================
+ * report_fix -- report an error fixed
+ *  increment fixed count
+ *======================================*/
+static void
+report_fix (INT err, STRING fmt, ...)
+{
+	va_list args;
+	++errs[err].fix_count;
 	va_start(args, fmt);
 	printf("! ");
 	vprintf(fmt, args);
@@ -325,7 +361,7 @@ check_ghosts (void)
 		}
 	}
 	
-	remove_list(tofix, NULL); 
+	destroy_list(tofix);
 	tofix=0;
 }
 /*============================================
@@ -424,8 +460,9 @@ static void
 finish_and_delete_nameset (void)
 {
 	char prevkey[8];
-	STRING name="";
-	TABLE table = create_table(DONTFREE);
+	CNSTRING name="";
+	CNSTRING skey="";
+	TABLE table = create_table_int();
 	prevkey[0]=0;
 	calc_indiseq_names(soundexseq);
 	keysort_indiseq(soundexseq);
@@ -435,19 +472,20 @@ finish_and_delete_nameset (void)
 	duplicates
 	*/
 	FORINDISEQ(soundexseq, el, num)
-		name = sval(el).w;
-		if (!eqstr(skey(el), prevkey)) {
+		name = element_sval(el);
+		skey = element_skey(el);
+		if (!eqstr(skey, prevkey)) {
 			/* new person, start over */
 			destroy_table(table);
-			table = create_table(DONTFREE);
+			table = create_table_int();
 		}
 		if (in_table(table, name)) {
 			report_error(ERR_DUPNAME, _("Duplicate name for %s (%s)")
-				, skey(el), name);
+				, skey, name);
 		} else {
 			insert_table_int(table, name, 1);
 		}
-		strcpy(prevkey, skey(el));
+		strcpy(prevkey, skey);
 	ENDINDISEQ
 	remove_indiseq(soundexseq);
 	soundexseq = NULL;
@@ -463,33 +501,34 @@ static void
 finish_and_delete_refnset (void)
 {
 	char prevkey[8];
-	STRING refn="";
-	TABLE table = create_table(DONTFREE);
+	CNSTRING refn="";
+	CNSTRING skey="";
+	TABLE table = create_table_int();
 	prevkey[0]=0;
 	canonkeysort_indiseq(soundexseq);
 	FORINDISEQ(soundexseq, el, num)
-		refn = sval(el).w;
-		if (!eqstr(skey(el), prevkey)) {
+		refn = element_sval(el);
+		skey = element_skey(el);
+		if (!eqstr(skey, prevkey)) {
 			/* new person, start over */
 			destroy_table(table);
-			table = create_table(DONTFREE);
+			table = create_table_int();
 		}
 		if (in_table(table, refn)) {
 			report_error(ERR_DUPREFN, _("Duplicate refn for %s (%s)")
-				, skey(el), refn);
+				, skey, refn);
 		} else {
 			insert_table_int(table, refn, 1);
 		}
-		strcpy(prevkey, skey(el));
+		strcpy(prevkey, skey);
 	ENDINDISEQ
 	remove_indiseq(soundexseq);
 	soundexseq = NULL;
 	destroy_table(table);
 }
 /*=================================
- * check_nodes -- Process all nodes
- *  checking and/or fixing as requested
- * Created: 2001/01/14, Perry Rapp
+ * check_nodes -- Check all nodes in database
+ * (as selected in struct work)
  *================================*/
 static void
 check_nodes (void)
@@ -512,6 +551,36 @@ check_nodes (void)
 	remove_indiseq(seq_evens);
 	remove_indiseq(seq_othes);
 }
+/*=================================
+ * check_and_fix_records -- 
+ *================================*/
+static void
+check_and_fix_records (void)
+{
+	tofix = create_list();
+	todo.pass = 1;
+	check_nodes();
+	todo.pass = 2;
+	fix_nodes();
+	destroy_empty_list(tofix);
+}
+/*=================================
+ * fix_nodes -- Fix all nodes on fix list
+ *  (uses list from check_nodes)
+ *================================*/
+static void
+fix_nodes (void)
+{
+	if (!tofix) return;
+	while (!is_empty_list(tofix)) {
+		STRING key = dequeue_list(tofix);
+		RECORD rec = key_to_record(key);
+		strfree(&key);
+		lock_record_in_cache(rec);
+		process_record(rec);
+		unlock_record_from_cache(rec);
+	}
+}
 /*=============================================
  * nodes_callback -- callback for node traversal
  *  for checking indis, fams, sours, evens, othes
@@ -520,107 +589,214 @@ check_nodes (void)
 static BOOLEAN
 nodes_callback (CNSTRING key, RECORD rec, void *param)
 {
-	param=param; /* unused */
+	param=param;	/* NOTUSED */
 	if (noisy)
 		report_progress("Node: %s", key);
 	switch (key[0]) {
-	case 'I': return todo.check_indis ? check_indi(key, rec) : TRUE;
-	case 'F': return todo.check_fams ? check_fam(key, rec) : TRUE;
-	case 'S': return todo.check_sours ? check_sour(key, rec) : TRUE;
-	case 'E': return todo.check_evens ? check_even(key, rec) : TRUE;
-	case 'X': return todo.check_othes ? check_othe(key, rec) : TRUE;
+	case 'I': return todo.check_indis ? check_indi((CNSTRING)key, (RECORD)rec) : TRUE;
+	case 'F': return todo.check_fams ? check_fam((CNSTRING)key, (RECORD)rec) : TRUE;
+	case 'S': return todo.check_sours ? check_sour((CNSTRING)key, (RECORD)rec) : TRUE;
+	case 'E': return todo.check_evens ? check_even((CNSTRING)key, (RECORD)rec) : TRUE;
+	case 'X': return todo.check_othes ? check_othe((CNSTRING)key, (RECORD)rec) : TRUE;
 	}
 	ASSERT(0); /* traverse_db_key_recs is broken */
 	return TRUE;
 }
 /*=====================================
- * check_indi -- process indi record
- *  checking and/or fixing as requested
- * Created: 2001/01/14, Perry Rapp
+ * process_record -- process record
+ *  Called during pass2 (fix)
+ *===================================*/
+static void
+process_record (RECORD rec)
+{
+	switch(nztype(rec)) {
+	case 'I': process_indi(rec); break;
+	case 'F': process_fam(rec); break;
+		/* TODO */
+	}
+}
+/*=====================================
+ * check_indi -- check indi record
+ *  and record any records needing fixing
+ *  in tofix list
  *===================================*/
 static BOOLEAN
 check_indi (CNSTRING key, RECORD rec)
 {
-	static char prevkey[9];
-	NODE indi1, name1, refn1, sex1, body1, famc1, fams1;
-	NODE node1;
-	CACHEEL icel1;
+	static char prevkey[MAXKEYWIDTH+1];
+	CACHEEL icel1=0;
+	RECORD recx=0;
+
 	if (eqstr(key, prevkey)) {
 		report_error(ERR_DUPINDI, _("Duplicate individual for %s"), key);
 	}
 	icel1 = indi_to_cacheel(rec);
 	lock_cache(icel1);
-	indi1 = nztop(rec);
+
+	recx = get_record_for_cel(icel1);
+	ASSERT(todo.pass == 1);
+	process_indi(recx);
+
+	unlock_cache(icel1);
+	check_pointers(key, rec);
+	append_indiseq_null(seq_indis, strsave(key), NULL, TRUE, TRUE);
+	release_record(recx);
+	return TRUE;
+}
+/*=====================================
+ * process_indi -- process indi record
+ *  checking in pass 1, fixing in pass 2
+ *===================================*/
+static void
+process_indi (RECORD rec)
+{
+	NODE indi0, indi1;
+	NODE name1, refn1, sex1, body1, famc1, fams1;
+	NODE node1;
+	BOOLEAN altered=FALSE;
+	BOOLEAN needfix=FALSE;
+	CNSTRING key = nzkey(rec);
+
+	indi0 = nztop(rec);
+	if (todo.pass==1) {
+		indi1 = indi0;
+	} else {
+		indi1 = copy_node_subtree(indi0);
+	}
 	split_indi_old(indi1, &name1, &refn1, &sex1, &body1, &famc1, &fams1);
-	/* check names */
-	for (node1 = name1; node1; node1 = nsibling(node1)) {
-		STRING name=nval(node1);
-		if (!valid_name(name)) {
-			report_error(ERR_BADNAME, _("Bad name for individual %s: %s"), key, name);
-		} else {
-		/* TO DO: verify that name is in db */
+
+	if (todo.pass == 1) {
+		/* check names */
+		for (node1 = name1; node1; node1 = nsibling(node1)) {
+			STRING name=nval(node1);
+			if (!valid_name(name)) {
+				report_error(ERR_BADNAME, _("Bad name for individual %s: %s"), key, name);
+			} else {
+			/* TO DO: verify that name is in db */
+			}
+		}
+		/* check refns */
+		for (node1 = refn1; node1; node1 = nsibling(node1)) {
+			/* STRING refn=nval(node1); */
+			/* TO DO: verify that refn is in db */
 		}
 	}
-	/* check refns */
-	for (node1 = refn1; node1; node1 = nsibling(node1)) {
-		STRING refn=nval(node1);
-		/* TO DO: verify that refn is in db */
-	}
+
 	/* check parents */
 	for (node1 = famc1; node1; node1 = nsibling(node1)) {
 		STRING famkey=rmvat(nval(node1));
 		NODE fam2 = qkey_to_fam(famkey);
 		if (!fam2) {
-			report_error(ERR_BADFAMREF, _("Bad family reference (%s) individual %s"), famkey, key);
+			if (todo.pass == 1) {
+				report_error(ERR_BADFAMREF, _("Bad family reference (%s) individual %s"), famkey, key);
+			}
 		} else {
 			/* look for indi1 (key) in fam2's children */
 			if (!find_xref(key, fam2, "CHIL", NULL)) {
-				report_error(ERR_MISSINGCHILD, _("Missing child (%s) in family (%s)"), key, famkey);
+				if (todo.pass == 1) {
+					report_error(ERR_MISSINGCHILD, _("Missing child (%s) in family (%s)"), key, famkey);
+					needfix=TRUE;
+				} else {
+					if (fix_bad_pointer(key, rec, node1)) {
+						report_fix(ERR_MISSINGCHILD, _("Fixed missing child (%s) in family (%s)"), key, famkey);
+						altered=TRUE;
+					}
+				}
 			}
 		}
 	}
+	
 	/* check spouses */
 	for (node1 = fams1; node1; node1 = nsibling(node1)) {
 		STRING famkey=rmvat(nval(node1));
 		NODE fam2 = qkey_to_fam(famkey);
 		if (!fam2) {
-			report_error(ERR_BADFAMREF, _("Bad family reference (%s) individual %s"), famkey, key);
+			if (todo.pass == 1) {
+				report_error(ERR_BADFAMREF, _("Bad family reference (%s) individual %s"), famkey, key);
+			}
 		} else {
 			/* look for indi1 (key) in fam2's spouses */
 			if (!find_xref(key, fam2, "HUSB", "WIFE")) {
-				report_error(ERR_MISSINGSPOUSE, _("Missing spouse (%s) in family (%s)"), key, famkey);
+				if (todo.pass == 1) {
+					report_error(ERR_MISSINGSPOUSE, _("Missing spouse (%s) in family (%s)"), key, famkey);
+					needfix=TRUE;
+				} else {
+					if (fix_bad_pointer(key, rec, node1)) {
+						report_fix(ERR_MISSINGSPOUSE, _("Fixed missing spouse (%s) in family (%s)"), key, famkey);
+						altered=TRUE;
+					}
+				}
 			}
 		}
 	}
+
 	join_indi(indi1, name1, refn1, sex1, body1, famc1, fams1);
-	unlock_cache(icel1);
-	check_pointers(key, rec);
-	append_indiseq_null(seq_indis, strsave(key), NULL, TRUE, TRUE);
-	return TRUE;
+	if (altered) {
+		/* must normalize, as some lineage references may have been 
+		altered to non-lineage tags to fix broken pointers */
+		normalize_indi(indi1);
+
+		/* write to database */
+		replace_indi(indi0, indi1);
+
+	} else if (needfix) {
+		enqueue_list(tofix, strsave(key));
+	}
 }
 /*=====================================
- * check_fam -- process fam record
- *  checking and/or fixing as requested
- * Created: 2001/01/14, Perry Rapp
+ * check_fam -- check family record
+ *  and record any records needing fixing
+ *  in tofix list
  *===================================*/
 static BOOLEAN
 check_fam (CNSTRING key, RECORD rec)
 {
-	static char prevkey[9];
-	NODE fam1, fref1, husb1, wife1, chil1, rest1;
-	NODE node1;
-	CACHEEL fcel1;
-	INT members = 0;
+	static char prevkey[MAXKEYWIDTH+1]="";
+	CACHEEL fcel1=0;
+	RECORD recx=0;
+
 	if (eqstr(key, prevkey)) {
 		report_error(ERR_DUPFAM, _("Duplicate family for %s"), key);
 	}
+	
 	fcel1 = fam_to_cacheel(rec);
 	lock_cache(fcel1);
-	fam1 = nztop(rec);
+
+	recx = get_record_for_cel(fcel1);
+	ASSERT(todo.pass == 1);
+	process_fam(recx);
+
+	unlock_cache(fcel1);
+	check_pointers(key, rec);
+	append_indiseq_null(seq_fams, strsave(key), NULL, TRUE, TRUE);
+	return TRUE;
+}
+/*=====================================
+ * process_fam -- process indi record
+ *  checking in pass 1, fixing in pass 2
+ *===================================*/
+static void
+process_fam (RECORD rec)
+{
+	NODE fam0, fam1;
+	NODE fref1, husb1, wife1, chil1, rest1;
+	NODE node1;
+	INT members = 0;
+	BOOLEAN altered=FALSE;
+	BOOLEAN needfix=FALSE;
+	CNSTRING key = nzkey(rec);
+
+	fam0 = nztop(rec);
+	if (todo.pass==1) {
+		fam1 = fam0;
+	} else {
+		fam1 = copy_node_subtree(fam0);
+	}
 	split_fam(fam1, &fref1, &husb1, &wife1, &chil1, &rest1);
+
 	/* check refns */
 	for (node1 = fref1; node1; node1 = nsibling(node1)) {
-		STRING refn=nval(node1);
+		/* STRING refn=nval(node1); */
 		/* TO DO: verify that refn is in db */
 	}
 	/* check husbs */
@@ -629,9 +805,19 @@ check_fam (CNSTRING key, RECORD rec)
 		NODE husb = qkey_to_indi(husbkey);
 		members++;
 		if (!husb) {
-			report_error(ERR_BADHUSBREF
-				, _("Bad HUSB reference (%s) in family %s")
-				, husbkey, key);
+			if (todo.pass == 1) {
+				report_error(ERR_BADHUSBREF
+					, _("Bad HUSB reference (%s) in family %s")
+					, husbkey, key);
+				needfix=TRUE;
+			} else {
+				if (fix_bad_pointer(key, rec, node1)) {
+					report_fix(ERR_BADHUSBREF
+						, _("Fixed Bad HUSB reference (%s) in family %s")
+						, husbkey, key);
+					altered=TRUE;
+				}
+			}
 		} else {
 			/* look for family (key) in husb */
 			if (!find_xref(key, husb, "FAMS", NULL)) {
@@ -647,15 +833,25 @@ check_fam (CNSTRING key, RECORD rec)
 		NODE wife = qkey_to_indi(wifekey);
 		members++;
 		if (!wife) {
-			report_error(ERR_BADWIFEREF
-				, _("Bad wife reference (%s) in family %s")
-				, wifekey, key);
+			if (todo.pass == 1) {
+				report_error(ERR_BADWIFEREF
+					, _("Bad wife reference (%s) in family %s")
+					, wifekey, key);
+				needfix=TRUE;
+			} else {
+				if (fix_bad_pointer(key, rec, node1)) {
+					report_fix(ERR_BADWIFEREF
+						, _("Fixed Bad wife reference (%s) in family %s")
+						, printkey(wifekey), key);
+					altered=TRUE;
+				}
+			}
 		} else {
 			/* look for family (key) in wife */
 			if (!find_xref(key, wife, "FAMS", NULL)) {
 				report_error(ERR_EXTRAWIFE
 					, _("Improper wife (%s) in family (%s)")
-					, wifekey, key);
+					, printkey(wifekey), key);
 			}
 		}
 	}
@@ -665,15 +861,25 @@ check_fam (CNSTRING key, RECORD rec)
 		NODE child = qkey_to_indi(chilkey);
 		members++;
 		if (!child) {
-			report_error(ERR_BADCHILDREF
-				, _("Bad child reference (%s) in family %s")
-				, chilkey, key);
+			if (todo.pass == 1) {
+				report_error(ERR_BADCHILDREF
+					, _("Bad child reference (%s) in family %s")
+					, printkey(chilkey), key);
+				needfix=TRUE;
+			} else {
+				if (fix_bad_pointer(key, rec, node1)) {
+					report_fix(ERR_BADCHILDREF
+						, _("Fixed bad child reference (%s) in family %s")
+						, printkey(chilkey), key);
+					altered=TRUE;
+				}
+			}
 		} else {
 			/* look for family (key) in child */
 			if (!find_xref(key, child, "FAMC", NULL)) {
 				report_error(ERR_EXTRACHILD
 					, _("Improper child (%s) in family (%s)")
-					, chilkey, key);
+					, printkey(chilkey), key);
 			}
 		}
 	}
@@ -684,10 +890,19 @@ check_fam (CNSTRING key, RECORD rec)
 	} else if (members == 1) {
 		report_error(ERR_SOLOFAM, _("Single person family (%s)"), key);
 	}
-	unlock_cache(fcel1);
-	check_pointers(key, rec);
-	append_indiseq_null(seq_fams, strsave(key), NULL, TRUE, TRUE);
-	return TRUE;
+
+
+	if (altered) {
+		/* must normalize, as some lineage references may have been 
+		altered to non-lineage tags to fix broken pointers */
+		normalize_fam(fam1);
+
+		/* write to database */
+		replace_fam(fam0, fam1);
+
+	} else if (needfix) {
+		enqueue_list(tofix, strsave(key));
+	}
 }
 /*=====================================
  * check_sour -- process sour record
@@ -697,7 +912,7 @@ check_fam (CNSTRING key, RECORD rec)
 static BOOLEAN
 check_sour (CNSTRING key, RECORD rec)
 {
-	static char prevkey[9];
+	static char prevkey[MAXKEYWIDTH+1];
 	if (!strcmp(key, prevkey)) {
 		report_error(ERR_DUPSOUR, _("Duplicate source for %s"), key);
 	}
@@ -713,7 +928,7 @@ check_sour (CNSTRING key, RECORD rec)
 static BOOLEAN
 check_even (CNSTRING key, RECORD rec)
 {
-	static char prevkey[9];
+	static char prevkey[MAXKEYWIDTH+1];
 	if (!strcmp(key, prevkey)) {
 		report_error(ERR_DUPEVEN, _("Duplicate event for %s"), key);
 	}
@@ -729,7 +944,7 @@ check_even (CNSTRING key, RECORD rec)
 static BOOLEAN
 check_othe (CNSTRING key, RECORD rec)
 {
-	static char prevkey[9];
+	static char prevkey[MAXKEYWIDTH+1];
 	if (!strcmp(key, prevkey)) {
 		report_error(ERR_DUPOTHE, _("Duplicate record for %s"), key);
 	}
@@ -753,7 +968,7 @@ find_xref (CNSTRING key, NODE node, CNSTRING tag1, CNSTRING tag2)
 		if (eqstr(tag1, ntag(node2))
 			|| (tag2 && eqstr(tag2, ntag(node2)))) {
 			STRING key2 = rmvat(nval(node2));
-			if (eqstr(key, key2)) {
+			if (key2 && eqstr(key, key2)) {
 				found = TRUE;
 				goto exit_find;
 			}
@@ -812,6 +1027,21 @@ check_node (CNSTRING n0key, NODE node, INT level)
 	if (nsibling(node))
 		check_node(n0key, nsibling(node), level);
 }
+/*===================================
+ * fix_bad_pointer -- Fix bad pointer if requested
+ *=================================*/
+static BOOLEAN
+fix_bad_pointer (CNSTRING key, RECORD rec, NODE node)
+{
+	key=key; /* unused */
+	rec=rec; /* unused */
+
+	if (todo.fix_alter_pointers) {
+		change_node_tag(node, "_badptr");
+		return TRUE;
+	}
+	return FALSE;
+}
 /*===================================
  * check_set -- Validate set of nodes
  * Created: 2001/01/14, Perry Rapp
@@ -823,19 +1053,44 @@ check_set (INDISEQ seq, char ctype)
 	keysort_indiseq(seq); /* spri now valid */
 	i = xref_next(ctype, i);
 	FORINDISEQ(seq, el, num)
-		while (i<spri(el)) {
-			report_error(ERR_MISSING
+		while (i<element_ikey(el)) {
+			report_error(ERR_UNDELETED
 				, _("Missing undeleted record %c%d")
 				, ctype, i);
+			if (todo.fix_deletes) {
+				char key[33];
+				sprintf(key, "%c%d", ctype, i);
+				if (mark_deleted_record_as_deleted(key)) {
+					report_fix(ERR_UNDELETED
+						, _("Fixed missing undeleted record %c%d")
+						, ctype, i);
+				} else {
+					report_error(-1
+						, _("Failed to fix missing undeleted record %c%d")
+						, ctype, i);
+				}
+			}
 			i = xref_next(ctype, i);
 		}
-		if (i == spri(el)) {
+		if (i == element_ikey(el)) {
 			/* in synch */
 			i = xref_next(ctype, i);
 		} else { /* spri(el) < i */
+			CNSTRING key = element_skey(el);
 			report_error(ERR_DELETED
 				, _("Delete set contains valid record %s")
-				, skey(el));
+				, key);
+			if (todo.fix_deletes) {
+				if (mark_live_record_as_live(key)) {
+					report_fix(ERR_DELETED
+						, _("Fixed delete set contains valid record %s")
+						, key);
+				} else {
+					report_error(-1
+						, _("Failed to fix missing undeleted record %s")
+						, key);
+				}
+			}
 			/* fall thru and only advance seq */
 		}
 	ENDINDISEQ
@@ -848,7 +1103,7 @@ static BOOLEAN
 check_btree (BTREE btr)
 {
 	/* keep track of which files we've visited */
-	TABLE fkeytab = create_table(FREEKEY);
+	TABLE fkeytab = create_table_int();
 	/* check that master index has its fkey correct */
 	INDEX index = bmaster(BTR);
 	INT mk1 = bkfile(btr).k_mkey;
@@ -857,7 +1112,7 @@ check_btree (BTREE btr)
 		return FALSE;
 	}
 	check_index(btr, index, fkeytab, NULL, NULL);
-	remove_table(fkeytab, FREEKEY);
+	destroy_table(fkeytab);
 	return TRUE;
 }
 /*=========================================
@@ -869,22 +1124,36 @@ check_index (BTREE btr, INDEX index, TABLE fkeytab, RKEY * lo, RKEY * hi)
 {
 	INT n = nkeys(index);
 	INT i;
-	if (!check_keys(btr, (BLOCK)index, lo, hi))
+	if (!check_keys((BLOCK)index, lo, hi))
 		return FALSE;
 	for (i = 0; i <= n; i++) {
-		INDEX newix = readindex(bbasedir(btr), fkeys(index, i), TRUE);
+		INDEX newix=0;
+		char scratch[200];
+		FKEY fkey = fkeys(index, i);
 		RKEY *lox, *hix;
+
+		get_index_file(scratch, btr, fkey);
+		if (in_table(fkeytab, scratch)) {
+			printf(_("Cycle in indexes, file %s found again!\n"), scratch);
+			return FALSE;
+		} else {
+			insert_table_int(fkeytab, scratch, 1);
+		}
+		newix = readindex(btr, fkey, TRUE);
 		if (!newix) {
 			printf(_("Error loading index at key %d\n"), i);
-			printblock(btr, (BLOCK)index);
+			printblock((BLOCK)index);
 		}
 		/* figure upper & lower bounds of what keys should be in the child */
 		lox = (i==0 ? lo : &rkeys(index, i));
 		hix = (i==n ? hi : &rkeys(index, i+1));
-		if (ixtype(newix) == BTINDEXTYPE)
-			check_index(btr, newix, fkeytab, lox, hix);
-		else
-			check_block(btr, (BLOCK)newix, lox, hix);
+		if (ixtype(newix) == BTINDEXTYPE) {
+			if (!check_index(btr, newix, fkeytab, lox, hix))
+				return FALSE;
+		} else {
+			if (!check_block((BLOCK)newix, lox, hix))
+				return FALSE;
+		}
 	}
 	/* TODO: use fkeytab */
 	return TRUE;
@@ -894,10 +1163,9 @@ check_index (BTREE btr, INDEX index, TABLE fkeytab, RKEY * lo, RKEY * hi)
  * Created: 2003/09/05, Perry Rapp
  *=======================================*/
 static BOOLEAN
-check_block (BTREE btr, BLOCK block, RKEY * lo, RKEY * hi)
+check_block (BLOCK block, RKEY * lo, RKEY * hi)
 {
-	INT n = nkeys(block);
-	if (!check_keys(btr, block, lo, hi))
+	if (!check_keys(block, lo, hi))
 		return FALSE;
 	return TRUE;
 }
@@ -906,7 +1174,7 @@ check_block (BTREE btr, BLOCK block, RKEY * lo, RKEY * hi)
  * Created: 2003/09/05, Perry Rapp
  *=======================================*/
 static BOOLEAN
-check_keys (BTREE btr, BLOCK block, RKEY * lo, RKEY * hi)
+check_keys (BLOCK block, RKEY * lo, RKEY * hi)
 {
 	INT n = nkeys(block);
 	INT i = 0;
@@ -918,26 +1186,26 @@ check_keys (BTREE btr, BLOCK block, RKEY * lo, RKEY * hi)
 		--n; /* keys are 0..n-1 for block */
 	for (i=start ; i <= n; i++) {
 		if (i==start && lo) {
-			INT rel = cmpkeys(btr, lo, &rkeys(block, i));
+			INT rel = cmpkeys(lo, &rkeys(block, i));
 			if (rel < 0) {
 				printf(_("First key in block below parent's limit\n"));
-				printblock(btr, block);
+				printblock(block);
 				ok = FALSE;
 			}
 		}
 		if (i==n && hi) {
-			INT rel = cmpkeys(btr, &rkeys(block, i), hi);
+			INT rel = cmpkeys(&rkeys(block, i), hi);
 			if (rel > 0) {
 				printf(_("Last key in block above parent's limit\n"));
-				printblock(btr, block);
+				printblock(block);
 				ok = FALSE;
 			}
 		}
 		if (i<n) {
-			INT rel = cmpkeys(btr, &rkeys(block, i), &rkeys(block, i+1));
+			INT rel = cmpkeys(&rkeys(block, i), &rkeys(block, i+1));
 			if (rel >= 0) {
 				printf(_("Key %d not below next key\n"), i);
-				printblock(btr, block);
+				printblock(block);
 				ok = FALSE;
 			}
 		}
@@ -949,7 +1217,7 @@ check_keys (BTREE btr, BLOCK block, RKEY * lo, RKEY * hi)
  * Created: 2003/09/05, Perry Rapp
  *=======================================*/
 static void
-printblock (BTREE btr, BLOCK block)
+printblock (BLOCK block)
 {
 	FKEY fkey = ixself(block);
 	STRING tname = _("data block");
@@ -959,7 +1227,7 @@ printblock (BTREE btr, BLOCK block)
 }
 /*=========================================
  * validate_errs -- Validate the errs array
- * Created: 2001/01/13, Perry Rapp
+ * (fatal assert if finds problem)
  *=======================================*/
 static void
 validate_errs (void)
@@ -974,9 +1242,48 @@ validate_errs (void)
 		}
 	}
 }
+/*=========================================
+ * check_missing_data_records -- Check for records
+ *  present in index but not in data
+ *=======================================*/
+static void
+check_missing_data_records (void)
+{
+	check_typed_missing_data_records('I');
+	check_typed_missing_data_records('F');
+	check_typed_missing_data_records('S');
+	check_typed_missing_data_records('E');
+	check_typed_missing_data_records('X');
+}
+/*=========================================
+ * check_typed_missing_data_records -- Check all records 
+ *  of specified type for orphans
+ *=======================================*/
+static void
+check_typed_missing_data_records (char ntype)
+{
+	int keynum=0;
+	while (TRUE) {
+		char key[33];
+		keynum = xref_next(ntype, keynum);
+		if (!keynum) return;
+		sprintf(key, "%c%d", ntype, keynum);
+		if (noisy)
+			report_progress("Check data record presence: %s", key);
+		if (is_record_missing_data_entry(key)) {
+			if (todo.pass == 1) {
+				report_error(ERR_MISSINGREC, _("Missing data record (%s)"), key);
+			} else {
+				if (todo.fix_missing_data_records) {
+					delete_record_missing_data_entry(key);
+					report_fix(ERR_MISSINGREC, _("Fixed missing data record (%s)"), key);
+				}
+			}
+		}
+	}
+}
 /*===========================================
  * main -- Main procedure of dbverify command
- * Created: 2001/01/01, Perry Rapp
  *=========================================*/
 int
 main (int argc,
@@ -991,15 +1298,23 @@ main (int argc,
 	BOOLEAN allchecks=FALSE; /* if user requested all checks */
 	INT returnvalue=1;
 
+	/* initialize all the low-level library code */
+	init_stdlib();
+
 	validate_errs();
 
-	save_original_locales();
+#if HAVE_SETLOCALE
+	/* initialize locales */
+	setlocale(LC_ALL, "");
+#endif /* HAVE_SETLOCALE */
 
 #if ENABLE_NLS
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 #endif
 
+	save_original_locales();
+
 #ifdef WIN32
 	/* TO DO - research if this is necessary */
 	_fmode = O_BINARY;	/* default to binary rather than TEXT mode */
@@ -1011,9 +1326,9 @@ main (int argc,
 	}
 	flags = argv[1];
 	dbname = argv[2];
-	for (ptr=&flags[1]; *ptr; ptr++) {
+	for (ptr=&flags[1]; *ptr; ++ptr) {
 		switch(*ptr) {
-		case 'L': todo.check_dbstructure=TRUE; break;
+		case 'l': todo.check_dbstructure=TRUE; break;
 		case 'g': todo.find_ghosts=TRUE; break;
 		case 'G': todo.fix_ghosts=TRUE; break;
 		case 'i': todo.check_indis=TRUE; break;
@@ -1023,6 +1338,10 @@ main (int argc,
 		case 'x': todo.check_othes=TRUE; break;
 		case 'n': noisy=TRUE; break;
 		case 'a': allchecks=TRUE; break;
+		case 'F': todo.fix_alter_pointers=TRUE; break;
+		case 'm': todo.check_missing_data_records=TRUE; break;
+		case 'M': todo.fix_missing_data_records=TRUE; break;
+		case 'D': todo.fix_deletes=TRUE; break;
 		default: print_usage(); goto done;
 		}
 	}
@@ -1031,12 +1350,16 @@ main (int argc,
 	/* This is unnecessary -- it is off anyway, Perry, 2001/10/28 */
 	alloclog  = FALSE;
 
+	/* Enable any checks needed for fixes selected */
+	if (todo.fix_missing_data_records)
+		todo.check_missing_data_records = 1;
+
 	/* initialize options & misc. stuff */
 	if (!init_lifelines_global(0, &msg, 0)) {
 		printf("%s\n", msg);
 		goto done;
 	}
-	if (!(BTR = openbtree(dbname, cflag, writ, immut))) {
+	if (!(BTR = bt_openbtree(dbname, cflag, writ, immut))) {
 		char buffer[256];
 		describe_dberror(bterrno, buffer, ARRSIZE(buffer));
 		puts(buffer);
@@ -1060,35 +1383,33 @@ main (int argc,
 	}
 	printf("Checking %s\n", dbname);
 
-	if (todo.find_ghosts || todo.fix_ghosts)
-		check_ghosts();
-
-	/* this simplifies later logic */
-	if (todo.fix_indis) todo.check_indis=TRUE;
-	if (todo.fix_fams) todo.check_fams=TRUE;
-	if (todo.fix_sours) todo.check_sours=TRUE;
-	if (todo.fix_evens) todo.check_evens=TRUE;
-	if (todo.fix_othes) todo.check_othes=TRUE;
-
-	/* all fixes - if any new ones, have to update this */
+	/* all checks - if any new ones, have to update this */
 	if (allchecks) {
 		todo.check_indis=todo.check_fams=todo.check_sours=TRUE;
 		todo.check_evens=todo.check_othes=TRUE;
 		todo.find_ghosts=TRUE;
 	}
 
-	if (!(bwrite(BTR))) {
-		todo.fix_indis = todo.fix_fams = todo.fix_sours = FALSE;
-		todo.fix_evens = todo.fix_othes = FALSE;
-	}
+	if (todo.find_ghosts || todo.fix_ghosts)
+		check_ghosts();
 
 
+	if (!(bwrite(BTR))) {
+		todo.fix_alter_pointers = FALSE;
+		todo.fix_ghosts = FALSE;
+	}
+
 	if (todo.check_indis
 		|| todo.check_fams
 		|| todo.check_sours
 		|| todo.check_evens
-		|| todo.check_othes)
-		check_nodes();
+		|| todo.check_othes) {
+		check_and_fix_records();
+	}
+
+	if (todo.check_missing_data_records) {
+		check_missing_data_records();
+	}
 
 	report_results();
 
@@ -1132,7 +1453,7 @@ report_results (void)
  *  handles null or empty details input
  *===========================*/
 void
-__fatal (STRING file, int line, STRING details)
+__fatal (STRING file, int line, CNSTRING details)
 {
 	printf(_("FATAL ERROR: "));
 	if (details && details[0]) {
@@ -1154,3 +1475,12 @@ crashlog (STRING fmt, ...)
 	vprintf(fmt, args);
 	va_end(args);
 }
+/*===============================
+ * printkey -- Displayable version of key
+ *  same as key, except for NULL keys
+ *=============================*/
+static CNSTRING
+printkey (CNSTRING key)
+{
+	return key ? key : "NULL";
+}
